(* generated by Ott 0.21.2 from: lang.ott *)

Require Import Arith.
Require Import Bool.
Require Import List.


Require Import Omega ClassicalChoice basis. 
Set Implicit Arguments. 

Definition termvar := nat.
Lemma eq_termvar: forall (x y : termvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_termvar : ott_coq_equality.
Definition typename := nat.
Lemma eq_typename: forall (x y : typename), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typename : ott_coq_equality.
Definition typebvar := nat.
Lemma eq_typebvar: forall (x y : typebvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typebvar : ott_coq_equality.
Definition location := nat.

Inductive exp : Set :=  (*r expressions *)
 | Evar (x:termvar) (*r Variable *)
 | Eunit : exp
 | Eint (n:nat)
 | Etrue : exp
 | Efalse : exp
 | Epair (e1:exp) (e2:exp)
 | Einl (e:exp)
 | Einr (e:exp)
 | Efst (e:exp)
 | Esnd (e:exp)
 | Eloc (loc:location)
 | Eif (e0:exp) (e1:exp) (e2:exp) (*r if *)
 | Ecase (x:termvar) (e0:exp) (e1:exp) (e2:exp) (*r case *)
 | Eroll (e:exp) (*r roll *)
 | Eunroll (e:exp) (*r unroll *)
 | Efix (f:termvar) (x:termvar) (e:exp) (*r Recursive function *)
 | Eapp (e:exp) (e':exp) (*r Function application *)
 | Egen (e:exp) (*r Type generalisation *)
 | Einst (e:exp) (*r Type instantiation *)
 | Epack (e:exp)
 | Eunpack (x:termvar) (e1:exp) (e2:exp)
 | Enew (e:exp)
 | Eread (e:exp)
 | Ewrite (e1:exp) (e2:exp)
 | Erefeq (e1:exp) (e2:exp).

Inductive typ : Set :=  (*r types (in locally nameless style) *)
 | Tname (aa:typename) (*r Free name *)
 | Tbvar (bb:typebvar) (*r Bound variable *)
 | Tunit : typ (*r Unit *)
 | Tint : typ (*r Integers *)
 | Tbool : typ (*r Booleans *)
 | Tprod (ty1:typ) (ty2:typ) (*r products *)
 | Tsum (ty1:typ) (ty2:typ) (*r Sums *)
 | Tarrow (ty1:typ) (ty2:typ) (*r Functions *)
 | Tref (ty:typ) (*r References *)
 | Tmu (ty:typ) (*r Recursive types *)
 | Tforall (ty:typ)
 | Texists (ty:typ).

Inductive val : Set :=  (*r values *)
 | Vvar (x:termvar)
 | Vunit : val
 | Vint (n:nat)
 | Vtrue : val
 | Vfalse : val
 | Vpair (v1:val) (v2:val)
 | Vinl (v:val)
 | Vinr (v:val)
 | Vroll (v:val)
 | Vfix (f:termvar) (x:termvar) (e:exp)
 | Vgen (e:exp)
 | Vpack (v:val)
 | Vloc (loc:location).

Inductive prg : Set :=  (*r Programs *)
 | Pvar (x:termvar)
 | Punit : prg
 | Pint (n:nat)
 | Ptrue : prg
 | Pfalse : prg
 | Ppair (p1:prg) (p2:prg)
 | Pinl (ty:typ) (p:prg)
 | Pinr (ty:typ) (p:prg)
 | Pfst (p:prg)
 | Psnd (p:prg)
 | Pif (p0:prg) (p1:prg) (p2:prg)
 | Pcase (x:termvar) (p0:prg) (p1:prg) (p2:prg)
 | Proll (ty:typ) (p:prg)
 | Punroll (p:prg)
 | Pfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (p:prg)
 | Papp (p:prg) (p':prg) (*r Function application *)
 | Pgen (p:prg)
 | Pinst (p:prg) (ty:typ)
 | Ppack (ty1:typ) (p:prg) (ty:typ)
 | Punpack (x:termvar) (p1:prg) (p2:prg)
 | Pnew (p:prg)
 | Pread (p:prg)
 | Pwrite (p1:prg) (p2:prg)
 | Prefeq (p1:prg) (p2:prg).

Inductive ctx : Set :=  (*r Contexts *)
 | Chole : ctx
 | Cpair1 (C:ctx) (p2:prg)
 | Cpair2 (p1:prg) (C:ctx)
 | Cinl (ty:typ) (C:ctx)
 | Cinr (ty:typ) (C:ctx)
 | Cfst (C:ctx)
 | Csnd (C:ctx)
 | Cif0 (C:ctx) (p1:prg) (p2:prg)
 | Cif1 (p0:prg) (C:ctx) (p2:prg)
 | Cif2 (p0:prg) (p1:prg) (C:ctx)
 | Ccase0 (x:termvar) (C:ctx) (p1:prg) (p2:prg)
 | Ccase1 (x:termvar) (p0:prg) (C:ctx) (p2:prg)
 | Ccase2 (x:termvar) (p0:prg) (p1:prg) (C:ctx)
 | Croll (ty:typ) (C:ctx)
 | Cunroll (C:ctx)
 | Cfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (C:ctx)
 | Capp1 (C:ctx) (p:prg)
 | Capp2 (p:prg) (C:ctx)
 | Cgen (C:ctx)
 | Cinst (C:ctx) (ty:typ)
 | Cpack (ty1:typ) (C:ctx) (ty:typ)
 | Cunpack1 (x:termvar) (C:ctx) (p:prg)
 | Cunpack2 (x:termvar) (p:prg) (C:ctx)
 | Cnew (C:ctx)
 | Cread (C:ctx)
 | Cwrite1 (C:ctx) (p:prg)
 | Cwrite2 (p:prg) (C:ctx)
 | Crefeq1 (C:ctx) (p:prg)
 | Crefeq2 (p:prg) (C:ctx).

Definition typeenv : Set := (list typename).

Definition termenv : Set := (list (termvar * typ)).


(*----------------------------------------------------------------------------*)
(** Coercions *)
(*----------------------------------------------------------------------------*)

Fixpoint exp_of_val (v : val) : exp :=
  match v with
  | Vvar x      => Evar x
  | Vunit       => Eunit
  | Vint n      => Eint n
  | Vtrue       => Etrue 
  | Vfalse      => Efalse
  | Vpair v1 v2 => Epair (exp_of_val v1) (exp_of_val v2)
  | Vinl v      => Einl (exp_of_val v)
  | Vinr v      => Einr (exp_of_val v)
  | Vroll v     => Eroll (exp_of_val v)
  | Vfix f x e  => Efix f x e
  | Vgen e      => Egen e
  | Vpack v     => Epack (exp_of_val v)
  | Vloc loc    => Eloc loc
  end.

Structure VoE := mkVoE { e_of : exp ; v_of : val ; VoE_pf : e_of = exp_of_val v_of }.


Inductive scont : Set :=  (*r Single-step evaluation contexts *)
 | SKif (e1:exp) (e2:exp)
 | SKpairL (e:exp)
 | SKpairR (v:val)
 | SKfst : scont
 | SKsnd : scont
 | SKinl : scont
 | SKinr : scont
 | SKcase (x:termvar) (e1:exp) (e2:exp)
 | SKroll : scont
 | SKunroll : scont
 | SKpack : scont
 | SKappL (e:exp)
 | SKappR (v:val)
 | SKinst : scont
 | SKunpack (x:termvar) (e:exp)
 | SKnew : scont
 | SKread : scont
 | SKwriteL (e:exp)
 | SKwriteR (v:val)
 | SKrefeqL (e:exp)
 | SKrefeqR (v:val).

Definition cont : Set := list scont.

Definition voe : Set := VoE.

(*----------------------------------------------------------------------------*)
(** Definition of a ``locally closed'' type *)
(*----------------------------------------------------------------------------*)

Fixpoint lc_typ (n : nat) (ty : typ) :=
  match ty with
  | Tbvar aa     => aa < n
  | Tname aa    => True
  | Tunit       => True 
  | Tint        => True 
  | Tbool       => True
  | (Tprod ty1 ty2)  => lc_typ n ty1 /\ lc_typ n ty2
  | (Tsum ty1 ty2)   => lc_typ n ty1 /\ lc_typ n ty2
  | (Tarrow ty1 ty2) => lc_typ n ty1 /\ lc_typ n ty2
  | (Tref ty)    => lc_typ n ty
  | (Tmu ty)     => lc_typ (S n) ty
  | (Tforall ty) => lc_typ (S n) ty
  | (Texists ty) => lc_typ (S n) ty
  end.

(*----------------------------------------------------------------------------*)
(** Substitution of bounded type variables *)
(*----------------------------------------------------------------------------*)

Fixpoint tsubst1 (x: typebvar) (tyA ty: typ) :=
  match ty with
  | Tbvar aa      =>
    match nat_compare aa x with
      | Lt => Tbvar aa
      | Eq => tyA
      | Gt => Tbvar (pred aa)
    end
  | Tname aa    => Tname aa
  | Tunit       => Tunit 
  | Tint        => Tint 
  | Tbool       => Tbool 
  | (Tprod ty1 ty2)  => Tprod  (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tsum ty1 ty2)   => Tsum   (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tref ty)    => Tref (tsubst1 x tyA ty)
  | (Tmu ty)     => Tmu     (tsubst1 (S x) tyA ty)
  | (Tforall ty) => Tforall (tsubst1 (S x) tyA ty)
  | (Texists ty) => Texists (tsubst1 (S x) tyA ty)
  end.

(*----------------------------------------------------------------------------*)
(** Definition of heaps *)
(*----------------------------------------------------------------------------*)

Definition finite_domain A B (f : A -> option B) := 
  exists d, forall x y, f x = Some y -> In x d.

Structure heap := 
  { heap_f :> location -> option val 
  ; heap_findom : finite_domain heap_f }.

Program Definition hemp : heap := {| heap_f := fun l => None |}.
Next Obligation. exists nil; discriminate. Qed.

Program Definition hext (h : heap) (loc : location) (v: val) : heap :=
  {| heap_f := fun l => if eq_nat_dec l loc then Some v else h l |}.  
Next Obligation. 
  destruct (heap_findom h) as [d X]; exists (loc :: d); intros; simpl.
  destruct (eq_nat_dec); eauto.
Qed.

Definition hplus (h1 h2 h: heap) := 
  forall x, h x = h2 x /\ h1 x = None \/ h x = h1 x /\ h2 x = None.


Parameter allocator : heap -> location.

Axiom allocator_unalloc : forall h : heap, h (allocator h) = None.

Lemma allocator_consistency: 
  exists allocator : heap -> location, forall h : heap, h (allocator h) = None.
Proof.
  cut (forall h : heap, exists l, h l = None).
    intros FRESH; apply (choice _ FRESH).
  intros [hf [d FIN] ]; simpl.
  set (l := S (fold_right plus 0 d)); specialize (FIN l); exists l.
  destruct (hf l); auto.
  cut (forall n (GE: n >= l) (IN: In n d), False).
    intros; exfalso; eauto.
  unfold l in *; clear -d; induction d; intros; eauto.
  simpl in GE, IN; destruct IN; [subst n | apply (IHd n)]; auto; omega.
Qed.

(*----------------------------------------------------------------------------*)
(** Operations *)
(*----------------------------------------------------------------------------*)

Fixpoint app_ctx (C : ctx) (p0: prg) : prg :=
  match C with
  | Chole => p0                         
  | (Cpair1 C p2) => (Ppair  (  (app_ctx  C   p0 )  )  p2)
  | (Cpair2 p1 C) => (Ppair p1  (  (app_ctx  C   p0 )  ) )
  | (Cinl ty C) => (Pinl ty  (  (app_ctx  C   p0 )  ) )
  | (Cinr ty C) => (Pinr ty  (  (app_ctx  C   p0 )  ) )
  | (Cfst C) => (Pfst  (  (app_ctx  C   p0 )  ) )
  | (Csnd C) => (Psnd  (  (app_ctx  C   p0 )  ) )
  | (Cif0 C p1 p2) => (Pif  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Cif1 p C p2) => (Pif p  (  (app_ctx  C   p0 )  )  p2)
  | (Cif2 p p1 C) => (Pif p p1  (  (app_ctx  C   p0 )  ) )
  | (Ccase0 x C p1 p2) => (Pcase x  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Ccase1 x p C p2) => (Pcase x p  (app_ctx  C   p0 )  p2)
  | (Ccase2 x p p1 C) => (Pcase x p p1  (app_ctx  C   p0 ) )
  | (Croll ty C) => (Proll ty  (  (app_ctx  C   p0 )  ) )
  | (Cunroll C) => (Punroll  (  (app_ctx  C   p0 )  ) )
  | (Cfix f x ty ty' C) => (Pfix f x ty ty'  (  (app_ctx  C   p0 )  ) )
  | (Capp1 C p) => (Papp  (  (app_ctx  C   p0 )  )  p)
  | (Capp2 p C) => (Papp p  (  (app_ctx  C   p0 )  ) )
  | (Cgen C) => (Pgen  (  (app_ctx  C   p0 )  ) )
  | (Cinst C ty) => (Pinst  (  (app_ctx  C   p0 )  )  ty)
  | (Cpack ty1 C ty) => (Ppack ty1  (app_ctx  C   p0 )  ty)
  | (Cunpack1 x C p) => (Punpack x  (app_ctx  C   p0 )  p)
  | (Cunpack2 x p C) => (Punpack x p  (  (app_ctx  C   p0 )  ) )
  | (Cnew C) => (Pnew  (  (app_ctx  C   p0 )  ) )
  | (Cread C) => (Pread  (  (app_ctx  C   p0 )  ) )
  | (Cwrite1 C p) => (Pwrite  (  (app_ctx  C   p0 )  )  p)
  | (Cwrite2 p C) => (Pwrite p  (  (app_ctx  C   p0 )  ) )
  | (Crefeq1 C p) => (Prefeq  (  (app_ctx  C   p0 )  )  p)
  | (Crefeq2 p C) => (Prefeq p  (  (app_ctx  C   p0 )  ) )
  end.



(** subrules *)
Fixpoint is_value_of_exp (e_5:exp) : Prop :=
  match e_5 with
  | (Evar x) => (True)
  | Eunit => (True)
  | (Eint n) => (True)
  | Etrue => (True)
  | Efalse => (True)
  | (Epair e1 e2) => ((is_value_of_exp e1) /\ (is_value_of_exp e2))
  | (Einl e) => ((is_value_of_exp e))
  | (Einr e) => ((is_value_of_exp e))
  | (Efst e) => False
  | (Esnd e) => False
  | (Eloc loc) => (True)
  | (Eif e0 e1 e2) => False
  | (Ecase x e0 e1 e2) => False
  | (Eroll e) => ((is_value_of_exp e))
  | (Eunroll e) => False
  | (Efix f x e) => (True)
  | (Eapp e e') => False
  | (Egen e) => (True)
  | (Einst e) => False
  | (Epack e) => ((is_value_of_exp e))
  | (Eunpack x e1 e2) => False
  | (Enew e) => False
  | (Eread e) => False
  | (Ewrite e1 e2) => False
  | (Erefeq e1 e2) => False
end.


(** context application *)
Definition appctx_scont_exp (scont5:scont) (exp5:exp) : exp :=
  match scont5 with
  | (SKif e1 e2) => (Eif exp5 e1 e2)
  | (SKpairL e) => (Epair exp5 e)
  | (SKpairR v) => (Epair  (exp_of_val  v )  exp5)
  | SKfst => (Efst exp5)
  | SKsnd => (Esnd exp5)
  | SKinl => (Einl exp5)
  | SKinr => (Einr exp5)
  | (SKcase x e1 e2) => (Ecase x exp5 e1 e2)
  | SKroll => (Eroll exp5)
  | SKunroll => (Eunroll exp5)
  | SKpack => (Epack exp5)
  | (SKappL e) => (Eapp exp5 e)
  | (SKappR v) => (Eapp  (exp_of_val  v )  exp5)
  | SKinst => (Einst exp5)
  | (SKunpack x e) => (Eunpack x exp5 e)
  | SKnew => (Enew exp5)
  | SKread => (Eread exp5)
  | (SKwriteL e) => (Ewrite exp5 e)
  | (SKwriteR v) => (Ewrite  (exp_of_val  v )  exp5)
  | (SKrefeqL e) => (Erefeq exp5 e)
  | (SKrefeqR v) => (Erefeq  (exp_of_val  v )  exp5)
end.



(*----------------------------------------------------------------------------*)
(** Definition of app_cont *)
(*----------------------------------------------------------------------------*)

Definition app_cont (k : cont) (e : exp) : exp :=
  fold_left (fun e0 k0 => appctx_scont_exp k0 e0) k e.

Definition app_cont_r k e := fold_right appctx_scont_exp e k.

Lemma app_cont_r_reverse: forall k e,
  app_cont_r k e = app_cont (rev k) e.
Proof. 
 intros; unfold app_cont, app_cont_r.
 rewrite <-fold_left_rev_right, rev_involutive; auto.
Qed.

Lemma app_cont_reverse: forall k e,
  app_cont k e = app_cont_r (rev k) e.
Proof. 
 intros; rewrite <-(rev_involutive k) at 1; rewrite app_cont_r_reverse; auto.
Qed.



(** free variables *)
Fixpoint ftn_typ (ty_5:typ) : list typename :=
  match ty_5 with
  | (Tname aa) => (cons aa nil)
  | (Tbvar bb) => nil
  | Tunit => nil
  | Tint => nil
  | Tbool => nil
  | (Tprod ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tsum ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tarrow ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tref ty) => ((ftn_typ ty))
  | (Tmu ty) => ((ftn_typ ty))
  | (Tforall ty) => ((ftn_typ ty))
  | (Texists ty) => ((ftn_typ ty))
end.

Fixpoint fv_exp (e_5:exp) : list termvar :=
  match e_5 with
  | (Evar x) => (cons x nil)
  | Eunit => nil
  | (Eint n) => nil
  | Etrue => nil
  | Efalse => nil
  | (Epair e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Einl e) => ((fv_exp e))
  | (Einr e) => ((fv_exp e))
  | (Efst e) => ((fv_exp e))
  | (Esnd e) => ((fv_exp e))
  | (Eloc loc) => nil
  | (Eif e0 e1 e2) => (app (fv_exp e0) (app (fv_exp e1) (fv_exp e2)))
  | (Ecase x e0 e1 e2) => (app (fv_exp e0) (app (list_minus eq_termvar (fv_exp e1) (cons x nil)) (list_minus eq_termvar (fv_exp e2) (cons x nil))))
  | (Eroll e) => ((fv_exp e))
  | (Eunroll e) => ((fv_exp e))
  | (Efix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
  | (Eapp e e') => (app (fv_exp e) (fv_exp e'))
  | (Egen e) => ((fv_exp e))
  | (Einst e) => ((fv_exp e))
  | (Epack e) => ((fv_exp e))
  | (Eunpack x e1 e2) => (app (fv_exp e1) (list_minus eq_termvar (fv_exp e2) (cons x nil)))
  | (Enew e) => ((fv_exp e))
  | (Eread e) => ((fv_exp e))
  | (Ewrite e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Erefeq e1 e2) => (app (fv_exp e1) (fv_exp e2))
end.

Fixpoint ftn_prg (p_5:prg) : list typename :=
  match p_5 with
  | (Pvar x) => nil
  | Punit => nil
  | (Pint n) => nil
  | Ptrue => nil
  | Pfalse => nil
  | (Ppair p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Pinl ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Pinr ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Pfst p) => ((ftn_prg p))
  | (Psnd p) => ((ftn_prg p))
  | (Pif p0 p1 p2) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_prg p2)))
  | (Pcase x p0 p1 p2) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_prg p2)))
  | (Proll ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Punroll p) => ((ftn_prg p))
  | (Pfix f x ty ty' p) => (app (ftn_typ ty) (app (ftn_typ ty') (ftn_prg p)))
  | (Papp p p') => (app (ftn_prg p) (ftn_prg p'))
  | (Pgen p) => ((ftn_prg p))
  | (Pinst p ty) => (app (ftn_prg p) (ftn_typ ty))
  | (Ppack ty1 p ty) => (app (ftn_typ ty1) (app (ftn_prg p) (ftn_typ ty)))
  | (Punpack x p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Pnew p) => ((ftn_prg p))
  | (Pread p) => ((ftn_prg p))
  | (Pwrite p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Prefeq p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
end.

Fixpoint fv_val (v_5:val) : list termvar :=
  match v_5 with
  | (Vvar x) => nil
  | Vunit => nil
  | (Vint n) => nil
  | Vtrue => nil
  | Vfalse => nil
  | (Vpair v1 v2) => (app (fv_val v1) (fv_val v2))
  | (Vinl v) => ((fv_val v))
  | (Vinr v) => ((fv_val v))
  | (Vroll v) => ((fv_val v))
  | (Vfix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
  | (Vgen e) => ((fv_exp e))
  | (Vpack v) => ((fv_val v))
  | (Vloc loc) => nil
end.

Fixpoint ftn_ctx (C5:ctx) : list typename :=
  match C5 with
  | Chole => nil
  | (Cpair1 C p2) => (app (ftn_ctx C) (ftn_prg p2))
  | (Cpair2 p1 C) => (app (ftn_prg p1) (ftn_ctx C))
  | (Cinl ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cinr ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cfst C) => ((ftn_ctx C))
  | (Csnd C) => ((ftn_ctx C))
  | (Cif0 C p1 p2) => (app (ftn_ctx C) (app (ftn_prg p1) (ftn_prg p2)))
  | (Cif1 p0 C p2) => (app (ftn_prg p0) (app (ftn_ctx C) (ftn_prg p2)))
  | (Cif2 p0 p1 C) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_ctx C)))
  | (Ccase0 x C p1 p2) => (app (ftn_ctx C) (app (ftn_prg p1) (ftn_prg p2)))
  | (Ccase1 x p0 C p2) => (app (ftn_prg p0) (app (ftn_ctx C) (ftn_prg p2)))
  | (Ccase2 x p0 p1 C) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_ctx C)))
  | (Croll ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cunroll C) => ((ftn_ctx C))
  | (Cfix f x ty ty' C) => (app (ftn_typ ty) (app (ftn_typ ty') (ftn_ctx C)))
  | (Capp1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Capp2 p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Cgen C) => ((ftn_ctx C))
  | (Cinst C ty) => (app (ftn_ctx C) (ftn_typ ty))
  | (Cpack ty1 C ty) => (app (ftn_typ ty1) (app (ftn_ctx C) (ftn_typ ty)))
  | (Cunpack1 x C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Cunpack2 x p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Cnew C) => ((ftn_ctx C))
  | (Cread C) => ((ftn_ctx C))
  | (Cwrite1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Cwrite2 p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Crefeq1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Crefeq2 p C) => (app (ftn_prg p) (ftn_ctx C))
end.

Definition fv_scont (SK5:scont) : list termvar :=
  match SK5 with
  | (SKif e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (SKpairL e) => ((fv_exp e))
  | (SKpairR v) => ((fv_val v))
  | SKfst => nil
  | SKsnd => nil
  | SKinl => nil
  | SKinr => nil
  | (SKcase x e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | SKroll => nil
  | SKunroll => nil
  | SKpack => nil
  | (SKappL e) => ((fv_exp e))
  | (SKappR v) => ((fv_val v))
  | SKinst => nil
  | (SKunpack x e) => ((fv_exp e))
  | SKnew => nil
  | SKread => nil
  | (SKwriteL e) => ((fv_exp e))
  | (SKwriteR v) => ((fv_val v))
  | (SKrefeqL e) => ((fv_exp e))
  | (SKrefeqR v) => ((fv_val v))
end.


(** substitutions *)
Fixpoint tsubst_typ (sub:list (typename*typ)) (ty__6:typ) {struct ty__6} : typ :=
  match ty__6 with
  | (Tname aa) => (match list_assoc eq_typename aa sub with None => (Tname aa) | Some ty_5 => ty_5 end)
  | (Tbvar bb) => Tbvar bb
  | Tunit => Tunit 
  | Tint => Tint 
  | Tbool => Tbool 
  | (Tprod ty1 ty2) => Tprod (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tsum ty1 ty2) => Tsum (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tref ty) => Tref (tsubst_typ sub ty)
  | (Tmu ty) => Tmu (tsubst_typ sub ty)
  | (Tforall ty) => Tforall (tsubst_typ sub ty)
  | (Texists ty) => Texists (tsubst_typ sub ty)
end.

Fixpoint subst_exp (sub:list (termvar*exp)) (e__6:exp) {struct e__6} : exp :=
  match e__6 with
  | (Evar x) => (match list_assoc eq_termvar x sub with None => (Evar x) | Some e_5 => e_5 end)
  | Eunit => Eunit 
  | (Eint n) => Eint n
  | Etrue => Etrue 
  | Efalse => Efalse 
  | (Epair e1 e2) => Epair (subst_exp sub e1) (subst_exp sub e2)
  | (Einl e) => Einl (subst_exp sub e)
  | (Einr e) => Einr (subst_exp sub e)
  | (Efst e) => Efst (subst_exp sub e)
  | (Esnd e) => Esnd (subst_exp sub e)
  | (Eloc loc) => Eloc loc
  | (Eif e0 e1 e2) => Eif (subst_exp sub e0) (subst_exp sub e1) (subst_exp sub e2)
  | (Ecase x e0 e1 e2) => Ecase x (subst_exp sub e0) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e1) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e2)
  | (Eroll e) => Eroll (subst_exp sub e)
  | (Eunroll e) => Eunroll (subst_exp sub e)
  | (Efix f x e) => Efix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
  | (Eapp e e') => Eapp (subst_exp sub e) (subst_exp sub e')
  | (Egen e) => Egen (subst_exp sub e)
  | (Einst e) => Einst (subst_exp sub e)
  | (Epack e) => Epack (subst_exp sub e)
  | (Eunpack x e1 e2) => Eunpack x (subst_exp sub e1) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e2)
  | (Enew e) => Enew (subst_exp sub e)
  | (Eread e) => Eread (subst_exp sub e)
  | (Ewrite e1 e2) => Ewrite (subst_exp sub e1) (subst_exp sub e2)
  | (Erefeq e1 e2) => Erefeq (subst_exp sub e1) (subst_exp sub e2)
end.

Fixpoint tsubst_prg (sub:list (typename*typ)) (p_5:prg) {struct p_5} : prg :=
  match p_5 with
  | (Pvar x) => Pvar x
  | Punit => Punit 
  | (Pint n) => Pint n
  | Ptrue => Ptrue 
  | Pfalse => Pfalse 
  | (Ppair p1 p2) => Ppair (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pinl ty p) => Pinl (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Pinr ty p) => Pinr (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Pfst p) => Pfst (tsubst_prg sub p)
  | (Psnd p) => Psnd (tsubst_prg sub p)
  | (Pif p0 p1 p2) => Pif (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pcase x p0 p1 p2) => Pcase x (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Proll ty p) => Proll (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Punroll p) => Punroll (tsubst_prg sub p)
  | (Pfix f x ty ty' p) => Pfix f x (tsubst_typ sub ty) (tsubst_typ sub ty') (tsubst_prg sub p)
  | (Papp p p') => Papp (tsubst_prg sub p) (tsubst_prg sub p')
  | (Pgen p) => Pgen (tsubst_prg sub p)
  | (Pinst p ty) => Pinst (tsubst_prg sub p) (tsubst_typ sub ty)
  | (Ppack ty1 p ty) => Ppack (tsubst_typ sub ty1) (tsubst_prg sub p) (tsubst_typ sub ty)
  | (Punpack x p1 p2) => Punpack x (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pnew p) => Pnew (tsubst_prg sub p)
  | (Pread p) => Pread (tsubst_prg sub p)
  | (Pwrite p1 p2) => Pwrite (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Prefeq p1 p2) => Prefeq (tsubst_prg sub p1) (tsubst_prg sub p2)
end.

Fixpoint subst_val (sub:list (termvar*exp)) (v_5:val) {struct v_5} : val :=
  match v_5 with
  | (Vvar x) => Vvar x
  | Vunit => Vunit 
  | (Vint n) => Vint n
  | Vtrue => Vtrue 
  | Vfalse => Vfalse 
  | (Vpair v1 v2) => Vpair (subst_val sub v1) (subst_val sub v2)
  | (Vinl v) => Vinl (subst_val sub v)
  | (Vinr v) => Vinr (subst_val sub v)
  | (Vroll v) => Vroll (subst_val sub v)
  | (Vfix f x e) => Vfix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
  | (Vgen e) => Vgen (subst_exp sub e)
  | (Vpack v) => Vpack (subst_val sub v)
  | (Vloc loc) => Vloc loc
end.

Fixpoint tsubst_ctx (sub:list (typename*typ)) (C5:ctx) {struct C5} : ctx :=
  match C5 with
  | Chole => Chole 
  | (Cpair1 C p2) => Cpair1 (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Cpair2 p1 C) => Cpair2 (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Cinl ty C) => Cinl (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cinr ty C) => Cinr (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cfst C) => Cfst (tsubst_ctx sub C)
  | (Csnd C) => Csnd (tsubst_ctx sub C)
  | (Cif0 C p1 p2) => Cif0 (tsubst_ctx sub C) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Cif1 p0 C p2) => Cif1 (tsubst_prg sub p0) (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Cif2 p0 p1 C) => Cif2 (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Ccase0 x C p1 p2) => Ccase0 x (tsubst_ctx sub C) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Ccase1 x p0 C p2) => Ccase1 x (tsubst_prg sub p0) (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Ccase2 x p0 p1 C) => Ccase2 x (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Croll ty C) => Croll (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cunroll C) => Cunroll (tsubst_ctx sub C)
  | (Cfix f x ty ty' C) => Cfix f x (tsubst_typ sub ty) (tsubst_typ sub ty') (tsubst_ctx sub C)
  | (Capp1 C p) => Capp1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Capp2 p C) => Capp2 (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Cgen C) => Cgen (tsubst_ctx sub C)
  | (Cinst C ty) => Cinst (tsubst_ctx sub C) (tsubst_typ sub ty)
  | (Cpack ty1 C ty) => Cpack (tsubst_typ sub ty1) (tsubst_ctx sub C) (tsubst_typ sub ty)
  | (Cunpack1 x C p) => Cunpack1 x (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Cunpack2 x p C) => Cunpack2 x (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Cnew C) => Cnew (tsubst_ctx sub C)
  | (Cread C) => Cread (tsubst_ctx sub C)
  | (Cwrite1 C p) => Cwrite1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Cwrite2 p C) => Cwrite2 (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Crefeq1 C p) => Crefeq1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Crefeq2 p C) => Crefeq2 (tsubst_prg sub p) (tsubst_ctx sub C)
end.

Definition subst_scont (sub:list (termvar*exp)) (SK5:scont) : scont :=
  match SK5 with
  | (SKif e1 e2) => SKif (subst_exp sub e1) (subst_exp sub e2)
  | (SKpairL e) => SKpairL (subst_exp sub e)
  | (SKpairR v) => SKpairR (subst_val sub v)
  | SKfst => SKfst 
  | SKsnd => SKsnd 
  | SKinl => SKinl 
  | SKinr => SKinr 
  | (SKcase x e1 e2) => SKcase x (subst_exp sub e1) (subst_exp sub e2)
  | SKroll => SKroll 
  | SKunroll => SKunroll 
  | SKpack => SKpack 
  | (SKappL e) => SKappL (subst_exp sub e)
  | (SKappR v) => SKappR (subst_val sub v)
  | SKinst => SKinst 
  | (SKunpack x e) => SKunpack x (subst_exp sub e)
  | SKnew => SKnew 
  | SKread => SKread 
  | (SKwriteL e) => SKwriteL (subst_exp sub e)
  | (SKwriteR v) => SKwriteR (subst_val sub v)
  | (SKrefeqL e) => SKrefeqL (subst_exp sub e)
  | (SKrefeqR v) => SKrefeqR (subst_val sub v)
end.



Definition fv_cont (k: cont) : list termvar :=
  fold_right (fun sk fvs => fv_exp (app_cont (sk::nil) Eunit) ++ fvs) nil k.


(** definitions *)

(* defns OpSemJudg *)
Inductive reduce : heap -> exp -> heap -> exp -> Prop :=    (* defn reduce *)
 | Rtrue : forall (h:heap) (e1 e2:exp),
     reduce h (Eif Etrue e1 e2) h e1
 | Rfalse : forall (h:heap) (e1 e2:exp),
     reduce h (Eif Efalse e1 e2) h e2
 | Rfst : forall (h:heap) (ve1 ve2:voe),
     reduce h (Efst (Epair  (e_of  ve1 )   (e_of  ve2 ) )) h  (e_of  ve1 ) 
 | Rsnd : forall (h:heap) (ve1 ve2:voe),
     reduce h (Esnd (Epair  (e_of  ve1 )   (e_of  ve2 ) )) h  (e_of  ve2 ) 
 | Rinl : forall (h:heap) (x:termvar) (ve:voe) (e1 e2:exp),
     reduce h (Ecase x (Einl  (e_of  ve ) ) e1 e2) h  (subst_exp(( x ,  (e_of  ve )  )::nil) e1 ) 
 | Rinr : forall (h:heap) (x:termvar) (ve:voe) (e1 e2:exp),
     reduce h (Ecase x (Einr  (e_of  ve ) ) e1 e2) h  (subst_exp(( x ,  (e_of  ve )  )::nil) e2 ) 
 | Rbeta : forall (h:heap) (f x:termvar) (e:exp) (ve:voe),
     reduce h (Eapp  ( (Efix f x e) )   (e_of  ve ) ) h  (subst_exp(( x ,  (e_of  ve )  )::( f , (Efix f x e) )::nil) e ) 
 | Rtybeta : forall (h:heap) (e:exp),
     reduce h (Einst  ( (Egen e) ) ) h e
 | Rpack : forall (h:heap) (x:termvar) (ve:voe) (e:exp),
     reduce h (Eunpack x (Epack  (e_of  ve ) ) e) h  (subst_exp(( x ,  (e_of  ve )  )::nil) e ) 
 | Rroll : forall (h:heap) (ve:voe),
     reduce h (Eunroll  ( (Eroll  (e_of  ve ) ) ) ) h  (e_of  ve ) 
 | Rnew : forall (h:heap) (ve:voe) (loc:location)
     (ALLOC:  (allocator  h  =  loc ) ),
     reduce h (Enew  (e_of  ve ) )  (hext  h   loc    (v_of  ve )  )  (Eloc loc)
 | Rread : forall (h:heap) (loc:location) (ve:voe)
     (LOAD:  ( h   loc  = Some   (v_of  ve )  ) ),
     reduce h (Eread (Eloc loc)) h  (e_of  ve ) 
 | Rwrite : forall (h:heap) (loc:location) (ve ve':voe),
     reduce  (hext  h   loc    (v_of  ve )  )  (Ewrite (Eloc loc)  (e_of  ve' ) )  (hext  h   loc    (v_of  ve' )  )  Eunit
 | RrefeqT : forall (h:heap) (loc:location),
     reduce h (Erefeq (Eloc loc) (Eloc loc)) h Etrue
 | RrefeqF : forall (h:heap) (loc loc':location)
     (NEQ:  ( loc  <>  loc' ) ),
     reduce h (Erefeq (Eloc loc) (Eloc loc')) h Efalse
 | Rsctx : forall (h:heap) (SK:scont) (e:exp) (h':heap) (e':exp)
     (RED: reduce h e h' e'),
     reduce h  (appctx_scont_exp  SK   e )  h'  (appctx_scont_exp  SK   e' ) .


Lemma Rctx :
  forall h e h' e' (RED: reduce h e h' e') k,
     reduce h (app_cont k e) h' (app_cont k e').
Proof. intros; revert h e h' e' RED; induction k; intros; eauto using Rsctx. Qed.

Hint Constructors reduce. 
Hint Resolve Rctx.


Coercion exp_of_val : val >-> exp.

Implicit Arguments finite_domain.
Hint Unfold fv_cont.
Arguments fv_cont k: simpl never.

Arguments Vvar x.
Arguments Vunit.
Arguments Vint n.
Arguments Vtrue.
Arguments Vfalse.
Arguments Vloc loc.

Arguments Evar x.
Arguments Eunit.
Arguments Eint n.
Arguments Etrue.
Arguments Efalse.
Arguments Eloc loc.

Arguments SKfst.
Arguments SKsnd.
Arguments SKinl.
Arguments SKinr.
Arguments SKroll.
Arguments SKunroll.
Arguments SKinst.
Arguments SKpack.
Arguments SKnew.
Arguments SKread.

Arguments hemp.

(*----------------------------------------------------------------------------*)
(** ** Canonical structure for values *)
(*----------------------------------------------------------------------------*)

Canonical Structure VoE_var x := @mkVoE (Evar x) (Vvar x) refl_equal.
Canonical Structure VoE_unit := @mkVoE (Eunit) (Vunit) refl_equal.
Canonical Structure VoE_int n := @mkVoE (Eint n) (Vint n) refl_equal.
Canonical Structure VoE_true := @mkVoE (Etrue) (Vtrue) refl_equal.
Canonical Structure VoE_false := @mkVoE (Efalse) (Vfalse) refl_equal.
Program 
Canonical Structure VoE_pair ve1 ve2 := 
  @mkVoE (Epair (e_of ve1) (e_of ve2)) (Vpair (v_of ve1) (v_of ve2)) _.
Next Obligation. destruct ve1, ve2; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_inl ve1 :=
  @mkVoE (Einl (e_of ve1)) (Vinl (v_of ve1)) _.
Next Obligation. destruct ve1; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_inr ve2 :=
  @mkVoE (Einr (e_of ve2)) (Vinr (v_of ve2)) _.
Next Obligation. destruct ve2; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_roll ve :=
  @mkVoE (Eroll (e_of ve)) (Vroll (v_of ve)) _.
Next Obligation. destruct ve; subst; simpl; auto. Qed.
Canonical Structure VoE_fix f x e := @mkVoE (Efix f x e) (Vfix f x e) refl_equal.
Canonical Structure VoE_gen e := @mkVoE (Egen e) (Vgen e) refl_equal.
Program 
Canonical Structure VoE_pack ve :=
  @mkVoE (Epack (e_of ve)) (Vpack (v_of ve)) _.
Next Obligation. destruct ve; subst; simpl; auto. Qed.
Canonical Structure VoE_loc l :=  @mkVoE (Eloc l) (Vloc l) refl_equal.
Canonical Structure VoE_val v :=  @mkVoE (exp_of_val v) v refl_equal.

Coercion VoE_val : val >-> VoE.

(*----------------------------------------------------------------------------*)
(** ** Embedding *)
(*----------------------------------------------------------------------------*)

Fixpoint e2l (e: exp) : exp :=
  match e with
  | (Evar x) => Evar x
  | Eunit => Eunit 
  | (Eint n) => Eint n
  | Etrue => Etrue 
  | Efalse => Efalse 
  | (Epair e1 e2) => Epair (e2l e1) (e2l e2)
  | (Einl e) => Einl (e2l e)
  | (Einr e) => Einr (e2l e)
  | (Efst e) => Efst (e2l e)
  | (Esnd e) => Esnd (e2l e)
  | (Eloc loc) => Eloc loc
  | (Eif e0 e1 e2) => Eif (e2l e0) (e2l e1) (e2l e2)
  | (Ecase x e0 e1 e2) => Ecase x (e2l e0) (e2l e1) (e2l e2)
  | (Eroll e) => Eroll (e2l e)
  | (Eunroll e) => Eunroll (e2l e)
  | (Efix f x e) => Efix f x (e2l e)
  | (Eapp e e') => Eapp (e2l e) (e2l e')
  | (Egen e) => Egen (e2l e)
  | (Einst e) => Einst (e2l e)
  | (Epack e) => Epack (e2l e)
  | (Eunpack x e1 e2) => Eunpack x (e2l e1) (e2l e2)
  | (Enew e) => Enew (e2l e)
  | (Eread e) => Eread (e2l e)
  | (Ewrite e1 e2) => Ewrite (e2l e1) (e2l e2)
  | (Erefeq e1 e2) => Erefeq (e2l e1) (e2l e2)
end.

Fixpoint v2l (v: val) : val :=
  match v with
  | (Vvar x) => Vvar x
  | Vunit => Vunit 
  | (Vint n) => Vint n
  | Vtrue => Vtrue 
  | Vfalse => Vfalse 
  | (Vpair v1 v2) => Vpair (v2l v1) (v2l v2)
  | (Vinl v) => Vinl (v2l v)
  | (Vinr v) => Vinr (v2l v)
  | (Vroll v) => Vroll (v2l v)
  | (Vfix f x e) => Vfix f x (e2l e)
  | (Vgen e) => Vgen (e2l e)
  | (Vpack v) => Vpack (v2l v)
  | (Vloc loc) => Vloc loc
end.

Program Definition h2l (h: heap) : heap :=
  {| heap_f l := option_map v2l (h l) |}.
Next Obligation.
  destruct h; destruct heap_findom0.
  simpl; eexists x; intros. 
  specialize (i x0); destruct (heap_f0 x0); eauto; discriminate H.
Qed.


(** definitions *)

(* defns EnvOKjudg *)
Inductive wfenv : typeenv -> termenv -> Prop :=    (* defn wfenv *)
 | wfenv_ok : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (lc_typ 0  ty )  )   )  )  ->
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (list_minus eq_typename   (ftn_typ  ty )    D  = nil)  )   )  )  ->
     wfenv D G.
(** definitions *)

(* defns TypeOKjudg *)
Inductive wftyp : typeenv -> termenv -> typ -> Prop :=    (* defn wftyp *)
 | wftyp_ok : forall (D:typeenv) (G:termenv) (ty:typ),
     wfenv D G ->
      (lc_typ 0  ty )  ->
      (list_minus eq_typename   (ftn_typ  ty )    D  = nil)  ->
     wftyp D G ty.
(** definitions *)

(* defns TypingJudg *)
Inductive hastype : typeenv -> termenv -> exp -> typ -> Prop :=    (* defn hastype *)
 | HTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     hastype D G (Evar x) ty
 | HTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Eunit Tunit
 | HTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     hastype D G (Eint n) Tint
 | HTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Etrue Tbool
 | HTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Efalse Tbool
 | HTpair : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty1 ty2:typ),
     hastype D G e1 ty1 ->
     hastype D G e2 ty2 ->
     hastype D G (Epair e1 e2) (Tprod ty1 ty2)
 | HTfst : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Efst e) ty1
 | HTsnd : forall (D:typeenv) (G:termenv) (e:exp) (ty2 ty1:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Esnd e) ty2
 | HTinl : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty1 ->
     wftyp D G ty2 ->
     hastype D G (Einl e) (Tsum ty1 ty2)
 | HTinr : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty2 ->
     wftyp D G ty1 ->
     hastype D G (Einr e) (Tsum ty1 ty2)
 | HTif : forall (D:typeenv) (G:termenv) (e0 e1 e2:exp) (ty:typ),
     hastype D G e0 Tbool ->
     hastype D G e1 ty ->
     hastype D G e2 ty ->
     hastype D G (Eif e0 e1 e2) ty
 | HTcase : forall (D:typeenv) (G:termenv) (x:termvar) (e0 e1 e2:exp) (ty ty1 ty2:typ),
     hastype D G e0 (Tsum ty1 ty2) ->
     hastype D  (cons ( x , ty1 )  G )  e1 ty ->
     hastype D  (cons ( x , ty2 )  G )  e2 ty ->
     hastype D G (Ecase x e0 e1 e2) ty
 | HTroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e  (tsubst1 0  (Tmu ty)   ty )  ->
     wftyp D G (Tmu ty) ->
     hastype D G (Eroll e) (Tmu ty)
 | HTunroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tmu ty) ->
     hastype D G (Eunroll e)  (tsubst1 0  (Tmu ty)   ty ) 
 | HTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (e:exp) (ty1 ty2:typ),
     hastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  e ty2 ->
     wftyp D G ty1 ->
     wftyp D G ty2 ->
     hastype D G (Efix f x e) (Tarrow ty1 ty2)
 | HTapp : forall (D:typeenv) (G:termenv) (e e':exp) (ty2 ty1:typ),
     hastype D G e (Tarrow ty1 ty2) ->
     hastype D G e' ty1 ->
     hastype D G (Eapp e e') ty2
 | HTgen : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ) (aa:typename),
     hastype  (cons  aa   D )  G e  (tsubst1 0  (Tname aa)   ty )  ->
     wftyp D G (Tforall ty) ->
      (~ In  aa   D )  ->
     hastype D G (Egen e) (Tforall ty)
 | HTinst : forall (D:typeenv) (G:termenv) (e:exp) (ty' ty:typ),
     hastype D G e (Tforall ty) ->
     wftyp D G ty' ->
     hastype D G (Einst e)  (tsubst1 0  ty'   ty ) 
 | HTpack : forall (D:typeenv) (G:termenv) (e:exp) (ty ty':typ),
     hastype D G e  (tsubst1 0  ty'   ty )  ->
     wftyp D G ty' ->
     hastype D G (Epack e) (Texists ty)
 | HTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (e1 e2:exp) (ty2 ty:typ) (aa:typename),
     hastype D G e1 (Texists ty) ->
     hastype  (cons  aa   D )   (cons ( x ,  (tsubst1 0  (Tname aa)   ty )  )  G )  e2 ty2 ->
     wftyp D G ty2 ->
      (~ In  aa   D )  ->
     hastype D G (Eunpack x e1 e2) ty2
 | HTnew : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e ty ->
     hastype D G (Enew e) (Tref ty)
 | HTread : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tref ty) ->
     hastype D G (Eread e) ty
 | HTwrite : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 ty ->
     hastype D G (Ewrite e1 e2) Tunit
 | HTrefeq : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 (Tref ty) ->
     hastype D G (Erefeq e1 e2) Tbool.
(** definitions *)

(* defns ProgTypingJudg *)
Inductive phastype : typeenv -> termenv -> prg -> typ -> Prop :=    (* defn phastype *)
 | PHTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     phastype D G (Pvar x) ty
 | PHTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Punit Tunit
 | PHTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     phastype D G (Pint n) Tint
 | PHTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Ptrue Tbool
 | PHTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Pfalse Tbool
 | PHTpair : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty1 ty2:typ),
     phastype D G p1 ty1 ->
     phastype D G p2 ty2 ->
     phastype D G (Ppair p1 p2) (Tprod ty1 ty2)
 | PHTfst : forall (D:typeenv) (G:termenv) (p:prg) (ty1 ty2:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Pfst p) ty1
 | PHTsnd : forall (D:typeenv) (G:termenv) (p:prg) (ty2 ty1:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Psnd p) ty2
 | PHTinl : forall (D:typeenv) (G:termenv) (ty2:typ) (p:prg) (ty1:typ),
     phastype D G p ty1 ->
     wftyp D G ty2 ->
     phastype D G (Pinl ty2 p) (Tsum ty1 ty2)
 | PHTinr : forall (D:typeenv) (G:termenv) (ty1:typ) (p:prg) (ty2:typ),
     phastype D G p ty2 ->
     wftyp D G ty1 ->
     phastype D G (Pinr ty1 p) (Tsum ty1 ty2)
 | PHTif : forall (D:typeenv) (G:termenv) (p0 p1 p2:prg) (ty:typ),
     phastype D G p0 Tbool ->
     phastype D G p1 ty ->
     phastype D G p2 ty ->
     phastype D G (Pif p0 p1 p2) ty
 | PHTcase : forall (D:typeenv) (G:termenv) (x:termvar) (p0 p1 p2:prg) (ty ty1 ty2:typ),
     phastype D G p0 (Tsum ty1 ty2) ->
     phastype D  (cons ( x , ty1 )  G )  p1 ty ->
     phastype D  (cons ( x , ty2 )  G )  p2 ty ->
     phastype D G (Pcase x p0 p1 p2) ty
 | PHTroll : forall (D:typeenv) (G:termenv) (ty:typ) (p:prg),
     phastype D G p  (tsubst1 0  (Tmu ty)   ty )  ->
     wftyp D G (Tmu ty) ->
     phastype D G (Proll ty p) (Tmu ty)
 | PHTunroll : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tmu ty) ->
     phastype D G (Punroll p)  (tsubst1 0  (Tmu ty)   ty ) 
 | PHTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (ty1 ty2:typ) (p:prg),
     phastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  p ty2 ->
     wftyp D G ty1 ->
     wftyp D G ty2 ->
     phastype D G (Pfix f x ty1 ty2 p) (Tarrow ty1 ty2)
 | PHTapp : forall (D:typeenv) (G:termenv) (p p':prg) (ty2 ty1:typ),
     phastype D G p (Tarrow ty1 ty2) ->
     phastype D G p' ty1 ->
     phastype D G (Papp p p') ty2
 | PHTgen : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ) (aa:typename),
     phastype  (cons  aa   D )  G p  (tsubst1 0  (Tname aa)   ty )  ->
     wftyp D G (Tforall ty) ->
      (~ In  aa   D )  ->
     phastype D G (Pgen p) (Tforall ty)
 | PHTinst : forall (D:typeenv) (G:termenv) (p:prg) (ty' ty:typ),
     phastype D G p (Tforall ty) ->
     wftyp D G ty' ->
     phastype D G (Pinst p ty')  (tsubst1 0  ty'   ty ) 
 | PHTpack : forall (D:typeenv) (G:termenv) (ty':typ) (p:prg) (ty:typ),
     phastype D G p  (tsubst1 0  ty'   ty )  ->
     wftyp D G ty' ->
     phastype D G (Ppack ty' p ty) (Texists ty)
 | PHTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (p1 p2:prg) (ty2 ty:typ) (aa:typename),
     phastype D G p1 (Texists ty) ->
     phastype  (cons  aa   D )   (cons ( x ,  (tsubst1 0  (Tname aa)   ty )  )  G )  p2 ty2 ->
     wftyp D G ty2 ->
      (~ In  aa   D )  ->
     phastype D G (Punpack x p1 p2) ty2
 | PHTnew : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p ty ->
     phastype D G (Pnew p) (Tref ty)
 | PHTread : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tref ty) ->
     phastype D G (Pread p) ty
 | PHTwrite : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 ty ->
     phastype D G (Pwrite p1 p2) Tunit
 | PHTrefeq : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 (Tref ty) ->
     phastype D G (Prefeq p1 p2) Tbool.
(** definitions *)

(* defns CtxJudg *)
Inductive wfctx : ctx -> typeenv -> termenv -> typ -> typeenv -> termenv -> typ -> Prop :=    (* defn wfctx *)
 | wfctx_hole : forall (D:typeenv) (G:termenv) (ty:typ),
     wftyp D G ty ->
     wfctx Chole D G ty D G ty
 | wfctx_pair1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty2:typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty2 ->
     wfctx (Cpair1 C p) D G ty D' G' (Tprod ty' ty2)
 | wfctx_pair2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty1 ->
     wfctx (Cpair2 p C) D G ty D' G' (Tprod ty1 ty')
 | wfctx_inl : forall (ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' G' ty2 ->
     wfctx (Cinl ty2 C) D G ty D' G' (Tsum ty' ty2)
 | wfctx_inr : forall (ty1:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' G' ty1 ->
     wfctx (Cinr ty1 C) D G ty D' G' (Tsum ty1 ty')
 | wfctx_fst : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty2:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Cfst C) D G ty D' G' ty1
 | wfctx_snd : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Csnd C) D G ty D' G' ty2
 | wfctx_if0 : forall (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' Tbool ->
     phastype D' G' p1 ty' ->
     phastype D' G' p2 ty' ->
     wfctx (Cif0 C p1 p2) D G ty D' G' ty'
 | wfctx_if1 : forall (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p2 ty' ->
     wfctx (Cif1 p0 C p2) D G ty D' G' ty'
 | wfctx_if2 : forall (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p1 ty' ->
     wfctx (Cif2 p0 p1 C) D G ty D' G' ty'
 | wfctx_case0 : forall (x:termvar) (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     wfctx C D G ty D' G' (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase0 x C p1 p2) D G ty D' G' ty'
 | wfctx_case1 : forall (x:termvar) (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     wfctx C D G ty D'  (cons ( x , ty1 )  G' )  ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase1 x p0 C p2) D G ty D' G' ty'
 | wfctx_case2 : forall (x:termvar) (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     wfctx C D G ty D'  (cons ( x , ty2 )  G' )  ty' ->
     wfctx (Ccase2 x p0 p1 C) D G ty D' G' ty'
 | wfctx_roll : forall (ty':typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wftyp D' G' (Tmu ty') ->
     wfctx C D G ty D' G'  (tsubst1 0  (Tmu ty')   ty' )  ->
     wfctx (Croll ty' C) D G ty D' G' (Tmu ty')
 | wfctx_unroll : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tmu ty') ->
     wfctx (Cunroll C) D G ty D' G'  (tsubst1 0  (Tmu ty')   ty' ) 
 | wfctx_fix : forall (f x:termvar) (ty1 ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D'  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G' )  )  ty2 ->
     wftyp D' G' ty1 ->
     wftyp D' G' ty2 ->
     wfctx (Cfix f x ty1 ty2 C) D G ty D' G' (Tarrow ty1 ty2)
 | wfctx_app1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tarrow ty1 ty2) ->
     phastype D' G' p ty1 ->
     wfctx (Capp1 C p) D G ty D' G' ty2
 | wfctx_app2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     phastype D' G' p (Tarrow ty1 ty2) ->
     wfctx C D G ty D' G' ty1 ->
     wfctx (Capp2 p C) D G ty D' G' ty2
 | wfctx_gen : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ) (aa:typename),
     wfctx C D G ty  (cons  aa   D' )  G'  (tsubst1 0  (Tname aa)   ty' )  ->
     wftyp D' G' (Tforall ty') ->
      (~ In  aa   D' )  ->
     wfctx (Cgen C) D G ty D' G' (Tforall ty')
 | wfctx_inst : forall (C:ctx) (ty1:typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tforall ty') ->
     wftyp D' G' ty1 ->
     wfctx (Cinst C ty1) D G ty D' G'  (tsubst1 0  ty1   ty' ) 
 | wfctx_pack : forall (ty1:typ) (C:ctx) (ty':typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D' G'  (tsubst1 0  ty1   ty' )  ->
     wftyp D' G' ty1 ->
     wfctx (Cpack ty1 C ty') D G ty D' G' (Texists ty')
 | wfctx_unpack1 : forall (x:termvar) (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     wfctx C D G ty D' G' (Texists ty') ->
     phastype  (cons  aa   D' )   (cons ( x ,  (tsubst1 0  (Tname aa)   ty' )  )  G' )  p ty2 ->
     wftyp D' G' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack1 x C p) D G ty D' G' ty2
 | wfctx_unpack2 : forall (x:termvar) (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     phastype D' G' p (Texists ty') ->
     wfctx C D G ty  (cons  aa   D' )   (cons ( x ,  (tsubst1 0  (Tname aa)   ty' )  )  G' )  ty2 ->
     wftyp D' G' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack2 x p C) D G ty D' G' ty2
 | wfctx_new : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wfctx (Cnew C) D G ty D' G' (Tref ty')
 | wfctx_read : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Cread C) D G ty D' G' ty'
 | wfctx_write1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p ty' ->
     wfctx (Cwrite1 C p) D G ty D' G' Tunit
 | wfctx_write2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' ty' ->
     wfctx (Cwrite2 p C) D G ty D' G' Tunit
 | wfctx_refeq1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p (Tref ty') ->
     wfctx (Crefeq1 C p) D G ty D' G' Tbool
 | wfctx_refeq2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Crefeq2 p C) D G ty D' G' Tbool.

Fixpoint erase_prg (p : prg) : exp :=
  match p with
  | (Pvar x) => Evar x
  | Punit => Eunit
  | (Pint n) => Eint n
  | Ptrue => Etrue
  | Pfalse => Efalse
  | (Ppair p1 p2) => (Epair  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pinl ty p) => (Einl  (erase_prg  p ) )
  | (Pinr ty p) => (Einr  (erase_prg  p ) )
  | (Pfst p) => (Efst  (erase_prg  p ) )
  | (Psnd p) => (Esnd  (erase_prg  p ) )
  | (Pif p0 p1 p2) => (Eif  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pcase x p0 p1 p2) => (Ecase x  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Proll ty p) => (Eroll  (erase_prg  p ) )
  | (Punroll p) => (Eunroll  (erase_prg  p ) )
  | (Pfix f x ty ty' p) => (Efix f x  (erase_prg  p ) )
  | (Papp p p') => (Eapp  (erase_prg  p )   (erase_prg  p' ) )
  | (Pgen p) => (Egen  (erase_prg  p ) )
  | (Pinst p ty) => (Einst  (erase_prg  p ) )
  | (Ppack ty1 p ty) => (Epack  (erase_prg  p ) )
  | (Punpack x p1 p2) => (Eunpack x  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pnew p) => (Enew  (erase_prg  p ) )
  | (Pread p) => (Eread  (erase_prg  p ) )
  | (Pwrite p1 p2) => (Ewrite  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Prefeq p1 p2) => (Erefeq  (erase_prg  p1 )   (erase_prg  p2 ) )
  end.

Hint Constructors hastype : hastype.
Hint Constructors phastype : hastype.

Coercion erase_prg : prg >-> exp.

Lemma hastype_erase :
  forall D G p ty, phastype D G p ty -> hastype D G (erase_prg p) ty.
Proof. induction 1; simpl in *; eauto with hastype. Qed.

Lemma phastype_app_ctx :
  forall D G C ty D' G' ty', 
    wfctx C D G ty D' G' ty' -> forall p,
    phastype D G p ty -> 
    phastype D' G' (app_ctx C p) ty'.
Proof. induction 1; intros; simpl in *; eauto with hastype. Qed.




