Require Import vgtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.

Open Scope string_scope.
Open Scope list_scope.

Set Implicit Arguments.

Local Notation cid := constructive_indefinite_description.

Section trans_part1. (*********************************************************************)

Variables
  (pD : list typename)
  (pG : list (termvar * typ))
  (pty : typ)
  (names: typename -> Prop) 
  (Nci : countably_infinite names).

Definition ftn_G_ty :=
  ftn_typ pty ++ fold_right (@app _) [] (map (fun x => ftn_typ (snd x)) pG).

Definition names1 : typename -> Prop := 
  proj1_sig (cid _ (countably_infinite_split Nci)).

Definition names1r : typename -> Prop := 
  proj1_sig (cid _ 
    (proj2_sig (cid _ (countably_infinite_split Nci)))).

Lemma names_split:
  <<CI1: countably_infinite names1>> /\ <<CI1r: countably_infinite names1r>> /\
  <<UN1r: names = nunion names1 names1r>> /\ <<DJ1r: ndisj names1 names1r>>.
Proof.
  apply (proj2_sig (cid _
    (proj2_sig (cid _ (countably_infinite_split Nci))))).
Qed.

Definition names2 : typename -> Prop :=
  proj1_sig (cid _ 
    (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty)).

Definition namesnew : typename -> Prop :=
  proj1_sig (cid _ 
    (proj2_sig (cid _ 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty)))).

Lemma names1r_split : 
  <<CI2: countably_infinite names2>> /\ <<CInew: countably_infinite namesnew>> /\
  <<UN2new: names1r = nunion names2 namesnew>> /\ <<DJ2new: ndisj names2 namesnew>> /\
  <<FRESH: forall a (INL: In a ftn_G_ty) (INN: namesnew a), False>>.
Proof.
  apply (proj2_sig (cid _
    (proj2_sig (cid _ 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty))))).
Qed.

Variables (lpw1 lpw2: world_local).

Definition pw1 := wlift lpw1.
Definition pw2 := wlift lpw2.

Variables (Nam1 : wnam pw1 <1= names1) (Nam2 : wnam pw2 <1= names2)
  (Wih1 : winhabited pw1) (Wih2 : winhabited pw2)
  (Wcs1 : wconsistent pw1) (Wcs2 : wconsistent pw2).

Fixpoint decomp proj ty : typ :=
  match ty with
    | Tname a => match LEM (namesnew a) with left pf => proj a | _ => Tname a end
    | Tbvar x => Tbvar x
    | Tunit => Tunit
    | Tint => Tint
    | Tbool => Tbool
    | Tprod ty1 ty2 => Tprod (decomp proj ty1) (decomp proj ty2)
    | Tsum ty1 ty2 => Tsum (decomp proj ty1) (decomp proj ty2)
    | Tarrow ty1 ty2 => Tarrow (decomp proj ty1) (decomp proj ty2)
    | Tforall ty1 => Tforall (decomp proj ty1)
    | Texists ty1 => Texists (decomp proj ty1)
    | Tmu ty1 => Tmu (decomp proj ty1)
    | Tref ty1 => Tref (decomp proj ty1)
  end.

Variables 
  (nfst nsnd : typename -> typ)
  (nfst_closed : forall a, lc_typ 0 (nfst a))
  (nsnd_closed : forall a, lc_typ 0 (nsnd a))
  (decomp_inj : forall ty1 ty2
                (EQ1 : decomp nfst ty1 = decomp nfst ty2)
                (EQ2 : decomp nsnd ty1 = decomp nsnd ty2),
              ty1 = ty2)
  (decomp_surj : forall ty1 ty2 (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2),
               exists ty, decomp nfst ty = ty1 /\ decomp nsnd ty = ty2).

Lemma Nci1 : countably_infinite names1.
Proof. apply names_split. Qed.

Lemma Nci2 : countably_infinite names2.
Proof. apply names1r_split. Qed.

Lemma Ncinew : countably_infinite namesnew.
Proof. apply names1r_split. Qed.

Lemma Nunion : forall a, names a <-> names1 a \/ names2 a \/ namesnew a.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by rewrite UN1r, UN2new; split; intros.
Qed.

Lemma Ndisj12 : ndisj names1 names2.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; left.
Qed.

Lemma Ndisj1new : ndisj names1 namesnew.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; right.
Qed.

Lemma Ndisj2new : ndisj names2 namesnew.
Proof.
  apply names1r_split.
Qed.

Lemma NdisjTy : forall a (IN : In a (ftn_typ pty)), namesnew a -> False.
Proof.
  intros; eapply names1r_split; eauto; apply in_or_app; auto.
Qed.

Lemma NdisjGa : forall p (IN' : In p pG) a (IN : In a (ftn_typ (snd p))), namesnew a -> False.
Proof.
  intros; eapply names1r_split; eauto; apply in_or_app; right.
  clear -IN IN' H; induction pG; [by contradiction|].
  by simpls; apply in_or_app; des; subst; eauto.
Qed.

Lemma eenc_definition:
  exists EEnc : typ -> nat -> val -> nat -> val -> nat,
    forall ty ty' n1 v1 n2 v2 n1' v1' n2' v2' 
      (EQ : EEnc ty n1 v1 n2 v2 = EEnc ty' n1' v1' n2' v2'),
    ty = ty' /\ n1 = n1' /\ v1 = v1' /\ n2 = n2' /\ v2 = v2'.
Proof.
  assert (X := cntinf_prod typ_cntinf (cntinf_prod (cntinf_prod location_cntinf val_cntinf) (cntinf_prod location_cntinf val_cntinf))).
  eapply countably_infinite_set_impl_set2 in X; red in X; des.
  exists (fun ty n1 v1 n2 v2 => f (ty,((n1,v1),(n2,v2)))).
  by intros; eapply X in EQ; inv EQ.
Qed.

Definition EEnc := proj1_sig (cid _ eenc_definition).

Lemma EEnc_inj :
  forall ty ty' n1 v1 n2 v2 n1' v1' n2' v2' (EQ : EEnc ty n1 v1 n2 v2 = EEnc ty' n1' v1' n2' v2'),
    ty = ty' /\ n1 = n1' /\ v1 = v1' /\ n2 = n2' /\ v2 = v2'.
Proof. apply (proj2_sig (cid _ eenc_definition)). Qed.

Lemma decomp_fst_tsubst1 : forall ty ty' n, 
  tsubst1 n (decomp nfst ty') (decomp nfst ty) = decomp nfst (tsubst1 n ty' ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N']. 
      by apply tsubst1_id; eapply lc_mon with (n:=0); [omega|apply nfst_closed].
    by apply tsubst1_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_snd_tsubst1 : forall ty ty' n, 
  tsubst1 n (decomp nsnd ty') (decomp nsnd ty) = decomp nsnd (tsubst1 n ty' ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N'].
      by apply tsubst1_id; eapply lc_mon with (n:=0); [omega|apply nsnd_closed].
    by apply tsubst1_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_lc_typ : forall ty n,
  lc_typ n (decomp nfst ty) \/ lc_typ n (decomp nsnd ty) -> lc_typ n ty.
Proof. by induction ty; simpl; intros; auto; try destruct H; des; auto. Qed.

Lemma decomp_lc_typ' : forall ty n,
  lc_typ n ty -> lc_typ n (decomp nfst ty) /\ lc_typ n (decomp nsnd ty).
Proof.
  induction ty; simpls; intros; des; hdes; eauto 10.

  case (LEM (namesnew aa)); intros; eauto.
  by split; (eapply lc_mon with (n := 0); [omega|eauto using nfst_closed, nsnd_closed]).
Qed.

Lemma decomp_tsubst_ty : forall proj d (dOK: forall x (IN: In x d), lc_typ 0 (snd x)),
  decomp proj (tsubst_typ d pty) = tsubst_typ (map_pair id (decomp proj) d) pty.
Proof.
  assert (NdisjTy := NdisjTy). (* otherwise generated IH is too weak *)
  induction pty; intros; simpls.
  - induction d; simpl.
      destruct (LEM (namesnew aa)) as [NN|NN]; auto.
      by exfalso; eapply NdisjTy; [left; reflexivity|].
    destruct a; simpl; unfold id; case (eq_typename t aa); eauto.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
Qed.

Lemma decomp_id : forall ty (tyN : forall a, In a (ftn_typ ty) -> namesnew a -> False) proj,
  decomp proj ty = ty.
Proof.
  induction ty; intros; simpl; auto.
  - destruct (LEM (namesnew aa)) as [N|N]; auto.
    by exfalso; eapply tyN; eauto; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
Qed.

Lemma decomp_tsubst : forall d (dOK: forall x (IN: In x d), lc_typ 0 (snd x))
    ty (tyN : forall a, In a (ftn_typ ty) -> namesnew a -> False) proj,
  decomp proj (tsubst_typ d ty) = tsubst_typ (map_pair id (decomp proj) d) ty.
Proof.
  induction ty; intros; simpls.
  - induction d; simpl.
      destruct (LEM (namesnew aa)); auto.
      by exfalso; eauto; eapply tyN; left; reflexivity.
    destruct a; simpl; unfold id; case (eq_typename t aa); auto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
Qed.

Inductive vrelf_fst (r: vrelf) : vrelf :=
 | vrelf_fst_intro : forall (ty ty' : flextyp)
           (E : decomp nfst ty = ty')
           (NN : forall a, names a -> ty <> FTname a)
           (NR : forall ty', ty <> FTref ty')
           n1 v1 n3 v3 (IN : r ty n1 v1 n3 v3) n1' (LEN: n1 <= n1') n2',
         vrelf_fst r ty' n1' v1 n2' (Vint (EEnc ty n1 v1 n3 v3))
.
Hint Constructors vrelf_fst.

Inductive vrelf_snd (r: vrelf) : vrelf :=
 | vrelf_snd_intro : forall (ty ty': flextyp)
           (E : decomp nsnd ty = ty')
           (NN : forall a, names a -> ty <> FTname a)
           (NR : forall ty', ty <> FTref ty')
           n1 v1 n3 v3 (IN : r ty n1 v1 n3 v3) n3' (LEN: n3 <= n3') n2',
         vrelf_snd r ty' n2' (Vint (EEnc ty n1 v1 n3 v3)) n3' v3
.
Hint Constructors vrelf_snd.

Lemma vrelf_fst_mon : forall r r' ty n1 n1' v1 n2 n2' v2 
    (REL: vrelf_fst r ty n1 v1 n2 v2) (LE: r <5= r') (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
  vrelf_fst r' ty n1' v1 n2' v2.
Proof. intros; destruct REL; econs; eauto; omega. Qed.

Lemma vrelf_snd_mon : forall r r' ty n1 n1' v1 n2 n2' v2 
    (REL: vrelf_snd r ty n1 v1 n2 v2) (LE: r <5= r') (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
  vrelf_snd r' ty n1' v1 n2' v2.
Proof. intros; destruct REL; econs; eauto; omega. Qed.

Lemma vrelf_fst_closed : forall r (CL : vrelf_closed r), vrelf_closed (vrelf_fst r).
Proof.
  red; intros; rhdes CL; unnw; hdes; destruct V; eauto 10.
Qed.

Lemma vrelf_snd_closed : forall r (CL : vrelf_closed r), vrelf_closed (vrelf_snd r).
Proof.
  red; intros; rhdes CL; unnw; hdes; destruct V; eauto 10.
Qed.

Definition comp2 A (r1 r2 : relation A) : relation A :=
  fun v1 v3 => exists v2, r1 v1 v2 /\ r2 v2 v3.

Inductive bullet A (r1 r2 : typ -> nat -> A -> nat -> A -> Prop) ty n1 v1 n3 v3 : Prop :=
_bullet n2 n2' v2
  (R1: r1 (decomp nfst ty) n1 v1 n2  v2)
  (R2: r2 (decomp nsnd ty) n2' v2 n3 v3)
.
Hint Constructors bullet.

(* next: construction of w *)

Inductive pw1_vrel (r : vrelf) (s1 : wst pw1) (ty: flextyp) n1 (v1: val) n2 (v2: val) : Prop := 
| w1_vrel_base (REL: vrelf_fst r ty n1 v1 n2 v2)
| w1_vrel_step r1 (REL: wval pw1 () s1 r1 ty n1 v1 n2 v2) (MON: r1 <5= pw1_vrel r s1).
Hint Constructors pw1_vrel.

Inductive pw2_vrel (r : vrelf) (s2 : wst pw2) (ty: flextyp) n1 (v1: val) n2 (v2: val) : Prop := 
| w2_vrel_base (REL: vrelf_snd r ty n1 v1 n2 v2)
| w2_vrel_step r2 (REL: wval pw2 () s2 r2 ty n1 v1 n2 v2) (MON: r2 <5= pw2_vrel r s2).
Hint Constructors pw2_vrel.

Inductive pw_val (s : wst pw1 * wst pw2) (r : vrelf) ty n1 v1 n2 v2 : Prop :=
_pw_val  n1' n2'
  (NAM: forall a (NAME: ty = FTname a), namesnew a)
  (R: bullet (vclos (pw1_vrel r (fst s))) (vclos (pw2_vrel r (snd s))) ty n1' v1 n2' v2)
  (NEQ1: n1 >= S n1') (NEQ1: n2 >= S n2').
Hint Constructors pw_val.

Definition pw_heap (s : wst pw1 * wst pw2) (r : vrelf) : hrel :=
  comp2 (wheap pw1 () (fst s) (pw1_vrel r (fst s))) (wheap pw2 () (snd s) (pw2_vrel r (snd s))).

Definition prodrel A B (r1: relation A) (r2: relation B) : relation (A*B) :=
  fun ab ab' => r1 (fst ab) (fst ab') /\ r2 (snd ab) (snd ab').

Lemma bullet_mon : forall A r1 r1' r2 r2' ty n1 n1' v1 n2 n2' v2
    (BUL: @bullet A r1 r2 ty n1 v1 n2 v2)
    (LE1: forall n1 n1' v1 n2 n2' v2 
            (REL: r1 (decomp nfst ty) n1 v1 n2 v2:Prop) (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
          r1' (decomp nfst ty) n1' v1 n2' v2:Prop) 
    (LE2: forall n1 n1' v1 n2 n2' v2
            (REL: r2 (decomp nsnd ty) n1 v1 n2 v2:Prop) (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
          r2' (decomp nsnd ty) n1' v1 n2' v2:Prop)
    (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
  bullet r1' r2' ty n1' v1 n2' v2.
Proof. intros; destruct BUL; eauto. Qed.

Lemma pw1_vrel_mon : forall r r' s s' ty n1 v1 n2 v2 
    (REL: pw1_vrel r s ty n1 v1 n2 v2) 
    (LE: r <5= r') (EXT : wext _ s' s) n1' (LEN1: n1 <= n1') n2' (LEN2: n2 <= n2'),
  pw1_vrel r' s' ty n1' v1 n2' v2.
Proof.
  intros until v2; induction 1; eauto using wval_mon, vrelf_fst_mon.
Qed.

Lemma pw2_vrel_mon : forall r r' s s' ty n1 v1 n2 v2 
    (REL: pw2_vrel r s ty n1 v1 n2 v2) 
    (LE: r <5= r') (EXT : wext _ s' s) n1' (LEN1: n1 <= n1') n2' (LEN2: n2 <= n2'),
  pw2_vrel r' s' ty n1' v1 n2' v2.
Proof.
  intros until v2; induction 1; eauto using wval_mon, vrelf_snd_mon.
Qed.

Lemma pw_val_mon: forall r r' s s' ty n1 n1' v1 n2 n2' v2 
    (REL: pw_val s r ty n1 v1 n2 v2) 
    (LE : r <5= r') (EXT : prodrel (wext pw1) (wext pw2) s' s)
    (LEN1: n1' >= n1) (LEN2: n2' >= n2),
  pw_val s' r' ty n1' v1 n2' v2.
Proof.
  i; rdes EXT; destruct REL.
  econs; eauto; [eapply bullet_mon
  ; eauto using vclos_mon, vclos_mon_lv, pw1_vrel_mon, pw2_vrel_mon with arith| |]
  ; omega.
Qed.

Lemma pw1_vrel_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw1_vrel r s).
Proof.
  repeat intro; induction V; [eapply vrelf_fst_closed | eapply wval_closed]; eauto.
Qed.

Lemma pw2_vrel_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw2_vrel r s).
Proof.
  repeat intro; induction V; [eapply vrelf_snd_closed | eapply wval_closed]; eauto.
Qed.

Lemma pw_val_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw_val s r).
Proof.
  ii; destruct V; subst.
  destruct R; split; [eapply vclos_closed in R1|eapply vclos_closed in R2]; des
  ; eauto using pw1_vrel_closed, pw2_vrel_closed.
Qed.

Lemma pw_heap_mon : forall s r r' h1 h2 (REL: pw_heap s r h1 h2) (LE: r <5= r'),
  pw_heap s r' h1 h2.
Proof.
  intros; rrdes REL; repeat red.
  eauto 10 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon. 
Qed.

Lemma pw_val_wf : forall s r ty n1 v1 n2 v2,
  pw_val s r ty n1 v1 n2 v2 -> n1 > 0 /\ n2 > 0.
Proof. i; destruct H; subst; omega. Qed.

Definition pw : world.
  refine 
    {| wnam := names
     ; wst := wst pw1 * wst pw2
     ; wext := prodrel (wext pw1) (wext pw2)
     ; wextpub := prodrel (wextpub pw1) (wextpub pw2)
     ; wval := fun _ => pw_val
     ; wheap := fun _ => pw_heap
     |};
  unfold prodrel in *; try by intros; des; unnw;
   eauto using wext_trans, wextpub_trans, pw_val_wf.
  - by intros; eapply pw_val_mon; eauto.
  - by intros; eapply pw_val_closed; eauto.
  - by intros; eapply pw_heap_mon; eauto.
  - intros; revert WV; try destruct (LEM (namesnew a)); intros. 
    + by apply Nunion; eauto.
    + destruct WV; exfalso; eauto.
Defined.

Definition grel_fst g (s2 : wst pw2) (s1 : wst pw1) := pw1_vrel (g (s1,s2)) s1.

Definition grel_snd g (s1 : wst pw1) (s2 : wst pw2) := pw2_vrel (g (s1,s2)) s2.

Hint Unfold grel_fst grel_snd.

Lemma pw_ext: forall s1 s2 s1' s2' (EXT1: wext _ s1 s1') (EXT2: wext _ s2 s2'),
  wext pw (s1,s2) (s1',s2').
Proof. split; eauto. Qed.

Lemma grel_fst_closed : forall g (G : wglob pw g) s2 s1, vrelf_closed (grel_fst g s2 s1).
Proof.
  red; intros; rdes G; induction V; [eapply vrelf_fst_closed | eapply wval_closed]; eauto.
Qed.

Lemma grel_snd_closed : forall g (G : wglob pw g) s1 s2, vrelf_closed (grel_snd g s1 s2).
Proof.
  red; intros; rdes G; induction V; [eapply vrelf_snd_closed | eapply wval_closed]; eauto.
Qed.

Lemma grel_fst_lerefn : forall g (G : wglob pw g) s2 s1,
  lerefn (wnam pw1) (wval pw1 () s1 (grel_fst g s2 s1)) (grel_fst g s2 s1).
Proof.
  red; split; try split; unnw; intros; eauto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty0; try (by inversion E); [by edestruct NR|].
    simpls; revert E; destruct (LEM (namesnew a)); intro.
      by edestruct NN; auto; apply Nunion; auto.
    eapply G in IN.
      by apply wnam_supports in IN.
    by edestruct NN; auto; apply Nunion; auto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty; simpls.
    destruct (LEM (namesnew a0)).
      by edestruct NN; auto; apply Nunion; auto. 
    inversion E; subst.
    by apply Nam1 in NM; edestruct NN; auto; apply Nunion; auto. 
Qed.

Lemma grel_snd_lerefn : forall g (G : wglob pw g) s1 s2,
  lerefn (wnam pw2) (wval pw2 () s2 (grel_snd g s1 s2)) (grel_snd g s1 s2).
Proof.
  red; split; try split; unnw; intros; eauto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty0; try (by inversion E); [by edestruct NR|].
    simpls; revert E; destruct (LEM (namesnew a)) as [N|N] _eqn : E'; intro.
      by edestruct NN; auto; apply Nunion; auto.
    eapply G in IN.
      by apply wnam_supports in IN.
    by edestruct NN; auto; apply Nunion; auto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty; simpls.
    destruct (LEM (namesnew a0)).
      by edestruct NN; auto; apply Nunion; auto. 
    inversion E; subst.
    by apply Nam2 in NM; edestruct NN; auto; apply Nunion; auto. 
Qed.

Lemma grel_fst_wglob : forall g (G : wglob pw g) s2, wglob pw1 (grel_fst g s2).
Proof.
  by intros; red; assert (G':=G); rdes G'; split; [| split]; intros; try destruct sp;
     eauto 10 using grel_fst_closed, grel_fst_lerefn, pw1_vrel_mon, pw_ext.
Qed.

Lemma grel_snd_wglob : forall g (G : wglob pw g) s1, wglob pw2 (grel_snd g s1).
Proof.
  by intros; red; assert (G':=G); rdes G'; split; [| split]; intros; try destruct sp;
     eauto 10 using grel_snd_closed, grel_snd_lerefn, pw2_vrel_mon, pw_ext.
Qed.

Lemma bullet_grel_flex1 : forall g (G : wglob pw g) (ty : flextyp) s1 s2 n1 v1 n2 v2
    (IN: bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty n1 v1 n2 v2),
  vclos (g (s1, s2)) ty (S n1) v1 (S n2) v2.
Proof.
  intros; depdes ty; [M|M|M|]; Mdo by
    destruct IN; simpl in R1, R2; destruct R1; inv REL; destruct R2; inv REL0; clear REL REL0;
    (esplits; eauto; do 2 econs; apply G; econs; eauto; [done|]);
    (econs; eauto; do 2 econs; eauto).

  destruct (LEM (namesnew a)) as [N'|N'] _eqn: E'.
    esplits; try reflexivity.
    by do 2 econs; eapply G; econs; eauto; i; depdes NAME.
  destruct (LEM (names1 a)).
    exfalso; destruct IN; simpls; rewrite E' in R1, R2.
    destruct R1; inv REL; destruct R2; inv REL0; clear REL REL0; destruct VR0. 
    2: by apply wnam_supports in REL; simpls; apply Nam2 in REL; eapply Ndisj12; eauto.
    inv REL; destruct ty; try inversion E; revert H0; destruct (LEM (namesnew a0)) as [N0|N0] _eqn : E0.
      intro; eapply G in IN.
        by apply wnam_supports in IN; simpls; edestruct NN; eauto.
      by apply Nunion; auto.
    intro; inversion H0; subst a0; eapply G in IN; simpl in IN.
      by eapply NN; eauto; apply Nunion; eauto.
    by apply Nunion; auto.
  destruct (LEM (names2 a)).
    exfalso; destruct IN; simpls; rewrite E' in R1, R2.
    destruct R1; inv REL; destruct R2; inv REL0; clear REL REL0; destruct VR.
    2: by apply wnam_supports in REL; simpls; apply Nam1 in REL; eapply Ndisj12; eauto.
    inv REL; destruct ty; try inversion E; revert H0; destruct (LEM (namesnew a0)) as [N0|N0] _eqn : E0.
      intro; eapply G in IN.
        by apply wnam_supports in IN; simpls; edestruct NN; eauto.
      by apply Nunion; auto.
    intro; inversion H0; subst a0; eapply G in IN; simpl in IN.
      by eapply NN; eauto; apply Nunion; eauto.
    by apply Nunion; auto.
  destruct (LEM (names a)) as [N|N] _eqn : E.
    by edestruct Nunion; specialize (H N); destruct H as [H|[H|H]]; contradiction.
  destruct IN; simpls; rewrite E' in R1, R2.
  destruct R1; inv REL; destruct R2; inv REL0; clear REL REL0; destruct VR, VR0;
    try (by apply wnam_supports in REL; simpls; eapply Nam1 in REL);
    try (by apply wnam_supports in REL0; simpls; eapply Nam2 in REL0).
  depdes REL REL0; eapply EEnc_inj in x; des; apply unflex_inj in x; subst.
  do 2 econs; simpls; replace (FTname a) with ty. 
  - by rdes G; eapply GMON; eauto.
  - by apply unflex_inj; eapply decomp_inj; simpl; rewrite E'.
Grab Existential Variables. exact 0. exact 0. exact 0. exact 0.
Qed.

Lemma bullet_grel1 : forall g (G : wglob pw g) s1 s2 ty n1 v1 n2 v2
    (R: bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty n1 v1 n2 v2),
  vclos (g (s1, s2)) ty (S n1) v1 (S n2) v2.
Proof.
  i; destruct R; remember (decomp nfst ty) in R1; move R1 before G; revert_until R1.
  pattern t, n1, v1, n0, v0; apply (vclos_induction_strong R1).
  clear t n1 v1 n0 v0 R1.
  i; destruct (classic (exists fty : flextyp, ty0 = fty)) as [(fty, E')|NF]; subst.
  - eapply (bullet_grel_flex1 G fty).
    econstructor; eauto.
    by econs; eapply vclos_gen_mon; eauto; inst; s; i; des; eauto.
  - intros; depdes ty0.
    + exfalso; apply NF; eexists (FTname _); s; eauto.
    + depdes IH.
    + depdes IH; destruct R2; depdes REL; esplits; eauto.
    + depdes IH; destruct R2; depdes REL; esplits; eauto.
    + depdes IH; destruct R2; depdes REL; esplits; eauto.
    + depdes IH; des; destruct R2; depdes REL; eauto.
    + depdes IH; des; destruct R2; depdes REL; eauto.
    + exfalso; apply NF; eexists (FTarrow _ _); s; eauto.
    + exfalso; apply NF; eexists (FTref _); s; eauto.
    + depdes IH; des; destruct R2; depdes REL.
      eapply mp; [eapply (VR0 (tsubst1 0 (Tmu ty0) ty0))|].
      * eby rewrite <- decomp_fst_tsubst1.
      * eby rewrite <- decomp_snd_tsubst1.
      * eauto.
    + exfalso; apply NF; eexists (FTforall _); s; eauto.
    + depdes IH; des; destruct R2; depdes REL.
      destruct decomp_surj with ty'0 ty' as (ty'' & E1 & E2); auto; subst.
      eapply mp; [eapply (VR0 (tsubst1 0 ty'' ty0))|].
      * eby rewrite <- decomp_fst_tsubst1.
      * eby rewrite <- decomp_snd_tsubst1.
      * i; des; esplits; econs; econs; [instantiate (1 := ty''); apply decomp_lc_typ|]; eauto.
Qed.

Lemma bullet_grel_flex2 : forall g (G : wglob pw g) (ty : flextyp) s1 s2 n1 v1 n2 v2
    (R: vclos (g (s1, s2)) ty n1 v1 n2 v2),
  bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty n1 v1 n2 v2.
Proof.
  intros; depdes ty; rr; destruct R; depdes REL.
  - econstructor 1 with (v2:=Vint (EEnc (Tarrow ty1 ty2) n v n' v')); eauto;
    s; do 3 econs; fold (unflex (FTarrow ty1 ty2)); econs; vauto.
  - edestruct G as (A1 & A2 & A3); eauto; eapply A3 in VR.
    destruct VR; destruct R; subst. 
    destruct R1, R2; depdes REL REL0; do 3 econs; eauto.
  - econstructor 1 with (v2:=Vint (EEnc (Tforall ty) n v n' v')); eauto;
    s; do 3 econs; fold (unflex (FTforall ty)); econs; vauto.
  - edestruct G as (A1 & A2 & A3).
    destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
      apply A3 in VR. 
        by destruct VR; subst; eapply bullet_mon; 
           eauto using vclos_mon, vclos_mon_lv, pw1_vrel_mon, pw2_vrel_mon; omega.
      by apply Nunion; des; eauto.
    destruct (LEM (names1 a)) as [N1|N1] _eqn : E1.
      exfalso; eapply A3 in VR.
        by destruct VR; des; subst; destruct R; eauto.
      by apply Nunion; des; eauto.
    destruct (LEM (names2 a)) as [N2|N2] _eqn : E2.
      exfalso; eapply A3 in VR.
        by destruct VR; subst; eauto using bullet_mon with arith.
      by apply Nunion; des; eauto.
    destruct (LEM (names a)) as [N|N] _eqn : E; 
      [by edestruct Nunion; specialize (H N); destruct H as [H|[H|H]]; contradiction|].
    by econstructor 1 with (v2:=Vint (EEnc (Tname a) n v n' v')); eauto; 
       s; rewrite E'; econstructor; fold (unflex (FTname a));
       do 3 econs; vauto; s; try rewrite E'; eauto; intros ? ? X; inv X. 
Grab Existential Variables. 
exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
Qed.

Lemma bullet_grel_nonflex2 : forall r1 r2 r s1 s2
    (ty: typ) (NF : ~ exists ty' : flextyp, ty = ty'),
  vclos_gen r (bullet s1 s2) ty <4= bullet (vclos_gen r1 s1) (vclos_gen r2 s2) ty.
Proof.
  intros; depdes ty; simpls; inv PR; try by econs; eauto.
  - by contradict NF; exists (FTname aa).
  - by destruct VR1, VR2; des; econs; eauto; s; eapply vclos_gen_prod; eauto.
  - by destruct VR; des; econs; eauto; s; eapply vclos_gen_sum_inl; eauto;
      apply decomp_lc_typ'; eauto.
  - by destruct VR; des; econs; eauto; s; eapply vclos_gen_sum_inr; eauto;
      apply decomp_lc_typ'; eauto.
  - by exfalso; eapply NF; eexists (FTarrow _ _); eauto.
  - by exfalso; eapply NF; eexists (FTref _); eauto.
  - destruct VR; rewrite <- decomp_fst_tsubst1 in R1; rewrite <- decomp_snd_tsubst1 in R2.
    econs; eauto; s; eapply vclos_gen_mu; eauto.
  - by exfalso; eapply NF; eexists (FTforall _); eauto.
  - destruct VR; rewrite <- decomp_fst_tsubst1 in R1; rewrite <- decomp_snd_tsubst1 in R2.
    econs; eauto; s
    ; [eapply vclos_gen_exists with (ty':=decomp nfst ty')
      |eapply vclos_gen_exists with (ty':=decomp nsnd ty')]
    ; try apply decomp_lc_typ'; eauto.
Grab Existential Variables.
  exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
  exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
Qed.

Lemma bullet_grel2 : forall g (G : wglob pw g) s1 s2,
  vclos (g (s1, s2)) <5= bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)).
Proof.
  intros g G s1 s2 ty n1 v1 n2 v2 PR; apply (vclos_induction PR).
  clear ty n1 v1 n2 v2 PR. i.
  destruct (classic (exists fty : flextyp, ty = fty)) as [(fty, E')|NF]; subst.
  - eapply bullet_grel_flex2; auto; dependent destruction fty; inversion IH; simpl; eauto.
  - eapply bullet_mon
    ; eauto using bullet_grel_nonflex2, vclos_mon, vclos_mon_lv, pw1_vrel_mon, pw2_vrel_mon.
Qed.

Lemma not_reduce_app_int : forall h n (v : val) h' e',
  ~ reduce h (Eapp (Vint n) v) h' e'.
Proof.
  i; depdes FH; destruct SK; depdes x; eauto using not_reduce_val.
Qed.

Lemma not_reduce_inst_int : forall h n h' e',
  ~ reduce h (Einst (Vint n)) h' e'.
Proof.
  i; depdes FH; destruct SK; depdes x; eauto using not_reduce_val.
Qed.

Lemma atom_app_val_val : forall (v1 v2: voe), is_atom (Eapp (v_of v1) (v_of v2)).
Proof.
  i; destruct v1, v2; subst; ss.
  ii; apply NV; clear NV.
  destruct sk; inv EQ; eauto using val_is_voe.
Qed.

Lemma atom_inst_val : forall (v: voe), is_atom (Einst (v_of v)).
Proof.
  i; destruct v; subst; ss.
  ii; apply NV; clear NV.
  destruct sk; inv EQ; eauto using val_is_voe.
Qed.

Lemma wexp_trans : forall g (gOK: wglob pw g) s1 s2 s1' s2' ty n1 e1 n2 n2' e2 n3 e3
    (EQ1: wexp pw1 (grel_fst g s2') s1 s1' (decomp nfst ty) n1 e1 n2 e2)
    (EQ2: wexp pw2 (grel_snd g s1') s2 s2' (decomp nsnd ty) n2' e2 n3 e3)
    n1e (GUARD1: is_value_of_exp e1 -> n1e > 0)
    n3e (GUARD3: is_value_of_exp e3 -> n3e > 0),
  wexp pw g (s1, s2) (s1', s2') ty (n1+n1e) e1 (n3+n3e) e3.
Proof.
  pcofix CIH with z.
  rr; i; assert (gOK':=gOK); rdes gOK'.
  destruct HREL as (h'' & HR1 & HR2).
  eapply wexp_core_external in EQ1;
    try assumption; try (by apply grel_fst_wglob); try (by apply HR1); eauto.
    2: intros; unfold grel_fst; split; intro IN. 
    2:   by destruct IN; [right; apply REL | left; eauto using wval_mon].
    2: by destruct IN; eauto.
  eapply wexp_core_external in EQ2;
    try assumption; try (by apply grel_snd_wglob); try (by apply HR2); eauto.
    2: intros; unfold grel_snd; split; intro IN.
    2:   by destruct IN; [right; apply REL | left; eauto using wval_mon].
    2: by destruct IN; eauto.
  pdes EQ1; pdes EQ2.

  Case "div & div".
  by pfold; apply wexp_Diverge; eauto using ldiv_div, div_ldiv.

  Case "div & term".
  exfalso.
  eapply reduces_val_no_diverges; [eapply lreds_reds, RED|eapply ldiv_div, DIV'].

  Case "div & call".
  {
  exfalso.
  eapply diverges_no_stuck; [by eapply ldiv_div, DIV'|].
  econs; [M| |M]; Mdo eauto using lreds_reds, wstuck_not_value1, value_cont.
  i; eapply not_reduce_cont in FH; eauto using wstuck_not_value1. 
  i; depdes EREL; subst; inv FREL; eauto using not_reduce_app_int, not_reduce_inst_int.
  }

  Case "term & div".
  exfalso.
  by eapply reduces_val_no_diverges; [eapply lreds_reds, RED' | eapply ldiv_div, DIV].

  Case "term & term".
  {
  edestruct reduces_val_det; [by eapply lreds_reds, RED'| |by eapply lreds_reds, RED0| |]; eauto; des; subst.
  eapply hplus_hemp_rE in PLUS0; eapply hplus_hemp_rE in PLUS2; subst.
  exploit (@bullet_grel1 _ gOK s' s'0).
    econs; eauto; eapply vclos_mon; [apply VREL| |apply VREL0|]; eauto.
      by i; eapply pw1_vrel_mon; eauto; i; eapply GMON; eauto; split; eauto.
    by i; eapply pw2_vrel_mon; eauto; i; eapply GMON; eauto; split; eauto.
  i; des; pfold; eapply wexp_Terminate with (s' := (s', s'0)); eauto; try edone.
  - by rr; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
  - destruct (classic (is_value_of_exp e1)).
    + apply GUARD1 in H.
      eapply lreds_mon; [apply lreds_plus with (n':=1)| |]; eauto.
    + eauto 10 using reds_lreds_val, lreds_reds, val_is_voe.
  - destruct (classic (is_value_of_exp e3)).
    + apply GUARD3 in H.
      eapply lreds_mon; [apply lreds_plus with (n':=1)| |]; eauto.
    + eauto 10 using reds_lreds_val, lreds_reds, val_is_voe.
  }

  Case "term & call".
  {
  exfalso.
  eapply reduces_val_no_stuck; [by eapply lreds_reds, RED' |].
  econs; [M| |M]; Mdo eauto using lreds_reds, wstuck_not_value1, value_cont.
  i; eapply not_reduce_cont in FH; eauto using wstuck_not_value1. 
  i; depdes EREL; subst; inv FREL; eauto using not_reduce_app_int, not_reduce_inst_int, val_is_voe.
  }

  Case "call & div".
  {
  exfalso.
  eapply diverges_no_stuck; [by eapply ldiv_div, DIV|].
  econs; [M| |M]; Mdo eauto using lreds_reds, wstuck_not_value2, value_cont.
  i; eapply not_reduce_cont in FH; eauto using wstuck_not_value2. 
  i; depdes EREL; subst; inv FREL; eauto using not_reduce_app_int, not_reduce_inst_int, val_is_voe.
  }

  Case "call & term".
  {
  exfalso.
  eapply reduces_val_no_stuck; [by eapply lreds_reds, RED0 |].
  econs; [eapply lreds_reds, RED'| |eauto using lreds_reds, wstuck_not_value2, value_cont].
  i; eapply not_reduce_cont in FH; eauto using wstuck_not_value2. 
  i; depdes EREL; subst; inv FREL; eauto using not_reduce_app_int, not_reduce_inst_int, val_is_voe.
  }

  Case "call & call".
  depdes EREL EREL0 FREL FREL0; subst.

    SCase "arrow & arrow". 
    {
    edestruct reduces_stuck_det;
    [ by eapply lreds_reds, RED'
    | by i; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by i; eapply value_cont in FH
    | by eapply lreds_reds, RED0
    | by i; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by i; eapply value_cont in FH
    | ]. 
    des; apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    depdes H1; eapply EEnc_inj in x; des; simpl_injs; subst.
    assert (FT : exists tty1 tty2, ty1 = FTarrow tty1 tty2).
      destruct ty1; ss; eauto.
      destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
        by exfalso; eapply NN; [apply Nunion; right |]; eauto.
      by inversion E.
    des; subst; depdes E.
    pfold; eapply wexp_Call with (s':=(s',s'0)); [M|M| | | | | |]
    ; (Mdo by apply hplus_hemp_rE in PLUS2; apply hplus_hemp_rE in PLUS0; subst
            ; rr; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext)
    ; [| |M|M| |]; Mdo by eauto.
    - eapply mp; [eapply (@bullet_grel1 _ gOK s' s'0)|].
        econs; eauto; eapply vclos_mon; [apply VREL| |apply VREL0|].
          by i; eapply pw1_vrel_mon; eauto; i; eapply GMON; eauto; split; eauto.
        by i; eapply pw2_vrel_mon; eauto; i; eapply GMON; eauto; split; eauto.
      i; des; eapply wstuck_App; eauto using pw_ext.
    - instantiate (2:= nk3+1); instantiate (3:= nk1+1).
      ii; assert (X := g'OK); rdes X; rrdes EXTP.
      destruct s'1 as (s''1, s''2).
      apply bullet_grel2 in VREL1; auto; destruct VREL1.
      replace (nk1+1+nv) with (nk1+nv+1) by omega.
      replace (nk3+1+nv') with (nk3+nv'+1) by omega.
      right; eapply CIH; eauto.
      + eapply wexp_mon; [eapply wexp_u_mon, KREL; [by i; destruct PR| | | | |]| | |]
        ; eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                         pw_ext, wextpub_in_wext, wext_trans; omega.
      + eapply wexp_mon; [eapply wexp_u_mon, KREL0; [by i; destruct PR| | | | |]| | |]
        ; eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                         pw_ext, wextpub_in_wext, wext_trans; omega.
    - eauto using lreds_mon.
    - eauto using lreds_mon.
    }

    SCase "arrow & forall".
    edestruct reduces_stuck_det;
    [ by eapply lreds_reds, RED'
    | by intros; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply value_cont in FH
    | by eapply lreds_reds, RED0
    | by intros; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply value_cont in FH
    | ].
    des; apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    by inversion H1.
  
    SCase "forall & arrow".
    edestruct reduces_stuck_det;
    [ by eapply lreds_reds, RED'
    | by intros; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply value_cont in FH
    | by eapply lreds_reds, RED0
    | by intros; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply value_cont in FH
    | ].
    des; apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    by inversion H1.

    SCase "forall & forall".
    {
    edestruct reduces_stuck_det;
    [ by eapply lreds_reds, RED'
    | by intros; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply value_cont in FH
    | by eapply lreds_reds, RED0
    | by intros; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply value_cont in FH
    | ].
    des; apply app_cont_eq in H0; des; eauto using atom_app_val_val, atom_inst_val.
    depdes H1; eapply EEnc_inj in x; des; simpl_injs.
    assert (X := decomp_surj _ _ LC LC0); des; subst.
    assert (FT : exists tty, ty3 = FTforall tty).
      destruct ty3; simpls; eauto.
      destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
        by exfalso; eapply NN; [apply Nunion; right |]; eauto.
      by inversion E.
    des; subst; depdes E.
    pfold; eapply wexp_Call with (s':=(s',s'0)); [M| | | |M|M| |]; Mdo by eauto. 
    - apply hplus_hemp_rE in PLUS2; apply hplus_hemp_rE in PLUS0; subst.
      by repeat red; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
    - eapply wstuck_Inst with (ty := tsubst1 0 ty4 tty); eauto using pw_ext, decomp_lc_typ.
    - instantiate (2:= nk3+1); instantiate (3:= nk1+1).
      ii; assert (X := g'OK); rdes X; rrdes EXTP.
      destruct s'1 as (s''1, s''2).
      apply bullet_grel2 in VREL; auto; destruct VREL.
      replace (nk1+1+nv) with (nk1+nv+1) by omega.
      replace (nk3+1+nv') with (nk3+nv'+1) by omega.
      rewrite <-decomp_fst_tsubst1 in *; rewrite <-decomp_snd_tsubst1 in *; eauto using decomp_lc_typ.
      right; eapply CIH; eauto.
      + eapply wexp_mon; [eapply wexp_u_mon, KREL; [by i; destruct PR| | | | |]| | |]
        ; eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                         pw_ext, wextpub_in_wext, wext_trans; omega.
      + eapply wexp_mon; [eapply wexp_u_mon, KREL0; [by i; destruct PR| | | | |]| | |]
        ; eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                         pw_ext, wextpub_in_wext, wext_trans; omega.
    - eauto using lreds_mon.
    - eauto using lreds_mon.
    }
Qed.

Lemma pw_inhabited : winhabited pw.
Proof.
  intros g Gg.
  edestruct Wih1 as (s1 & IN1); try by apply wvalmin_glob.
  edestruct Wih2 as (s2 & IN2); try by apply wvalmin_glob.
  exists (s1, s2); exists hemp; simpl decomp.
  by split; eapply wheap_mon; [by apply IN1 | | by apply IN2 |];
     intros; induction PR; eauto using wval_mon.
Qed.

Lemma pw_consistent : wconsistent pw.
Proof.
  ii; rename x0 into ty, x1 into n1, x2 into e1, x3 into n3, x4 into e3.
  destruct s as (s1,s2).
  assert (P : exists n2 n2' e2, 
    n1 > 0 /\ n3 > 0 /\
    wstuck (grel_fst g s2 s1) (grel_fst g s2 s1) (decomp nfst ty) (n1-1) e1 n2 e2 /\
    wstuck (grel_snd g s1 s2) (grel_snd g s1 s2) (decomp nsnd ty) n2' e2 (n3-1) e3).
  {
    depdes PR; subst.
    - (* app *)
      apply bullet_grel2 in VREL; eauto; depdes VREL.
      inv FREL; depdes R; esplits; eauto; econs; eauto.
      + destruct R0; depdes REL; eapply (grel_fst_wglob gOK _) in VR; eauto.
      + destruct R3; depdes REL; eapply (grel_snd_wglob gOK _) in VR; eauto.
    - (* inst *)
      inv FREL; depdes R; esplits; eauto.
      + rewrite <-decomp_fst_tsubst1; econs 2; eauto.
        * by eapply decomp_lc_typ'. 
        * destruct R1; depdes REL; eapply (grel_fst_wglob gOK _) in VR; eauto.
      + rewrite <-decomp_snd_tsubst1; econs 2; eauto.
        * by eapply decomp_lc_typ'. 
        * destruct R2; depdes REL; eapply (grel_snd_wglob gOK _) in VR; eauto.
  }

  des; econs.
  - eapply wexp_trans with (n1e:=0) (n3e:=0); eauto.
    + ii; pfold; eapply wexp_Call with (nk1:=0)(nk2:=0)(K1:=[])(K2:=[])
      ; eauto using lreduces_nil.
      left; eapply (@wexp_val pw1); eauto.
    + ii; pfold; eapply wexp_Call with (nk1:=0)(nk2:=0)(K1:=[])(K2:=[])
      ; eauto using lreduces_nil.
      left; eapply (@wexp_val pw2); eauto.
    + s; i; exfalso; eauto using wstuck_not_value1.
    + s; i; exfalso; eauto using wstuck_not_value2.
  - econs; omega.
  - econs; omega.
Qed.

Lemma weq_trans : forall fn1 fn2 fn2' fn3 pe1 pe2 pe3
    (WEQ1 : weq pD pG (wlift lpw1) pty fn1 pe1 fn2 pe2)
    (WEQ2 : weq pD pG (wlift lpw2) pty fn2' pe2 fn3 pe3),
  weq pD pG pw pty (fun x => fn1 x + 1) pe1 (fun x => fn3 x + 1) pe3.
Proof.
  intros; cut (wequiv pD pG pw pty (fun x => fn1 x + 1) pe1 (fun x => fn3 x + 1) pe3).
    by eauto 10 using pw_inhabited, pw_consistent.

  rdes WEQ1; rdes WEQ2.
  destruct EQ, EQ0; econs; eauto.
  i; destruct s as (s1, s2).

  (**********************************************************************)
  (* in order to construct the mediating gammas, we use classical choice *)
  (* to do so, we first bring bullet_grel into the right form: *)
  assert (bullet_grel' :
      forall (t : typ * ((nat * val) * (nat * val))),
      exists nnv2 : (nat * nat) * val,
        match t with pair ty (pair (pair n1 v1) (pair n3 v3)) =>
        match nnv2 with pair (pair n2 n2') v2 =>
        forall (IN : vclos (g (s1, s2)) ty n1 v1 n3 v3),
          vclos (grel_fst g s2 s1) (decomp nfst ty) n1 v1 n2 v2 /\
          vclos (grel_snd g s1 s2) (decomp nsnd ty) n2' v2 n3 v3
        end end).
    intros; destruct t as (ty,((n1,v1),(n3,v3))).
    destruct (classic (vclos (g (s1, s2)) ty n1 v1 n3 v3)) as [IN|IN]
    ; [| by exists (0,0,Vunit)].
    apply bullet_grel2 in IN; auto.
    destruct IN; exists (n2,n2',v2); eauto.

  apply choice in bullet_grel'; destruct bullet_grel' as (f & Pf).
  set (gg1 := map (fun p => 
      (fst p, (fst(snd p),
      let x:=f (tsubst_typ d (snd(fst p)),snd p) in (fst(fst x),snd x)))) gg).
  set (gg2 := map (fun p =>
      (fst p, 
      (let x:=f (tsubst_typ d (snd(fst p)),snd p) in (snd(fst x),snd x),
      snd(snd p)))) gg).
  set (d1 := map_pair id (decomp nfst) d).
  set (d2 := map_pair id (decomp nsnd) d).

  (**********************************************************************)

  eapply wexp_trans; [eauto| | |eauto|eauto]; unfold map_pair in *. 

  Case "w1".
  {
    replace (decomp nfst (tsubst_typ d pty)) with (tsubst_typ d1 pty) 
      by (symmetry; eapply decomp_tsubst; eauto using NdisjGa, NdisjTy).
    replace (map (fun x => (fst(fst x),exp_of_val(snd(fst(snd x))))) gg)
      with  (map (fun x => (fst(fst x),exp_of_val(snd(fst(snd x))))) gg1)
      by (unfold gg1; rewrite map_map; eauto).
    replace (map (fun x => fst (fst (snd x))) gg)
      with  (map (fun x => fst (fst (snd x))) gg1)
      by (unfold gg1; rewrite map_map; eauto).
    eapply (WEXP s1 _ (grel_fst_wglob gOK s2) d1) with (gg:=gg1).
    + unfold d1; rewrite map_map, <-DEQ; eauto.
    + i; unfold d1 in IN; apply in_map_iff in IN; des; subst x.
      s; eapply decomp_lc_typ'; eauto.
    + by rewrite <- GEQ; unfold gg1; rewrite map_map.
    + i; assert (X : decomp nfst (tsubst_typ d ty0) = tsubst_typ d1 ty0).
        apply decomp_tsubst; intros; [by apply dOK|].
        assert (In (x, ty0) pG).
          rewrite <- GEQ; revert IN.
          clear dependent gg2; clear ggOK GEQ.
          induction gg; auto; intros; destruct IN.
            by left; destruct a0 as [[? ?] [? ?]]; inversion H1.
          by right; auto.
        by eapply NdisjGa; eauto.
      rewrite <- X.
      assert (V : exists nv3 v3, In (x, ty0, ((nv1,v1), (nv3,v3))) gg /\ 
                  nv2 = fst(fst(f (tsubst_typ d ty0, ((nv1,v1), (nv3,v3))))) /\
                  v2 = snd (f (tsubst_typ d ty0, ((nv1,v1), (nv3,v3))))).
        clear dependent gg2; clear ggOK GEQ.
        induction gg; inversion IN.
          destruct a as [[] [[] []]]; depdes H.
          by esplits; eauto; left.
        edestruct IHgg as (nv3,(v3,IHgg')); eauto.
        by des; subst; esplits; eauto; right.
      des; subst v2 nv2.
      specialize (Pf (tsubst_typ d ty0, ((nv1,v1), (nv3,v3)))).
      remember (f (tsubst_typ d ty0, (nv1, v1, (nv3, v3)))) in *; clear Heqp.
      destruct p as ((n2,n2'),v2).
      by eapply Pf, ggOK, V.
  }

  Case "w2".
  {
    replace (decomp nsnd (tsubst_typ d pty)) with (tsubst_typ d2 pty) 
      by (symmetry; eapply decomp_tsubst; eauto using NdisjGa, NdisjTy).
    replace (map (fun x => (fst(fst x),exp_of_val(snd(snd(snd x))))) gg1)
      with  (map (fun x => (fst(fst x),exp_of_val(snd(fst(snd x))))) gg2)
      by (unfold gg1, gg2; rewrite !map_map; eauto).
    replace (map (fun x => (fst(fst x),exp_of_val(snd(snd(snd x))))) gg)
      with  (map (fun x => (fst(fst x),exp_of_val(snd(snd(snd x))))) gg2)
      by (unfold gg2; rewrite map_map; eauto).
    replace (map (fun x => fst (snd (snd x))) gg)
      with  (map (fun x => fst (snd (snd x))) gg2)
      by (unfold gg2; rewrite map_map; eauto).
    eapply (WEXP0 s2 _ (grel_snd_wglob gOK s1) d2) with (gg:=gg2).
    + unfold d2; rewrite map_map, <-DEQ; eauto.
    + i; unfold d2 in IN; apply in_map_iff in IN; des; subst x.
      s; eapply decomp_lc_typ'; eauto.
    + by rewrite <- GEQ; unfold gg2; rewrite map_map.
    + intros x ty0 nv2 v2 nv3 v3 IN.
      assert (X : decomp nsnd (tsubst_typ d ty0) = tsubst_typ d2 ty0).
        apply decomp_tsubst; intros; [by apply dOK|].
        assert (In (x, ty0) pG).
          rewrite <- GEQ; revert IN.
          clear dependent gg1; clear ggOK GEQ.
          induction gg; auto; intros; destruct IN.
            by left; destruct a0 as [[? ?] [? ?]]; inversion H1.
          by right; auto.
        by eapply NdisjGa; eauto.
      rewrite <- X.
      assert (V : exists nv1 v1, In (x, ty0, ((nv1,v1), (nv3,v3))) gg /\ 
                  nv2 = snd(fst(f (tsubst_typ d ty0, ((nv1,v1), (nv3,v3))))) /\
                  v2 = snd (f (tsubst_typ d ty0, ((nv1,v1), (nv3,v3))))).
        clear dependent gg1; clear ggOK GEQ.
        induction gg; inversion IN.
          destruct a as [[] [[] []]]; depdes H.
          by esplits; eauto; left.
        edestruct IHgg as (nv1,(v1,IHgg')); eauto.
        by des; subst; esplits; eauto; right.
      des; subst v2 nv2.
      specialize (Pf (tsubst_typ d ty0, ((nv1,v1), (nv3,v3)))).
      remember (f (tsubst_typ d ty0, (nv1, v1, (nv3, v3)))) in *; clear Heqp.
      destruct p as ((n2,n2'),v2).
      by eapply Pf, ggOK, V.
  }
Qed.

End trans_part1.
