Require Import vgtac Program List Arith Even Div2 Relations String Max.
Require Import basis countable lang subs model basiclemmas transitivity_part1.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.

Open Scope string_scope.
Open Scope list_scope.

Set Implicit Arguments.

Local Notation cid := constructive_indefinite_description.

Section trans_part2.

Variables
  (pD : list typename)
  (pG : list (termvar * typ))
  (pty : typ)
  (names: typename -> Prop) 
  (Nci : countably_infinite names).

Variables
  (lpw1 lpw2: world_local)
  (Nam1 : wnam (wlift lpw1) <1= names1 Nci) (Nam2 : wnam (wlift lpw2) <1= names2 pG pty Nci)
  (Wih1 : winhabited (wlift lpw1)) (Wih2 : winhabited (wlift lpw2))
  (Wcs1 : wconsistent (wlift lpw1)) (Wcs2 : wconsistent (wlift lpw2))
  (Wst1 : wstable lpw1) (Wst2 : wstable lpw2).

Variables 
  (nfst nsnd : typename -> typ)
  (nfst_closed : forall a, lc_typ 0 (nfst a))
  (nsnd_closed : forall a, lc_typ 0 (nsnd a))
  (decomp_inj : forall ty1 ty2
                (EQ1 : decomp pG pty Nci nfst ty1 = decomp pG pty Nci nfst ty2)
                (EQ2 : decomp pG pty Nci nsnd ty1 = decomp pG pty Nci nsnd ty2),
              ty1 = ty2)
  (decomp_surj : forall ty1 ty2 (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2),
               exists ty, decomp pG pty Nci nfst ty = ty1 /\ decomp pG pty Nci nsnd ty = ty2).

Notation bullet := (@bullet pG pty _ Nci nfst nsnd).
Notation bullet_grel1 := (@bullet_grel1 pG pty _ Nci lpw1 lpw2 Nam1 Nam2 nfst nsnd nfst_closed nsnd_closed decomp_inj decomp_surj).
Notation bullet_grel2 := (@bullet_grel2 pG pty _ Nci lpw1 lpw2 nfst nsnd nfst_closed nsnd_closed).
Notation decomp := (@decomp pG pty _ Nci).
Notation grel_fst := (@grel_fst pG pty _ Nci lpw1 lpw2 nfst).
Notation grel_snd := (@grel_snd pG pty _ Nci lpw1 lpw2 nsnd).
Notation decomp_inj := (@decomp_inj pG pty _ Nci).
Notation decomp_surj := (@decomp_surj pG pty _ Nci).
Notation namesnew := (@namesnew pG pty _ Nci).
Notation Nunion := (@Nunion pG pty _ Nci).
Notation pw := (@pw pG pty _ Nci lpw1 lpw2 nfst nsnd).
Notation pw1 := (@pw1 lpw1).
Notation pw2 := (@pw2 lpw2).
Notation pw_val := (@pw_val pG pty _ Nci lpw1 lpw2 nfst nsnd).
Notation pw1_vrel := (@pw1_vrel pG pty _ Nci lpw1 nfst).
Notation pw2_vrel := (@pw2_vrel pG pty _ Nci lpw2 nsnd).
Notation pw_val_closed := (@pw_val_closed pG pty _ Nci lpw1 lpw2 nfst nsnd).
Notation decomp_lc_typ := (@decomp_lc_typ pG pty _ Nci nfst nsnd).
Notation pw_consistent := (@pw_consistent pG pty _ Nci lpw1 lpw2 Nam1 Nam2 Wcs1 Wcs2 nfst nsnd nfst_closed nsnd_closed decomp_inj decomp_surj).
Notation pw_inhabited := (@pw_inhabited pG pty _ Nci lpw1 lpw2 Wih1 Wih2 nfst nsnd).

(* cartesian product kind of thing *)
Definition join A B la lb : list (A * B) :=
  fold_right (fun b acc => map (fun a => (a, b)) la ++ acc) nil lb.

Lemma in_join : forall A B
    la (a : A) (INa : In a la)
    lb (b : B) (INb : In b lb),
  In (a, b) (join la lb).
Proof.
  induction lb; intros; auto; simpl.
  destruct (classic (b = a0)); subst; apply in_or_app; [left | right].
    clear IHlb; induction la; auto; simpls; destruct (classic (a1 = a)); subst; auto.
    by right; apply IHla; inversion INa; auto.
  by apply IHlb; rrdes INb; vauto.
Qed.

Inductive wrefst_filter_fst' (s s' : wst wref) ty l1 l2 :=
  wrefst_filter_fst_intro (IN : (`s) ty l1 l2)
                   (NIN : forall ty' l2', ~ (`s') ty' l1 l2').

Inductive wrefst_filter_snd' (s s' : wst wref) ty l1 l2 :=
  wrefst_filter_snd_intro (IN : (`s) ty l1 l2)
                   (NIN : forall ty' l1', ~ (`s') ty' l1' l2).

Definition wrefst_filter_fst (s1 s2 : wst wref) : wst wref.
  refine (exist _ (wrefst_filter_fst' s1 s2) _).

  repeat red; intros; unfold wrefst; do 2 try (split; intros).
  - destruct IN, s1; eapply w; eauto.
  - by destruct IN, IN', s1, s2; simpls; subst; rrdes w; auto.
  - by destruct s1, s2; rrdes w; exists tydom ldom; intros; destruct IN; apply w2.
Defined.

Definition wrefst_filter_snd (s1 s2 : wst wref) : wst wref.
  refine (exist _ (wrefst_filter_snd' s1 s2) _).

  repeat red; intros; unfold wrefst; do 2 try (split; intros).
  - by destruct IN, s1; eapply w; eauto.
  - by destruct IN, IN', s1, s2; simpls; subst; rrdes w; auto.
  - by destruct s1, s2; rrdes w; exists tydom ldom; intros; destruct IN; apply w2.
Defined.

Definition wrefst_bullet (s1 s2 : wst wref) : wst wref.
  refine (exist _ (fun ty => comp2 (`s1 (decomp nfst ty)) (`s2 (decomp nsnd ty))) _).

  repeat split; intros; destruct s1, s2; try destruct IN as (l, A);
  try destruct IN' as (l', A'); rrdes w; rrdes w0; simpls.
  - by eapply decomp_lc_typ; left; eapply w; eauto.
  - by hdes; eapply decomp_inj; eauto.
  - by hdes; eauto.
  - by hdes; eapply decomp_inj; eauto.
  - by hdes; eauto.
  - set (tydom' := map
                     (fun p => match LEM (lc_typ 0 (fst p) /\ lc_typ 0 (snd p)) with
                                  | left LC => ` (cid _ (decomp_surj _ _ (proj1 LC) (proj2 LC)))
                                  | right _ => Tunit (* hack *)
                               end)
                     (join tydom tydom0)).
    exists tydom' (ldom0 ++ ldom); intros; rrdes IN; split.
      edestruct w2; [by apply IN |]; edestruct w4; [by apply IN0 |]; des.
      apply in_map_iff; exists (decomp nfst ty, decomp nsnd ty); simpl.
      case_eq (LEM (lc_typ 0 (decomp nfst ty) /\ lc_typ 0 (decomp nsnd ty))); intros; des; simpl.
        split; auto.
          set (X := (cid (fun ty1 : typ => decomp nfst ty1 = decomp nfst ty /\
            decomp nsnd ty1 = decomp nsnd ty)
            (decomp_surj (decomp nfst ty) (decomp nsnd ty) (proj1 (conj a a0))
            (proj2 (conj a a0))))).
          by destruct X; simpls; eapply decomp_inj; des; eauto.
        by apply in_join.
      by exfalso; apply f; split; [eapply w | eapply w0]; eauto.
    by split; eapply in_or_app; hdes; eauto.
Defined.

Lemma wrefst_bullet_mon : forall s1 s1' (X1 : wext wref s1' s1) s2 s2' (X2 : wext wref s2' s2),
  wext wref (wrefst_bullet s1' s2') (wrefst_bullet s1 s2).
Proof.
  repeat red; intros; destruct PR; des.
  apply X1 in H; apply X2 in H0.
  by exists x.
Qed.

(* next: construction of lpw *)

Definition lpw_val (s : wst pw1 * wst pw2) (r : vrelf) : vrelf :=
  fun ty => match ty with
              | FTref _ => fun _ _ _ _ => False
              | _ => pw_val s r ty
            end.

Lemma lpw_val_mon : forall r r' s s' ty n1 n1' v1 n2 n2' v2
    (IN: lpw_val s r ty n1 v1 n2 v2) 
    (LE : r <5= r') (EXT : prodrel (wext pw1) (wext pw2) s' s)
    (LEN1: n1 <= n1') (LEN2: n2 <= n2'),
  lpw_val s' r' ty n1' v1 n2' v2.
Proof. intros; destruct ty; unfold lpw_val in *; eauto using pw_val_mon. Qed.

Lemma lpw_val_closed : forall s r (CL : vrelf_closed r), vrelf_closed (lpw_val s r).
Proof.
  by repeat intro; unfold lpw_val in *; destruct ty; eapply pw_val_closed; eauto.
Grab Existential Variables. exact 0. exact 0. exact (FTname 0). exact s.
Qed.

Lemma lpw_val_wf : forall s r ty n1 f1 n2 f2
    (IN : lpw_val s r ty n1 f1 n2 f2),
  n1 > 0 /\ n2 > 0.
Proof.
  by i; destruct ty; try done; eapply pw_val_wf; eauto.
Qed.

Inductive lpw_heap (sr : wst wref) : wst pw1 * wst pw2 -> vrelf -> hrel :=
  lpw_heap_intro : forall
      (sr1 sr2 : wst wref) (SR : sr = wrefst_bullet sr1 sr2)
      (s1 : wst lpw1) (s2 : wst lpw2) r
      h1 h1g h1l (H1 : hplus h1g h1l h1) h2ab
      h2ag h2al (H2a : hplus h2ag h2al h2ab)
      h2bg h2bl (H2b : hplus h2bg h2bl h2ab)
      h3 h3g h3l (H3 : hplus h3g h3l h3)
      (DJ1 : forall l2 (INH : h2al l2 <> None) (INS : exists ty1 l1, `sr1 ty1 l1 l2), False)
      (DJ2 : forall l1 (INH : h2bl l1 <> None) (INS : exists ty2 l2, `sr2 ty2 l1 l2), False)
      (HR1 : wheap wref () (wrefst_filter_fst sr1 sr) (pw1_vrel r (sr1,s1)) h1g h2ag)
      (HW1 : wheap lpw1 sr1 s1 (pw1_vrel r (sr1,s1)) h1l h2al)
      (HR2 : wheap wref () (wrefst_filter_snd sr2 sr) (pw2_vrel r (sr2,s2)) h2bg h3g)
      (HW2 : wheap lpw2 sr2 s2 (pw2_vrel r (sr2,s2)) h2bl h3l),
    lpw_heap sr ((sr1, s1), (sr2, s2)) r h1 h3.
          
Lemma lpw_heap_mon : forall sr s r r' h1 h2 
    (HR: lpw_heap sr s r h1 h2) (MON: r <5= r'),
  lpw_heap sr s r' h1 h2.
Proof.
  intros; destruct s as [[sr1  s1] [sr2 s2]].
  destruct HR; subst; econstructor.
  - done. - by apply H1. - by apply H2a. - by apply H2b. - by apply H3. - done. - done.
  - by eauto using wheap_mon, pw1_vrel_mon.
  - by eauto using wheap_mon, pw1_vrel_mon.
  - by eauto using wheap_mon, pw2_vrel_mon.
  - by eauto using wheap_mon, pw2_vrel_mon.
Qed.

(*
Program Definition lpw : world_local :=
  {| wldlocal :=
    {| wnam := names
     ; wst := wst pw1 * wst pw2
     ; wext := prodrel (wext pw1) (wext pw2)
     ; wextpub := prodrel (wextpub pw1) (wextpub pw2)
     ; wval := fun _ => lpw_val
     ; wheap := lpw_heap
    |}
  |}.
Next Obligation. unfold prodrel in *; des; eauto using wext_trans, wextpub_trans. Qed.
Next Obligation. unfold prodrel in *; des; eauto using wext_trans, wextpub_trans. Qed.
Next Obligation. unfold prodrel in *; des; eauto using wext_trans, wextpub_trans. Qed.
Next Obligation. unfold prodrel in *; des; eauto using wext_trans, wextpub_trans. Qed.
Next Obligation. intros; eapply lpw_val_mon; eauto; eapply lpw_val_mon; eauto. Qed.
Next Obligation. intros; eapply lpw_val_closed; eauto. Qed.
Next Obligation. eapply lpw_heap_mon; eauto. Qed.
Next Obligation. by intros; revert WV; simpl; destruct (LEM (namesnew a)) as [] _eqn : E; intros; try apply Nunion; auto. Qed.
Next Obligation. intros; eapply lpw_val_arrow_wf; eauto. Qed.
Next Obligation. intros; eapply lpw_val_forall_wf; eauto. Qed.
*)
Definition lpw : world_local.
refine (
  {| wldlocal :=
    {| wnam := names
     ; wst := wst pw1 * wst pw2
     ; wext := prodrel (wext pw1) (wext pw2)
     ; wextpub := prodrel (wextpub pw1) (wextpub pw2)
     ; wval := fun _ => lpw_val
     ; wheap := lpw_heap
    |}
  |}).

eauto.
Grab Existential Variables.
- i; revert WV; s; destruct (LEM (namesnew a)) as [] _eqn : E; i; [apply Nunion|destruct WV; exfalso]; eauto.
- i; eapply lpw_heap_mon; eauto.
- i; eapply lpw_val_closed; eauto.
- i; eapply lpw_val_mon; eauto; eapply lpw_val_mon; eauto.
- i; unfold prodrel in *; des; eauto using wext_trans, wextpub_trans.
- i; unfold prodrel in *; des; eauto using wext_trans, wextpub_trans.
- i; unfold prodrel in *; des; eauto using wext_trans, wextpub_trans.
- i; unfold prodrel in *; des; eauto using wext_trans, wextpub_trans.
Defined.

Definition grel_pw (g : wst (wlift lpw) -> vrelf) (s : wst pw) : vrelf :=
  g (wrefst_bullet (fst (fst s)) (fst (snd s)), s).

Lemma wval_pw_lpw : forall sr1 s1 sr2 s2,
  wval pw () ((sr1, s1), (sr2, s2)) =
  wval (wlift lpw) () (wrefst_bullet sr1 sr2, ((sr1, s1), (sr2, s2))).
Proof.
  intros; extensionalities g ty' n1 v1 n3 v3.
  apply prop_ext; split; intro IN.
  - depdes ty'; simpls; auto.
    destruct IN; destruct R; subst.
    destruct R1, R2; depdes REL REL0.
    destruct VR.
    + depdes REL; destruct ty0; depdes E; [eby exfalso; eapply NR|].
      s in x; destruct (LEM (namesnew a)) _eqn: Ha; [|depdes x].
      exfalso; eapply NN; eauto.
      assert (X:= names_split Nci); des; rewrite UN1r; right.
      by assert (X:= names1r_split pG pty Nci); des; rewrite UN2new; right.
    + destruct VR0.
      * depdes REL0; destruct ty0; depdes E; [eby exfalso; eapply NR|].
        s in x; destruct (LEM (namesnew a)) _eqn: Ha; [|depdes x].
        exfalso; eapply NN; eauto.
        assert (X:= names_split Nci); des; rewrite UN1r; right.
        by assert (X:= names1r_split pG pty Nci); des; rewrite UN2new; right.
      * rrdes REL; [|by apply wlocal_nonref in REL].
        rrdes REL0; [|by apply wlocal_nonref in REL0].
        rrdes REL; rrdes REL0; subst; depdes TEQ0 VEQ0.
        unfold comp2; left; esplits; eauto; omega. 
  - depdes ty'.
    + destruct IN as [IN|IN]; [|by ss; eauto].
        by destruct IN; des; depdes TEQ.
    + destruct IN as [IN|IN]; [|by destruct IN; des; subst; depdes TEQ].
      rrdes IN; subst; depdes TEQ; rrdes TLL.
      econs; eauto; [by i; depdes NAME|].
      do 3 econs; eright; eauto; left.
      * by exists loc v2 (decomp nfst ty'); esplits; eauto.
      * by exists v2 loc' (decomp nsnd ty'); esplits; eauto.
    + destruct IN as [IN|IN]; [|by ss; eauto].
      by destruct IN as (?&?&?&?); des; depdes TEQ.
    + destruct IN as [IN|IN]; [|by ss; eauto].
      by destruct IN as (?&?&?&?); des; depdes TEQ.
Grab Existential Variables. exact 0. exact 0.
Qed.

Lemma grel_pw_glob : forall g (G : wglob (wlift lpw) g), wglob pw (grel_pw g).
Proof.
  intros; split; unnw; intros; [by eapply G; eauto | split]. 
  - intros; destruct G; des; eapply GMON; eauto.
    destruct s, s'; simpl @fst; repeat red; unnw; split; auto.
    by eapply wrefst_bullet_mon; destruct EXT as [[] []].

  - split; try split; unnw; intros.
    + by destruct s as [[] []]; apply G; rewrite <- wval_pw_lpw.
    + by destruct s as [[] []]; rewrite wval_pw_lpw; apply G.
    + by destruct s as [[] []]; rewrite wval_pw_lpw; apply G; [by right |].
Qed.

Lemma lpw_names : wnam lpw <1= names.
Proof. done. Qed.  

Lemma freshloc :
  exists freshloc : list location -> typ -> location -> location -> location,
    <<freshloc_inj : forall ctx ty l1 l2 ty' l1' l2',
        freshloc ctx ty l1 l2 = freshloc ctx ty' l1' l2' -> ty = ty' /\ l1 = l1' /\ l2 = l2'>>
    /\
    <<freshloc_fresh : forall ctx ty l1 l2, ~ In (freshloc ctx ty l1 l2) ctx>>.
Proof.
  assert (X := cntinf_prod typ_cntinf (cntinf_prod location_cntinf location_cntinf)).
  eapply countably_infinite_set_impl_set2 in X; red in X; des.
  exists (fun ctx ty l1 l2 => (fold_right max 0 ctx) + 1 + f (ty,(l1,l2))).
  split; unnw; intros.
  - by apply plus_reg_l in H; eapply X in H; inv H.
  - revert FH.
    assert (MAX : forall a, ~ In (max a (fold_right max 0 ctx) + 1 + f (ty, (l1, l2))) ctx).
      induction ctx; auto; intros; simpls; destruct FH.
        apply (lt_irrefl a); red; rewrite H at 2.
        eapply (le_trans _ (max a (fold_right max 0 ctx) + 1 + f (ty, (l1, l2)))).
          rewrite <- plus_assoc, plus_comm, <- plus_assoc; apply le_n_S; rewrite plus_comm.
          by apply le_plus_trans, le_max_l.
        by do 2 apply plus_le_compat_r; apply le_max_r.
      by eapply IHctx; rewrite max_assoc in H; eauto.
    by intros; eapply MAX with 0.
Qed.

Lemma lpw_stable : wstable lpw.
Proof.
  assert (X := freshloc); des; rename freshloc0 into freshloc.

  red; intros; inversion HR; subst.
  set (sr := wrefst_bullet sr1 sr2); rename sp' into sr'; rename hs2 into hs3.
  destruct (proj2_sig sr1) as (A1 & B1 & (tydom1 & ldom1 & DOM1)).
  destruct (proj2_sig sr2) as (A2 & B2 & (tydom2 & ldom2 & DOM2)).
  destruct (heap_findom h2ag) as (domh2ag, DOMh2ag).
  destruct (heap_findom h2bg) as (domh2bg, DOMh2bg).
  destruct (heap_findom h1l) as (domh1l, DOMh1l).
  destruct (heap_findom h2al) as (domh2al, DOMh2al).
  destruct (heap_findom h2bl) as (domh2bl, DOMh2bl).
  destruct (heap_findom h3l) as (domh3l, DOMh3l).
  set (ctx := ldom1 ++ ldom2 ++ domh2ag ++ domh2bg ++
              domh1l ++ domh2al ++ domh2bl ++ domh3l).

  set_prop sr1' (exist wrefst (fun ty1 l1 l2 => `sr1 ty1 l1 l2 \/ exists ty l3,
         ty1 = decomp nfst ty /\ l2 = freshloc ctx ty l1 l3 /\
         `sr' ty l1 l3 /\ ~ `sr ty l1 l3)) sr1'_refst.
  {
    red; intros; split; intros; unnw; subst.
      rrdes IN; rrdes IN'; subst.
        by destruct sr1; eapply w; eauto.
      by eapply decomp_lc_typ'; eauto; destruct sr'; eapply w; eauto.
    destruct (proj2_sig sr') as (A' & B' & (tydom' & ldom' & DOM')); split; intros.
    {
      rrdes IN; rrdes IN'; subst.
      - by eapply B1.
      - split; intros; subst.
          exfalso; assert (IN1' := IN1); apply SHR in IN1; des.
          case_eq (h1 loc1'); intro; (try by eapply hplus_disjoint; eauto).
          eapply hplus_none in H; [| by apply H1]; des.
          destruct (classic (exists ty1 loc2',
            ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 loc1' loc2')).
            by des; rrdes HR1; edestruct HSAT; simpl; eauto; rrdes H4; vauto.
          eapply H2; exists ty' loc2'; econstructor; eauto; intros.
          assert (A'' := FH); apply EXT in FH.
          by edestruct wrefst_inj1; [apply IN1' | apply FH | subst; vauto].
        exfalso; eapply DOM1 in IN'; des; eapply (freshloc_fresh ctx).
        by unfold ctx; subst; apply in_or_app; left; eauto.
      - split; intros; subst.
          exfalso; assert (IN'1' := IN'1); apply SHR in IN'1; des.
          case_eq (h1 loc1'); intro; (try by eapply hplus_disjoint; eauto).
          eapply hplus_none in H; [| by apply H1]; des.
          destruct (classic (exists ty1 loc2',
            ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 loc1' loc2')).
            by des; rrdes HR1; edestruct HSAT; simpl; eauto; rrdes H4; vauto.
          eapply H2; exists ty loc2; econstructor; eauto; intros.
          assert (A'' := FH); apply EXT in FH.
          by edestruct wrefst_inj1; [apply IN'1' | apply FH | subst; vauto].
        exfalso; eapply DOM1 in IN; des; eapply (freshloc_fresh ctx).
        by unfold ctx; subst; apply in_or_app; left; eauto.
      - split; intros; subst.
          edestruct B' as (B'1, B'2); [by apply IN'1 | by apply IN1 |].
          by specialize (B'1 eq_refl); des; specialize (B'2 LEQ); des; subst.
        by apply freshloc_inj in EQ; des; subst.
    }

    exists (tydom1 ++ map (decomp nfst) tydom').
    exists (ldom1 ++ ldom' ++
      map (fun x => freshloc ctx (fst x) (fst (snd x)) (snd (snd x)))
          (join tydom' (join ldom' ldom'))).
    intros; rrdes IN; repeat split;
    try (by apply in_or_app; left; edestruct DOM1; des; eauto).
    - by apply in_or_app; right; edestruct DOM'; des; eauto; subst; apply in_map; eauto.
    - by apply in_or_app; right; apply in_or_app; left; edestruct DOM'; des; eauto.
    - do 2 (apply in_or_app; right); subst; change (freshloc ctx ty0 l1 l3) with
        ((fun x => freshloc ctx (fst x) (fst (snd x)) (snd (snd x))) (ty0, (l1, l3))).
      by apply in_map; repeat try apply in_join; edestruct DOM'; des; eauto.
  }

  rename h2 into h3.
  set_prop sr2' (exist wrefst (fun ty2 l2 l3 => `sr2 ty2 l2 l3 \/ exists ty l1,
         ty2 = decomp nsnd ty /\ l2 = freshloc ctx ty l1 l3 /\
         `sr' ty l1 l3 /\ ~ `sr ty l1 l3)) sr2'_refst.
  {
    red; intros; split; intros; unnw; subst.
      rrdes IN; rrdes IN'; subst.
        by destruct sr2; eapply w; eauto.
      by eapply decomp_lc_typ'; eauto; destruct sr'; eapply w; eauto.
    destruct (proj2_sig sr') as (A' & B' & (tydom' & ldom' & DOM')); split; intros;
    try rename loc2 into loc3; try rename loc2' into loc3'.
    {
      rrdes IN; rrdes IN'; subst.
      - by eapply B2.
      - split; intros; subst.
          exfalso; eapply DOM2 in IN'; des; eapply (freshloc_fresh ctx).
          by unfold ctx; subst; apply in_or_app; right; apply in_or_app; left; eauto.
        exfalso; assert (IN1' := IN1); apply SHR in IN1; des.
        case_eq (h3 loc3'); intro; (try by eapply hplus_disjoint; eauto).
        eapply hplus_none in H; [| by apply H3]; des.
        destruct (classic (exists ty2 loc2',
          ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 loc2' loc3')).
          by des; rrdes HR2; edestruct HSAT; simpl; eauto; rrdes H4; vauto.
        eapply H2; exists ty' loc1'; econstructor; eauto; intros.
        assert (A'' := FH); apply EXT in FH.
        by edestruct wrefst_inj2; [apply IN1' | apply FH | subst; vauto].
      - split; intros; subst.
          exfalso; eapply DOM2 in IN; des; eapply (freshloc_fresh ctx).
          by unfold ctx; subst; apply in_or_app; right; apply in_or_app; left; eauto.
        exfalso; assert (IN'1' := IN'1); apply SHR in IN'1; des.
        case_eq (h3 loc3'); intro; (try by eapply hplus_disjoint; eauto).
        eapply hplus_none in H; [| by apply H3]; des.
        destruct (classic (exists ty2 loc2',
          ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 loc2' loc3')).
          by des; rrdes HR2; edestruct HSAT; simpl; eauto; rrdes H4; vauto.
        eapply H2; exists ty loc1; econstructor; eauto; intros.
        assert (A'' := FH); apply EXT in FH.
        by edestruct wrefst_inj2; [apply IN'1' | apply FH | subst; vauto].
      - split; intros; subst.
          by apply freshloc_inj in EQ; des; subst.
        edestruct B' as (B'1, B'2); [by apply IN'1 | by apply IN1 |].
        by specialize (B'2 eq_refl); des; specialize (B'1 LEQ); des; subst.
    }
    exists (tydom2 ++ map (decomp nsnd) tydom').
    exists (ldom2 ++ ldom' ++
      map (fun x => freshloc ctx (fst x) (fst (snd x)) (snd (snd x)))
          (join tydom' (join ldom' ldom'))).
    intros; rename l2 into l3, l1 into l2; rrdes IN; repeat split;
    try (by apply in_or_app; left; edestruct DOM2; des; eauto).
    - by apply in_or_app; right; edestruct DOM'; des; eauto; subst; apply in_map; eauto.
    - do 2 (apply in_or_app; right); subst; change (freshloc ctx ty0 l1 l3) with
        ((fun x => freshloc ctx (fst x) (fst (snd x)) (snd (snd x))) (ty0, (l1, l3))).
      by apply in_map; repeat try apply in_join; edestruct DOM'; des; eauto.
    - by apply in_or_app; right; apply in_or_app; left; edestruct DOM'; des; eauto.
  }

  assert (SR' : sr' = wrefst_bullet sr1' sr2').
  {
    eapply (@proj1_sig_inj _ wrefst).
    extensionality ty; extensionality l1; extensionality l3; apply prop_ext;
    unfold sr1', sr2'; split; intros; simpl; destruct (classic (`sr ty l1 l3)) as [IN|NIN].
    - by rrdes IN; eexists; vauto.
    - by exists (freshloc ctx ty l1 l3); split; right; repeat eexists; eauto.
    - by auto.
    - rrdes H; simpls; des; subst.
      + by exfalso; eapply NIN; eexists; eauto.
      + exfalso; eapply DOM2 in H0; des; eapply (freshloc_fresh ctx).
        by apply in_or_app; right; apply in_or_app; left; eauto.
      + exfalso; eapply DOM1 in H; des; eapply (freshloc_fresh ctx).
        by apply in_or_app; left; eauto.
      + apply freshloc_inj in H2; des; subst.
        by assert (E := decomp_inj _ _ H H0); subst; eauto.
  }

  eapply wheap_mon in SHR.
  2 : intros; instantiate (1 := g (sr', (sr1', s1, (sr2', s2)))).
  2 : eapply gOK in PR; apply PR; simpl; unnw; eauto.
  2 : by split; intros; red; repeat split; intros; auto; simpls; auto.

  set (hs1g := hunion hs1 h1g).
  assert (Ehs1g : hplus hs1 h1g hs1g).
    intro l1; unfold hs1g; simpl; unfold hunion'.
    case_eq (hs1 l1); case_eq (h1g l1); intros; vauto.
    by exfalso; eapply hplus_inl in H; eauto; eapply (hplus_disjoint DIS); eauto.

  assert (mediate : forall ty l2
        l1 (IN1 : ` sr1' (decomp nfst ty) l1 l2)
        l3 (IN2 : ` sr2' (decomp nsnd ty) l2 l3),
      exists v2 v1 v3 n1 n2 n2' n3,
        hs1 l1 = Some v1 /\ hs3 l3 = Some v3 /\
        vclos (grel_fst (grel_pw g) (sr2', s2) (sr1', s1)) (decomp nfst ty) n1 v1 n2 v2 /\
        vclos (grel_snd (grel_pw g) (sr1', s1) (sr2', s2)) (decomp nsnd ty) n2' v2 n3 v3).
    intros; destruct SHR as (WH1 & WH2 & WH3).
    edestruct WH1 as (n1 & v1 & n3 & v3 & ?); des.
      by rewrite SR'; exists l2; eauto.
    exploit @bullet_grel2.
        by apply grel_pw_glob; eauto.
      by rewrite SR' in VC; instantiate (7:=(_,_)); instantiate (6:=(_,_)); apply VC.
    by i; destruct x0; esplits; eauto.

  assert (HS2 : forall l2, exists sv2, 
      (forall ty l1 l3
          (IN1 : (`sr1') (decomp nfst ty) l1 l2) 
          (IN2 : (`sr2') (decomp nsnd ty) l2 l3),
        exists v2 v1 v3 n1 n2 n2' n3,
          sv2 = Some v2 /\ hs1 l1 = Some v1 /\ hs3 l3 = Some v3 /\
          vclos (grel_fst (grel_pw g) (sr2', s2) (sr1', s1)) (decomp nfst ty) n1 v1 n2 v2 /\
          vclos (grel_snd (grel_pw g) (sr1', s1) (sr2', s2)) (decomp nsnd ty) n2' v2 n3 v3)
      /\
      (forall v2 (ES : sv2 = Some v2), exists ty l1 l3,
        (`sr1') (decomp nfst ty) l1 l2 /\ (`sr2') (decomp nsnd ty) l2 l3)).
  {
    intros; destruct (classic (exists ty l1 l3, (`sr1') (decomp nfst ty) l1 l2
                                             /\ (`sr2') (decomp nsnd ty) l2 l3)) as [IN|NIN].
    - des; assert (X:= mediate _ _ _ IN _ IN0); des; exists (Some v2); split; eauto; intros;
      destruct (wrefst_inj2 IN IN1); destruct (wrefst_inj1 IN0 IN2); subst l1 l3.
      do 7 eexists; split; [| split]; eauto; split; [| split];
      try rewrite <- H; try rewrite <- H2; eauto.
    - by eexists None; split; [intros; exfalso; apply NIN; eauto | vauto].
  }

  destruct (choice _ HS2) as (hs2f & HS2F); clear HS2.
  set_prop hs2 (@Build_heap hs2f) hs2f_fin.
    destruct (proj2_sig sr1') as (? & ? & (tydom1' & ldom1' & DOM1')); exists ldom1'; intros.
    by apply HS2F in H2; des; eapply DOM1' in H2; des.

  set (hs2ag := hunion hs2 h2ag).
  assert (Ehs2ag : hplus hs2 h2ag hs2ag).
  {
    intro l2; simpl; unfold hunion'.
    case_eq (hs2 l2); case_eq (h2ag l2); des; intros v2 E2; auto; intros v2' E2'.
    exfalso; unfold hs2 in E2'; simpls; hdesH HS2F; eapply HS2F1 in E2'; des.
    - apply HDOM5 in E2; des; destruct E2.
      destruct (wrefst_inj2 IN E2'); subst l.
      by eapply NIN; exists l2; eauto.
    - subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; do 2 (apply in_or_app; right); apply in_or_app; left.
      by apply DOMh2ag in E2; eauto.
    - subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; do 2 (apply in_or_app; right); apply in_or_app; left.
      by apply DOMh2ag in E2; eauto.
    - apply HDOM5 in E2; des; destruct E2.
      subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; apply in_or_app; left.
      by hdesH DOM1; hdesH DOM0; apply INL2 in IN; eauto.
  }

  set (hs2bg := hunion hs2 h2bg).
  assert (Ehs2bg : hplus hs2 h2bg hs2bg).
  {
    intro l2; simpl; unfold hunion'.
    case_eq (hs2 l2); case_eq (h2bg l2); des; intros v2 E2; auto; intros v2' E2'.
    exfalso; unfold hs2 in E2'; simpls; hdesH HS2F; eapply HS2F1 in E2'; des.
    - apply HDOM0 in E2; des; destruct E2.
      destruct (wrefst_inj1 IN E2'0); subst l'.
      by eapply NIN; exists l2; eauto.
    - subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; do 3 (apply in_or_app; right); apply in_or_app; left.
      by apply DOMh2bg in E2; eauto.
    - subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; do 3 (apply in_or_app; right); apply in_or_app; left.
      by apply DOMh2bg in E2; eauto.
    - apply HDOM0 in E2; des; destruct E2.
      subst l2; eapply (freshloc_fresh ctx).
      unfold ctx in *; apply in_or_app; right; apply in_or_app; left.
      by hdesH DOM2; hdesH DOM0; apply INL1 in IN; eauto.
  }

  set (hs3g := hunion hs3 h3g).
  assert (Ehs3g : hplus hs3 h3g hs3g).
    intro l3; unfold hs3g; simpl; unfold hunion'.
    case_eq (hs3 l3); case_eq (h3g l3); intros; vauto.
    by exfalso; eapply hplus_inl in H; eauto; eapply (hplus_disjoint DIS0); eauto.

  eapply wheap_mon in HR1.
  2 : intros; change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2)))) with
    (grel_pw g (sr1, s1, (sr2, s2))) in PR.
  2 : instantiate (1 := pw1_vrel (grel_pw g (sr1', s1, (sr2', s2))) (sr1', s1)).
  2 : eapply pw1_vrel_mon; eauto; intros; edestruct grel_pw_glob; eauto; des.
  2 :   by eapply GMON; eauto; repeat split; eauto; simpl; unnw; intros; left.
  2 : by split; auto; simpl; unnw; intros; left.

  eapply wheap_mon in HW1.
  2 : intros; change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2)))) with
    (grel_pw g (sr1, s1, (sr2, s2))) in PR.
  2 : instantiate (1 := pw1_vrel (grel_pw g (sr1, s1, (sr2', s2))) (sr1, s1)).
  2 : eapply pw1_vrel_mon; eauto; intros; edestruct grel_pw_glob; eauto; des.
  2 : by eapply GMON; eauto; repeat split; eauto; simpl; unnw; intros; left.

  eapply wheap_mon in HR2.
  2 : intros; change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2)))) with
    (grel_pw g (sr1, s1, (sr2, s2))) in PR.
  2 : instantiate (1 := pw2_vrel (grel_pw g (sr1', s1, (sr2', s2))) (sr2', s2)).
  2 : eapply pw2_vrel_mon; eauto; intros; edestruct grel_pw_glob; eauto; des.
  2 :   by eapply GMON; eauto; repeat split; eauto; simpl; unnw; intros; left.
  2 : by split; auto; simpl; unnw; intros; left.

  eapply wheap_mon in HW2.
  2 : intros; change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2)))) with
    (grel_pw g (sr1, s1, (sr2, s2))) in PR.
  2 : instantiate (1 := pw2_vrel (grel_pw g (sr1', s1, (sr2, s2))) (sr2, s2)).
  2 : eapply pw2_vrel_mon; eauto; intros; edestruct grel_pw_glob; eauto; des.
  2 : by eapply GMON; eauto; repeat split; eauto; simpl; unnw; intros; left.

  assert (WHR1 : wheap wref () sr1'
                   (pw1_vrel (grel_pw g (sr1', s1, (sr2', s2))) (sr1', s1)) hs1g hs2ag).
  {
    repeat red; unnw; intros; split; intros.
    - assert (R' := R); destruct R.
        destruct (classic (` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty loc loc')).
          eapply HR1 in H0; des; eapply hplus_inr in HL; eauto.
          by eapply hplus_inr in HL'; esplits; eauto.
        assert (X : exists ty' loc'', ` (wrefst_bullet sr1 sr2) ty' loc loc'').
          apply not_all_not_ex; intro A; apply H0; constructor; auto.
          by intros; eapply not_ex_all_not in A; eauto.
        do 2 rrdes X; hdesH HS2F; edestruct HS2F0; try (by left; eauto); des.
        edestruct (wrefst_inj1 H X); subst ty loc'; eapply hplus_inl in H4; eauto.
        by change (hs2f v2) with (hs2 v2) in H2; apply (hplus_inl Ehs2ag) in H2; esplits; eauto.
      des; rewrite SR' in H2; destruct H2; des.
      assert (H2' := H2); eapply HS2F in H2; eauto; des.
      change (hs2f x) with (hs2 x) in H2; apply (hplus_inl Ehs2ag) in H2; eauto.
      by eapply hplus_inl in H6; eauto; edestruct (wrefst_inj1 R' H2'); subst ty x; esplits; eauto.
    - split; intros.
        unfold hs1g in IN; simpl in IN; unfold hunion' in IN.
        revert IN; case_eq (hs1 l); intros.
          inversion IN; subst v0.
          by apply SHR in H; des; rewrite SR' in H; rrdes H; eauto.
        by apply HR1 in IN; des; inversion IN; simpl; eauto.
      unfold hs2ag in IN; simpl in IN; unfold hunion' in IN.
        revert IN; case_eq (hs2 l'); intros.
          inversion IN; subst v'.
          by eapply HS2F in H; des; eauto.
        by apply HR1 in IN; des; inversion IN; simpl; eauto.
  }

  assert (WHR2 : wheap wref () sr2'
                   (pw2_vrel (grel_pw g (sr1', s1, (sr2', s2))) (sr2', s2)) hs2bg hs3g).
  {
    repeat red; unnw; intros; split; intros.
    - assert (R' := R); destruct R.
        destruct (classic (` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty loc loc')).
          eapply HR2 in H0; des; eapply hplus_inr in HL; eauto.
          by eapply hplus_inr in HL'; esplits; eauto.
        assert (X : exists ty' loc'', ` (wrefst_bullet sr1 sr2) ty' loc'' loc').
          apply not_all_not_ex; intro A; apply H0; constructor; auto.
          by intros; eapply not_ex_all_not in A; eauto.
        do 2 rrdes X; hdesH HS2F; edestruct HS2F0; try (by left; eauto); des.
        edestruct (wrefst_inj2 H X0); subst ty loc; eapply hplus_inl in H5; eauto.
        by change (hs2f v2) with (hs2 v2) in H2; apply (hplus_inl Ehs2bg) in H2; esplits; eauto.
      des; rewrite SR' in H2; destruct H2; des.
      assert (H2' := H2); eapply HS2F in H2; eauto; des.
      change (hs2f x) with (hs2 x) in H2; apply (hplus_inl Ehs2bg) in H2; eauto.
      by eapply hplus_inl in H7; eauto; edestruct (wrefst_inj2 R' H5); subst ty x; esplits; eauto.
    - split; intros.
        unfold hs2bg in IN; simpl in IN; unfold hunion' in IN.
          revert IN; case_eq (hs2 l); intros.
            inversion IN; subst v0.
            by eapply HS2F in H; des; eauto.
          by apply HR2 in IN; des; inversion IN; simpl; eauto.
      unfold hs3g in IN; simpl in IN; unfold hunion' in IN.
      revert IN; case_eq (hs3 l'); intros.
        inversion IN; subst v'.
        by apply SHR in H; des; rewrite SR' in H; rrdes H; eauto.
      by apply HR2 in IN; des; inversion IN; simpl; eauto.
  }

  assert (DIS1' : exists hh1', hplus hs1g h1l hh1').
    exists (hunion hs1g h1l).
    intro l1; simpl; unfold hs1g; unfold hunion; unfold hunion'; simpl.
    case_eq (hs1 l1); case_eq (h1g l1); case_eq (h1l l1); intros; auto;
    try (by exfalso; eapply (hplus_disjoint H1); eauto).
    by eapply hplus_inr in H; eauto; exfalso; eapply (hplus_disjoint DIS); eauto.
  rrdes DIS1'.

  assert (DIS2a' : exists hh2a', hplus hs2ag h2al hh2a').
    exists (hunion hs2ag h2al).
    intro l2; simpl; unfold hs2ag; unfold hunion; unfold hunion'; simpl.
    case_eq (hs2f l2); case_eq (h2ag l2); case_eq (h2al l2); intros; auto;
    try (by exfalso; eapply (hplus_disjoint H2a); eauto).
    apply HS2F in H2; des; exfalso; eapply DJ1.
      by instantiate (1 := l2); intro; vauto.
    rrdes H2; eauto; subst l2.
    exfalso; eapply (freshloc_fresh ctx); unfold ctx.
    do 5 (apply in_or_app; right); apply in_or_app; left.
    by apply DOMh2al in H; eauto.
  rrdes DIS2a'.

  assert (DIS2b' : exists hh2b', hplus hs2bg h2bl hh2b').
    exists (hunion hs2bg h2bl).
    intro l2; simpl; unfold hs2bg; unfold hunion; unfold hunion'; simpl.
    case_eq (hs2f l2); case_eq (h2bg l2); case_eq (h2bl l2); intros; auto;
    try (by exfalso; eapply (hplus_disjoint H2b); eauto).
    apply HS2F in H2; des; exfalso; eapply DJ2.
      by instantiate (1 := l2); intro; vauto.
    rrdes H4; eauto; subst l2.
    exfalso; eapply (freshloc_fresh ctx); unfold ctx.
    do 6 (apply in_or_app; right); apply in_or_app; left.
    by apply DOMh2bl in H; eauto.
  rrdes DIS2b'.

  assert (DIS3' : exists hh3', hplus hs3g h3l hh3').
    exists (hunion hs3g h3l).
    intro l3; simpl; unfold hs3g; unfold hunion; unfold hunion'; simpl.
    case_eq (hs3 l3); case_eq (h3g l3); case_eq (h3l l3); intros; auto;
    try (by exfalso; eapply (hplus_disjoint H3); eauto).
    by eapply hplus_inr in H; eauto; exfalso; eapply (hplus_disjoint DIS0); eauto.
  rrdes DIS0'.

  edestruct Wst1; des.
  : instantiate (1 := grel_fst (grel_pw g) (sr2', s2)).
    by eauto using grel_fst_wglob, grel_pw_glob.
  : by eauto.
  : by instantiate (1 := sr1'); left.
  : by eauto.
  : by eauto.
  : by eauto.

  edestruct Wst2; des.
  : instantiate (1 := grel_snd (grel_pw g) (sr1', s1)).
    by eauto using grel_snd_wglob, grel_pw_glob.
  : by eauto.
  : by instantiate (1 := sr2'); left.
  : by eauto.
  : by eauto.
  : by eauto.

  assert (FI1 : wrefst_filter_fst sr1' (wrefst_bullet sr1' sr2') =
                wrefst_filter_fst sr1  (wrefst_bullet sr1  sr2)).
  {
    eapply (@proj1_sig_inj _ wrefst); simpl.
    extensionality ty; extensionality l1; extensionality l2; apply prop_ext; split; intros;
    inversion H; econstructor; vauto.
    - destruct IN; vauto; des.
      by exfalso; eapply NIN; rewrite <- SR'; eauto.
    - by intros; eapply NIN; rewrite <- SR'; apply EXT; eauto.
    - intros; rrdes FH; assert (B := FH); rrdes FH; rrdes FH0.
      + by destruct (wrefst_inj1 IN FH); subst l2 ty; eapply NIN; vauto.
      + eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; left.
        by subst v2; apply DOM1 in FH; des; eauto.
      + eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; right; apply in_or_app; left.
        by subst v2; apply DOM2 in FH0; des; eauto.
      + edestruct DOM1; eauto; des.
        edestruct (wrefst_inj1 B); [by left; eauto |]; subst v2 l2.
        by eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; left; eauto.
  }

  assert (FI2 : wrefst_filter_snd sr2' (wrefst_bullet sr1' sr2') =
                wrefst_filter_snd sr2  (wrefst_bullet sr1  sr2)).
  {
    eapply (@proj1_sig_inj _ wrefst); simpl.
    extensionality ty; extensionality l1; extensionality l2; apply prop_ext; split; intros;
    inversion H; econstructor; vauto.
    - destruct IN; vauto; des.
      by exfalso; eapply NIN; rewrite <- SR'; eauto.
    - by intros; eapply NIN; rewrite <- SR'; apply EXT; eauto.
    - intros; rrdes FH; assert (B := FH0); rrdes FH; rrdes FH0.
      + by destruct (wrefst_inj2 IN FH0); subst l1 ty; eapply NIN; vauto.
      + eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; left.
        by subst v2; apply DOM1 in FH; des; eauto.
      + eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; right; apply in_or_app; left.
        by subst v2; apply DOM2 in FH0; des; eauto.
      + edestruct DOM2; eauto; des.
        edestruct (wrefst_inj2 B); [by left; eauto |]; subst v2 l1.
        by eapply (freshloc_fresh ctx); unfold ctx; apply in_or_app; right;
          apply in_or_app; left; eauto.
  }

  rename x into s1', x0 into s2'; exists (sr1', s1', (sr2', s2')); split.
    by repeat split; auto; simpl; auto using @left.
  econstructor.
  - auto.
  - apply H1.
  - apply H2a.
  - apply H2b.
  - apply H3.
  - intros; des; destruct INS; des.
      by eapply DJ1; eauto.
    revert INH; case_eq (h2al l2); vauto; intros; apply DOMh2al in H5; subst l2.
    eapply (freshloc_fresh ctx); unfold ctx; do 5 (apply in_or_app; right).
    apply in_or_app; left; eauto.
  - intros; des; destruct INS; des.
      by eapply DJ2; eauto.
    revert INH; case_eq (h2bl l1); vauto; intros; apply DOMh2bl in H5; subst l1.
    eapply (freshloc_fresh ctx); unfold ctx; do 6 (apply in_or_app; right).
    apply in_or_app; left; eauto.
  - rewrite SR', FI1; eapply wheap_mon; eauto; intros.
    eapply pw1_vrel_mon; eauto; intros.
      by rrdes gOK; eapply GMON; eauto; repeat split; auto.
    repeat split; auto.
  - eapply wheap_mon; eauto; intros.
    eapply pw1_vrel_mon; eauto; intros.
    rewrite SR'; rrdes gOK; eapply GMON; eauto; repeat split; auto.
  - rewrite SR', FI2; eapply wheap_mon; eauto; intros.
    eapply pw2_vrel_mon; eauto; intros.
      by rrdes gOK; eapply GMON; eauto; repeat split; auto.
    repeat split; auto.
  - eapply wheap_mon; eauto; intros.
    eapply pw2_vrel_mon; eauto; intros.
    rewrite SR'; rrdes gOK; eapply GMON; eauto; repeat split; auto.
Qed.

Lemma wheap_pw_lpw_1 :
  forall g (G : wglob pw g) sr1 s1 sr2 s2 h1 h3
         (WH : wheap pw () ((sr1,s1),(sr2,s2)) (g ((sr1,s1),(sr2,s2))) h1 h3),
  wheap (wlift lpw) () (wrefst_bullet sr1 sr2, ((sr1,s1),(sr2,s2))) (g ((sr1,s1),(sr2,s2))) h1 h3.
Proof.
  i.
  destruct WH as (h2 & (h1c & h2ac & h1l & h2al & ?) & (h2bc & h3c & h2bl & h3l & ?)).
  des.

  set_prop h1d (@Build_heap (fun l1 =>
      match LEM (exists ty1 l2,
                   ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 l1 l2) with
      | right p => h1c l1
      | _ => None
      end)) h1d_fin.
    destruct (heap_findom h1c) as (h1c_dom, INh1c); exists h1c_dom; intros l1 v1 E.
    revert E; case_eq (LEM (exists ty1 l2,
                              (` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2))) ty1 l1 l2));
    intros; try (by inversion E).
    by eapply INh1c; eauto.

  set_prop h1g (@Build_heap (fun l1 =>
      match LEM (exists ty1 l2,
                   ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 l1 l2) with
      | left p => h1c l1
      | _ => None
      end)) h1'_fin.
    destruct (heap_findom h1c) as (h1c_dom, INh1c); exists h1c_dom; intros l1 v1 E.
    revert E; case_eq (LEM (exists ty1 l2,
                              (` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2))) ty1 l1 l2));
    intros; try (by inversion E).
    by eapply INh1c; eauto.

  set_prop h2ad (@Build_heap (fun l2 =>
      match LEM (exists ty1 l1,
                   ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 l1 l2) with
      | right p => h2ac l2
      | _ => None
      end)) h2ad_fin.
    destruct (heap_findom h2ac) as (h2ac_dom, INh2ac); exists h2ac_dom; intros l2 v2 E.
    revert E; case_eq (LEM (exists ty1 l1,
                              (` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2))) ty1 l1 l2));
    intros; try (by inversion E).
    by eapply INh2ac; eauto.

  set_prop h2ag (@Build_heap (fun l2 =>
      match LEM (exists ty1 l1,
                   ` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2)) ty1 l1 l2) with
      | left p => h2ac l2
      | _ => None
      end)) h2a'_fin.
    destruct (heap_findom h2ac) as (h2ac_dom, INh2ac); exists h2ac_dom; intros l2 v2 E.
    revert E; case_eq (LEM (exists ty1 l1,
                              (` (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2))) ty1 l1 l2));
    intros; try (by inversion E).
    by eapply INh2ac; eauto.

  set_prop h2bd (@Build_heap (fun l2 =>
      match LEM (exists ty2 l3,
                   ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 l2 l3) with
      | right p => h2bc l2
      | _ => None
      end)) h2bd_fin.
    destruct (heap_findom h2bc) as (h2bc_dom, INh2bc); exists h2bc_dom; intros l2 v2 E.
    revert E; case_eq (LEM (exists ty2 l3,
                              (` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2))) ty2 l2 l3));
    intros; try (by inversion E).
    by eapply INh2bc; eauto.

  set_prop h2bg (@Build_heap (fun l2 =>
      match LEM (exists ty2 l3,
                   ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 l2 l3) with
      | left p => h2bc l2
      | _ => None
      end)) h2b'_fin.
    destruct (heap_findom h2bc) as (h2bc_dom, INh2bc); exists h2bc_dom; intros l2 v2 E.
    revert E; case_eq (LEM (exists ty2 l3,
                              (` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2))) ty2 l2 l3));
    intros; try (by inversion E).
    by eapply INh2bc; eauto.

  set_prop h3d (@Build_heap (fun l3 =>
      match LEM (exists ty2 l2,
                   ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 l2 l3) with
      | right p => h3c l3
      | _ => None
      end)) h3d_fin.
    destruct (heap_findom h3c) as (h3c_dom, INh3c); exists h3c_dom; intros l3 v3 E.
    revert E; case_eq (LEM (exists ty2 l2,
                              (` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2))) ty2 l2 l3));
    intros; try (by inversion E).
    by eapply INh3c; eauto.

  set_prop h3g (@Build_heap (fun l3 =>
      match LEM (exists ty2 l2,
                   ` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2)) ty2 l2 l3) with
      | left p => h3c l3
      | _ => None
      end)) h3'_fin.
    destruct (heap_findom h3c) as (h3c_dom, INh3c); exists h3c_dom; intros l3 v3 E.
    revert E; case_eq (LEM (exists ty2 l2,
                              (` (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2))) ty2 l2 l3));
    intros; try (by inversion E).
    by eapply INh3c; eauto.

  assert (Eh1c : hplus h1d h1g h1c).
    intro l1; unfold h1d; simpl.
    by case_eq (LEM (exists ty1 l2, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 l2));
    intros; [left | right].

  assert (Eh2ac : hplus h2ad h2ag h2ac).
    intro l2; unfold h2ad; simpl.
    by case_eq (LEM (exists ty1 l1, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 l2));
    intros; [left | right].

  assert (Eh2bc : hplus h2bd h2bg h2bc).
    intro l2; unfold h2bd; simpl.
    by case_eq (LEM (exists ty2 l3, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l3));
    intros; [left | right].

  assert (Eh3c : hplus h3d h3g h3c).
    intro l3; unfold h3d; simpl.
    by case_eq (LEM (exists ty2 l2, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l3));
    intros; [left | right].

  exists h1d h3d (hunion h1g h1l) (hunion h3g h3l); unnw; split.

  Case "1".
  repeat red; intros; unnw; split; intros.
    rrdes R; assert (R' := R); eapply WH0 in R; assert (R0' := R0); eapply WH in R0'; des.
    eapply hplus_inl in HL'0; eauto; eapply hplus_inl in HL; eauto.
    rewrite HL'0 in HL; inversion HL; subst.
    exploit (bullet_grel1 G); [by econs; eauto|i; des].
    repeat eexists; simpl.
        case_eq (LEM (exists ty1 l2, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 loc l2));
        intros; eauto; exfalso; destruct H; rrdes e.
        by inversion e; eapply NIN; eexists; eauto.
      case_eq (LEM (exists ty2 l2, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 loc'));
      intros; eauto; exfalso; destruct H; rrdes e.
      by inversion e; eapply NIN; eexists; eauto.
    by destruct x0; eauto.

  split; intros; assert (IN' := IN).
    edestruct WH0; des; eapply hplus_inl in IN'; eauto; apply HDOM1 in IN'; des; simpl in IN'.
    simpl in IN; revert IN; case (LEM (exists ty1 l2,
               wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l l2)); intros; vauto.
    do 2 (eapply not_ex_all_not in f); instantiate (1 := l') in f; instantiate (1 := ty) in f.
    assert (X : exists ty' l'', ty = decomp nfst ty' /\ `sr2 (decomp nsnd ty') l' l'').
      apply not_all_not_ex; intro X; apply f.
      apply wrefst_filter_fst_intro; auto; intros.
      specialize (X ty'); destruct FH; apply X; exists l2'.
      assert (EE : ty = decomp nfst ty' /\ l' = x).
        edestruct sr1 as (?, WR); simpls; des.
         rdes WR; des; edestruct WR0; [apply IN' | apply H0 | auto].
      by des; subst.
    by simpl; des; do 2 eexists; subst; eexists; eauto.
  edestruct WH; des; eapply hplus_inl in IN'; eauto; apply HDOM2 in IN'; des; simpl in IN'.
  simpl in IN; revert IN; case (LEM (exists ty2 l2,
             wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l')); intros; vauto.
  do 2 (eapply not_ex_all_not in f); instantiate (1 := l) in f; instantiate (1 := ty) in f.
  assert (X : exists ty' l'', ty = decomp nsnd ty' /\ `sr1 (decomp nfst ty') l'' l).
    apply not_all_not_ex; intro X; apply f.
    apply wrefst_filter_snd_intro; auto; intros.
    specialize (X ty'); destruct FH; apply X; exists l1'; des.
    assert (EE : ty = decomp nsnd ty' /\ l = x).
      edestruct sr2 as (?, WR); simpls; des.
      by rdes WR; des; edestruct WR0; [apply IN' | apply H1 | auto].
    by des; subst.
  by simpl; des; do 2 eexists; subst; eexists; eauto.

  Case "2".
  apply and_comm; split; try split.
      intro l1; simpl; unfold hunion'; simpl.
      case_eq (LEM (exists ty1 l2, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 l2));
      intros; vauto.
      case_eq (h1c l1); intros; [by eapply hplus_inl in H0; eauto |].
      by left; destruct (HP0 l1); des; auto; rewrite H2, H1.
    intro l3; simpl; unfold hunion'; simpl.
    case_eq (LEM (exists ty2 l2, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l3));
    intros; vauto.
    case_eq (h3c l3); intros; [by eapply hplus_inl in H0; eauto |].
    by left; destruct (HP' l3); des; auto; rewrite H2, H1.
  eapply lpw_heap_intro with (h1g := h1g) (h2ag := h2ag) (h3g := h3g) (h2bg := h2bg)
    (h2ab := hrestrict h2 (fun l2 => forall ty l1 l3
                                       (IN1 : `sr1 (decomp nfst ty) l1 l2)
                                       (IN2 : `sr2 (decomp nsnd ty) l2 l3), 
                                     False)); eauto.

  - intro l1; simpl; unfold hunion'; case_eq (h1g l1); case_eq (h1l l1); intros; vauto.
    by exfalso; eapply hplus_inr in H0; eauto; eapply (hplus_disjoint HP0); eauto.

  - intro l2; unfold h2ag; simpl.
    case_eq (h2ac l2); case_eq (h2al l2);
    case_eq (LEM (exists ty1 l1, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 l2));
    unfold hrestrict'; case_eq (LEM
             (forall (ty0 : typ) (l1 l3 : location),
              (`sr1) (decomp nfst ty0) l1 l2 ->
              (`sr2) (decomp nsnd ty0) l2 l3 -> False)); intros; vauto;
    try (by edestruct (hplus_disjoint HP'0); des; eauto);
    try (by eapply hplus_inl in H2; eauto);
    try (by eapply hplus_inr in H1; eauto);
    try (by left; edestruct (HP'0 l2); des; rewrite H3; eauto).
          clear H; do 3 (apply not_all_ex_not in f; des); exfalso; eapply f; intros.
          by destruct e; eapply NIN; destruct (wrefst_inj2 IN H); subst; vauto.
        apply WH0 in H2; des; simpl in H2; clear H0; do 2 (eapply not_ex_all_not in f0).
        exfalso; eapply f0; econstructor; [by apply H2 |].
        by intros; rrdes FH; destruct (wrefst_inj1 H2 FH); subst; eapply f; eauto.
      des; destruct e; exfalso; eapply f; intros; eapply NIN.
      by destruct (wrefst_inj2 IN H3); subst; exists l2; eauto.
    by exfalso; eapply f; intros; eapply WH0 in H3; des; rewrite H2 in HL'.

  - intro l2; unfold h2bg; simpl.
    case_eq (h2bc l2); case_eq (h2bl l2);
    case_eq (LEM (exists ty2 l3, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l3));
    unfold hrestrict'; case_eq (LEM
             (forall (ty0 : typ) (l1 l3 : location),
              (`sr1) (decomp nfst ty0) l1 l2 ->
              (`sr2) (decomp nsnd ty0) l2 l3 -> False)); intros; vauto;
    try (by edestruct (hplus_disjoint HP); des; eauto);
    try (by eapply hplus_inl in H2; eauto);
    try (by eapply hplus_inr in H1; eauto);
    try (by left; edestruct (HP l2); des; rewrite H3; eauto).
          clear H; do 3 (apply not_all_ex_not in f; des); exfalso; eapply f; intros.
          by destruct e; eapply NIN; destruct (wrefst_inj1 IN H3); subst; vauto.
        apply WH in H2; des; simpl in H2; clear H0; do 2 (eapply not_ex_all_not in f0).
        exfalso; eapply f0; econstructor; [by apply H2 |].
          by intros; rrdes FH; destruct (wrefst_inj2 H2 FH0); subst; eapply f; eauto.
      des; destruct e; exfalso; eapply f; intros; eapply NIN.
      by destruct (wrefst_inj1 IN H4); subst; exists l2; eauto.
    by exfalso; eapply f; intros; eapply WH in H4; des; rewrite H2 in HL.

  - intro l3; simpl; unfold hunion'; case_eq (h3g l3); case_eq (h3l l3); intros; vauto.
    by exfalso; eapply hplus_inr in H0; eauto; eapply (hplus_disjoint HP'); eauto.

  - by intros; des; specialize (HP'0 l2); des;
       [ destruct WH0 as (A, ?); unnw; apply A in INS; des; rewrite HP'1 in INS0
       | rewrite HP'1 in INH].
  - by intros; des; specialize (HP l1); des;
       [ by destruct WH as (A, ?); unnw; apply A in INS; des; rewrite HP1 in INS
       | by rewrite HP1 in INH].

  - repeat split; unnw; intros.
    + edestruct R; destruct WH0 as (A, ?); unnw; apply A in IN; des.
      exists n v n' v'; unfold h1g, h2ag; simpl.
      case_eq (LEM (exists ty1 l2, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 loc l2)); intros.
        repeat split; eauto; [|by destruct IN1].
        case_eq (LEM (exists ty1 l1, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 loc')); intros; auto.
        exfalso; eapply f; des; exists ty1 loc; edestruct e, R; edestruct sr1 as (?, WR); simpls; des.
        rdes WR; des; edestruct WR0; [apply IN2 | apply IN3 | auto].
        by specialize (a eq_refl); unnw; des; rewrite <- a1.
      by exfalso; eapply f; vauto.
    + revert IN; simpl.
      by case_eq (LEM (exists ty1 l2, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l l2)); intros; vauto.
    + revert IN; simpl.
      by case_eq (LEM (exists ty1 l1, wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty1 l1 l')); intros; vauto.

  - repeat split; unnw; intros.
    + edestruct R; destruct WH as (A, ?); unnw; apply A in IN; des.
      exists n v n' v'; unfold h3g, h2bg; simpl.
      case_eq (LEM (exists ty2 l3, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 loc l3)); intros.
        repeat split; eauto; [|by destruct IN1].
        case_eq (LEM (exists ty2 l2, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 loc')); intros; auto.
        exfalso; eapply f; des; exists ty2 loc; edestruct e, R; edestruct sr2 as (?, WR); simpls; des.
        rdes WR; des; edestruct WR0; [apply IN2 | apply IN3 | auto].
        by specialize (a eq_refl); unnw; des; rewrite <- a1.
      by exfalso; eapply f; vauto.
    + revert IN; simpl.
      by case_eq (LEM (exists ty2 l3, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l l3)); intros; vauto.
    + revert IN; simpl.
      by case_eq (LEM (exists ty2 l2, wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty2 l2 l')); intros; vauto.
Qed.

Definition wrefst_minus (s1 s2 : wst wref) : wst wref.
Proof.
  refine (exist _ (fun ty l1 l2 => `s1 ty l1 l2 /\ ~ `s2 ty l1 l2) _).

  repeat split; intros; des; destruct s1;
  try (by eapply w; eauto);
  try (by red in w; simpls; des; edestruct w0; [apply IN | apply IN' |]; hdes; eauto).
  rrdes w; exists tydom ldom; intros; apply w1; apply IN.
Defined.

Lemma wheap_pw_lpw_2 :
  forall g (G : wglob (wlift lpw) g) sr1 s1 sr2 s2 h1 h3
         (WH : wheap (wlift lpw) () (wrefst_bullet sr1 sr2, ((sr1,s1),(sr2,s2))) (g (wrefst_bullet sr1 sr2, ((sr1,s1),(sr2,s2)))) h1 h3),
  wheap pw () ((sr1,s1),(sr2,s2)) (g (wrefst_bullet sr1 sr2, ((sr1,s1),(sr2,s2)))) h1 h3.
Proof.
  i; destruct WH as (h1d & h3d & h1c & h3c & ?); des.
  inversion WH'; subst; simpl in SR.

  assert (mediate : forall ty l2
        l1 (IN1 : ` sr1 (decomp nfst ty) l1 l2)
        l3 (IN2 : ` sr2 (decomp nsnd ty) l2 l3),
      exists v2 v1 v3 n1 n2 n2' n3,
        h1d l1 = Some v1 /\ h3d l3 = Some v3 /\
        vclos (grel_fst (grel_pw g) (sr2, s2) (sr1, s1)) (decomp nfst ty) n1 v1 n2 v2 /\
        vclos (grel_snd (grel_pw g) (sr1, s1) (sr2, s2)) (decomp nsnd ty) n2' v2 n3 v3).
    intros; destruct WH as (WH1 & WH2 & WH3).
    edestruct WH1 as (v1 & v3 & ?); des.
      by simpl; exists l2; eauto.
    exploit @bullet_grel2.
        by apply grel_pw_glob; eauto.
      by fold (fst (sr1, s1)) in VC; fold (fst (sr2, s2)) in VC; apply VC.
    by i; destruct x0; esplits; eauto.

  set (sr1' := wrefst_minus sr1 (wrefst_filter_fst sr1 (wrefst_bullet sr1 sr2))).
  set (sr2' := wrefst_minus sr2 (wrefst_filter_snd sr2 (wrefst_bullet sr1 sr2))).

  assert (H2D : forall l2, exists sv2, 
      (forall ty l1 l3
          (IN1 : (`sr1) (decomp nfst ty) l1 l2) 
          (IN2 : (`sr2) (decomp nsnd ty) l2 l3),
        exists v2 v1 v3 n1 n2 n2' n3,
          sv2 = Some v2 /\ h1d l1 = Some v1 /\ h3d l3 = Some v3 /\
          vclos (grel_fst (grel_pw g) (sr2, s2) (sr1, s1)) (decomp nfst ty) n1 v1 n2 v2 /\
          vclos (grel_snd (grel_pw g) (sr1, s1) (sr2, s2)) (decomp nsnd ty) n2' v2 n3 v3)
      /\
      (forall v2 (ES : sv2 = Some v2), exists ty l1 l3,
        (`sr1) (decomp nfst ty) l1 l2 /\ (`sr2) (decomp nsnd ty) l2 l3)).
  {
    intros; destruct (classic (exists ty l1 l3, (`sr1) (decomp nfst ty) l1 l2
                                             /\ (`sr2) (decomp nsnd ty) l2 l3)) as [IN|NIN].
    - des; assert (X:= mediate _ _ _ IN _ IN0); des; exists (Some v2); split; eauto; intros.
      destruct sr1, sr2; simpls; destruct w as (? & B1 & ?); destruct w0 as (? & B2 & ?).
      edestruct B1; [by apply IN | by apply IN1 |].
      edestruct B2; [by apply IN0 | by apply IN2 |].
      specialize (a0 eq_refl); specialize (a1 eq_refl).
      specialize (a (proj2 a0)); specialize (a2 (proj2 a1)).
      by des; assert (Y := decomp_inj _ _ TEQ2 TEQ0); subst; esplits; eauto.
    - by eexists None; split; [intros; exfalso; apply NIN; eauto | vauto].
  }

  destruct (choice _ H2D) as (h2df & H2DF); clear H2D.

  assert (h2df_fin_aux : forall l1 (IN : h1d l1 <> None), exists l2 l3 ty,
                           `sr1 (decomp nfst ty) l1 l2 /\ `sr2 (decomp nsnd ty) l2 l3).
    intros; destruct WH as (WH1 & WH2 & WH3).
    case_eq (h1d l1); vauto; intros.
    by edestruct WH2 as (ty' & l3 & (l2 & IN1 & IN2)); eauto.

  set_prop h2d (@Build_heap h2df) h2df_fin.
    destruct (heap_findom h1d) as (h1d_dom, DOMh1d).
    exists (fold_right (fun l1 ls => match LEM (h1d l1 <> None) with
                                     | left p => proj1_sig (cid _ (h2df_fin_aux l1 p)) :: ls
                                     | _ => ls
                                     end) nil h1d_dom).
    intros l2 v2 IN; apply H2DF in IN; des.
    edestruct H2DF as (A, _); specialize (A _ _ _ IN IN0); des.
    specialize (DOMh1d _ _ A0); apply in_split in DOMh1d; des.
    rewrite DOMh1d; clear DOMh1d; induction l0; simpl.
      case (LEM (h1d l1 <> None)); intro.
      2 : by exfalso; eapply f; rewrite A0; intro; inversion A3.
      left.
      set (X := cid _ (h2df_fin_aux l1 n)).
      destruct X as (l2' & l3' & ty' & IN1 & IN2); simpl.
      edestruct sr1 as (?, WR); simpls.
      rdes WR; des; edestruct WR0; [by apply IN | by apply IN1 |].
      by unnw; specialize (a eq_refl); des; rewrite a1.
    by case (LEM (h1d a <> None)); intro; auto; right.

  assert (WH1d2d : wheap wref () sr1' (grel_fst (grel_pw g) (sr2, s2) (sr1, s1)) h1d h2d).
  {
    repeat split; unnw; intros.
    - specialize (H2DF loc'); rrdes R; simpl in R0.
      assert (X : exists ty' loc'', ty = decomp nfst ty' /\ `sr2 (decomp nsnd ty') loc' loc'').
        apply not_all_not_ex; intro X; apply R0.
        apply wrefst_filter_fst_intro; auto; intros.
        specialize (X ty'); destruct FH; apply X; exists l2'.
        assert (EE : ty = decomp nfst ty' /\ loc' = x).
          edestruct sr1 as (?, WR); simpls; des.
          rdes WR; des; edestruct WR0; [by apply R | by apply H |].
          by auto.
        by des; subst.
      by des; subst; edestruct H2DF; des; esplits; eauto.
    - eapply WH in IN; rrdes IN; simpls.
      rdes IN; eexists; eexists; split; eauto; intro X.
      by destruct X; eapply NIN; eexists; eauto.
    - eapply H2DF in IN; des; simpl.
      eexists; eexists; split; eauto; intro X.
      by destruct X; eapply NIN; eexists; eauto.
  }

  assert (WH2d3d : wheap wref () sr2' (grel_snd (grel_pw g) (sr1, s1) (sr2, s2)) h2d h3d).
  {
    repeat split; unnw; intros.
    - specialize (H2DF loc); rrdes R; simpl in R0.
      assert (X : exists ty' loc'', ty = decomp nsnd ty' /\ `sr1 (decomp nfst ty') loc'' loc).
        apply not_all_not_ex; intro X; apply R0.
        apply wrefst_filter_snd_intro; auto; intros.
        specialize (X ty'); destruct FH; apply X; exists l1'.
        assert (EE : ty = decomp nsnd ty' /\ loc = x).
          edestruct sr2 as (?, WR); simpls; des.
          rdes WR; des; edestruct WR0; [by apply R | by apply H0 |].
          by auto.
        by des; subst.
       by des; subst; edestruct H2DF; des; esplits; eauto.
    - eapply H2DF in IN; des; simpl.
      eexists; eexists; split; eauto; intro X.
      by destruct X; eapply NIN; eexists; eauto.
    - eapply WH in IN; rrdes IN; simpls.
      rdes IN; eexists; eexists; split; eauto; intro X.
      by destruct X; eapply NIN; eexists; eauto.
  }

  set (h1gd := hunion h1g h1d).
  assert (Eh1gd : hplus h1g h1d h1gd).
    intro l1; simpl; unfold hunion'.
    case_eq (h1g l1); intros; case_eq (h1d l1); intros; try (by right); try (by left).
    exfalso; destruct HR1 as (A1 & A2 & A3); unnw.    
    edestruct A2 as (? & ? & ?); eauto; destruct WH1d2d as (B1 & B2 & B3); unnw.
    edestruct B2 as (? & ? & ?); eauto; rrdes H2; edestruct H1.
    destruct sr1; simpls; edestruct w as (? & B & ?).
    edestruct B; [by apply H2 | by apply IN |].
    specialize (a eq_refl); des; specialize (a0 LEQ); des; subst.
    by contradiction. 

  assert (Eh1 : hplus h1gd h1l h1).
    eapply hplus_comm, hplus_assoc.
        by apply hplus_comm, H7.
      by apply hplus_comm, HP.
    by auto.

  set (h3gd := hunion h3g h3d).
  assert (Eh3gd : hplus h3g h3d h3gd).
    intro l3; simpl; unfold hunion'.
    case_eq (h3g l3); intros; case_eq (h3d l3); intros; try (by right); try (by left).
    exfalso; destruct HR2 as (A1 & A2 & A3); unnw.    
    edestruct A3 as (? & ? & ?); eauto; destruct WH2d3d as (B1 & B2 & B3); unnw.    
    edestruct B3 as (? & ? & ?); eauto; rrdes H2; edestruct H1.
    destruct sr2; simpls; edestruct w as (? & B & ?).
    edestruct B; [by apply H2 | by apply IN |].
    specialize (a0 eq_refl); des; specialize (a LEQ); des; subst.
    by contradiction. 

  assert (Eh3 : hplus h3gd h3l h3).
    eapply hplus_comm, hplus_assoc.
        by apply hplus_comm, H8.
      by apply hplus_comm, HP'.
    by auto.

  set (h2agd := hunion h2ag h2d).
  assert (Eh2agd : hplus h2ag h2d h2agd).
    intro l2; simpl; unfold hunion'.
    case_eq (h2ag l2); intros; case_eq (h2df l2); intros; try (by right); try (by left).
    exfalso; destruct HR1 as (A1 & A2 & A3); unnw.
    edestruct A3 as (? & ? & ?); eauto; apply H2DF in H0; des.
    destruct H1; simpls; eapply NIN; edestruct (proj2_sig sr1) as (? & B & ?).
    edestruct B as (X & Y); [by apply H0 | by apply IN |].
    specialize (Y eq_refl); des; specialize (X LEQ); des; subst.
    by red; eauto.

  set (h2bgd := hunion h2bg h2d).
  assert (Eh2bgd : hplus h2bg h2d h2bgd).
    intro l2; simpl; unfold hunion'.
    case_eq (h2bg l2); intros; case_eq (h2df l2); intros; try (by right); try (by left).
    exfalso; destruct HR2 as (A1 & A2 & A3); unnw.
    edestruct A2 as (? & ? & ?); eauto; apply H2DF in H0; des.
    destruct H1; simpls; eapply NIN; edestruct (proj2_sig sr2) as (? & B & ?).
    edestruct B as (X & Y); [by apply H2 | by apply IN |].
    specialize (X eq_refl); des; specialize (Y LEQ); des; subst.
    by red; eauto.

  set (h2 := hunion h2agd h2al).
  assert (EAh2 : hplus h2agd h2al h2).
    intro l2; specialize (Eh2agd l2); revert Eh2agd.
    unfold h2, h2agd; repeat (simpl; unfold hunion').
    case_eq (h2ag l2); case_eq (h2al l2); case_eq (h2df l2); intros; des; vauto.
      by exfalso; specialize (H2a l2); des; vauto.
    edestruct DJ1 with l2.
      by intro; rewrite H0 in H2; vauto.
    by apply H2DF in H; des; eauto.
  assert (EBh2 : hplus h2bgd h2bl h2).
    eapply hplus_comm, hplus_assoc.
        by apply hplus_comm, H2b.
      eapply hplus_comm, hplus_assoc; [| | by apply H2a].
        by apply hplus_comm, Eh2agd.
      by auto.
    by auto.

  exists h2; split; [exists h1gd h2agd h1l h2al | exists h2bgd h3gd h2bl h3l]; split; unnw.

  - repeat red; unnw; repeat split; intros; simpls.
    + destruct (classic (wrefst_filter_fst' sr1 (wrefst_bullet sr1 sr2) ty loc loc')).
        apply HR1 in H; des.
        esplits.
            by eapply (hunion_inl h1g h1d loc); eauto.
          by eapply (hunion_inl h2ag h2d loc'); eauto.
        by destruct VC; eauto.
      des; edestruct HSAT0; [by split; eauto | des].
      exists x v n' v'; repeat split.
          by eapply (@hunion_inr h1g h1d loc); eauto.
        by eapply (@hunion_inr h2ag h2d loc'); eauto.
      by destruct VC; eauto.
    + revert IN; unfold hunion'; case_eq (h1g l); intros.
        by apply HR1 in H; des; inversion H; eauto.
      by apply WH1d2d in IN; des; eauto.
    + revert IN; unfold hunion'.
      case_eq (h2ag l'); intros.
        apply HR1 in H; des.
        by inversion H; eauto.
      by apply WH1d2d in IN; des; eauto.

  - by split.

  - repeat red; unnw; repeat split; intros; simpls.
    + destruct (classic (wrefst_filter_snd' sr2 (wrefst_bullet sr1 sr2) ty loc loc')).
        apply HR2 in H; des.
        esplits.
            by eapply (hunion_inl h2bg h2d loc); eauto.
          by eapply (hunion_inl h3g h3d loc'); eauto.
        by destruct VC; eauto.
      des; edestruct HSAT; [by split; eauto | des].
      exists x v n' v'; repeat split.
          by eapply (@hunion_inr h2bg h2d loc); eauto.
        by eapply (@hunion_inr h3g h3d loc'); eauto.
      by destruct VC; eauto.
    + revert IN; unfold hunion'; case_eq (h2bg l); intros.
        by apply HR2 in H; des; inversion H; eauto.
      by apply WH2d3d in IN; des; eauto.
    + revert IN; unfold hunion'.
      case_eq (h3g l'); intros.
        apply HR2 in H; des.
        by inversion H; eauto.
      by apply WH2d3d in IN; des; eauto.

  - by split.
Qed.

Lemma wexp_pw_lpw :
  forall g (G : wglob (wlift lpw) g) sr1 sr2 s1 s2 sr1' sr2' s1' s2' ty nf1 e1 nf3 e3
         (WE : wexp pw (grel_pw g) ((sr1,s1),(sr2,s2)) ((sr1',s1'),(sr2',s2')) ty nf1 e1 nf3 e3),
    wexp (wlift lpw) g (wrefst_bullet sr1 sr2, ((sr1,s1),(sr2,s2))) (wrefst_bullet sr1' sr2', ((sr1',s1'),(sr2',s2'))) ty nf1 e1 nf3 e3.
Proof.
  pcofix CIH with z.
  ii; eapply wheap_pw_lpw_2 in HREL; eauto.
  edestruct WE; eauto.
  destruct SIM; [by eauto| |]; destruct s' as [[sr1'' s1''] [sr2'' s2'']].
  - pfold; econs 2.
    + instantiate (1 := (wrefst_bullet sr1'' sr2'', ((sr1'', s1''), (sr2'', s2'')))).
      destruct EXT as [[] []]; repeat split; eauto using wrefst_bullet_mon.
    + destruct EXTP as [[] []]; repeat split; eauto using wrefst_bullet_mon.
      eapply wrefst_bullet_mon; eauto.
    + eauto.
    + change (g (wrefst_bullet sr1'' sr2'', (sr1'', s1'', (sr2'', s2''))))
      with (grel_pw g (sr1'', s1'', (sr2'', s2''))) in *. (* ???!?!?! *)
      eapply wheap_pw_lpw_1; eauto using grel_pw_glob.
    + eauto. + eauto. + eauto. + eauto.
  - pfold; econs 3.
    + instantiate (1 := (wrefst_bullet sr1'' sr2'', ((sr1'', s1''), (sr2'', s2'')))).
      destruct EXT as [[] []]; repeat split; eauto using wrefst_bullet_mon.
    + change (g (wrefst_bullet sr1'' sr2'', (sr1'', s1'', (sr2'', s2''))))
      with (grel_pw g (sr1'', s1'', (sr2'', s2''))) in *. (* ???!?!?! *)
      eapply wheap_pw_lpw_1; eauto using grel_pw_glob.
    + inversion EREL; [econs 1; [eapply EQ1|eapply EQ2| |]| ]; eauto.
    + ii; destruct s' as [srr [[sr11 s11] [sr22 s22]]].
      assert (EE : srr = wrefst_bullet sr11 sr22).
      { edestruct HREL1; des; inversion WH'; ss; subst; eauto. }
      rewrite EE in *.
      right; eapply CIH; eauto.
      assert (WX : wexp_u pw pco (grel_pw g') (grel_pw g') (sr1,s1,(sr2,s2))
              (sr11,s11,(sr22,s22)) ty (nk1+nv) (app_cont K1 v) (nk2+nv') (app_cont K2 v')).
      {
        edestruct EXTP, EXT0; eapply KREL; eauto using grel_pw_glob.
        i; eapply LT; eauto.
      }
      assert (KREL' : wcont_u pw (wexp_core pw) (grel_pw g) (sr1, s1, (sr2, s2))
           (sr1'', s1'', (sr2'', s2'')) (sr1'', s1'', (sr2'', s2'')) ty' ty
           nk1 K1 nk2 K2).
      {
        ii. r.
edestruct LE.
eapply KREL; eauto. eauto. vauto.
      }
      edestruct EXTP, EXT0; eapply KREL'; eauto using grel_pw_glob.
      i; eapply LT; eauto.
    + eauto. + eauto. + eauto. + eauto.
Qed.

Lemma wequiv_pw_lpw : wequiv pD pG pw pty <4= wequiv pD pG (wlift lpw) pty.
Proof.
  i; inv PR; econs; eauto.
  ii; destruct s as [sr [[sr1 s1] [sr2 s2]]].
  assert (SR : sr = wrefst_bullet sr1 sr2).
    by inv HREL; des; inv WH'.
  rewrite SR in *; eapply wexp_pw_lpw; eauto.
  eapply WEXP; eauto; apply (grel_pw_glob gOK).
Qed.

Lemma lpw_consistent : wconsistent (wlift lpw).
Proof.
  ii; rename x0 into ty, x1 into n1, x2 into e1, x3 into n3, x4 into e3.
  assert (GT: n1 > 0 /\ n3 > 0).
    by depdes PR; s in FREL; depdes FREL; des; try destruct H; omega.
  des; econs; [|instantiate (2:=n1-1); eauto|instantiate (2:=n3-1); eauto].
  ii; destruct s as [sr [[sr1 s1] [sr2 s2]]].
  assert (SR : sr = wrefst_bullet sr1 sr2).
    by inv HREL; des; inv WH'.
  subst; eapply wexp_pw_lpw; eauto.
  rewrite <- wval_pw_lpw in PR.
  change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2))))
  with (grel_pw g (sr1, s1, (sr2, s2))) in *. (* ???!?!?! *)
  eapply pw_consistent in PR; [|by eapply grel_pw_glob].
  destruct PR; ii. 
  eapply wexp_core_lreds; [eapply R| | |]; eauto using lred_lreds.
Qed.

Lemma lpw_inhabited : winhabited (wlift lpw).
Proof.
  ii.
  edestruct pw_inhabited; eauto using grel_pw_glob.
  destruct x as [[sr1 s1] [sr2 s2]].
  exists (wrefst_bullet sr1 sr2, ((sr1, s1), (sr2, s2))).
  change (g (wrefst_bullet sr1 sr2, (sr1, s1, (sr2, s2))))
  with (grel_pw g (sr1, s1, (sr2, s2))) in *. (* ???!?!?! *)
  eapply wheap_pw_lpw_1; eauto using grel_pw_glob.
Qed.

Lemma weq_pw_lpw : weq pD pG pw pty <4= weq pD pG (wlift lpw) pty.
Proof.
  ii; inv PR; rr; esplits; des
  ; eauto using lpw_consistent, lpw_inhabited, wequiv_pw_lpw.
Qed.

End trans_part2. (************************************************************************)
