Require Import vgtac Program List Arith Even Div2 Relations.
Require Import basis countable lang subs model.
Require Import Classical FunctionalExtensionality.

Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** ** automatically apply omega  *)
(*----------------------------------------------------------------------------*)

Hint Extern 3 (@eq nat _ _) => omega.
Hint Extern 3 (le _ _) => omega.
Hint Extern 3 (ge _ _) => omega.
Hint Extern 3 (lt _ _) => omega.
Hint Extern 3 (gt _ _) => omega.

(*----------------------------------------------------------------------------*)
(** ** Macros about Paco  *)
(*----------------------------------------------------------------------------*)

Tactic Notation "pdes" ident(H) := punfold H; destruct H; pclearbot.
Tactic Notation "pinv" ident(H) := punfold H; inv H; pclearbot.
Tactic Notation "pdepdes" ident(H) := punfold H; dependent destruction H; pclearbot.

Lemma or_false: forall P, P \/ False -> P.
Proof. intuition. Qed.

Lemma wcgen_elim_bot: forall (w: world_data ()) 
    (r: (dwst w -> know) -> dwst w -> 
        typ -> nat -> val + cont -> nat -> val + cont -> Prop),
  r \7/ bot7 = r.
Proof.
  i; extensionalities a b c d e f; extensionality h; apply prop_ext.
  by split; i; des; eauto. 
Qed.

(*----------------------------------------------------------------------------*)
(** ** Injectivities *)
(*----------------------------------------------------------------------------*)

Lemma exp_of_val_inj: forall v v', exp_of_val v = exp_of_val v' -> v = v'.
Proof. 
  induction v; destruct v'; intross; clarify; try by rewrite (IHv _ H).
  by rewrite (IHv1 _ H1), (IHv2 _ H).
Qed.

Lemma e_of_inj :
  forall v v', e_of v = e_of v' -> v_of v = v_of v'.
Proof. by intros; destruct v, v'; subst; apply exp_of_val_inj in H; simpls; subst. Qed.

Lemma v_of_inj :
  forall v v', v_of v = v_of v' -> v = v'.
Proof. by intros; destruct v, v'; simpls; subst. Qed.

Ltac simpl_inj :=
  repeat match goal with 
  | H : v_of _ = _ |- _ => 
      try apply v_of_inj in H; let P := type of H in fold (__HID__ P) in H
  | H : _ = v_of _ |- _ => 
      try apply v_of_inj in H; let P := type of H in fold (__HID__ P) in H
  | H : _ = _ |- _ => 
      try apply e_of_inj in H; let P := type of H in fold (__HID__ P) in H
  end; unfold __HID__ in *.
Ltac simpl_injs := simpl_inj; simpl v_of in *.

(*----------------------------------------------------------------------------*)
(** ** Properties of heaps *)
(*----------------------------------------------------------------------------*)

Definition hunion' (h1 h2 : heap) l :=
  match h1 l with Some v1 => Some v1 | _ => h2 l end.

Definition hunion (h1 h2 : heap) : heap.
  refine (@Build_heap (hunion' h1 h2) _).
  intros.
  destruct (heap_findom h1) as (dom1, ?); destruct (heap_findom h2) as (dom2, ?).
  exists (app dom1 dom2); intros l ?.
  by unfold hunion'; case_eq (h1 l); intros; apply in_or_app; [left | right]; eauto.
Defined.

Definition hrestrict' (h : heap) (dom : location -> Prop) l :=
  match LEM (dom l) with
  | left _ => h l
  | right _ => None
  end.

Definition hrestrict (h  : heap )(dom : location -> Prop) : heap.
  refine (@Build_heap (hrestrict' h dom) _).
  destruct (heap_findom h) as (hdom, ?); exists hdom; unfold hrestrict'.
  by intro l; case_eq (LEM (dom l)); intros; eauto.
Defined.

Lemma heap_eq : forall h1 h2, (forall x, heap_f h1 x = heap_f h2 x) -> h1 = h2.
Proof. 
  intros [] [] ?; simpl in *; apply functional_extensionality in H.
  subst; f_equal; apply proof_irr.
Qed. 

Lemma hplus_det : 
  forall h1 h2 h (H: hplus h1 h2 h) h' (H': hplus h1 h2 h'), h' = h.
Proof. 
  intros; apply heap_eq; intro x.
  destruct (H x); destruct (H' x); des; congruence.
Qed.

Lemma hplus_hemp_l : forall h, hplus hemp h h.
Proof. vauto. Qed.

Lemma hplus_hemp_r : forall h, hplus h hemp h.
Proof. vauto. Qed.

Lemma hplus_hemp_lE : forall h h', hplus hemp h h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_rE : forall h h', hplus h hemp h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_inv : forall h h', hplus h h' hemp -> h = hemp /\ h' = hemp.
Proof.
  by intros; split; eapply heap_eq; intros x; specialize (H x); intuition.
Qed.

Lemma hplus_comm : forall h1 h2 h, hplus h1 h2 h -> hplus h2 h1 h.
Proof. by red; intros; edestruct H; vauto. Qed.

Lemma hplus_assoc_l : 
  forall h1 h2 h12 (R: hplus h1 h2 h12) h3 h (R': hplus h12 h3 h),
    exists h23, hplus h1 h23 h /\ hplus h2 h3 h23. 
Proof.
  intros; eexists (hunion h2 h3); split; intro x; simpl; unfold hunion';
    specialize (R x); specialize (R' x); destruct (h2 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_assoc_r : 
  forall h2 h3 h23 (R: hplus h2 h3 h23) h1 h (R': hplus h1 h23 h),
    exists h12, hplus h1 h2 h12 /\ hplus h12 h3 h. 
Proof.
  intros; eexists (hunion h1 h2); split; intro x; simpl; unfold hunion';
    specialize (R x); specialize (R' x); destruct (h1 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_none :
  forall h1 h2 h (P: hplus h1 h2 h) loc (N: h loc = None),
    h1 loc = None /\ h2 loc = None.
Proof.
  intros; specialize (P loc); simpls; des; split; congruence.
Qed.

Lemma hplus_some1 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h1 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_some2 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h2 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_inl : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h1 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP0 in EH | rewrite HP].
Qed.

Lemma hplus_inr : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h2 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP | rewrite HP0 in EH].
Qed.

Lemma hplus_assoc : forall h1 h2 h3 h12 h23 h123
    (HP12 : hplus h1 h2 h12)
    (HP123 : hplus h12 h3 h123)
    (HP23 : hplus h2 h3 h23),
  hplus h1 h23 h123.
Proof.
  intros; assert (HP := hplus_assoc_l HP12 HP123); des.
  by rewrite (hplus_det HP0 HP23).
Qed.

Lemma hplus_disjoint : forall h1 h2 h12 (HP : hplus h1 h2 h12) l
    v1 (IN1 : h1 l = Some v1) v2 (IN2 : h2 l = Some v2),
  False.
Proof.
  by intros; destruct (HP l); des; [rewrite H0 in IN1 | rewrite H0 in IN2].
Qed.

Lemma hunion_inl : forall (h1 h2 : heap) l v (INL : h1 l = Some v),
  hunion h1 h2 l = Some v.
Proof.
  by intros; simpl; unfold hunion'; rewrite INL.
Qed.

Lemma hunion_inr : forall (h1 h2 : heap) l v (INR : h2 l = Some v) h12 (DJ : hplus h1 h2 h12),
  hunion h1 h2 l = Some v.
Proof.
  intros; simpl; unfold hunion'; specialize (DJ l); revert DJ.
  by rewrite INR; case_eq (h1 l); intros; des; vauto.
Qed.

Hint Immediate hplus_hemp_l hplus_hemp_r hplus_comm. 

Lemma hext_lookup_new: forall h hh v,
  hext h (allocator hh) v (allocator hh) = Some v.
Proof.
  by i; s; destruct eq_nat_dec.
Qed.

Lemma hext_lookup_old: forall h hf hh loc v v'
    (HPLUS: hplus h hf hh)
    (LOOKUP: h loc = Some v),
  hext h (allocator hh) v' loc = Some v.
Proof.
  i; s; destruct eq_nat_dec; eauto. 
  subst; eapply hplus_some1 in LOOKUP; eauto. 
  by rewrite allocator_unalloc in LOOKUP.
Qed.

Lemma hext_eq: forall h hf hh loc v
    (HPLUS: hplus h hf hh)
    (LOOKUP: h loc = Some v),
  hh = hext hh loc v.
Proof.
  i; eapply heap_eq; i.
  s; destruct eq_nat_dec; subst; eauto using hplus_some1.
Qed.

Lemma hext_plus: forall h hf hh loc v v'
    (HPLUS: hplus h hf hh)
    (LOOKUP: h loc = Some v),
  hplus (hext h loc v') hf (hext hh loc v').
Proof.
  ii; specialize (HPLUS x).
  s; destruct eq_nat_dec; subst; eauto.
  des; eauto.
  by rewrite LOOKUP in HPLUS0.
Qed.

Lemma allocator_unalloc': forall h hf hh
    (HPLUS: hplus h hf hh),
  h (allocator hh) = None.
Proof.
  i; assert (X:= allocator_unalloc hh).
  eapply hplus_none in X; des; eauto.
Qed.

Ltac clarify_heap :=
  clarify; repeat match goal with
  | H: hplus hemp _ _ |- _ => apply hplus_hemp_lE in H; clarify
  | H: hplus _ hemp _ |- _ => apply hplus_hemp_rE in H; clarify
  | H: hplus ?x ?y _, H' : hplus ?x ?y _ |- _ => apply (hplus_det H) in H'; clarify
  end.

(*----------------------------------------------------------------------------*)
(** ** Properties of the underlying untyped language *)
(*----------------------------------------------------------------------------*)

Inductive singknow ty n1 vk1 n2 vk2 : know := 
_singknow : forall n1' n2' (NLE1: n1' >= n1) (NLE2: n2' >= n2), 
            singknow ty n1 vk1 n2 vk2 ty n1' vk1 n2' vk2.
Hint Constructors singknow.

Definition sysknow : know := fun ty _ vk1 _ vk2 => 
  ty = Tint /\ vk1 = inr [] /\ vk2 = inr [].
Hint Unfold sysknow.

Definition is_val e := exists ve : voe, e = e_of ve.
Hint Unfold is_val.

Definition is_atom e := 
  forall sk e' (EQ: e = appctx_scont_exp sk e') (NV: ~ is_value_of_exp e'), False.

Inductive reduces h e : heap -> exp -> Prop :=
| reduces_nil : reduces h e h e
| reduces_step : forall h1 e1 h2 e2
    (STEP: reduce h e h1 e1)
    (RED: reduces h1 e1 h2 e2),
  reduces h e h2 e2.
Hint Resolve reduces_step.

Inductive diverges_gen diverges h e : Prop :=
| diverges_step : forall h' e' 
    (STEP: reduce h e h' e')
    (DIV: diverges h' e': Prop), 
  diverges_gen diverges h e.
Hint Constructors diverges_gen.

Definition diverges h e := paco2 diverges_gen bot2 h e.
Hint Unfold diverges.
Lemma diverges_gen_mon: monotone2 diverges_gen. Proof. pmonauto. Qed.
Hint Resolve diverges_gen_mon : paco.

Lemma singknow_closed: forall ty n vk n' vk'
    (CLO1: forall x, ~ fv_exp (vk2exp vk) x)
    (CLO2: forall x, ~ fv_exp (vk2exp vk') x),
  know_closed (singknow ty n vk n' vk').
Proof. ii; depdes V; des; eauto. Qed.

Lemma voe_is_val: forall e, is_value_of_exp e -> is_val e.
Proof.
  i; induction e; eauto; ss; des
  ; try edestruct IHe; try edestruct IHe1; try edestruct IHe2; subst; eauto.
Qed.  

Lemma val_is_voe: forall e, is_val e -> is_value_of_exp e.
Proof.
  i; rdes H; subst; destruct ve; subst; s; induction v_of; s; eauto.
Qed.

Lemma value_cont : forall k e (V: is_value_of_exp (app_cont k e)), is_value_of_exp e.
Proof.
  i; apply voe_is_val in V; apply val_is_voe; revert e V.
  induction k; eauto; i; ss; eapply IHk in V.
  rdes V; destruct ve; subst; destruct a, v_of; ss; inv V; eauto.
Qed.

Lemma app_cont_assoc: forall k1 k2 e,
   app_cont (k1 ++ k2) e = app_cont k2 (app_cont k1 e).
Proof. by intros; unfold app_cont; rewrite fold_left_app. Qed.

Lemma app_scont_eq: forall sk sk' e e'
    (EQ: appctx_scont_exp sk e = appctx_scont_exp sk' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e'),
  sk = sk' /\ e = e'.
Proof.
  i; assert (NOV1 := fun x => NOV (val_is_voe x)); clear NOV.
  assert (NOV2 := fun x => NOV' (val_is_voe x)); clear NOV'.
  by intros; destruct sk, sk'; try inv EQ;
   try (exfalso; first [by apply NOV1; eauto|by apply NOV2; eauto]); 
   simpl_injs; subst.
Qed.

Lemma atom_generalize: forall e (ATM: is_atom e) 
    k e' (EQ: e = app_cont k e') (NV: ~ is_value_of_exp e'),
  k = [].
Proof.
  induction k; eauto; intros.
  exploit (IHk _ EQ); eauto using (value_cont [a]).
  intros; subst; exfalso; eapply ATM; simpls. 
Qed.

Lemma rev_revert: forall A l (P: list A -> Prop) (ASM: forall l, P (rev l)), P l.
Proof. by intros; specialize (ASM (rev l)); rewrite rev_involutive in ASM. Qed.

Lemma app_cont_r_eq : forall k k' e e'
    (EQ: app_cont_r k e = app_cont_r k' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e')
    (ATM1 : is_atom e)
    (ATM2 : is_atom e'),
  k = k' /\ e = e'.
Proof.
  intros; assert (ATOM1 := atom_generalize ATM1); assert (ATOM2 := atom_generalize ATM2).
  clear ATM1 ATM2; revert_until k; induction k; intros.
  - destruct k'; eauto.
    exploit (ATOM1 (rev (s::k')) e'); eauto. 
    : by rewrite <-app_cont_r_reverse; eauto.
    intro X; assert (Y:= rev_involutive (s::k')); rewrite X in Y; inv X.
  - destruct k'.
    + exploit (ATOM2 (rev (a::k)) e); eauto. 
      : by rewrite <-app_cont_r_reverse; eauto.
      intro X; assert (Y:= rev_involutive (a::k)); rewrite X in Y; inv X.
    + simpl in EQ; eapply app_scont_eq in EQ.
      * des; subst; eapply IHk in EQ0; des; subst; eauto. 
      * rewrite app_cont_r_reverse; eauto using value_cont.
      * rewrite app_cont_r_reverse; eauto using value_cont.
Qed.

Lemma app_cont_eq : forall k k' e e'
    (EQ: app_cont k e = app_cont k' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e')
    (ATM1 : is_atom e)
    (ATM2 : is_atom e'),
  k = k' /\ e = e'.
Proof.
  intro k; apply (rev_revert k); clear k; intro k.
  intro k'; apply (rev_revert k'); clear k'; intro k'.
  setoid_rewrite <- app_cont_r_reverse.
  by intros; exploit app_cont_r_eq; eauto; intro; des; subst.
Qed.

Lemma not_reduceb_val :
   forall h ve h' e', ~ reduceb h (e_of ve) h' e'.
Proof.
  intros; remember (e_of ve) as e; revert ve Heqe.
  induction FH; try by intros []; intros; subst; destruct v_of.
  intros.
  cut (exists v: val, e = v); [by intro; des; eauto|].
  revert ve Heqe; clear.
  destruct SK; simpls; intros; eauto; 
   try by destruct ve; subst; destruct v_of; simpls; clarify; eauto.
Qed.

Lemma not_reduce_val :
   forall h ve h' e', ~ reduce h (e_of ve) h' e'.
Proof.
  i; depdes FH; eauto using not_reduceb_val.
  - cut (is_value_of_exp (Ecallcc x0 e)); eauto.
    eapply (value_cont K); rewrite x; eauto using val_is_voe.
  - cut (is_value_of_exp (Ethrow v (Econt K'))); eauto.
    eapply (value_cont K); rewrite x; eauto using val_is_voe.
Qed.

Lemma reduceb_not_val :
  forall h e h' e', reduceb h e h' e' -> ~ is_value_of_exp e.
Proof.
  i; apply voe_is_val in FH; rdes FH; eapply not_reduceb_val; subst; eauto.
Qed.

Lemma reduce_not_val :
  forall h e h' e', reduce h e h' e' -> ~ is_value_of_exp e.
Proof. 
  i; depdes H; [eauto using reduceb_not_val| |].
  - apply value_cont in FH; eauto using val_is_voe.
  - apply value_cont in FH; eauto using val_is_voe.
Qed.

Lemma reduceb_app_cont_inv: forall K h e h' e' 
    (R : reduceb h (app_cont K e) h' e') 
    (EXP: ~ is_value_of_exp e),
  exists e'', reduceb h e h' e''.
Proof.
  induction K; eauto.
  intros; simpls; exploit IHK; eauto using (value_cont [a]).
  intros; des; depdes x0 
  ; try by destruct a; depdes x; ss; exfalso; apply EXP; esplits; eauto using val_is_voe.
  apply app_scont_eq in x; des; subst; eauto using reduce_not_val.
Qed.

Lemma reduce_app_cont_inv: forall K h e h' e' 
    (R : reduce h (app_cont K e) h' e') 
    (EXP: ~ is_value_of_exp e),
  exists e'', reduce h e h' e''.
Proof.
  induction K; eauto.
  intros; simpls; exploit IHK; eauto using (value_cont [a]).
  intros; des; depdes x0.
  - depdes RED; try by destruct a; depdes x; ss
                       ; exfalso; apply EXP; esplits; eauto using val_is_voe.
    apply app_scont_eq in x; des; subst; eauto using reduce_not_val.
  - rewrite app_cont_reverse in x.
    destruct (rev K0); [by ss; destruct a; done|].
    destruct a, s; depdes x; try rewrite app_cont_r_reverse in *; eauto
    ; try (by exfalso; eauto using val_is_voe)
    ; by exfalso; assert (X: is_value_of_exp v) by eauto using val_is_voe 
      ; rewrite <- x in X; apply value_cont in X.
  - rewrite app_cont_reverse in x.
    destruct (rev K0).
      by ss; destruct a; try done; depdes x; ss; exfalso; eauto using val_is_voe.
    destruct a, s; depdes x; try rewrite app_cont_r_reverse in *; eauto
    ; try (by exfalso; eauto using val_is_voe)
    ; by exfalso; assert (X: is_value_of_exp v0) by eauto using val_is_voe
         ; rewrite <- x in X; apply value_cont in X.
Qed.

Lemma not_reduceb_cont : forall k h e h' e'
    (NR : forall h' e', ~ reduceb h e h' e') 
    (NV : ~ is_value_of_exp e),
  ~ reduceb h (app_cont k e) h' e'.
Proof. intros; edestruct reduceb_app_cont_inv; eauto. Qed.

Lemma not_reduce_cont : forall k h e h' e'
    (NR : forall h' e', ~ reduce h e h' e') 
    (NV : ~ is_value_of_exp e),
  ~ reduce h (app_cont k e) h' e'.
Proof. intros; edestruct reduce_app_cont_inv; eauto. Qed.

(** The operational semantics is deterministic. *)

Lemma reduceb_det :
  forall h e h1 e1 (R1: reduceb h e h1 e1) h2 e2 (R2: reduceb h e h2 e2),
    h1 = h2 /\ e1 = e2.
Proof.
  induction 1;inversion 1; repeat simpl_inj; subst; eauto; try done
  ; try by destruct SK; ss; depdes H1
           ; try destruct (reduceb_not_val RED)
           ; try destruct (reduceb_not_val R1); eauto using val_is_voe.
- rewrite LOAD in LOAD0; depdes LOAD0; repeat simpl_inj; subst; eauto.
- esplits; eauto; apply heap_eq.
  i; assert (X:=equal_f H0 x); ss.
  destruct (eq_nat_dec _ _); eauto.
- apply app_scont_eq in H1; eauto using reduceb_not_val.
  by des; subst; destruct (IHR1 _ _ RED); subst.
Qed.

Lemma callcc_nored: forall K h x e h' e',
  ~ reduceb h (app_cont K (Ecallcc x e)) h' e'.
Proof.
  i; eapply not_reduceb_cont; eauto; eauto.
  by inversion 1; destruct SK.
Qed.

Lemma throw_nored: forall K h (ve: voe) K' h' e',
  ~ reduceb h (app_cont K (Ethrow (e_of ve) (Econt K'))) h' e'.
Proof.
  i; eapply not_reduceb_cont; eauto; eauto.
  inversion 1; destruct SK; try done
  ; by ss; depdes H1; eauto using not_reduceb_val.
Qed.

Lemma callcc_atom: forall x e, is_atom (Ecallcc x e).
Proof. by ii; edestruct sk. Qed.

Lemma throw_atom: forall ve ve', is_atom (Ethrow (e_of ve) (e_of ve')).
Proof. 
  by ii; edestruct sk; try done; ss; depdes EQ; eauto using val_is_voe.
Qed.

Lemma throw_callcc_conflict: forall K K' (ve ve': voe) x e,
  ~ app_cont K (Ethrow (e_of ve) (e_of ve')) = app_cont K' (Ecallcc x e).
Proof.
  by i; eapply app_cont_eq in FH; des; s; eauto using callcc_atom, throw_atom.
Qed.

Lemma reduce_det :
  forall h e h1 e1 (R1: reduce h e h1 e1) h2 e2 (R2: reduce h e h2 e2),
    h1 = h2 /\ e1 = e2.
Proof.
  i; depdes R1.
  - depdes R2; try by exfalso; eauto using callcc_nored, throw_nored.
    eauto using reduceb_det.
  - depdes R2; try by exfalso; eauto using throw_callcc_conflict, callcc_nored.
    esplits; eauto. 
    eapply app_cont_eq in x; des; s; eauto using callcc_atom.
    by depdes x2.
  - depdes R2; try by exfalso; eauto using throw_callcc_conflict, throw_nored.
    esplits; eauto.
    eapply app_cont_eq in x; des; s; eauto using throw_atom.
    by depdes x0; simpl_inj; ss; subst.
Qed.

Lemma lreds_reds: forall n h e n1 h1 e1 
    (REDS: lreduces n h e n1 h1 e1),
  reduces h e h1 e1.
Proof.
  i; induction REDS; [econs|depdes STEP]; eauto.
Qed.

Lemma reds_lreds: forall n h e h1 e1 
    (REDS: reduces h e h1 e1),
  lreduces n h e n h1 e1.
Proof.
  i; induction REDS; [econs|depdes STEP]; eauto.
Qed.

Lemma div_ldiv: forall n h e (DIV: diverges h e), ldiverges n h e.
Proof. 
  pcofix CIH; i; pdes DIV; pfold; eauto.
Grab Existential Variables. exact 0.
Qed.

Lemma ldiv_div: forall n h e (DIV: ldiverges n h e), diverges h e.
Proof.
  pcofix CIH. 
  intro n; generalize (le_n n); generalize n at -2; induction n.
  - i; pdes DIV; depdes STEP; [omega|pfold; eauto].
  - i; pdes DIV; depdes STEP.
    + eapply IHn, DIV; omega.
    + pfold; eauto.
Qed.

Lemma lreds_term_det: forall h e 
    n  n1 h1 e1 (R1: lreduces n h e n1 h1 e1) (NRS1 : forall h' e', ~ reduce h1 e1 h' e')
    n' n2 h2 e2 (R2: lreduces n' h e n2 h2 e2) (NRS2 : forall h' e', ~ reduce h2 e2 h' e'),
  h1 = h2 /\ e1 = e2.
Proof.
  i; apply lreds_reds in R1; apply lreds_reds in R2.
  revert_until e1; induction 1; intros.
    by destruct R2; subst; auto; exfalso; eapply NRS1; eauto.
  destruct R2; subst. 
    by exfalso; eapply NRS2; eauto.
  destruct (reduce_det STEP STEP0); subst.
  exploit IHR1; try apply R2; eauto; i; des; eauto.
Qed.

Lemma ldiv_lreds : 
  forall n h e n' h' e' (R: lreduces n h e n' h' e') ne (D: ldiverges ne h e) ne',
    ldiverges ne' h' e'.
Proof. 
  i; apply lreds_reds in R; apply ldiv_div in D; apply div_ldiv; clear ne ne'.
  revert_until e'; induction 1; subst; intross.
  pdes D; eapply (reduce_det STEP) in STEP0; des; clarify; eauto.
Qed.

Lemma lreds_ldiv : 
  forall r n h e n' h' e' (R: lreduces n h e n' h' e') (D: paco3 ldiverges_gen r n' h' e'),
    paco3 ldiverges_gen r n h e.
Proof. induction 1; subst; i; eauto using ldiverges_step. Qed.

Lemma lreds_trans : 
  forall n h e n' h' e' (R: lreduces n h e n' h' e') 
    n'' h'' e'' (R': lreduces n' h' e' n'' h'' e''),
    lreduces n h e n'' h'' e''.
Proof. 
  induction 1; subst; [|i; eapply lreduces_step; s]; eauto. 
Qed.

Lemma ldiv_mon: forall n n' h e (DIV: ldiverges n h e) (LE: n' >= n), ldiverges n' h e.
Proof.
  i; pdes DIV; pfold; depdes STEP; eauto. 
  econs; eauto; econs; omega.
Qed.

Lemma lreds_mon: forall n h e n1 h1 e1
    (DIV: lreduces n h e n1 h1 e1) 
    n' (GE: n' >= n) n1' (LE: n1' <= n1), 
  lreduces n' h e n1' h1 e1.
Proof.
  induction 1; i.
  - assert (n1' = n' \/ n1' < n') by omega.
    des; subst; eauto using lreduces_nil.  
  - econs; [|by eapply IHDIV; eauto].
    depdes STEP; [econs; omega|eauto].
Qed.

Lemma lred_lreds: forall n h e n' h' e',
  lreduce n h e n' h' e' -> lreduces (n - 1) h e n' h' e'.
Proof.
  i; destruct H.
  - eapply lreds_mon; eauto using lreduces_nil; omega.
  - eauto using lreduces_nil.
Qed.

Lemma lreds_val_noldiv :
  forall n h e n' h' (v: voe), lreduces n h e n' h' (e_of v) -> ~ ldiverges n h e.
Proof.
  intros; eapply ldiv_lreds in H; eauto.
  by apply ldiv_div in H; pinv H; eauto using not_reduce_val.
Grab Existential Variables. exact 0.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Tactic and Lemmas for language syntax (closedness, substitution, etc)  *)
(*----------------------------------------------------------------------------*)

Lemma fv_exp_val: forall (v: val) x, fv_exp v x -> fv_val v x.
Proof. induction v; s; i; des; eauto. Qed.

Lemma fv_val_exp: forall (v: val) x, fv_val v x -> fv_exp v x.
Proof. induction v; s; i; des; eauto. Qed.

Lemma fv_exp_app_cont: forall K e x,
  fv_exp (app_cont K e) x <-> fv_exp (Vcont K) x \/ fv_exp e x.
Proof.
  induction K; s; i; [by intuition|].
  split; i.
  - apply IHK in H; des; eauto.
    destruct a; ss; des; eauto using fv_exp_val.
  - apply IHK; des; eauto
    ; destruct a; ss; des; eauto using fv_val_exp.
Qed.

(*----------------------------------------------------------------------------*)
(** * wvfix *)
(*----------------------------------------------------------------------------*)

Inductive wvfix (w : world) r (s: wst w) ty n1 vk1 n2 vk2 : Prop := 
| wvfix_base
    (REL: gmon w r s ty n1 vk1 n2 vk2)
| wvfix_step g 
    (REL: wval w (fun _ => g) () s ty n1 vk1 n2 vk2) (MON: g <6= wvfix w r)
| wvfix_ksys (TY : ty = Tint) (E1 : vk1 = inr []) (E2 : vk2 = inr [])
.
Hint Constructors wvfix.

Lemma wvfix_induction: forall w r (P: _->know) 
    (BASE: forall s ty n1' vk1 n2' vk2
             (REL: gmon w r s ty n1' vk1 n2' vk2) 
             n1 (N1: n1' <= n1) n2 (N2: n2' <= n2),
           P s ty n1 vk1 n2 vk2)
    (STEP: wval w (fun _ => P) () <6= P)
    (KSYS: forall s n1 n2, P s Tint n1 (inr []) n2 (inr [])),
  wvfix w r <6= P. 
Proof.
  i; induction PR; subst; eauto using wval_mon.
Qed.

Lemma wvfix_mon: forall (w: world) r s ty n1 vk1 n2 vk2
    (REL: wvfix w r s ty n1 vk1 n2 vk2)
    r' (LE: r <6= r')
    s' (EXT: wextpub w s' s)
    n1' (N1: n1 <= n1') n2' (N2: n2 <= n2'),
  wvfix w r' s' ty n1' vk1 n2' vk2.
Proof.
  i; revert_until REL; pattern s,ty,n1,vk1,n2,vk2.
  eapply wvfix_induction, REL; [econs 1|econs 2|econs 3]
  ; eauto using gmon_mon, wval_mon; omega.
Qed.

Lemma wvfix_closed: forall (w: world) r s
    (CLO: forall s, know_closed (r s)),
  know_closed (wvfix w r s).
Proof.
  ii; revert_until V; induction V. 
  - destruct REL; eapply CLO; eauto.
  - eapply wval_closed; eauto; ii; eapply H; eauto.
  - by subst; s; i; des.
Qed.

Lemma wvfix_glob: forall (w : world), wglob w (wvfix w bot6).
Proof.
  econs; i.
  - r; i; revert x FH; pattern s,ty,n,vk,n',vk'. 
    eapply wvfix_induction; eauto; i; [by destruct REL| |by ss; des].
    apply wval_closed in PR; eauto.
  - eauto using wvfix_mon.
  - econs; unnw; i; eauto
    ;(destruct R; [by depdes REL | eauto using wval_mon| done]).
  - eauto.
Qed.

Lemma wvfix_init: forall (w: world) g (G: wglob w g), wvfix w bot6 <6= g.
Proof.
  s; i; induction PR. 
  - by destruct REL.
  - eapply refn_le; (inst; s; eauto using (wg_le G), wval_mon).
  - by subst; apply (wg_sc G).
Qed.

(*----------------------------------------------------------------------------*)
(** * Language-independent lemmas about worlds *)
(*----------------------------------------------------------------------------*)

Lemma wstuck_mon: forall rf rv rf' rv' n1 e1 n2 e2 
    (WS: wstuck rf rv n1 e1 n2 e2) (LEF: rf <5= rf') (LEV: rv <5= rv'),
  wstuck rf' rv' n1 e1 n2 e2.
Proof. i; inv WS; [econs 1|econs 2|econs 3]; eauto using vclos_mon. Qed.

Lemma sref_eq :
  forall (s s': wst wref) (EQ: forall ty l l', `s ty l l' <-> `s' ty l l'), s = s'. 
Proof.
  by intros; eapply (@sig_eq _ _ s); eauto using @functional_extensionality, prop_ext.
Qed.

Lemma sref0_hemp: forall s r (HR: wheap wref r () s hemp hemp), s = sref0.
Proof.
  intros; apply sref_eq; intuition; repeat red in HR; des.
  by specialize (HSAT _ _ _ H); des.
Qed.

Lemma wrefst_inj1 : forall (s : wst wref) l1
    ty l2 (IN : (`s) ty l1 l2)
    ty' l2' (IN' : (`s) ty' l1 l2'),
  ty = ty' /\ l2 = l2'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (A, ?); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Lemma wrefst_inj2 : forall (s : wst wref) l2
    ty l1 (IN : (`s) ty l1 l2)
    ty' l1' (IN' : (`s) ty' l1' l2),
  ty = ty' /\ l1 = l1'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (?, A); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Implicit Arguments wrefst_inj1 [s l1 ty l2 ty' l2'].
Implicit Arguments wrefst_inj2 [s l2 ty l1 ty' l1'].

Definition wdisj (w1 w2 : world_local) := ndisj (wnam w1) (wnam w2).

Lemma wdisj_comm : 
  forall w1 w2, wdisj w1 w2 -> wdisj w2 w1.
Proof. eby red; red; intros; eapply H. Qed.

Lemma wdisj_prod_l : 
  forall w1 w2 w (D: wdisj w1 w) (D': wdisj w2 w), 
  wdisj (wprod w1 w2) w.
Proof. by intros; intros n []; [apply D | apply D']. Qed.

Lemma wdisj_prod_r : 
  forall w1 w2 w (D: wdisj w w1) (D': wdisj w w2),
  wdisj w (wprod w1 w2).
Proof. intros; auto using wdisj_prod_l, wdisj_comm. Qed.

Hint Resolve wdisj_prod_l wdisj_prod_r.
Hint Immediate wdisj_comm.

(*----------------------------------------------------------------------------*)
(** ** Basic properties of equivalence definitions *)
(*----------------------------------------------------------------------------*)

Lemma wfdg_closed prj: forall G r dd gg
    (PRJ: prj = fst \/ prj = snd)
    (WFDG: wfdg G r dd gg)
    (CLO: know_closed r),
  subst_closed ((exp_of_val<*>snd <*> prj)<*>gg).
Proof.
  ii; des; subst; assert (X:=vclos_closed (wfdg_ggok WFDG x) CLO); des; eauto.
Qed.

Lemma wfdg_mon: forall G r r' dd gg
    (LE: r <5= r')
    (WF: wfdg G r dd gg),
  wfdg G r' dd gg.
Proof.
  i; destruct WF; eauto using vclos_mon.
Qed.

Lemma wstuck_union: forall r1 r2,
  wstuck (r1 \5/ r2) <5= (wstuck r1 \5/ wstuck r2).
Proof. i; inv PR; destruct CREL; vauto. Qed.

Lemma wstuck_lift:
  forall w sp s r r' n1 e1 n2 e2,
    wstuck (wval (wlift w) r sp s) r' n1 e1 n2 e2 <->
    wstuck (wval w (fun sp s => r () (sp,s)) (fst s) (snd s)) r' n1 e1 n2 e2.
Proof. 
  s; unfold wlift_val.
  split; i; destruct H; simpls; des; try discriminate; vauto. 
Qed.

Lemma wglob_prod1 : forall w1 w2 (DISJ : wdisj w1 w2)
    g (G : wglob (wlift (wprod w1 w2)) g) s2,
  wglob (wlift w1) (fun s => gmon w2 (fun s2' => g (fst s, (snd s, s2'))) s2).
Proof.
  s; ii; econs; ii.
  - inversion V; subst; eapply (wg_clo G); eauto.
  - eapply gmon_mon; eauto. 
    i; ss; des; eapply (wg_mon G); eauto; s; eauto.
  - econs; i.
    + eapply gmon_vk; eauto. 
      eapply (wg_le G).
      unfold wlift_val in *; simpl in R |- *. 
      destruct (is_ref _ _ _); s; eauto.
    + inversion R; subst.
      apply (wg_le G) in REL; rrdes REL; depdes TEQ VEQ VEQ'.
      rr; esplits; eauto; omega.
    + inversion R; subst.
      apply (wg_le G) in REL; [|by inversion NM; right; left; eauto].
      destruct REL.
      * ss; eapply wval_mon; eauto.
        s; eauto using gmon_mon.
      * destruct NM; [done|]; eapply wnam_supports in H.
        exfalso; eapply DISJ; eauto.
  - eauto using (wg_sc G).
Qed.

Lemma wglob_prod2 : forall w1 w2 (DISJ : wdisj w1 w2)
    g (G : wglob (wlift (wprod w1 w2)) g) s1,
  wglob (wlift w2) (fun s => gmon w1 (fun s1' => g (fst s, (s1', snd s))) s1).
Proof.
  s; ii; econs; ii.
  - inversion V; subst; eapply (wg_clo G); eauto.
  - eapply gmon_mon; eauto. 
    i; ss; des; eapply (wg_mon G); eauto; s; eauto.
  - econs; i.
    + eapply gmon_vk; eauto. 
      eapply (wg_le G).
      unfold wlift_val in *; simpl in R |- *. 
      destruct (is_ref _ _ _); s; eauto.
    + inversion R; subst.
      apply (wg_le G) in REL; rrdes REL; depdes TEQ VEQ VEQ'.
      rr; esplits; eauto; omega.
    + inversion R; subst.
      apply (wg_le G) in REL; [|by inversion NM; right; right; eauto].
      destruct REL.
      * destruct NM; [done|]; eapply wnam_supports in H.
        exfalso; eapply DISJ; eauto.
      * ss; eapply wval_mon; eauto.
        s; eauto using gmon_mon.
  - eauto using (wg_sc G).
Qed.

Lemma gmon_prod1: forall w1 w2 g (gOK: wglob (wlift (wprod w1 w2)) g) sr s1 s2,
  gmon w1 (fun s1 => g (sr, (s1, s2))) s1 <5= g (sr,(s1,s2)).
Proof.
  i; depdes PR.
  eapply (wg_mon gOK); s; eauto.
Qed.

Lemma gmon_prod2: forall w1 w2 g (gOK: wglob (wlift (wprod w1 w2)) g) sr s1 s2,
  gmon w2 (fun s2 => g (sr, (s1, s2))) s2 <5= g (sr,(s1,s2)).
Proof.
  i; depdes PR.
  eapply (wg_mon gOK); s; eauto.
Qed.

Lemma gmon_incl: forall pst pext (w: @world_dep pst pext) g, g <6= gmon w g.
Proof. i; destruct x3,x5; eauto. Qed.
Hint Resolve gmon_incl.

Lemma gmon_glob_eq: forall w g (G: wglob (wlift w) g), 
  g = fun s => gmon w (fun s' => g (fst s,s')) (snd s).
Proof.
  i; extensionalities s ty n1 x1 n2 x2; destruct s; apply prop_ext; split.
  - eauto.
  - i; depdes H; eapply (wg_mon G); s; eauto.
Qed.

Lemma wexp_core_lred: 
  forall (w: world) hf hf' hh hh' hh1 hh1' e e' e1 e1' s s1 r g n n' n1 n1'
    (R: wexp_core w r g s1 hf hh1 n1 e1 hf' hh1' n1' e1')
    (RED : lreduces n hh e n1 hh1 e1)
    (RED' : lreduces n' hh' e' n1' hh1' e1')
    (EXT : wextpub w s1 s),
  wexp_core w r g s hf hh n e hf' hh' n' e'.
Proof.
  i; destruct R.
  - eapply wexp_Diverge; eauto using lreds_ldiv.
  - eapply wexp_Stuck with (s':=s'); s; eauto using wextpub_trans, lreds_trans.
Qed.

Lemma wexp_core_mon :
  forall (w: world) r r' g g' s hf hh n n1 e hf' hh' n' n1' e'
    (WE: wexp_core w r g s hf hh n e hf' hh' n' e') 
    (LTR: r <6= r') (LTG: g <6= g') (LTN: n <= n1) (LTN': n' <= n1'), 
  wexp_core w r' g' s hf hh n1 e hf' hh' n1' e'. 
Proof.
  i; destruct WE; [eauto using ldiv_mon|].
  eapply wexp_Stuck; s; eauto using wstuck_mon, wheap_mon, lreds_mon.
Qed.

Lemma wexpgen_mon :
  forall (w: world) r g s n1 n1' e1 n2 n2' e2 
    (WE: wexpgen w r g s n1 e1 n2 e2) 
    (LTN1: n1 <= n1') (LTN2: n2 <= n2'), 
  wexpgen w r g s n1' e1 n2' e2. 
Proof. r; i; eapply wexp_core_mon; eauto. Qed.

Structure taggedExp := tExp0 { untagExp :> exp}.

Structure KESplit := mkKESplit 
  { term_of : taggedExp; scont_of : scont; body_of : exp; 
    KEsplit_pf : untagExp term_of = app_cont [scont_of] body_of }.

Lemma wexp_app_scont :
  forall (w: world) g s n n' k1 ke1 k2 ke2
    (CEQ: wexp w g s n (app_cont (scont_of ke1 :: k1) (body_of ke1))
                        n' (app_cont (scont_of ke2 :: k2) (body_of ke2))),
  wexp w g s n (app_cont k1 (term_of ke1)) n' (app_cont k2 (term_of ke2)).
Proof. by intros; destructs ke1 ke2 term_of0 term_of1; subst; simpls; subst. Qed.

Definition tExp1  e := @tExp0  e.
Definition tExp2  e := @tExp1  e.
Definition tExp3  e := @tExp2  e.
Definition tExp4  e := @tExp3  e.
Definition tExp5  e := @tExp4  e.
Definition tExp6  e := @tExp5  e.
Definition tExp7  e := @tExp6  e.
Definition tExp8  e := @tExp7  e.
Definition tExp9  e := @tExp8  e.
Definition tExp10 e := @tExp9  e.
Definition tExp11 e := @tExp10 e.
Definition tExp12 e := @tExp11 e.
Definition tExp13 e := @tExp12 e.
Definition tExp14 e := @tExp13 e.
Definition tExp15 e := @tExp14 e.
Definition tExp16 e := @tExp15 e.
Definition tExp17 e := @tExp16 e.
Definition tExp18 e := @tExp17 e.
Definition tExp19 e := @tExp18 e.
Definition tExp20 e := @tExp19 e.
Definition tExp21 e := @tExp20 e.
Definition tExp22 e := @tExp21 e.
Canonical Structure tExp23 e := @tExp22 e.

Canonical Structure KES_SKif ee e1 e2 :=
  @mkKESplit (tExp0 (Eif ee e1 e2)) (SKif e1 e2) ee refl_equal.
Program 
Canonical Structure KES_SKpairR ee ve :=
  @mkKESplit (tExp1 (Epair (e_of ve) ee)) (SKpairR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKpairL ee e :=
  @mkKESplit (tExp2 (Epair ee e)) (SKpairL e) ee refl_equal.
Canonical Structure KES_SKfst ee :=
  @mkKESplit (tExp3 (Efst ee)) SKfst ee refl_equal.
Canonical Structure KES_SKsnd ee :=
  @mkKESplit (tExp4 (Esnd ee)) SKsnd ee refl_equal.
Canonical Structure KES_SKinl ee :=
  @mkKESplit (tExp5 (Einl ee)) SKinl ee refl_equal.
Canonical Structure KES_SKinr ee :=
  @mkKESplit (tExp6 (Einr ee)) SKinr ee refl_equal.
Canonical Structure KES_SKcase ee x e1 e2 :=
  @mkKESplit (tExp7 (Ecase x ee e1 e2)) (SKcase x e1 e2) ee refl_equal.
Canonical Structure KES_SKroll ee :=
  @mkKESplit (tExp8 (Eroll ee)) SKroll ee refl_equal.
Canonical Structure KES_SKunroll ee :=
  @mkKESplit (tExp9 (Eunroll ee)) SKunroll ee refl_equal.
Program 
Canonical Structure KES_SKappR ee ve :=
  @mkKESplit (tExp10 (Eapp (e_of ve) ee)) (SKappR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKappL ee e :=
  @mkKESplit (tExp11 (Eapp ee e)) (SKappL e) ee refl_equal.
Canonical Structure KES_SKinst ee :=
  @mkKESplit (tExp12 (Einst ee)) SKinst ee refl_equal.
Canonical Structure KES_SKpack ee :=
  @mkKESplit (tExp13 (Epack ee)) SKpack ee refl_equal.
Canonical Structure KES_SKunpack ee x e :=
  @mkKESplit (tExp14 (Eunpack x ee e)) (SKunpack x e) ee refl_equal.
Canonical Structure KES_SKnew ee :=
  @mkKESplit (tExp15 (Enew ee)) SKnew ee refl_equal.
Canonical Structure KES_SKread ee :=
  @mkKESplit (tExp16 (Eread ee)) SKread ee refl_equal.
Program 
Canonical Structure KES_SKwriteR ee ve :=
  @mkKESplit (tExp17 (Ewrite (e_of ve) ee)) (SKwriteR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKwriteL ee e :=
  @mkKESplit (tExp18 (Ewrite ee e)) (SKwriteL e) ee refl_equal.
Program 
Canonical Structure KES_SKrefeqR ee ve :=
  @mkKESplit (tExp19 (Erefeq (e_of ve) ee)) (SKrefeqR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKrefeqL ee e :=
  @mkKESplit (tExp20 (Erefeq ee e)) (SKrefeqL e) ee refl_equal.
Program 
Canonical Structure KES_SKthrowR ee ve :=
  @mkKESplit (tExp21 (Ethrow (e_of ve) ee)) (SKthrowR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKthrowL ee e :=
  @mkKESplit (tExp22 (Ethrow ee e)) (SKthrowL e) ee refl_equal.
Canonical Structure KES_SKisolate ee :=
  @mkKESplit (tExp23 (Eisolate ee)) SKisolate ee refl_equal.

(*----------------------------------------------------------------------------*)
(** ** External calls *)
(*----------------------------------------------------------------------------*)

Inductive cowexp_gen (w: world) r g hf hf' cowexp : 
  wst w -> heap -> nat -> exp -> heap -> nat -> exp -> Prop :=
| cowexp_stop s hh n e hh' n' e'
    (OEQ: @wexp_core w r g s hf hh n e hf' hh' n' e') :
  cowexp_gen w r g hf hf' cowexp s hh n e hh' n' e'
| cowexp_steps s s1 hh hh' n e n' e' hh1 hh1' n1 e1 n1' e1' hh2 hh2' n2 e2 n2' e2'
    (REL: cowexp s1 hh2 n2 e2 hh2' n2' e2' : Prop)
    (EXT: wextpub w s1 s) 
    (STEP1: lreduces n hh e n1 hh1 e1) (STEPS1: lreduce n1 hh1 e1 n2 hh2 e2)
    (STEP2: lreduces n' hh' e' n1' hh1' e1') (STEPS2: lreduce n1' hh1' e1' n2' hh2' e2'):
  cowexp_gen w r g hf hf' cowexp s hh n e hh' n' e'
.
Hint Constructors cowexp_gen.

Definition cowexp w r g hf hf' s hh n e hh' n' e' := 
  paco7 (cowexp_gen w r g hf hf') bot7 s hh n e hh' n' e'.
Lemma cowexp_gen_mon: 
  forall w r g hf hf', monotone7 (cowexp_gen w r g hf hf'). 
Proof. pmonauto. Qed.
Hint Unfold cowexp.
Hint Resolve cowexp_gen_mon : paco.

Lemma cowexp_div_or_wexp: forall w r g hf hf' n n' s hh e hh' e'
    (REL: cowexp w r g hf hf' s hh n e hh' n' e'),
  (ldiverges n hh e /\ ldiverges n' hh' e') \/
  (wexp_core w r g s hf hh n e hf' hh' n' e').
Proof.
  i; destruct (classic(wexp_core w r g s hf hh n e hf' hh' n' e')); eauto.
  r in H; left; split;
  ( revert_until hf'; pcofix CIH with z 
  ; i; punfold REL; depdes REL; [by exfalso; eauto|]
  ; pclearbot; eapply lreds_ldiv; eauto 10 using wexp_core_lred, lreds_trans, lreduces_nil).
Qed.

Lemma wexp_core_external :
  forall w g r s hf hf' hh hh' n e n' e' 
    (CONS: wconsistent w) 
    (gOK: wglob w g)
    (gEQ: forall s, g s <5= (wval w (fun _ => g) () s \5/ r s))
    (REL: wexp_core w g g s hf hh n e hf' hh' n' e'),
  wexp_core w r g s hf hh n e hf' hh' n' e'.
Proof.
  i; cut (cowexp w r g hf hf' s hh n e hh' n' e').
    by intro X; eapply cowexp_div_or_wexp in X; des; eauto.
  revert s hh n e hh' n' e' REL; pcofix CIH with z.
  i; pfold; destruct REL; eauto.
  eapply wstuck_mon in EREL; [|apply gEQ|eauto]. 
  eapply wstuck_union in EREL; des; [|econs; eauto].
  destruct (CONS _ _ gOK _ _ _ _ EREL).
  econs 2; eauto; econs; omega.
Qed.

Lemma wexpgen_external :
  forall w g r s n e n' e' 
    (CONS: wconsistent w) 
    (gOK: wglob w g) 
    (gEQ: forall s, g s <5= (wval w (fun _ => g) () s \5/ r s))
    (REL: wexp w g s n e n' e'),
  wexpgen w r g s n e n' e'.
Proof. by repeat intro; eapply wexp_core_external; eauto. Qed.

Corollary wexp_no_external :
  forall w g s n e n' e'
    (CONS: wconsistent w) 
    (gOK: wglob w g)
    (gEQ: forall s, g s <5= (wval w (fun _ => g) () s \5/ sysknow))
    (REL: wexp w g s n e n' e'), 
  wexpgen w (fun _ => sysknow) g s n e n' e'.
Proof. i; eapply wexpgen_external; eauto. Qed.
