(* generated by Ott 0.21.2 from: lang.ott *)

Require Import Arith.
Require Import Bool.
Require Import List.


Require Import Omega ClassicalChoice basis. 
Set Implicit Arguments. 

Definition termvar := nat.
Lemma eq_termvar: forall (x y : termvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_termvar : ott_coq_equality.
Definition typename := nat.
Lemma eq_typename: forall (x y : typename), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typename : ott_coq_equality.
Definition typebvar := nat.
Lemma eq_typebvar: forall (x y : typebvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typebvar : ott_coq_equality.
Definition location := nat.

Inductive typ : Set :=  (*r types (in locally nameless style) *)
 | Tname (aa:typename) (*r Free name *)
 | Tbvar (bb:typebvar) (*r Bound variable *)
 | Tunit : typ (*r Unit *)
 | Tint : typ (*r Integers *)
 | Tbool : typ (*r Booleans *)
 | Tprod (ty1:typ) (ty2:typ) (*r products *)
 | Tsum (ty1:typ) (ty2:typ) (*r Sums *)
 | Tarrow (ty1:typ) (ty2:typ) (*r Functions *)
 | Tref (ty:typ) (*r References *)
 | Tmu (ty:typ) (*r Recursive types *)
 | Tforall (ty:typ)
 | Texists (ty:typ)
 | Tcont (ty:typ) (*r Continuations *).

Inductive val : Set :=  (*r values *)
 | Vvar (x:termvar)
 | Vunit : val
 | Vint (n:nat)
 | Vtrue : val
 | Vfalse : val
 | Vpair (v1:val) (v2:val)
 | Vinl (v:val)
 | Vinr (v:val)
 | Vroll (v:val)
 | Vfix (f:termvar) (x:termvar) (e:exp)
 | Vgen (e:exp)
 | Vpack (v:val)
 | Vloc (loc:location)
 | Vcont (K:list scont)
with exp : Set :=  (*r expressions *)
 | Evar (x:termvar) (*r Variable *)
 | Eunit : exp
 | Eint (n:nat)
 | Etrue : exp
 | Efalse : exp
 | Epair (e1:exp) (e2:exp)
 | Einl (e:exp)
 | Einr (e:exp)
 | Efst (e:exp)
 | Esnd (e:exp)
 | Eloc (loc:location)
 | Eif (e0:exp) (e1:exp) (e2:exp) (*r if *)
 | Ecase (x:termvar) (e0:exp) (e1:exp) (e2:exp) (*r case *)
 | Eroll (e:exp) (*r roll *)
 | Eunroll (e:exp) (*r unroll *)
 | Efix (f:termvar) (x:termvar) (e:exp) (*r Recursive function *)
 | Eapp (e1:exp) (e2:exp) (*r Function application *)
 | Egen (e:exp) (*r Type generalisation *)
 | Einst (e:exp) (*r Type instantiation *)
 | Epack (e:exp)
 | Eunpack (x:termvar) (e1:exp) (e2:exp)
 | Enew (e:exp)
 | Eread (e:exp)
 | Ewrite (e1:exp) (e2:exp)
 | Erefeq (e1:exp) (e2:exp)
 | Econt (K:list scont)
 | Ecallcc (x:termvar) (e:exp)
 | Ethrow (e1:exp) (e2:exp)
 | Eisolate (e:exp)
with scont : Set :=  (*r Continuations *)
 | SKif (e1:exp) (e2:exp)
 | SKpairL (e:exp)
 | SKpairR (v:val)
 | SKfst
 | SKsnd
 | SKinl
 | SKinr
 | SKcase (x:termvar) (e1:exp) (e2:exp)
 | SKroll
 | SKunroll
 | SKpack
 | SKappL (e:exp)
 | SKappR (v:val)
 | SKinst
 | SKunpack (x:termvar) (e:exp)
 | SKnew
 | SKread
 | SKwriteL (e:exp)
 | SKwriteR (v:val)
 | SKrefeqL (e:exp)
 | SKrefeqR (v:val)
 | SKthrowL (e:exp)
 | SKthrowR (v:val)
 | SKisolate.

Definition cont : Set := list scont.

(*----------------------------------------------------------------------------*)
(** Coercions *)
(*----------------------------------------------------------------------------*)

Fixpoint exp_of_val (v : val) : exp :=
  match v with
  | Vvar x      => Evar x
  | Vunit       => Eunit
  | Vint n      => Eint n
  | Vtrue       => Etrue 
  | Vfalse      => Efalse
  | Vpair v1 v2 => Epair (exp_of_val v1) (exp_of_val v2)
  | Vinl v      => Einl (exp_of_val v)
  | Vinr v      => Einr (exp_of_val v)
  | Vroll v     => Eroll (exp_of_val v)
  | Vfix f x e  => Efix f x e
  | Vgen e      => Egen e
  | Vpack v     => Epack (exp_of_val v)
  | Vloc loc    => Eloc loc
  | Vcont k     => Econt k
  end.

Structure VoE := mkVoE { e_of : exp ; v_of : val ; VoE_pf : e_of = exp_of_val v_of }.

Definition voe : Set := VoE.

(*----------------------------------------------------------------------------*)
(** Definition of a ``locally closed'' type *)
(*----------------------------------------------------------------------------*)

Fixpoint lc_typ (n : nat) (ty : typ) :=
  match ty with
  | Tbvar aa     => aa < n
  | Tname aa    => True
  | Tunit       => True 
  | Tint        => True 
  | Tbool       => True
  | (Tprod ty1 ty2)  => lc_typ n ty1 /\ lc_typ n ty2
  | (Tsum ty1 ty2)   => lc_typ n ty1 /\ lc_typ n ty2
  | (Tarrow ty1 ty2) => lc_typ n ty1 /\ lc_typ n ty2
  | (Tref ty)    => lc_typ n ty
  | (Tmu ty)     => lc_typ (S n) ty
  | (Tforall ty) => lc_typ (S n) ty
  | (Texists ty) => lc_typ (S n) ty
  | (Tcont ty) => lc_typ n ty
  end.

(*----------------------------------------------------------------------------*)
(** Operations *)
(*----------------------------------------------------------------------------*)

(** subrules *)
Fixpoint is_value_of_exp (e_5:exp) : Prop :=
  match e_5 with
  | (Evar x) => (True)
  | Eunit => (True)
  | (Eint n) => (True)
  | Etrue => (True)
  | Efalse => (True)
  | (Epair e1 e2) => ((is_value_of_exp e1) /\ (is_value_of_exp e2))
  | (Einl e) => ((is_value_of_exp e))
  | (Einr e) => ((is_value_of_exp e))
  | (Efst e) => False
  | (Esnd e) => False
  | (Eloc loc) => (True)
  | (Eif e0 e1 e2) => False
  | (Ecase x e0 e1 e2) => False
  | (Eroll e) => ((is_value_of_exp e))
  | (Eunroll e) => False
  | (Efix f x e) => (True)
  | (Eapp e e') => False
  | (Egen e) => (True)
  | (Einst e) => False
  | (Epack e) => ((is_value_of_exp e))
  | (Eunpack x e1 e2) => False
  | (Enew e) => False
  | (Eread e) => False
  | (Ewrite e1 e2) => False
  | (Erefeq e1 e2) => False
  | (Econt K) => (True)
  | (Ecallcc x e) => False
  | (Ethrow e1 e2) => False
  | (Eisolate e) => False
end.

(*----------------------------------------------------------------------------*)
(** Definition of app_cont *)
(*----------------------------------------------------------------------------*)

(** context application *)
Definition appctx_scont_exp (scont5:scont) (exp5:exp) : exp :=
  match scont5 with
  | (SKif e1 e2) => (Eif exp5 e1 e2)
  | (SKpairL e) => (Epair exp5 e)
  | (SKpairR v) => (Epair  (exp_of_val  v )  exp5)
  | SKfst => (Efst exp5)
  | SKsnd => (Esnd exp5)
  | SKinl => (Einl exp5)
  | SKinr => (Einr exp5)
  | (SKcase x e1 e2) => (Ecase x exp5 e1 e2)
  | SKroll => (Eroll exp5)
  | SKunroll => (Eunroll exp5)
  | SKpack => (Epack exp5)
  | (SKappL e) => (Eapp exp5 e)
  | (SKappR v) => (Eapp  (exp_of_val  v )  exp5)
  | SKinst => (Einst exp5)
  | (SKunpack x e) => (Eunpack x exp5 e)
  | SKnew => (Enew exp5)
  | SKread => (Eread exp5)
  | (SKwriteL e) => (Ewrite exp5 e)
  | (SKwriteR v) => (Ewrite  (exp_of_val  v )  exp5)
  | (SKrefeqL e) => (Erefeq exp5 e)
  | (SKrefeqR v) => (Erefeq  (exp_of_val  v )  exp5)
  | (SKthrowL e) => (Ethrow  exp5 e)
  | (SKthrowR v) => (Ethrow  (exp_of_val  v )   exp5)
  | (SKisolate) => (Eisolate  exp5)
end.

Definition app_cont (k : cont) (e : exp) : exp :=
  fold_left (fun e0 k0 => appctx_scont_exp k0 e0) k e.

Definition app_cont_r k e := fold_right appctx_scont_exp e k.

Lemma app_cont_r_reverse: forall k e,
  app_cont_r k e = app_cont (rev k) e.
Proof. 
 intros; unfold app_cont, app_cont_r.
 rewrite <-fold_left_rev_right, rev_involutive; auto.
Qed.

Lemma app_cont_reverse: forall k e,
  app_cont k e = app_cont_r (rev k) e.
Proof. 
 intros; rewrite <-(rev_involutive k) at 1; rewrite app_cont_r_reverse; auto.
Qed.

(*----------------------------------------------------------------------------*)
(** Well-formed Expression *)
(*----------------------------------------------------------------------------*)

Fixpoint wf_exp (e: exp) :=
  match e with
  | (Evar x) => True
  | Eunit => True
  | (Eint n) => True
  | Etrue => True
  | Efalse => True
  | (Epair e1 e2) => (wf_exp e1) /\ (wf_exp e2)
  | (Einl e) => ((wf_exp e))
  | (Einr e) => ((wf_exp e))
  | (Efst e) => ((wf_exp e))
  | (Esnd e) => ((wf_exp e))
  | (Eloc loc) => False
  | (Eif e0 e1 e2) => (wf_exp e0) /\ (wf_exp e1) /\ (wf_exp e2)
  | (Ecase x e0 e1 e2) => (wf_exp e0) /\ (wf_exp e1) /\ (wf_exp e2)
  | (Eroll e) => ((wf_exp e))
  | (Eunroll e) => ((wf_exp e))
  | (Efix f x e) => wf_exp e
  | (Eapp e e') => (wf_exp e) /\ (wf_exp e')
  | (Egen e) => ((wf_exp e))
  | (Einst e) => ((wf_exp e))
  | (Epack e) => ((wf_exp e))
  | (Eunpack x e1 e2) => (wf_exp e1) /\ (wf_exp e2)
  | (Enew e) => ((wf_exp e))
  | (Eread e) => ((wf_exp e))
  | (Ewrite e1 e2) => (wf_exp e1) /\ (wf_exp e2)
  | (Erefeq e1 e2) => (wf_exp e1) /\ (wf_exp e2)
  | (Econt K) => False
  | (Ecallcc x e) => wf_exp e
  | (Ethrow e1 e2) => (wf_exp e1) /\ (wf_exp e2)
  | (Eisolate e) => ((wf_exp e))
end.

(*----------------------------------------------------------------------------*)
(** Free Variables *)
(*----------------------------------------------------------------------------*)

Fixpoint ftn_typ (ty_5:typ) (a: typename) : Prop :=
  match ty_5 with
  | (Tname aa) => a = aa
  | (Tbvar bb) => False
  | Tunit => False
  | Tint => False
  | Tbool => False
  | (Tprod ty1 ty2) => ftn_typ ty1 a \/ ftn_typ ty2 a
  | (Tsum ty1 ty2) => ftn_typ ty1 a \/ ftn_typ ty2 a
  | (Tarrow ty1 ty2) => ftn_typ ty1 a \/ ftn_typ ty2 a
  | (Tref ty) => ftn_typ ty a
  | (Tmu ty) => ftn_typ ty a
  | (Tforall ty) => ftn_typ ty a
  | (Texists ty) => ftn_typ ty a
  | (Tcont ty) => ftn_typ ty a
end.

Fixpoint fv_exp (e_5:exp) (y: termvar) : Prop :=
  match e_5 with
  | (Evar x) => y = x
  | Eunit => False
  | (Eint n) => False
  | Etrue => False
  | Efalse => False
  | (Epair e1 e2) => (fv_exp e1 y) \/ (fv_exp e2 y)
  | (Einl e) => ((fv_exp e y))
  | (Einr e) => ((fv_exp e y))
  | (Efst e) => ((fv_exp e y))
  | (Esnd e) => ((fv_exp e y))
  | (Eloc loc) => False
  | (Eif e0 e1 e2) => (fv_exp e0 y) \/ (fv_exp e1 y) \/ (fv_exp e2 y)
  | (Ecase x e0 e1 e2) => (fv_exp e0 y) \/ ((y = x -> False) /\ (fv_exp e1 y \/ fv_exp e2 y))
  | (Eroll e) => ((fv_exp e y))
  | (Eunroll e) => ((fv_exp e y))
  | (Efix f x e) => (y = f -> False) /\ (y = x -> False) /\ fv_exp e y
  | (Eapp e e') => (fv_exp e y) \/ (fv_exp e' y)
  | (Egen e) => ((fv_exp e y))
  | (Einst e) => ((fv_exp e y))
  | (Epack e) => ((fv_exp e y))
  | (Eunpack x e1 e2) => (fv_exp e1 y) \/ ((y = x -> False) /\ fv_exp e2 y)
  | (Enew e) => ((fv_exp e y))
  | (Eread e) => ((fv_exp e y))
  | (Ewrite e1 e2) => (fv_exp e1 y) \/ (fv_exp e2 y)
  | (Erefeq e1 e2) => (fv_exp e1 y) \/ (fv_exp e2 y)
  | (Econt K) => fold_right (fun SK P => fv_scont SK y \/ P) False K
  | (Ecallcc x e) => (y = x -> False) /\ fv_exp e y
  | (Ethrow e1 e2) => (fv_exp e1 y) \/ (fv_exp e2 y)
  | (Eisolate e) => ((fv_exp e y))
end
with fv_val (v_5:val) (y: termvar) : Prop :=
  match v_5 with
  | (Vvar x) => y = x
  | Vunit => False
  | (Vint n) => False
  | Vtrue => False
  | Vfalse => False
  | (Vpair v1 v2) => (fv_val v1 y) \/ (fv_val v2 y)
  | (Vinl v) => ((fv_val v y))
  | (Vinr v) => ((fv_val v y))
  | (Vroll v) => ((fv_val v y))
  | (Vfix f x e) => (y = f -> False) /\ (y = x -> False) /\ fv_exp e y
  | (Vgen e) => ((fv_exp e y))
  | (Vpack v) => ((fv_val v y))
  | (Vloc loc) => False
  | (Vcont K) => fold_right (fun SK P => fv_scont SK y \/ P) False K
end
with fv_scont (K5:scont) (y: termvar) : Prop :=
  match K5 with
  | (SKif e1 e2) => fv_exp e1 y \/ fv_exp e2 y
  | (SKpairL e) => fv_exp e y
  | (SKpairR v) => fv_val v y
  | (SKfst) => False
  | (SKsnd) => False
  | (SKinl) => False
  | (SKinr) => False
  | (SKcase x e1 e2) => ((y = x -> False) /\ (fv_exp e1 y \/ fv_exp e2 y))
  | (SKroll) => False
  | (SKunroll) => False
  | (SKpack) => False
  | (SKappL e) => (fv_exp e y)
  | (SKappR v) => (fv_val v y)
  | (SKinst) => False
  | (SKunpack x e) => ((y = x -> False) /\ fv_exp e y)
  | (SKnew) => False
  | (SKread) => False
  | (SKwriteL e) => (fv_exp e y)
  | (SKwriteR v) => (fv_val v y)
  | (SKrefeqL e) => (fv_exp e y)
  | (SKrefeqR v) => (fv_val v y)
  | (SKthrowL e) => (fv_exp e y)
  | (SKthrowR v) => (fv_val v y)
  | (SKisolate) => False
end.

(*----------------------------------------------------------------------------*)
(** Substitutions *)
(*----------------------------------------------------------------------------*)

Fixpoint tsubst_bvar (x: typebvar) (tyA ty: typ) :=
  match ty with
  | Tbvar aa      =>
    match nat_compare aa x with
      | Lt => Tbvar aa
      | Eq => tyA
      | Gt => Tbvar (pred aa)
    end
  | Tname aa    => Tname aa
  | Tunit       => Tunit 
  | Tint        => Tint 
  | Tbool       => Tbool 
  | (Tprod ty1 ty2)  => Tprod  (tsubst_bvar x tyA ty1) (tsubst_bvar x tyA ty2)
  | (Tsum ty1 ty2)   => Tsum   (tsubst_bvar x tyA ty1) (tsubst_bvar x tyA ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst_bvar x tyA ty1) (tsubst_bvar x tyA ty2)
  | (Tref ty)    => Tref (tsubst_bvar x tyA ty)
  | (Tmu ty)     => Tmu     (tsubst_bvar (S x) tyA ty)
  | (Tforall ty) => Tforall (tsubst_bvar (S x) tyA ty)
  | (Texists ty) => Texists (tsubst_bvar (S x) tyA ty)
  | (Tcont ty)   => Tcont   (tsubst_bvar x tyA ty)
  end.

Fixpoint tsubst_name (sub:typename->typ) (ty__6:typ) {struct ty__6} : typ :=
  match ty__6 with
  | (Tname aa) => sub aa
  | (Tbvar bb) => Tbvar bb
  | Tunit => Tunit 
  | Tint => Tint 
  | Tbool => Tbool 
  | (Tprod ty1 ty2) => Tprod (tsubst_name sub ty1) (tsubst_name sub ty2)
  | (Tsum ty1 ty2) => Tsum (tsubst_name sub ty1) (tsubst_name sub ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst_name sub ty1) (tsubst_name sub ty2)
  | (Tref ty) => Tref (tsubst_name sub ty)
  | (Tmu ty) => Tmu (tsubst_name sub ty)
  | (Tforall ty) => Tforall (tsubst_name sub ty)
  | (Texists ty) => Texists (tsubst_name sub ty)
  | (Tcont ty) => Tcont (tsubst_name sub ty)
end.

Definition sid := fun x => Evar x.

Definition sext {A} x e (sub: termvar->A) := 
  fun y => if eq_termvar y x then e else sub y.

Fixpoint subst_exp (sub:termvar->exp) (e__6:exp) {struct e__6} : exp :=
  match e__6 with
  | (Evar x) => sub x
  | Eunit => Eunit 
  | (Eint n) => Eint n
  | Etrue => Etrue 
  | Efalse => Efalse 
  | (Epair e1 e2) => Epair (subst_exp sub e1) (subst_exp sub e2)
  | (Einl e) => Einl (subst_exp sub e)
  | (Einr e) => Einr (subst_exp sub e)
  | (Efst e) => Efst (subst_exp sub e)
  | (Esnd e) => Esnd (subst_exp sub e)
  | (Eloc loc) => Eloc loc
  | (Eif e0 e1 e2) => Eif (subst_exp sub e0) (subst_exp sub e1) (subst_exp sub e2)
  | (Ecase x e0 e1 e2) => Ecase x (subst_exp sub e0) (subst_exp (sext x (Evar x) sub) e1) (subst_exp (sext x (Evar x) sub) e2)
  | (Eroll e) => Eroll (subst_exp sub e)
  | (Eunroll e) => Eunroll (subst_exp sub e)
  | (Efix f x e) => Efix f x (subst_exp (sext x (Evar x) (sext f (Evar f) sub)) e)
  | (Eapp e e') => Eapp (subst_exp sub e) (subst_exp sub e')
  | (Egen e) => Egen (subst_exp sub e)
  | (Einst e) => Einst (subst_exp sub e)
  | (Epack e) => Epack (subst_exp sub e)
  | (Eunpack x e1 e2) => Eunpack x (subst_exp sub e1) (subst_exp (sext x (Evar x) sub) e2)
  | (Enew e) => Enew (subst_exp sub e)
  | (Eread e) => Eread (subst_exp sub e)
  | (Ewrite e1 e2) => Ewrite (subst_exp sub e1) (subst_exp sub e2)
  | (Erefeq e1 e2) => Erefeq (subst_exp sub e1) (subst_exp sub e2)
  | (Econt K) => Econt K (* !! *)
  | (Ecallcc x e) => Ecallcc x (subst_exp (sext x (Evar x) sub) e)
  | (Ethrow e1 e2) => Ethrow (subst_exp sub e1) (subst_exp sub e2)
  | (Eisolate e) => Eisolate (subst_exp sub e)
end.

(** definitions *)

(*----------------------------------------------------------------------------*)
(** Definition of heaps *)
(*----------------------------------------------------------------------------*)

Definition finite_domain A B (f : A -> option B) := 
  exists d, forall x y, f x = Some y -> In x d.

Structure heap := 
  { heap_f :> location -> option val 
  ; heap_findom : finite_domain heap_f }.

Program Definition hemp : heap := {| heap_f := fun l => None |}.
Next Obligation. exists nil; discriminate. Qed.

Program Definition hext (h : heap) (loc : location) (v: val) : heap :=
  {| heap_f := fun l => if eq_nat_dec l loc then Some v else h l |}.  
Next Obligation. 
  destruct (heap_findom h) as [d X]; exists (loc :: d); intros; simpl.
  destruct (eq_nat_dec); eauto.
Qed.

Definition hplus (h1 h2 h: heap) := 
  forall x, h x = h2 x /\ h1 x = None \/ h x = h1 x /\ h2 x = None.


Parameter allocator : heap -> location.

Axiom allocator_unalloc : forall h : heap, h (allocator h) = None.

Lemma allocator_consistency: 
  exists allocator : heap -> location, forall h : heap, h (allocator h) = None.
Proof.
  cut (forall h : heap, exists l, h l = None).
    intros FRESH; apply (choice _ FRESH).
  intros [hf [d FIN] ]; simpl.
  set (l := S (fold_right plus 0 d)); specialize (FIN l); exists l.
  destruct (hf l); auto.
  cut (forall n (GE: n >= l) (IN: In n d), False).
    intros; exfalso; eauto.
  unfold l in *; clear -d; induction d; intros; eauto.
  simpl in GE, IN; destruct IN; [subst n | apply (IHd n)]; auto; omega.
Qed.

(*----------------------------------------------------------------------------*)
(** Operational Semantics *)
(*----------------------------------------------------------------------------*)

(* defns OpSemJudgB *)
Inductive reduceb : heap -> exp -> heap -> exp -> Prop :=    (* defn reduceb *)
 | Rtrue : forall (h:heap) (e1 e2:exp),
     reduceb h (Eif Etrue e1 e2) h e1
 | Rfalse : forall (h:heap) (e1 e2:exp),
     reduceb h (Eif Efalse e1 e2) h e2
 | Rfst : forall (h:heap) (ve1 ve2:voe),
     reduceb h (Efst (Epair  (e_of  ve1 )   (e_of  ve2 ) )) h  (e_of  ve1 ) 
 | Rsnd : forall (h:heap) (ve1 ve2:voe),
     reduceb h (Esnd (Epair  (e_of  ve1 )   (e_of  ve2 ) )) h  (e_of  ve2 ) 
 | Rinl : forall (h:heap) (x:termvar) (ve:voe) (e1 e2:exp),
     reduceb h (Ecase x (Einl  (e_of  ve ) ) e1 e2) h  (subst_exp(sext x (e_of ve) sid) e1 ) 
 | Rinr : forall (h:heap) (x:termvar) (ve:voe) (e1 e2:exp),
     reduceb h (Ecase x (Einr  (e_of  ve ) ) e1 e2) h  (subst_exp(sext x (e_of ve) sid) e2 ) 
 | Rbeta : forall (h:heap) (f x:termvar) (e:exp) (ve:voe),
     reduceb h (Eapp  ( (Efix f x e) )   (e_of  ve ) ) h  (subst_exp(sext x (e_of ve) (sext f (Efix f x e) sid)) e ) 
 | Rtybeta : forall (h:heap) (e:exp),
     reduceb h (Einst  ( (Egen e) ) ) h e
 | Rpack : forall (h:heap) (x:termvar) (ve:voe) (e:exp),
     reduceb h (Eunpack x (Epack  (e_of  ve ) ) e) h  (subst_exp(sext x (e_of  ve) sid) e ) 
 | Rroll : forall (h:heap) (ve:voe),
     reduceb h (Eunroll  ( (Eroll  (e_of  ve ) ) ) ) h  (e_of  ve ) 
 | Rnew : forall (h:heap) (ve:voe) (loc:location)
     (ALLOC:  (allocator  h  =  loc ) ),
     reduceb h (Enew  (e_of  ve ) )  (hext  h   loc    (v_of  ve )  )  (Eloc loc)
 | Rread : forall (h:heap) (loc:location) (ve:voe)
     (LOAD:  ( h   loc  = Some   (v_of  ve )  ) ),
     reduceb h (Eread (Eloc loc)) h  (e_of  ve ) 
 | Rwrite : forall (h:heap) (loc:location) (ve ve':voe),
     reduceb  (hext  h   loc    (v_of  ve )  )  (Ewrite (Eloc loc)  (e_of  ve' ) )  (hext  h   loc    (v_of  ve' )  )  Eunit
 | RrefeqT : forall (h:heap) (loc:location),
     reduceb h (Erefeq (Eloc loc) (Eloc loc)) h Etrue
 | RrefeqF : forall (h:heap) (loc loc':location)
     (NEQ:  ( loc  <>  loc' ) ),
     reduceb h (Erefeq (Eloc loc) (Eloc loc')) h Efalse
 | Risolate : forall (h:heap) (ve:voe),
     reduceb h (Eisolate  (e_of  ve ) ) h (Econt  (SKappR (v_of ve) :: nil) )
 | Rsctx : forall (h:heap) (SK:scont) (e:exp) (h':heap) (e':exp)
     (RED: reduceb h e h' e'),
     reduceb h  (appctx_scont_exp  SK   e )  h'  (appctx_scont_exp  SK   e' ) 
.

(* defns OpSemJudg *)
Inductive reduce : heap -> exp -> heap -> exp -> Prop :=    (* defn reduce *)
 | Rnormal : forall (h:heap) (e:exp) (h':heap) (e':exp)
     (RED: reduceb h e h' e'),
     reduce h e h' e'
 | Rcallcc : forall (h:heap) (K:cont) (x:termvar) (e:exp),
     reduce h  (app_cont  K    ( (Ecallcc x e) )  )  h  (app_cont  K    (  (subst_exp(sext x (Econt K) sid) e )  )  ) 
 | Rthrow : forall (h:heap) (K:cont) (v:val) (K':cont),
     reduce h  (app_cont  K   (Ethrow  (exp_of_val  v )  (Econt K')) )  h  (app_cont  K'    (exp_of_val  v )  ) 
.

Lemma Rctx :
  forall h e h' e' (RED: reduceb h e h' e') K,
     reduceb h (app_cont K e) h' (app_cont K e').
Proof. intros; revert h e h' e' RED; induction K; intros; eauto using Rsctx. Qed.

Hint Constructors reduceb. 
Hint Constructors reduce. 
Hint Resolve Rctx.

Coercion exp_of_val : val >-> exp.

Implicit Arguments finite_domain.

(*----------------------------------------------------------------------------*)
(** ** Canonical structure for values *)
(*----------------------------------------------------------------------------*)

Canonical Structure VoE_var x := @mkVoE (Evar x) (Vvar x) refl_equal.
Canonical Structure VoE_unit := @mkVoE (Eunit) (Vunit) refl_equal.
Canonical Structure VoE_int n := @mkVoE (Eint n) (Vint n) refl_equal.
Canonical Structure VoE_true := @mkVoE (Etrue) (Vtrue) refl_equal.
Canonical Structure VoE_false := @mkVoE (Efalse) (Vfalse) refl_equal.
Program 
Canonical Structure VoE_pair ve1 ve2 := 
  @mkVoE (Epair (e_of ve1) (e_of ve2)) (Vpair (v_of ve1) (v_of ve2)) _.
Next Obligation. destruct ve1, ve2; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_inl ve1 :=
  @mkVoE (Einl (e_of ve1)) (Vinl (v_of ve1)) _.
Next Obligation. destruct ve1; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_inr ve2 :=
  @mkVoE (Einr (e_of ve2)) (Vinr (v_of ve2)) _.
Next Obligation. destruct ve2; subst; simpl; auto. Qed.
Program 
Canonical Structure VoE_roll ve :=
  @mkVoE (Eroll (e_of ve)) (Vroll (v_of ve)) _.
Next Obligation. destruct ve; subst; simpl; auto. Qed.
Canonical Structure VoE_fix f x e := @mkVoE (Efix f x e) (Vfix f x e) refl_equal.
Canonical Structure VoE_gen e := @mkVoE (Egen e) (Vgen e) refl_equal.
Program 
Canonical Structure VoE_pack ve :=
  @mkVoE (Epack (e_of ve)) (Vpack (v_of ve)) _.
Next Obligation. destruct ve; subst; simpl; auto. Qed.
Canonical Structure VoE_loc l := @mkVoE (Eloc l) (Vloc l) refl_equal.
Canonical Structure VoE_cont K := @mkVoE (Econt K) (Vcont K) refl_equal.
Canonical Structure VoE_val v :=  @mkVoE (exp_of_val v) v refl_equal.

Coercion VoE_val : val >-> VoE.


(*----------------------------------------------------------------------------*)
(** Definition of program and context *)
(*----------------------------------------------------------------------------*)

Inductive prg : Set :=  (*r Programs *)
 | Pvar (x:termvar)
 | Punit : prg
 | Pint (n:nat)
 | Ptrue : prg
 | Pfalse : prg
 | Ppair (p1:prg) (p2:prg)
 | Pinl (ty:typ) (p:prg)
 | Pinr (ty:typ) (p:prg)
 | Pfst (p:prg)
 | Psnd (p:prg)
 | Pif (p0:prg) (p1:prg) (p2:prg)
 | Pcase (x:termvar) (p0:prg) (p1:prg) (p2:prg)
 | Proll (ty:typ) (p:prg)
 | Punroll (p:prg)
 | Pfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (p:prg)
 | Papp (p:prg) (p':prg) (*r Function application *)
 | Pgen (p:prg)
 | Pinst (p:prg) (ty:typ)
 | Ppack (ty1:typ) (p:prg) (ty:typ)
 | Punpack (x:termvar) (p1:prg) (p2:prg)
 | Pnew (p:prg)
 | Pread (p:prg)
 | Pwrite (p1:prg) (p2:prg)
 | Prefeq (p1:prg) (p2:prg)
 | Pcallcc (x:termvar) (p:prg)
 | Pthrow (p1:prg) (p2:prg)
 | Pisolate (p:prg)
.

Inductive ctx : Set :=  (*r Contexts *)
 | Chole : ctx
 | Cpair1 (C:ctx) (p2:prg)
 | Cpair2 (p1:prg) (C:ctx)
 | Cinl (ty:typ) (C:ctx)
 | Cinr (ty:typ) (C:ctx)
 | Cfst (C:ctx)
 | Csnd (C:ctx)
 | Cif0 (C:ctx) (p1:prg) (p2:prg)
 | Cif1 (p0:prg) (C:ctx) (p2:prg)
 | Cif2 (p0:prg) (p1:prg) (C:ctx)
 | Ccase0 (x:termvar) (C:ctx) (p1:prg) (p2:prg)
 | Ccase1 (x:termvar) (p0:prg) (C:ctx) (p2:prg)
 | Ccase2 (x:termvar) (p0:prg) (p1:prg) (C:ctx)
 | Croll (ty:typ) (C:ctx)
 | Cunroll (C:ctx)
 | Cfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (C:ctx)
 | Capp1 (C:ctx) (p:prg)
 | Capp2 (p:prg) (C:ctx)
 | Cgen (C:ctx)
 | Cinst (C:ctx) (ty:typ)
 | Cpack (ty1:typ) (C:ctx) (ty:typ)
 | Cunpack1 (x:termvar) (C:ctx) (p:prg)
 | Cunpack2 (x:termvar) (p:prg) (C:ctx)
 | Cnew (C:ctx)
 | Cread (C:ctx)
 | Cwrite1 (C:ctx) (p:prg)
 | Cwrite2 (p:prg) (C:ctx)
 | Crefeq1 (C:ctx) (p:prg)
 | Crefeq2 (p:prg) (C:ctx)
 | Cthrow1 (C:ctx) (p:prg)
 | Cthrow2 (p:prg) (C:ctx)
 | Cisolate (C:ctx)
.

Fixpoint app_ctx (C : ctx) (p0: prg) : prg :=
  match C with
  | Chole => p0                         
  | (Cpair1 C p2) => (Ppair  (  (app_ctx  C   p0 )  )  p2)
  | (Cpair2 p1 C) => (Ppair p1  (  (app_ctx  C   p0 )  ) )
  | (Cinl ty C) => (Pinl ty  (  (app_ctx  C   p0 )  ) )
  | (Cinr ty C) => (Pinr ty  (  (app_ctx  C   p0 )  ) )
  | (Cfst C) => (Pfst  (  (app_ctx  C   p0 )  ) )
  | (Csnd C) => (Psnd  (  (app_ctx  C   p0 )  ) )
  | (Cif0 C p1 p2) => (Pif  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Cif1 p C p2) => (Pif p  (  (app_ctx  C   p0 )  )  p2)
  | (Cif2 p p1 C) => (Pif p p1  (  (app_ctx  C   p0 )  ) )
  | (Ccase0 x C p1 p2) => (Pcase x  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Ccase1 x p C p2) => (Pcase x p  (app_ctx  C   p0 )  p2)
  | (Ccase2 x p p1 C) => (Pcase x p p1  (app_ctx  C   p0 ) )
  | (Croll ty C) => (Proll ty  (  (app_ctx  C   p0 )  ) )
  | (Cunroll C) => (Punroll  (  (app_ctx  C   p0 )  ) )
  | (Cfix f x ty ty' C) => (Pfix f x ty ty'  (  (app_ctx  C   p0 )  ) )
  | (Capp1 C p) => (Papp  (  (app_ctx  C   p0 )  )  p)
  | (Capp2 p C) => (Papp p  (  (app_ctx  C   p0 )  ) )
  | (Cgen C) => (Pgen  (  (app_ctx  C   p0 )  ) )
  | (Cinst C ty) => (Pinst  (  (app_ctx  C   p0 )  )  ty)
  | (Cpack ty1 C ty) => (Ppack ty1  (app_ctx  C   p0 )  ty)
  | (Cunpack1 x C p) => (Punpack x  (app_ctx  C   p0 )  p)
  | (Cunpack2 x p C) => (Punpack x p  (  (app_ctx  C   p0 )  ) )
  | (Cnew C) => (Pnew  (  (app_ctx  C   p0 )  ) )
  | (Cread C) => (Pread  (  (app_ctx  C   p0 )  ) )
  | (Cwrite1 C p) => (Pwrite  (  (app_ctx  C   p0 )  )  p)
  | (Cwrite2 p C) => (Pwrite p  (  (app_ctx  C   p0 )  ) )
  | (Crefeq1 C p) => (Prefeq  (  (app_ctx  C   p0 )  )  p)
  | (Crefeq2 p C) => (Prefeq p  (  (app_ctx  C   p0 )  ) )
  | (Cthrow1 C p) => (Pthrow (app_ctx C p0) p)
  | (Cthrow2 p C) => (Pthrow p (app_ctx C p0))
  | (Cisolate C) => (Pisolate (app_ctx C p0))
  end.

(** definitions *)

(* defns EnvOKjudg *)

Definition typeenv : Set := (list typename).

Definition termenv : Set := (list (termvar * typ)).

Fixpoint ftn_typL (ty_5:typ) : list typename :=
  match ty_5 with
  | (Tname aa) => (cons aa nil)
  | (Tbvar bb) => nil
  | Tunit => nil
  | Tint => nil
  | Tbool => nil
  | (Tprod ty1 ty2) => (app (ftn_typL ty1) (ftn_typL ty2))
  | (Tsum ty1 ty2) => (app (ftn_typL ty1) (ftn_typL ty2))
  | (Tarrow ty1 ty2) => (app (ftn_typL ty1) (ftn_typL ty2))
  | (Tref ty) => ((ftn_typL ty))
  | (Tmu ty) => ((ftn_typL ty))
  | (Tforall ty) => ((ftn_typL ty))
  | (Texists ty) => ((ftn_typL ty))
  | (Tcont ty) => ((ftn_typL ty))
end.

Inductive wfenv : typeenv -> termenv -> Prop :=    (* defn wfenv *)
 | wfenv_ok : forall (D:typeenv) (G:termenv),
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (lc_typ 0  ty )  )   )  )  ->
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (list_minus eq_typename   (ftn_typL  ty )    D  = nil)  )   )  )  ->
     wfenv D G.
(** definitions *)

(* defns TypeOKjudg *)
Inductive wftyp_gen : typeenv -> nat -> typ -> Prop :=    (* defn wftyp *)
 | wftyp_ok : forall (D:typeenv) n (ty:typ),
      (lc_typ n ty )  ->
      (list_minus eq_typename   (ftn_typL  ty )    D  = nil)  ->
     wftyp_gen D n ty.

Definition wftyp D ty := wftyp_gen D 0 ty.

(** definitions *)

(* defns TypingJudg *)
Inductive hastype : typeenv -> termenv -> exp -> typ -> Prop :=    (* defn hastype *)
 | HTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     hastype D G (Evar x) ty
 | HTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Eunit Tunit
 | HTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     hastype D G (Eint n) Tint
 | HTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Etrue Tbool
 | HTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Efalse Tbool
 | HTpair : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty1 ty2:typ),
     hastype D G e1 ty1 ->
     hastype D G e2 ty2 ->
     hastype D G (Epair e1 e2) (Tprod ty1 ty2)
 | HTfst : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Efst e) ty1
 | HTsnd : forall (D:typeenv) (G:termenv) (e:exp) (ty2 ty1:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Esnd e) ty2
 | HTinl : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty1 ->
     wftyp D ty2 ->
     hastype D G (Einl e) (Tsum ty1 ty2)
 | HTinr : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty2 ->
     wftyp D ty1 ->
     hastype D G (Einr e) (Tsum ty1 ty2)
 | HTif : forall (D:typeenv) (G:termenv) (e0 e1 e2:exp) (ty:typ),
     hastype D G e0 Tbool ->
     hastype D G e1 ty ->
     hastype D G e2 ty ->
     hastype D G (Eif e0 e1 e2) ty
 | HTcase : forall (D:typeenv) (G:termenv) (x:termvar) (e0 e1 e2:exp) (ty ty1 ty2:typ),
     hastype D G e0 (Tsum ty1 ty2) ->
     hastype D  (cons ( x , ty1 )  G )  e1 ty ->
     hastype D  (cons ( x , ty2 )  G )  e2 ty ->
     hastype D G (Ecase x e0 e1 e2) ty
 | HTroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e  (tsubst_bvar 0  (Tmu ty)   ty )  ->
     wftyp D (Tmu ty) ->
     hastype D G (Eroll e) (Tmu ty)
 | HTunroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tmu ty) ->
     hastype D G (Eunroll e)  (tsubst_bvar 0  (Tmu ty)   ty ) 
 | HTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (e:exp) (ty1 ty2:typ),
     hastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  e ty2 ->
     wfenv D G ->
     hastype D G (Efix f x e) (Tarrow ty1 ty2)
 | HTapp : forall (D:typeenv) (G:termenv) (e e':exp) (ty2 ty1:typ),
     hastype D G e (Tarrow ty1 ty2) ->
     hastype D G e' ty1 ->
     hastype D G (Eapp e e') ty2
 | HTgen : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ) (aa:typename),
     hastype  (cons  aa   D )  G e  (tsubst_bvar 0  (Tname aa)   ty )  ->
     wfenv D G ->
     wftyp_gen D 1 ty ->
      (~ In  aa   D )  ->
     hastype D G (Egen e) (Tforall ty)
 | HTinst : forall (D:typeenv) (G:termenv) (e:exp) (ty' ty:typ),
     hastype D G e (Tforall ty) ->
     wftyp D ty' ->
     hastype D G (Einst e)  (tsubst_bvar 0  ty'   ty ) 
 | HTpack : forall (D:typeenv) (G:termenv) (e:exp) (ty ty':typ),
     hastype D G e  (tsubst_bvar 0  ty'   ty )  ->
     wftyp D ty' ->
     hastype D G (Epack e) (Texists ty)
 | HTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (e1 e2:exp) (ty2 ty:typ) (aa:typename),
     hastype D G e1 (Texists ty) ->
     hastype  (cons  aa   D )   (cons ( x ,  (tsubst_bvar 0  (Tname aa)   ty )  )  G )  e2 ty2 ->
     wftyp D ty2 ->
      (~ In  aa   D )  ->
     hastype D G (Eunpack x e1 e2) ty2
 | HTnew : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e ty ->
     hastype D G (Enew e) (Tref ty)
 | HTread : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tref ty) ->
     hastype D G (Eread e) ty
 | HTwrite : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 ty ->
     hastype D G (Ewrite e1 e2) Tunit
 | HTrefeq : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 (Tref ty) ->
     hastype D G (Erefeq e1 e2) Tbool
 | HTcallcc : forall (D:typeenv) (G:termenv) (x:termvar) (e:exp) (ty:typ),
     hastype D (cons (x, Tcont ty) G) e ty ->
     wfenv D G ->
     hastype D G (Ecallcc x e) ty
 | HTthrow : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty1 ty2:typ),
     hastype D G e1 ty1 ->
     hastype D G e2 (Tcont ty1) ->
     wftyp D ty2 ->
     hastype D G (Ethrow e1 e2) ty2
 | HTisolate : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tarrow ty Tint) ->
     hastype D G (Eisolate e) (Tcont ty)
.

(* defns ProgTypingJudg *)
Inductive phastype : typeenv -> termenv -> prg -> typ -> Prop :=    (* defn phastype *)
 | PHTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     phastype D G (Pvar x) ty
 | PHTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Punit Tunit
 | PHTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     phastype D G (Pint n) Tint
 | PHTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Ptrue Tbool
 | PHTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Pfalse Tbool
 | PHTpair : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty1 ty2:typ),
     phastype D G p1 ty1 ->
     phastype D G p2 ty2 ->
     phastype D G (Ppair p1 p2) (Tprod ty1 ty2)
 | PHTfst : forall (D:typeenv) (G:termenv) (p:prg) (ty1 ty2:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Pfst p) ty1
 | PHTsnd : forall (D:typeenv) (G:termenv) (p:prg) (ty2 ty1:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Psnd p) ty2
 | PHTinl : forall (D:typeenv) (G:termenv) (ty2:typ) (p:prg) (ty1:typ),
     phastype D G p ty1 ->
     wftyp D ty2 ->
     phastype D G (Pinl ty2 p) (Tsum ty1 ty2)
 | PHTinr : forall (D:typeenv) (G:termenv) (ty1:typ) (p:prg) (ty2:typ),
     phastype D G p ty2 ->
     wftyp D ty1 ->
     phastype D G (Pinr ty1 p) (Tsum ty1 ty2)
 | PHTif : forall (D:typeenv) (G:termenv) (p0 p1 p2:prg) (ty:typ),
     phastype D G p0 Tbool ->
     phastype D G p1 ty ->
     phastype D G p2 ty ->
     phastype D G (Pif p0 p1 p2) ty
 | PHTcase : forall (D:typeenv) (G:termenv) (x:termvar) (p0 p1 p2:prg) (ty ty1 ty2:typ),
     phastype D G p0 (Tsum ty1 ty2) ->
     phastype D  (cons ( x , ty1 )  G )  p1 ty ->
     phastype D  (cons ( x , ty2 )  G )  p2 ty ->
     phastype D G (Pcase x p0 p1 p2) ty
 | PHTroll : forall (D:typeenv) (G:termenv) (ty:typ) (p:prg),
     phastype D G p  (tsubst_bvar 0  (Tmu ty)   ty )  ->
     wftyp D (Tmu ty) ->
     phastype D G (Proll ty p) (Tmu ty)
 | PHTunroll : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tmu ty) ->
     phastype D G (Punroll p)  (tsubst_bvar 0  (Tmu ty)   ty ) 
 | PHTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (ty1 ty2:typ) (p:prg),
     phastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  p ty2 ->
     wfenv D G ->
     phastype D G (Pfix f x ty1 ty2 p) (Tarrow ty1 ty2)
 | PHTapp : forall (D:typeenv) (G:termenv) (p p':prg) (ty2 ty1:typ),
     phastype D G p (Tarrow ty1 ty2) ->
     phastype D G p' ty1 ->
     phastype D G (Papp p p') ty2
 | PHTgen : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ) (aa:typename),
     phastype  (cons  aa   D )  G p  (tsubst_bvar 0  (Tname aa)   ty )  ->
     wfenv D G ->
     wftyp_gen D 1 ty ->
      (~ In  aa   D )  ->
     phastype D G (Pgen p) (Tforall ty)
 | PHTinst : forall (D:typeenv) (G:termenv) (p:prg) (ty' ty:typ),
     phastype D G p (Tforall ty) ->
     wftyp D ty' ->
     phastype D G (Pinst p ty')  (tsubst_bvar 0  ty'   ty ) 
 | PHTpack : forall (D:typeenv) (G:termenv) (ty':typ) (p:prg) (ty:typ),
     phastype D G p  (tsubst_bvar 0  ty'   ty )  ->
     wftyp D ty' ->
     phastype D G (Ppack ty' p ty) (Texists ty)
 | PHTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (p1 p2:prg) (ty2 ty:typ) (aa:typename),
     phastype D G p1 (Texists ty) ->
     phastype  (cons  aa   D )   (cons ( x ,  (tsubst_bvar 0  (Tname aa)   ty )  )  G )  p2 ty2 ->
     wftyp D ty2 ->
      (~ In  aa   D )  ->
     phastype D G (Punpack x p1 p2) ty2
 | PHTnew : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p ty ->
     phastype D G (Pnew p) (Tref ty)
 | PHTread : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tref ty) ->
     phastype D G (Pread p) ty
 | PHTwrite : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 ty ->
     phastype D G (Pwrite p1 p2) Tunit
 | PHTrefeq : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 (Tref ty) ->
     phastype D G (Prefeq p1 p2) Tbool
 | PHTcallcc : forall (D:typeenv) (G:termenv) (x:termvar) (p:prg) (ty:typ),
     phastype D (cons (x, Tcont ty) G) p ty ->
     wfenv D G ->
     phastype D G (Pcallcc x p) ty
 | PHTthrow : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty1 ty2:typ),
     phastype D G p1 ty1 ->
     phastype D G p2 (Tcont ty1) ->
     wftyp D ty2 ->
     phastype D G (Pthrow p1 p2) ty2
 | PHTisolate : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tarrow ty Tint) ->
     phastype D G (Pisolate p) (Tcont ty)
.

(* defns CtxJudg *)
Inductive wfctx : ctx -> typeenv -> termenv -> typ -> typeenv -> termenv -> typ -> Prop :=    (* defn wfctx *)
 | wfctx_hole : forall (D:typeenv) (G:termenv) (ty:typ),
     wfenv D G ->
     wftyp D ty ->
     wfctx Chole D G ty D G ty
 | wfctx_pair1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty2:typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty2 ->
     wfctx (Cpair1 C p) D G ty D' G' (Tprod ty' ty2)
 | wfctx_pair2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty1 ->
     wfctx (Cpair2 p C) D G ty D' G' (Tprod ty1 ty')
 | wfctx_inl : forall (ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' ty2 ->
     wfctx (Cinl ty2 C) D G ty D' G' (Tsum ty' ty2)
 | wfctx_inr : forall (ty1:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' ty1 ->
     wfctx (Cinr ty1 C) D G ty D' G' (Tsum ty1 ty')
 | wfctx_fst : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty2:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Cfst C) D G ty D' G' ty1
 | wfctx_snd : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Csnd C) D G ty D' G' ty2
 | wfctx_if0 : forall (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' Tbool ->
     phastype D' G' p1 ty' ->
     phastype D' G' p2 ty' ->
     wfctx (Cif0 C p1 p2) D G ty D' G' ty'
 | wfctx_if1 : forall (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p2 ty' ->
     wfctx (Cif1 p0 C p2) D G ty D' G' ty'
 | wfctx_if2 : forall (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p1 ty' ->
     wfctx (Cif2 p0 p1 C) D G ty D' G' ty'
 | wfctx_case0 : forall (x:termvar) (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     wfctx C D G ty D' G' (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase0 x C p1 p2) D G ty D' G' ty'
 | wfctx_case1 : forall (x:termvar) (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     wfctx C D G ty D'  (cons ( x , ty1 )  G' )  ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase1 x p0 C p2) D G ty D' G' ty'
 | wfctx_case2 : forall (x:termvar) (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     wfctx C D G ty D'  (cons ( x , ty2 )  G' )  ty' ->
     wfctx (Ccase2 x p0 p1 C) D G ty D' G' ty'
 | wfctx_roll : forall (ty':typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wftyp D' (Tmu ty') ->
     wfctx C D G ty D' G'  (tsubst_bvar 0  (Tmu ty')   ty' )  ->
     wfctx (Croll ty' C) D G ty D' G' (Tmu ty')
 | wfctx_unroll : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tmu ty') ->
     wfctx (Cunroll C) D G ty D' G'  (tsubst_bvar 0  (Tmu ty')   ty' ) 
 | wfctx_fix : forall (f x:termvar) (ty1 ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D'  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G' )  )  ty2 ->
     wfenv D' G' ->
     wfctx (Cfix f x ty1 ty2 C) D G ty D' G' (Tarrow ty1 ty2)
 | wfctx_app1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tarrow ty1 ty2) ->
     phastype D' G' p ty1 ->
     wfctx (Capp1 C p) D G ty D' G' ty2
 | wfctx_app2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     phastype D' G' p (Tarrow ty1 ty2) ->
     wfctx C D G ty D' G' ty1 ->
     wfctx (Capp2 p C) D G ty D' G' ty2
 | wfctx_gen : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ) (aa:typename),
     wfctx C D G ty  (cons  aa   D' )  G'  (tsubst_bvar 0  (Tname aa)   ty' )  ->
     wfenv D' G' ->
     wftyp_gen D' 1 ty' ->
      (~ In  aa   D' )  ->
     wfctx (Cgen C) D G ty D' G' (Tforall ty')
 | wfctx_inst : forall (C:ctx) (ty1:typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tforall ty') ->
     wftyp D' ty1 ->
     wfctx (Cinst C ty1) D G ty D' G'  (tsubst_bvar 0  ty1   ty' ) 
 | wfctx_pack : forall (ty1:typ) (C:ctx) (ty':typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D' G'  (tsubst_bvar 0  ty1   ty' )  ->
     wftyp D' ty1 ->
     wfctx (Cpack ty1 C ty') D G ty D' G' (Texists ty')
 | wfctx_unpack1 : forall (x:termvar) (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     wfctx C D G ty D' G' (Texists ty') ->
     phastype  (cons  aa   D' )   (cons ( x ,  (tsubst_bvar 0  (Tname aa)   ty' )  )  G' )  p ty2 ->
     wftyp D' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack1 x C p) D G ty D' G' ty2
 | wfctx_unpack2 : forall (x:termvar) (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     phastype D' G' p (Texists ty') ->
     wfctx C D G ty  (cons  aa   D' )   (cons ( x ,  (tsubst_bvar 0  (Tname aa)   ty' )  )  G' )  ty2 ->
     wftyp D' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack2 x p C) D G ty D' G' ty2
 | wfctx_new : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wfctx (Cnew C) D G ty D' G' (Tref ty')
 | wfctx_read : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Cread C) D G ty D' G' ty'
 | wfctx_write1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p ty' ->
     wfctx (Cwrite1 C p) D G ty D' G' Tunit
 | wfctx_write2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' ty' ->
     wfctx (Cwrite2 p C) D G ty D' G' Tunit
 | wfctx_refeq1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p (Tref ty') ->
     wfctx (Crefeq1 C p) D G ty D' G' Tbool
 | wfctx_refeq2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Crefeq2 p C) D G ty D' G' Tbool
 | wfctx_throw1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ) (ty'':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p (Tcont ty') ->
     wftyp D' ty'' ->
     wfctx (Cthrow1 C p) D G ty D' G' ty''
 | wfctx_throw2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ) (ty'':typ),
     phastype D' G' p ty' ->
     wfctx C D G ty D' G' (Tcont ty') ->
     wftyp D' ty'' ->
     wfctx (Cthrow2 p C) D G ty D' G' ty''
 | wfctx_isolate : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tarrow ty' Tint) ->
     wfctx (Cisolate C) D G ty D' G' (Tcont ty')
.

Fixpoint erase_prg (p : prg) : exp :=
  match p with
  | (Pvar x) => Evar x
  | Punit => Eunit
  | (Pint n) => Eint n
  | Ptrue => Etrue
  | Pfalse => Efalse
  | (Ppair p1 p2) => (Epair  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pinl ty p) => (Einl  (erase_prg  p ) )
  | (Pinr ty p) => (Einr  (erase_prg  p ) )
  | (Pfst p) => (Efst  (erase_prg  p ) )
  | (Psnd p) => (Esnd  (erase_prg  p ) )
  | (Pif p0 p1 p2) => (Eif  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pcase x p0 p1 p2) => (Ecase x  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Proll ty p) => (Eroll  (erase_prg  p ) )
  | (Punroll p) => (Eunroll  (erase_prg  p ) )
  | (Pfix f x ty ty' p) => (Efix f x  (erase_prg  p ) )
  | (Papp p p') => (Eapp  (erase_prg  p )   (erase_prg  p' ) )
  | (Pgen p) => (Egen  (erase_prg  p ) )
  | (Pinst p ty) => (Einst  (erase_prg  p ) )
  | (Ppack ty1 p ty) => (Epack  (erase_prg  p ) )
  | (Punpack x p1 p2) => (Eunpack x  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pnew p) => (Enew  (erase_prg  p ) )
  | (Pread p) => (Eread  (erase_prg  p ) )
  | (Pwrite p1 p2) => (Ewrite  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Prefeq p1 p2) => (Erefeq  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pcallcc x p) => (Ecallcc x (erase_prg p))
  | (Pthrow p1 p2) => (Ethrow  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pisolate p) => (Eisolate (erase_prg p))
  end.

Hint Constructors hastype : hastype.
Hint Constructors phastype : hastype.

Coercion erase_prg : prg >-> exp.

Lemma hastype_erase :
  forall D G p ty, phastype D G p ty -> hastype D G (erase_prg p) ty.
Proof. induction 1; simpl in *; eauto with hastype. Qed.

Lemma phastype_app_ctx :
  forall D G C ty D' G' ty', 
    wfctx C D G ty D' G' ty' -> forall p,
    phastype D G p ty -> 
    phastype D' G' (app_ctx C p) ty'.
Proof. induction 1; intros; simpl in *; eauto with hastype. Qed.

Lemma phastype_fv: forall D G e ty
    (WF: phastype D G e ty)
    x (NOTIN: forall ty', In (x,ty') G -> False),
  fv_exp e x -> False.
Proof.
  induction 1; simpl; intros; eauto; try (repeat destruct H; eauto; fail).
  - subst; eauto using list_assoc_someD.
  - destruct H; eauto; destruct H. 
    destruct H0; [eapply IHWF2|eapply IHWF3]; eauto
    ; (intros; destruct H1; [inversion H1|]; eauto).
  - destruct H0 as (?&?&?).
    eapply IHWF; eauto; repeat (intros; destruct H3; [inversion H3|]; eauto).
  - destruct H1; eauto; destruct H1.
    eapply IHWF2; eauto; intros; destruct H3; [inversion H3|]; eauto.
  - destruct H0; eapply IHWF; eauto.
    intros; destruct H2; [inversion H2|]; eauto.
  - destruct H0; eauto.
Qed.

