Require Import vgtac Program List Arith Even Div2 Relations.
Require Import basis countable lang subs model basiclemmas wiso wcomplete.
Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** * Properties about extending worlds *)
(*----------------------------------------------------------------------------*)

Lemma wexp_extend1 : forall (w1: world_local) (w2: world_local) 
    (DISJ : wdisj w1 w2) (STA : wstable w2)
    g (G : wglob (wlift (wprod w1 w2)) g) sr s1 s2,
  wexp (wlift w1) (fun p => gmon w2 (fun s2' => g (fst p, (snd p, s2'))) s2) (sr, s1) <4=
  wexp (wlift (wprod w1 w2)) g (sr, (s1, s2)).
Proof.
  ii; rename h into hrab1, h' into hrab2, hh into hrabf1, hh' into hrafb2. 
  rename hf into hf1, hf' into hf2, HPLUS into Prab1f1, HPLUS' into Prab2f2.
  destruct HREL as (hr1&hr2&hab1&hab2&Hr1r2&Hab1ab2&Pr1ab1&Pr2ab2).
  destruct Hab1ab2 as (ha1&ha2&hb1&hb2&Ha1a2&Hb1b2&Pa1b1&Pa2b2).
  destruct (hplus_assoc_r Pa1b1 Pr1ab1) as (hra1 & Pr1a1 & Pra1b1).
  destruct (hplus_assoc_r Pa2b2 Pr2ab2) as (hra2 & Pr2a2 & Pra2b2).
  destruct (hplus_assoc_l Pra1b1 Prab1f1) as (hbf1 & Pra1bf1 & Pb1f1).
  destruct (hplus_assoc_l Pra2b2 Prab2f2) as (hbf2 & Pra2bf2 & Pb2f2).
  edestruct PR; [by rr; esplits; [eapply wheap_mon| | |]; eauto; inst; s; eauto| | | |]; eauto.
  destruct s' as (sr', s1'); rrdes EXT.
  destruct HREL as (hr1'&hr2'&hra1'&hra2'&Hr1'r2'&Ha1'a2'&Pr1'ra1'&Pr2'ra2').
  destruct (hplus_assoc_r Pb1f1 PLUS1) as (hhb1 & Ph1hb1 & Phb1f1).
  destruct (hplus_assoc_r Pb2f2 PLUS2) as (hhb2 & Ph2hb2 & Phb2f2).
  destruct (hplus_assoc_l Pr1'ra1' Ph1hb1) as (ra'b1 & Pr1'ra'b1 & Pra1'b1).
  destruct (hplus_assoc_l Pr2'ra2' Ph2hb2) as (ra'b2 & Pr2'ra'b2 & Pra2'b2).
  destruct (hplus_assoc_r (hplus_comm Pra1'b1) Pr1'ra'b1) as (hr1'b1 & Pr1'b1 & Pr1'b1ra1').
  destruct (hplus_assoc_r (hplus_comm Pra2'b2) Pr2'ra'b2) as (hr2'b2 & Pr2'b2 & Pr2'b2ra2').

  assert (wglob': wglob (wlift w2) (fun s=> gmon w1 (fun s1'=> g (fst s,(s1',snd s))) s1')).
    by apply wglob_prod2; eauto.

  edestruct (STA _ wglob') as [s2'].
  - eapply wheap_mon; eauto; instantiate; ii; eapply gmon_mon; ss; eauto.
  - eauto.
  - destruct G; eapply wheap_mon; [eauto|].
    i; ss; inversion PR0; vauto; econs; try eapply wg_mon; s; eauto.
  - eauto.
  - eauto.
  - des; eapply wexp_Stuck with (s':=(sr', (s1', s2'))); eauto.
    + by ss; eauto 10.
    + eapply wstuck_mon; eauto
      ; i; apply (gmon_prod2 G) in PR0; eapply (wg_mon G); s; eauto.
    + rr; esplits; [| |by apply Pr1'ra'b1|by apply Pr2'ra'b2].
      * eapply wheap_mon; [apply Hr1'r2'|].
        by s; i; eapply (gmon_prod2 G) in PR0; eapply (wg_mon G); s; eauto.
      * by rr; esplits; [eapply wheap_mon; [apply Ha1'a2'|]| | |]
           ; inst; s; eauto using gmon_mon.
Qed.

Lemma wexp_extend2 : forall (w1: world_local) (w2: world_local)
    (DISJ : wdisj w1 w2) (STA : wstable w1)
    g (G : wglob (wlift (wprod w1 w2)) g) sr s1 s2,
  wexp (wlift w2) (fun p => gmon w1 (fun s1' => g (fst p, (s1', snd p))) s1) (sr, s2) <4=
  wexp (wlift (wprod w1 w2)) g (sr, (s1, s2)).
Proof.
  ii; eapply (wexp_iso (swapiso w2 w1)); eauto.
  eapply wexp_extend1; eauto; eapply wiso_glob; eauto.
Qed.

Lemma wexp_lift: forall w g sr s n1 e1 n2 e2
    (STA: wstable w)
    (G: wglob (wlift w) g) 
    (EXP: wexp wref (fun sr => g (sr,s)) sr n1 e1 n2 e2),
  wexp (wlift w) g (sr, s) n1 e1 n2 e2.
Proof.
  i; eapply (wexp_iso (id_intro_iso w)); eauto.
  eapply wexp_extend2; eauto using (wiso_glob (id_intro_iso _)); try done.
  eapply (wexp_iso id_elim_iso); eauto.
  - rewrite (gmon_glob_eq G) in G.
    eapply wglob_wid in G; eauto.
  - rewrite (gmon_glob_eq G) in EXP.
    eauto.
Qed.

Lemma winhabited_prod: forall (w1: world_local) (w2: world_local)
    (DISJ: wdisj w1 w2) (I1: winhabited (wlift w1)) (I2: winhabited (wlift w2)),
  winhabited (wlift (wprod w1 w2)).
Proof.
  r; i; rrdes I1; rrdes I2; destruct s0 as (sr,s0), s1 as (sr',s1). 
  exists (sr, (s0, s1)); i.
  specialize (I1 _ (wglob_prod1 DISJ gOK s1)).
  specialize (I2 _ (wglob_prod2 DISJ gOK s0)).
  rrdes I1; apply hplus_hemp_inv in HP; apply hplus_hemp_inv in HP'; des; subst.
  rrdes I2; apply hplus_hemp_inv in HP; apply hplus_hemp_inv in HP'; des; subst.
  apply sref0_hemp in WH; apply sref0_hemp in WH0; unfold fst in *; subst.
  by do 4 eexists hemp; rr; esplits; rr; esplits; eauto.
Qed.

Lemma wconsistent_prod:
  forall (w1: world_local) (w2: world_local) 
    (WD: wdisj w1 w2)
    (C1: wconsistent (wlift w1)) (STA1: wstable w1)
    (C2: wconsistent (wlift w2)) (STA2: wstable w2),
  wconsistent (wlift (wprod w1 w2)).
Proof. 
  ii; destruct s as [sr [s1 s2]].
  apply wstuck_lift in PR; s in PR; eapply wstuck_union in PR; des; i.
  - r in C1; eapply mp; [eapply (C1 (fun p=> gmon w2 (fun s2'=> g (fst p,(snd p,s2'))) s2))|].
    + eapply wglob_prod1; eauto.
    + eapply wstuck_lift; instantiate (6 := (sr,s1)).
      eapply wstuck_mon; eauto; i; eapply wval_mon; eauto.
    + i; destruct H; econs; eauto.
      eapply wexp_extend1; eauto.
  - r in C2; eapply mp; [eapply (C2 (fun p=> gmon w1 (fun s1'=> g (fst p,(s1',snd p))) s1))|].
    + eapply wglob_prod2; eauto.
    + eapply wstuck_lift; instantiate (6 := (sr,s2)).
      eapply wstuck_mon; eauto; i; eapply wval_mon; eauto.
    + i; destruct H; econs; eauto.
      eapply wexp_extend2; eauto.
Qed.

Lemma wstable_prod : forall (w1: world_local) (w2: world_local) 
    (WD: wdisj w1 w2) (STA1: wstable w1) (STA2: wstable w2),
  wstable (wprod w1 w2).
Proof.
  r; i; destruct s as (s1 & s2); s in HR; des.
  destruct (hplus_assoc_r HP DIS) as (hsh1 & SHR1 & SHH1).
  destruct (hplus_assoc_r HP' DIS0) as (hsh2 & SHR2 & SHH2).
  apply hplus_comm in HP; eapply hplus_comm in HP'.
  destruct (hplus_assoc_r HP DIS) as (hsh1' & SHR1' & SHH1').
  destruct (hplus_assoc_r HP' DIS0) as (hsh2' & SHR2' & SHH2').
  refine (mp (STA1 (fun s => gmon w2 (fun s2 => g (fst s,(snd s,s2))) s2)
                   _ _ _ _ _ _ _ EXT _ _ _ _ _ _ _) _).
  - eauto using wglob_prod1.
  - eapply wheap_mon; eauto.
  - eapply wheap_mon; eauto; instantiate; i; ss; eauto.
  - eauto.
  - eauto.
  - refine (mp (STA2 (fun s => gmon w1 (fun s1 => g (fst s,(s1,snd s))) s1)
                     _ _ _ _ _ _ _ EXT _ _ _ _ _ _ _) _).
    + eauto using wglob_prod2.
    + eapply wheap_mon; eauto. 
    + eapply wheap_mon; eauto; instantiate; i; ss; eauto.
    + eauto.
    + eauto.
    + i; des; exists (s',s'0); splits; s; eauto.
      esplits; try (eapply wheap_mon; eauto; instantiate; ss); eauto using gmon_mon.
Qed.

Hint Resolve winhabited_prod wconsistent_prod wstable_prod.

(*----------------------------------------------------------------------------*)
(** * Compatibility *)
(*----------------------------------------------------------------------------*)

Create HintDb compat.

(** The next set of lemmas are helper lemmas for proving the compatibility
    lemmas. *)

Lemma meta0r : 
  forall G ty e e' 
    (WF: wfprog G ty e e')
    (OK: wequiv (wlift wid) G ty e e'),
  wreq G ty e e'.
Proof.
  split; eauto.
  ii; eexists wid; esplits; rr; ss; eauto.
Qed.

Lemma meta1r : 
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world_local) (STA: wstable w) (CONS: wconsistent (wlift w))
           (E1: wequiv (wlift (wcomp w)) G1 ty1 e1 e1'),
         wequiv (wlift (wcomp w)) G ty e e'),
  wreq G ty e e'.
Proof.
  i; rdes E1; split; eauto.
  ii; specialize (REQ _ INF); des; unfold weq; rdes WEQ.
  edestruct OK; eauto.
  - depdes EQ; econs; eauto.
    i; eapply WEXP; eauto; apply wcomp_glob; eauto.
  - exists (wcomp w); esplits.
    + eauto.
    + ii; eapply WSTA; eauto using wcomp_glob. 
    + rrdes INH; rr; esplits; i; eapply INH, wcomp_glob; eauto.
    + r; setoid_rewrite wlift_wcomp; apply (@wkcomp_consistent (wlift w)).
    + eauto.
Qed.

Lemma meta1 : 
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world)
           (E1: wequivgen w (wkcomp w) G1 ty1 e1 e1'),
         wequivgen w (wkcomp w) G ty e e'),
  wreq G ty e e'.
Proof.
  i; eapply meta1r; eauto.
  by unfold wequiv; setoid_rewrite wlift_wcomp; i; eapply (OK (wlift w)).
Qed.

Lemma meta2r : 
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G2 ty2 e2 e2' (E2: wreq G2 ty2 e2 e2')
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world_local)
           (STA: wstable w) (CONS: wconsistent (wlift w))
           (E1: wequiv (wlift (wcomp w)) G1 ty1 e1 e1')
           (E2: wequiv (wlift (wcomp w)) G2 ty2 e2 e2'),
         wequiv (wlift (wcomp w)) G ty e e'),
    wreq G ty e e'.
Proof.
  i; rdes E1; rdes E2; split; eauto.
  ii; apply countably_infinite_split in INF; des; subst.
  destruct (REQ _ Nci1) as [w1 PW].
  destruct (REQ0 _ Nci2) as [w2 PW']; des; clarify.
  assert (DIS: wdisj w1 w2) by (eby eapply ndisj_mon).
  rdes WEQ0; rdes WEQ; destruct EQ, EQ0.
  eapply mp; [eapply (OK (wprod w1 w2)); eauto; econs; i; eauto
              ; eapply wcomp_glob in gOK; eauto
              ; destruct s as [sref [s1 s2]] | ].
  - eapply wexp_extend1; eauto.
    eapply WEXP; eauto.
    + apply wglob_prod1; eauto using wconsistent_prod.
    + s; destruct dgOK; econs; eauto using vclos_mon.
  - eapply wexp_extend2; eauto.
    eapply WEXP0; eauto.
    + apply wglob_prod2; eauto using wconsistent_prod.
    + s; destruct dgOK; econs; eauto using vclos_mon.
  - unfold weq; i; eexists (wcomp (wprod w1 w2)); esplits.
     + s; i; r; rdes PR; eauto.
     + ii; eapply (wstable_prod DIS WSTA0 WSTA); eauto using wcomp_glob.
     + assert (X:=winhabited_prod DIS INH INH0).
       rrdes X; rr; esplits; i; eapply X; eauto.
       apply wcomp_glob; eauto.
     + r; setoid_rewrite wlift_wcomp. 
       apply (@wkcomp_consistent (wlift (wprod w1 w2))).
     + eauto.
Qed.

Lemma meta2 :
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G2 ty2 e2 e2' (E2: wreq G2 ty2 e2 e2')
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world)
           (E1: wequivgen w (wkcomp w) G1 ty1 e1 e1')
           (E2: wequivgen w (wkcomp w) G2 ty2 e2 e2'),
         wequivgen w (wkcomp w) G ty e e'),
  wreq G ty e e'.
Proof.
  i; eapply meta2r; [apply E1|apply E2|eauto|].
  by unfold wequiv; setoid_rewrite wlift_wcomp; i; eapply (OK (wlift w)).
Qed.

Lemma meta3r : 
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G2 ty2 e2 e2' (E2: wreq G2 ty2 e2 e2')
    G3 ty3 e3 e3' (E3: wreq G3 ty3 e3 e3') 
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world_local)
           (STA: wstable w) (CONS: wconsistent (wlift w))
           (E1: wequiv (wlift (wcomp w)) G1 ty1 e1 e1')
           (E2: wequiv (wlift (wcomp w)) G2 ty2 e2 e2')
           (E3: wequiv (wlift (wcomp w)) G3 ty3 e3 e3'),
         wequiv (wlift (wcomp w)) G ty e e'),
  wreq G ty e e'.
Proof.
  i; rdes E1; rdes E2; rdes E3; split; eauto.
  red; intros; apply countably_infinite_split in INF; des; subst.
  apply countably_infinite_split in Nci2; des; subst.
  destruct (REQ _ Nci1) as [w1 PWA].
  destruct (REQ0 _ Nci0) as [w2 PWB].
  destruct (REQ1 _ Nci2) as [w3 PWC]; des; clarify.
  assert (DISJ23: wdisj w2 w3) by (eapply (ndisj_mon Ndisj0); eauto).
  assert (DISJ: wdisj w1 (wprod w2 w3)).
    by eapply (ndisj_mon Ndisj); simpls; red; destruct 1; constructor (by auto).
  rdes WEQ; rdes WEQ0; rdes WEQ1; destruct EQ, EQ0, EQ1.
  eapply mp; [eapply (OK (wprod w1 (wprod w2 w3))); eauto; econs; i; eauto
              ; eapply wcomp_glob in gOK; eauto
              ; destruct s as [sref [s1 [s2 s3]]]; destruct dgOK | ].
  - eapply wexp_extend1; eauto.
    eapply WEXP1; eauto using vclos_mon.
    by apply wglob_prod1; eauto using wconsistent_prod.
  - eapply wexp_extend2; eauto.
    eapply wexp_extend1; eauto.
      by eapply wglob_prod2; eauto using wconsistent_prod.
    eapply WEXP0; eauto; [eapply wglob_prod2 in gOK; eapply wglob_prod1 in gOK|]
    ; eauto using vclos_mon.
  - eapply wexp_extend2; eauto.
    eapply wexp_extend2; eauto.
      by eapply wglob_prod2; eauto using wconsistent_prod.
    eapply WEXP; eauto; [eapply wglob_prod2 in gOK; eapply wglob_prod2 in gOK|]
    ; eauto using vclos_mon.
  - unfold weq; i; eexists (wcomp (wprod w1 (wprod w2 w3))); esplits.
     + i; repeat rrdes PR; [left|right;left|right;right]; eauto using NAM1, NAM0, NAM.
     + ii; eapply (wstable_prod DISJ WSTA1 (wstable_prod DISJ23 WSTA0 WSTA))
       ; eauto using wcomp_glob.
     + assert (X:=winhabited_prod DISJ INH1 (winhabited_prod DISJ23 INH0 INH)).
       rrdes X; rr; esplits; i; eapply X; eauto.
       apply wcomp_glob; eauto.
     + r; setoid_rewrite wlift_wcomp. 
       apply (@wkcomp_consistent (wlift (wprod w1 (wprod w2 w3)))).
     + eauto.
Grab Existential Variables. exact s1. exact (s2,s3).
Qed.

Lemma meta3 :
  forall
    G1 ty1 e1 e1' (E1: wreq G1 ty1 e1 e1')
    G2 ty2 e2 e2' (E2: wreq G2 ty2 e2 e2')
    G3 ty3 e3 e3' (E3: wreq G3 ty3 e3 e3') 
    G ty e e' (WF: wfprog G ty e e')
    (OK: forall (w: world)
           (E1: wequivgen w (wkcomp w) G1 ty1 e1 e1')
           (E2: wequivgen w (wkcomp w) G2 ty2 e2 e2')
           (E3: wequivgen w (wkcomp w) G3 ty3 e3 e3'),
         wequivgen w (wkcomp w) G ty e e'),
    wreq G ty e e'.
Proof.
  i; eapply meta3r; [apply E1|apply E2|apply E3|eauto|].
  by unfold wequiv; setoid_rewrite wlift_wcomp; i; eapply (OK (wlift w)).
Qed.

Ltac compat_init1 := 
let WR1 := fresh "WR1" in
let E1 := fresh "E1" in
  i; apply (meta1 WR1); i; [by eauto|];
  apply proj1 in WR1;
  destruct E1 as [nf1 nf1' E1];
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

Ltac compat_init2 := 
let WR1 := fresh "WR1" in let WR2 := fresh "WR2" in
let E1 := fresh "E1" in let E2 := fresh "E2" in
  i; apply (meta2 WR1 WR2); i; [by eauto|];
  apply proj1 in WR1; apply proj1 in WR2;
  destruct E1 as [nf1 nf1' E1], E2 as [nf2 nf2' E2];
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

Ltac compat_init3 := 
let WR1 := fresh "WR1" in let WR2 := fresh "WR2" in let WR3 := fresh "WR3" in
let E1 := fresh "E1" in let E2 := fresh "E2" in let E3 := fresh "E3" in
  i; apply (meta3 WR1 WR2 WR3); i; [by eauto|];
  apply proj1 in WR1; apply proj1 in WR2; apply proj1 in WR3;
  destruct E1 as [nf1 nf1' E1], E2 as [nf2 nf2' E2], E3 as [nf3 nf3' E3];
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

Ltac auto_clo w := let FH := fresh "FH" in
  ii; ss; des; revert1
  ; try (i; match goal with [H: wglobgen _ _ _ |- _ ] => 
            eapply (wg_clo H); (inst; s; eauto); fail
         end; fail)
  ; try (eapply subexp_closed_weakly; eauto with wfprog;
         [repeat (eapply weakclo_sext; eauto); eapply subclo_weakly
         |unfold sext; repeat (destruct eq_termvar; eauto)]
         ; eapply wfdg_closed; eauto using (@wg_clo w); fail)
  ; try (eauto using vclos_closed, fv_val_exp, subexp_closed, 
    (wfdg_closed (or_introl refl_equal)), (wfdg_closed (or_intror refl_equal)),
    (@wg_clo w) with wfprog; fail).

Ltac complete w := let X := fresh "X" in let V := fresh "V" in
    let LE := fresh "LE" in let na := fresh "na" in let nb:= fresh "nb" in 
    let STK := fresh "STK" in let CREL := fresh "CREL" in 
  try left; i; pfold; econs; [by ii; try depdes V; auto_clo w|];
  intros ? ? ? ? LE na ? nb ? STK; destruct STK; subst; depdes CREL.

Ltac ret :=
  ii; econs 2; [|econs 1| | | |M|M]; Mdo (by s; eauto using lreduces_nil).

Ltac step :=
  ii; s; eapply wexp_core_lred; [|M|M|]
  ; Mdo by econs 2; [econs 2|]; eauto using lreduces_nil.

Definition BF (nf: (termvar->nat)->nat) := nf.
Arguments BF /.
Ltac kill_budget nk nf :=
  instantiate (2 := nk); instantiate (1:= nf); s; omega.

(*----------------------------------------------------------------------------*)
(** ** Variables *)
(*----------------------------------------------------------------------------*)

Lemma compat_var : 
  forall G ty x (TEQ: ty = G x)
    (WF: wfprog G ty (Evar x) (Evar x)),
  wreq G ty (Evar x) (Evar x).
Proof.
  split; [by eauto|].
  ii; subst; eexists wid; esplits; try done; eauto.
  r; esplits; eauto; econs; econs 2; eauto.
  - econs 1; (inst; s; eauto using wfdg_ggok).
  - instantiate (1:= fun f => f x); apply lreduces_nil.
  - instantiate (1:= fun f => f x); apply lreduces_nil.
Qed.
Hint Resolve compat_var : compat.

(*----------------------------------------------------------------------------*)
(** ** Unit Type *)
(*----------------------------------------------------------------------------*)

Lemma compat_unit: forall G, 
  wfprog G Tunit Eunit Eunit -> wreq G Tunit Eunit Eunit.
Proof. 
  i; split; eauto.
  ii; exists wid; esplits; eauto using wstable_id; try done.
  r; esplits; eauto using winhabited_id, wconsistent_id.
  econs; ii; econs 2; try apply lreduces_nil; eauto; econs 1; (inst; s; eauto).
Grab Existential Variables.
i; exact 0. i; exact 0.
Qed.
Hint Resolve compat_unit : compat.

(*----------------------------------------------------------------------------*)
(** ** Int Type *)
(*----------------------------------------------------------------------------*)

Lemma compat_int: forall G n, 
  wfprog G Tint (Eint n) (Eint n) -> wreq G Tint (Eint n) (Eint n).
Proof.
  i; split; eauto.
  ii; exists wid; esplits; eauto using wstable_id; try done.
  r; esplits; eauto using winhabited_id, wconsistent_id.
  econs; ii; econs 2; try apply lreduces_nil; eauto; econs 1; (inst; s; eauto).
Grab Existential Variables.
i; exact 0. i; exact 0.
Qed.
Hint Resolve compat_int : compat.

(*----------------------------------------------------------------------------*)
(** ** Bool Type *)
(*----------------------------------------------------------------------------*)

Lemma compat_true: forall G 
    (WF: wfprog G Tbool Etrue Etrue),
  wreq G Tbool Etrue Etrue.
Proof.
  i; split; eauto.
  ii; exists wid; esplits; eauto using wstable_id; try done.
  r; esplits; eauto using winhabited_id, wconsistent_id.
  econs; ii; econs 2; try apply lreduces_nil; eauto; econs 1; (inst; s; eauto).
Grab Existential Variables.
i; exact 0. i; exact 0.
Qed.
Hint Resolve compat_true : compat.

Lemma compat_false: forall G 
    (WF: wfprog G Tbool Efalse Efalse),
  wreq G Tbool Efalse Efalse.
Proof.
  i; split; eauto.
  ii; exists wid; esplits; eauto using wstable_id; try done.
  r; esplits; eauto using winhabited_id, wconsistent_id.
  econs; ii; econs 2; try apply lreduces_nil; eauto; econs 1; (inst; s; eauto).
Grab Existential Variables.
i; exact 0. i; exact 0.
Qed.
Hint Resolve compat_false : compat.

Lemma compat_if :
  forall G ty e1 e1' e2 e2' e3 e3'
    (WR1: wreq G Tbool e1 e1')
    (WR2: wreq G ty e2 e2')
    (WR3: wreq G ty e3 e3')
    (WF: wfprog G ty (Eif e1 e2 e3) (Eif e1' e2' e3')),
  wreq G ty (Eif e1 e2 e3) (Eif e1' e2' e3').
Proof.
  compat_init3.
  s; eapply wexp_app_scont; eauto.
  eapply E1; eauto.
  s; apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst.

  - econs; [|M|M]; Mdo (by econs 2; eauto).
    eapply E2; try apply HREL; eauto using wfdg_mon, (wg_mon gOK).
    eapply wglobgen_mon; eauto; inst; s; eauto.

  - econs; [|M|M]; Mdo (by econs 2; eauto).
    eapply E3; try apply HREL; eauto using wfdg_mon, (wg_mon gOK).
    eapply wglobgen_mon; eauto; inst; s; eauto.
Qed.
Hint Resolve compat_if : compat.

(*----------------------------------------------------------------------------*)
(** ** Functions *)
(*----------------------------------------------------------------------------*)

Lemma compat_fix :
  forall G ty1 ty2 (f x: termvar) e e' 
    (WR1: wreq (sext x ty1 (sext f (Tarrow ty1 ty2) G)) ty2 e e')
    (WF: wfprog G (Tarrow ty1 ty2) (Efix f x e) (Efix f x e')),
  wreq G (Tarrow ty1 ty2) (Efix f x e) (Efix f x e').
Proof.
  compat_init1.

  ret; eauto.
  s; repeat econs; apply (wg_le gOK). 
  eapply lk_generalize; eauto.
  hide_evars; pcofix CIH; show_evars.
  complete w.
  simpl_inj; subst; s.
  econs; [|M|M]; Mdo (by econs 2; eauto).
  destruct v1 as [? v1 ?], v2 as [? v2 ?]; ss; subst.
  rewrite <-!subexp_sext_l2
  ; eauto using vclos_closed, subclo_weakly, wfdg_closed, (@wg_clo w)
  ; [|M|M]; Mdo by auto_clo w.

  eapply mp'; [|eapply E1 
    with (gg := sext x0 ((_,_),(_,_)) (sext f ((_,_),(_,_)) gg))].
  - hide_evars. 
    rewrite !sext_map; unfold compose; simpl @snd; simpl @fst. 
    show_evars.
    intro X; exp2val; apply X.
  - by eapply wglobgen_mon; eauto; left; eapply paco7_mon; eauto.
  - econs; [by apply (wfdg_ddok dgOK)|].
    i; unfold compose, sext; destruct (eq_termvar x x0); [by s; eauto|].
    destruct (eq_termvar x f).
    2: by eapply vclos_mon; eauto using (wg_mon g'OK0), (wfdg_ggok dgOK).
    subst; s.
    repeat econs; eapply (wg_le g'OK0); right.
    eapply CIH; eauto using wextpub_trans.
    eapply wglobgen_mon; eauto.
    by i; s; left; eapply paco7_mon; eauto.
  - by eauto.
Grab Existential Variables.
eauto. eauto.
Qed.
Hint Resolve compat_fix : compat.

Lemma compat_app :
  forall G ty ty' e1 e1' e2 e2' 
    (WR1: wreq G (Tarrow ty ty') e1 e1')
    (WR2: wreq G ty e2 e2')
    (WF: wfprog G ty' (Eapp e1 e2) (Eapp e1' e2')),
  wreq G ty' (Eapp e1 e2) (Eapp e1' e2').
Proof.
  compat_init2.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nf2(fst<*>fst<*>gg)+nk+2) (fun gg => nf1 gg+nf2 gg+2).
  2: by kill_budget (nf2'(fst<*>snd<*>gg)+nk'+2) (fun gg => nf1' gg+nf2' gg+2).

  s; apply (wg_le gOK).
  complete w.
  econs; [|M|M]; Mdo i; eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E2; eauto| |].
  { eapply wglobgen_mon; eauto; i; s; left; eapply paco7_mon; eauto. }
  { econs; [eapply wfdg_ddok|]; eauto using vclos_mon, wfdg_ggok, (wg_mon gOK). }
  2: by instantiate (1:= nv1+nk+1); omega.
  2: by instantiate (1:= nv2+nk'+1); omega.

  apply (wg_le g'OK).
  complete w.
  econs; [|M|M]; Mdo i; eapply lred_log.
  2: by instantiate (1:= na0 - 1); omega.
  2: by instantiate (1:= nb0 - 1); omega.

  destruct VREL; depdes REL.
  ii; econs 2; s; eauto using lreduces_nil. 
  econs 2; eauto using (wg_mon g'OK0), wextpub_trans.
Qed.
Hint Resolve compat_app : compat.

(*----------------------------------------------------------------------------*)
(** ** Products *)
(*----------------------------------------------------------------------------*)

Lemma compat_pair :
  forall G ty1 e1 e1' ty2 e2 e2'
    (WR1: wreq G ty1 e1 e1')
    (WR2: wreq G ty2 e2 e2')
    (WF: wfprog G (Tprod ty1 ty2) (Epair e1 e2) (Epair e1' e2')),
  wreq G (Tprod ty1 ty2) (Epair e1 e2) (Epair e1' e2').
Proof.
  compat_init2.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nf2(fst<*>fst<*>gg)+nk+2) (fun gg => nf1 gg+nf2 gg+2).
  2: by kill_budget (nf2'(fst<*>snd<*>gg)+nk'+2) (fun gg => nf1' gg+nf2' gg+2).

  s; apply (wg_le gOK).
  complete w.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E2; eauto| |].
  { eapply wglobgen_mon; eauto; i; s; left; eapply paco7_mon; eauto. }
  { econs; [eapply wfdg_ddok|]; eauto using vclos_mon, wfdg_ggok, (wg_mon gOK). }
  2: by instantiate (1:= nv1+nk+1); omega.
  2: by instantiate (1:= nv2+nk'+1); omega.

  apply (wg_le g'OK).
  complete w.
  econs; [|M|M]; Mdo i; eapply lred_log.
  2: by instantiate (1:= na0 - 1); omega.
  2: by instantiate (1:= nb0 - 1); omega.

  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK0), wextpub_trans).
  by instantiate (1:=0); omega.
  by instantiate (1:=0); omega.
Qed. 
Hint Resolve compat_pair : compat.

Lemma compat_fst : 
  forall G ty1 ty2 e e' 
    (WR1: wreq G (Tprod ty1 ty2) e e')
    (WF: wfprog G ty1 (Efst e) (Efst e')),
  wreq G ty1 (Efst e) (Efst e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply E1; eauto.
  s; apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst.

  econs; [|M|M]; Mdo (by econs 2; s; eauto).
  ret; eauto using (wg_mon g'OK).
Qed.
Hint Resolve compat_fst : compat.

Lemma compat_snd : 
  forall G ty1 ty2 e e' 
    (WR1: wreq G (Tprod ty1 ty2) e e')
    (WF: wfprog G ty2 (Esnd e) (Esnd e')),
  wreq G ty2 (Esnd e) (Esnd e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply E1; eauto.
  s; apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst.

  econs; [|M|M]; Mdo (by econs 2; s; eauto).
  ret; eauto using (wg_mon g'OK).
Qed.
Hint Resolve compat_snd : compat.

(*----------------------------------------------------------------------------*)
(** ** Sums *)
(*----------------------------------------------------------------------------*)

Lemma compat_inl : 
  forall G ty1 ty2 e e' 
    (WR1: wreq G ty1 e e')
    (WF: wfprog G (Tsum ty1 ty2) (Einl e) (Einl e')),
  wreq G (Tsum ty1 ty2) (Einl e) (Einl e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  s; apply (wg_le gOK).
  complete w.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK), wextpub_trans).
  by instantiate (1:=0); omega.
  by instantiate (1:=0); omega.
Qed.
Hint Resolve compat_inl : compat.

Lemma compat_inr : 
  forall G ty2 e e' 
    (WR1: wreq G ty2 e e') ty1 (ty1OK: lc_typ 0 ty1)
    (WF: wfprog G (Tsum ty1 ty2) (Einr e) (Einr e')),
  wreq G (Tsum ty1 ty2) (Einr e) (Einr e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  s; apply (wg_le gOK).
  complete w.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK), wextpub_trans).
  by instantiate (1:=0); omega.
  by instantiate (1:=0); omega.
Qed.
Hint Resolve compat_inr : compat.

Lemma wfdg_sext x ty G g dd n v n' v' gg:
  wfdg G g dd gg ->
  vclos g (tsubst_name dd ty) n v n' v' ->
  wfdg (sext x ty G) g dd (sext x (n, v, (n', v')) gg).
Proof.
  destruct 1; econs; eauto; i; unfold compose, sext.
  destruct eq_termvar; s; eauto. 
Qed.

Lemma compat_case : 
  forall G ty1 ty2 e e' e1 e1' e2 e2' 
    (WR1: wreq G (Tsum ty1 ty2) e e') ty (x: termvar)
    (WR2: wreq (sext x ty1 G) ty e1 e1')
    (WR3: wreq (sext x ty2 G) ty e2 e2')
    (WF: wfprog G ty (Ecase x e e1 e2) (Ecase x e' e1' e2')),
  wreq G ty (Ecase x e e1 e2) (Ecase x e' e1' e2').
Proof.
  compat_init3.

  s; eapply wexp_app_scont; eauto.
  eapply E1; eauto.
  s; apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst.

  - s; econs; [|M|M]; Mdo (by econs 2; eauto).
    rewrite <-!subexp_sext_l
    ; eauto using vclos_closed, subclo_weakly, wfdg_closed, (@wg_clo w).
  
    eapply mp'; [|eapply E2 with (gg := sext x0 ((_,_),(_,_)) gg)].
    + hide_evars. 
      rewrite !sext_map; unfold compose; simpl @snd; simpl @fst. 
      show_evars.
      intro X; exp2val; apply X.
    + by eapply wglobgen_mon; eauto; left; eapply paco7_mon; eauto.
    + econs; [by apply (wfdg_ddok dgOK)|].
      i; unfold compose, sext; destruct (eq_termvar x x0); s
      ; eauto using vclos_mon, (wg_mon g'OK), (wfdg_ggok dgOK).
    + eauto using (wg_mon g'OK).

  - s; econs; [|M|M]; Mdo (by econs 2; eauto).
    rewrite <-!subexp_sext_l
    ; eauto using vclos_closed, subclo_weakly, wfdg_closed, (@wg_clo w).
  
    eapply mp'; [|eapply E3 with (gg := sext x0 ((_,_),(_,_)) gg)].
    + hide_evars. 
      rewrite !sext_map; unfold compose; simpl @snd; simpl @fst. 
      show_evars.
      intro X; exp2val; apply X.
    + by eapply wglobgen_mon; eauto; left; eapply paco7_mon; eauto.
    + econs; [by apply (wfdg_ddok dgOK)|].
      i; unfold compose, sext; destruct (eq_termvar x x0); s
      ; eauto using vclos_mon, (wg_mon g'OK), (wfdg_ggok dgOK).
    + eauto using (wg_mon g'OK).
Qed.
Hint Resolve compat_case : compat.

(*----------------------------------------------------------------------------*)
(** ** References *)
(*----------------------------------------------------------------------------*)

Definition wrefst_ext s newty loc loc' := fun (ty: typ) (loc1 loc2: location) => 
  ty = newty /\ loc1 = loc /\ loc2 = loc' \/ s ty loc1 loc2.
Hint Unfold wrefst_ext.

Lemma wrefext_wf: forall s newty loc loc'
    (WF: wrefst s)
    (CL: lc_typ 0 newty)
    (PF : forall ty loc', (s ty loc loc': Prop) -> False)
    (PF' : forall ty loc, s ty loc loc' -> False),
  wrefst (wrefst_ext s newty loc loc').
Proof.
  i; ss; rdes WF; hdesH WF1; unfold wrefst_ext.
  split; [by intros; des; subst; eauto|].
  hdesH WF0; split.
  - split; i; des; do 2 subst; eauto; try (by exfalso; eauto).
  - exists (newty :: tydom) (loc::loc'::ldom); intros.
    des; hdes; subst; simpls; eauto 15.
Qed.

Lemma wref_update: forall g sr h1 h2 ty loc1 loc2 n1 n2 v1 v2
    (HR: wheap wref (fun _ => g) () sr h1 h2)
    (LOC: (` sr) ty loc1 loc2)
    (VAL: vclos (g sr) ty n1 v1 n2 v2),
  wheap wref (fun _ => g) () sr (hext h1 loc1 v1) (hext h2 loc2 v2).
Proof.
  i; rrdes HR; rr; esplits.
  - i; assert (X:= R); eapply HSAT in X; des.
    s; destruct eq_nat_dec.
    + subst; eapply (proj2_sig sr) in LOC; eauto.
      des; destruct (LOC refl_equal); unnw; subst. 
      by destruct eq_nat_dec; esplits; eauto.
    + destruct eq_nat_dec; eauto.
      subst; eapply (proj2_sig sr) in LOC; eauto.
      by des; destruct (LOC0 refl_equal); unnw; subst.
  - s; i; destruct eq_nat_dec; subst; eauto.
  - s; i; destruct eq_nat_dec; subst; eauto.
Grab Existential Variables.
exact 0. exact 0.
Qed.

Lemma compat_new : 
  forall G ty e e' 
    (WR1: wreq G ty e e')
    (WF: wfprog G (Tref ty) (Enew e) (Enew e')),
  wreq G (Tref ty) (Enew e) (Enew e').
Proof.
  i; apply (meta1r WR1); i; [by eauto|].
  destruct WR1. 
  
  destruct E1 as [nf1 nf1' E1].
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).
  s; apply (wg_le gOK).
  complete w.

  econs.
  2: by econs; instantiate (1:=na-1); omega.
  2: by econs; instantiate (1:=nb-1); omega.
  destruct s as (sr, s), s' as (sr',s').
  eapply wexp_lift; eauto using wkcomp_glob_gen.

  ii; exploit (@wrefext_wf (` sr') (tsubst_name dd ty) (allocator hh) (allocator hh')).
  { by apply proj2_sig. }
  { by eauto using lc_tsubst, wfdg_ddok, wfprog_tyok. }
  { by i; eapply HREL in H1; des; erewrite allocator_unalloc' in HL; eauto. }
  { by i; eapply HREL in H1; des; erewrite allocator_unalloc' in HL'; eauto. }
  intro Psr.

  step.
  2: by instantiate (1:= exist _ _ Psr); rr; eauto.
  rewrite !plus_0_r.

  ret; try apply le_n; eauto.
  - eapply (wg_mon g'OK), wextpub_trans, LES; s; eauto 6.
  - s; do 2 econs; apply (wg_le g'OK); rr; s; esplits; eauto.
  - rrdes HREL; rr; esplits.
    + ii; rrdes R.
      * subst; esplits
        ; try by instantiate (2:=hext _ (allocator _) _); eauto using hext_lookup_new.
        by eapply vclos_mon; eauto; i; eapply (wg_mon g'OK); s; eauto 6.
      * eapply HSAT in R; des; esplits
        ; try by erewrite hext_lookup_old; eauto.
        by eapply vclos_mon; eauto; i; eapply (wg_mon g'OK); s; eauto 6.
    + s; i; destruct eq_nat_dec; clarify.
      * by esplits; eauto.
      * by apply HDOM1 in IN; des; esplits; eauto.
    + s; i; destruct eq_nat_dec; clarify.
      * by esplits; eauto.
      * by apply HDOM2 in IN; des; esplits; eauto.
  - intro x; generalize (HPLUS x); simpl; destruct eq_nat_dec; clarify.
    by rewrite allocator_unalloc; intros; des; eauto.
  - intro x; generalize (HPLUS' x); simpl; destruct eq_nat_dec; clarify.
    by rewrite allocator_unalloc; intros; des; eauto.
Grab Existential Variables.
exact 0. exact 0.
Qed.

Hint Resolve compat_new : compat.

Lemma compat_read : 
  forall G ty e e' 
    (WR1: wreq G (Tref ty) e e')
    (WF: wfprog G ty (Eread e) (Eread e')),
  wreq G ty (Eread e) (Eread e').
Proof.
  i; eapply meta1r; eauto.
  i; apply proj1 in WR1.
  destruct E1 as [nf1 nf1' E1].
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |]; clear E1.
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  apply (wg_le gOK).
  complete (wlift (wcomp w)).
  ss; destruct VREL; depdes REL; simpl_inj; subst; ss.

  apply (wg_le g'OK) in VR; rr in VR; des; clarify; simpl_injs; subst.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  ii; rr in HREL; des.
  pose proof (proj1 WH _ _ _ TLL); des. 
  do 2 (eapply hplus_some1 in HL; eauto).
  do 2 (eapply hplus_some1 in HL'; eauto).

  eapply wexp_core_lred with (s1:=s'); [|M|M|]
  ; (Mdo econs 2; [econs 2; econs; eapply Rctx; econs|]; eauto using lreduces_nil)
  ; eauto using wextpub_refl.

  econs 2; eauto using lreduces_nil; [|eby do 5 eexists].
  by destruct s'; econs; try eapply LE, (wg_mon gOK); eauto.
Qed.
Hint Resolve compat_read : compat.

Lemma compat_write : 
  forall G ty e1 e1' e2 e2' 
    (WR1: wreq G (Tref ty) e1 e1')
    (WR2: wreq G ty e2 e2')
    (WF: wfprog G Tunit (Ewrite e1 e2) (Ewrite e1' e2')),
  wreq G Tunit (Ewrite e1 e2) (Ewrite e1' e2').
Proof.
  i; apply (meta2r WR1 WR2); i; [by eauto|].
  destruct WR1, WR2. 
  destruct E1 as [nf1 nf1' E1], E2 as [nf2 nf2' E2].
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nf2(fst<*>fst<*>gg)+nk+3) (fun gg => nf1 gg+nf2 gg+3).
  2: by kill_budget (nf2'(fst<*>snd<*>gg)+nk'+3) (fun gg => nf1' gg+nf2' gg+3).

  s; apply (wg_le gOK).
  complete (wlift (wcomp w)).
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E2; eauto| |].
  { by setoid_rewrite wlift_wcomp; rewrite (wcgen_elim_bot (wlift w)) in g'OK. }
  { by econs; [eapply wfdg_ddok|]; eauto using vclos_mon, wfdg_ggok, (wg_mon gOK). }
  2: by instantiate (1:= nv1+nk+2); omega.
  2: by instantiate (1:= nv2+nk'+2); omega.

  s; apply (wg_le g'OK).
  complete (wlift w).
  econs.
  2: by econs; instantiate (1:=na0-1); omega.
  2: by econs; instantiate (1:=nb0-1); omega.
  destruct s as (sr,s), s' as (sr',s'), s'0 as (sr'',s'').
  eapply wexp_lift; eauto using wkcomp_glob_gen.

  destruct VREL; depdes REL.
  eapply (wg_le g'OK) in VR; rrdes VR; clarify; simpl_injs; subst.
  ii; assert (X:= TLL); eapply LES0, HREL in X; des.
  rewrite (hext_eq _ HPLUS HL), (hext_eq _ HPLUS' HL').

  step; [|by eauto].
  
  ret
  ; try eapply hext_plus
  ; (try by eapply wref_update; eauto; des; eapply LES0; eauto)
  ; (try by eapply (wg_mon g'OK0); try eapply LE0, (LE (_,_)); eauto using wextpub_trans)
  ; eauto.
Grab Existential Variables.
exact 0. exact 0.
Qed.
Hint Resolve compat_write : compat.

Lemma compat_refeq : 
  forall G ty e1 e1' e2 e2' 
    (WR1: wreq G (Tref ty) e1 e1')
    (WR2: wreq G (Tref ty) e2 e2')
    (WF: wfprog G Tbool (Erefeq e1 e2) (Erefeq e1' e2')),
  wreq G Tbool (Erefeq e1 e2) (Erefeq e1' e2').
Proof.
  i; eapply (meta2r WR1 WR2); eauto.
  i; apply proj1 in WR1; apply proj1 in WR2.
  destruct E1 as [nf1 nf1' E1], E2 as [nf2 nf2' E2].
  econs; intros until gg; intros ? nk K nk' K' ?; unnw.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |]; clear E1.
  2: by kill_budget (nf2(fst<*>fst<*>gg)+nk+3) (fun gg => nf1 gg+nf2 gg+3).
  2: by kill_budget (nf2'(fst<*>snd<*>gg)+nk'+3) (fun gg => nf1' gg+nf2' gg+3).

  s; apply (wg_le gOK).
  complete (wlift (wcomp w)).
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  s; eapply wexp_app_scont.
  eapply wexpgen_mon; [eapply E2| |]; clear E2.
  { by setoid_rewrite wlift_wcomp; rewrite (wcgen_elim_bot (wlift w)) in g'OK. }
  { by econs; [eapply wfdg_ddok|]; eauto using vclos_mon, wfdg_ggok, (wg_mon gOK). }
  2: by instantiate (1:= nv1+nk+2); omega.
  2: by instantiate (1:= nv2+nk'+2); omega.

  ss; destruct VREL; depdes REL; simpl_inj; subst.
  assert (M := VR).
  apply (wg_le g'OK) in M; rr in M; des; clarify. 
  rewrite H0, H1.

  apply g'OK; ss.
  complete (wlift (wcomp w)).
  ss; destruct VREL; depdes REL; simpl_inj; subst; ss.
  assert (M := VR0).
  apply (wg_le g'OK0) in M; rr in M; des; clarify. 
  rewrite (VoE_pf v0), (VoE_pf v3), H0, H1.
  pose proof (proj1 (proj2 (proj2_sig (fst s'0))) _ _ _ (WREP0 _ _ _ TLL) _ _ _ TLL0); ss; unnw.
  destruct (eq_nat_dec loc loc0), (eq_nat_dec loc' loc'0); subst; [|tauto|tauto|].
  * s; (econs; [|M|M]; Mdo (econs 2; eauto)).
    ret; try apply le_n; eauto; eapply LE0, LE, (wg_mon gOK); eauto.
    by split; unnw; s; try done; eauto using wextpub_trans.
  * s; (econs; [|M|M]; Mdo (econs 2; eauto)).
    ret; try apply le_n; eauto; eapply LE0, LE, (wg_mon gOK); eauto.
    by split; unnw; s; try done; eauto using wextpub_trans.
Grab Existential Variables.
  exact 0. exact 0. exact 0. exact 0.
Qed.
Hint Resolve compat_refeq : compat.

(*----------------------------------------------------------------------------*)
(** ** Recursive types *)
(*----------------------------------------------------------------------------*)

Lemma compat_roll : 
  forall G ty e e' 
    (WR1: wreq G (tsubst_bvar 0 (Tmu ty) ty) e e')
    (WF: wfprog G (Tmu ty) (Eroll e) (Eroll e')),
  wreq G (Tmu ty) (Eroll e) (Eroll e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  s; apply (wg_le gOK).
  complete w.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  rewrite tsubst_name_bvar in VREL; [|by apply dgOK].
  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK), wextpub_trans).
    by instantiate (1:=0); omega.
    by instantiate (1:=0); omega.
Qed.
Hint Resolve compat_roll : compat.

Lemma compat_unroll : 
  forall G ty e e' 
    (WR1: wreq G (Tmu ty) e e')
    (WF: wfprog G (tsubst_bvar 0 (Tmu ty) ty) (Eunroll e) (Eunroll e')),
  wreq G (tsubst_bvar 0 (Tmu ty) ty) (Eunroll e) (Eunroll e').
Proof.
  compat_init1; s.
  eapply wexp_app_scont, E1; eauto; s.
  apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst; ss.
  econs; [|M|M]; Mdo (econs 2; eauto).
  ret; try apply le_n; (inst; s; eauto 10 using vclos_mon, (wg_mon gOK), wextpub_trans).
  rewrite tsubst_name_bvar; eauto. 
  by apply dgOK.
Qed.
Hint Resolve compat_unroll : compat.

(*----------------------------------------------------------------------------*)
(** ** Existential types *)
(*----------------------------------------------------------------------------*)

Lemma compat_pack : 
  forall G ty ty' e e'
    (WR1: wreq G (tsubst_bvar 0 ty' ty) e e')
    (WF: wfprog G (Texists ty) (Epack e) (Epack e'))
    (ty'OK: lc_typ 0 ty'),
  wreq G (Texists ty) (Epack e) (Epack e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  s; apply (wg_le gOK).
  complete w.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na-1); omega.
  2: by instantiate (1:= nb-1); omega.

  rewrite tsubst_name_bvar in VREL; [|by apply dgOK].
  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK), wextpub_trans).
  2: by instantiate (1:=0); omega.
  2: by instantiate (1:=0); omega.
  do 2 econs; eauto.
  by eapply lc_tsubst; eauto; r; apply dgOK.  
Qed.
Hint Resolve compat_pack : compat.

Lemma compat_unpack : 
  forall G ty ty' aa e1 e1' x e2 e2' 
    (WR1: wreq G (Texists ty) e1 e1')
    (WR2: wreq (sext x (tsubst_bvar 0 (Tname aa) ty) G) ty' e2 e2')
    (WF: wfprog G ty' (Eunpack x e1 e2) (Eunpack x e1' e2'))
    (FRESH1: forall x, ~ ftn_typ (G x) aa)
    (FRESH2: ~ ftn_typ ty aa)
    (FRESH3: ~ ftn_typ ty' aa),
  wreq G ty' (Eunpack x e1 e2) (Eunpack x e1' e2').
Proof.
  compat_init2.

  s; eapply wexp_app_scont; eauto.
  eapply E1; eauto.
  s; apply (wg_le gOK).
  complete w.
  ss; destruct VREL; depdes REL; simpl_inj; subst.

  s; econs; [|M|M]; Mdo (by econs 2; eauto).
  rewrite <-!subexp_sext_l
  ; eauto using vclos_closed, subclo_weakly, wfdg_closed, (@wg_clo w).

  eapply mp'; [|eapply E2 with (gg := sext x0 ((_,_),(_,_)) gg)].
  - hide_evars. 
    rewrite !sext_map; unfold compose; simpl @snd; simpl @fst. 
    show_evars.
    intro X; exp2val; apply X.
  - by eapply wglobgen_mon; eauto; left; eapply paco7_mon; eauto.
  - econs. 
    + instantiate (1:= sext aa ty'0 dd).
      by i; unfold sext; destruct eq_termvar; eauto using wfdg_ddok.
    + i; unfold compose; unfold sext; destruct (eq_termvar x x0); subst.
      * rewrite tsubst_name_bvar.
        2: by i; destruct eq_termvar; eauto using wfdg_ddok.
        s; erewrite eqtmv_true, tsubnam_agree; eauto.
        i; destruct eq_termvar; eauto.
        subst; exfalso; eauto.
      * erewrite tsubnam_agree.
          by eauto using vclos_mon, (wg_mon g'OK), (wfdg_ggok dgOK).
        i; destruct eq_termvar; eauto.
        subst; exfalso; eauto.
  - erewrite tsubnam_agree; eauto using (wg_mon g'OK).  
    i; unfold sext; destruct eq_termvar; eauto.
    subst; exfalso; eauto.
Qed.
Hint Resolve compat_unpack : compat.

(*----------------------------------------------------------------------------*)
(** ** Universal types *)
(*----------------------------------------------------------------------------*)

Lemma tsubst_name_simpl1 : 
  forall aa ty (FRESH : ~ ftn_typ ty aa) dd ty',
    tsubst_name (fun x => if eq_nat_dec x aa then ty' else dd x) ty = tsubst_name dd ty.
Proof.
  induction ty; ss; ii; f_equal; eauto.
  destruct eq_nat_dec; intuition. 
Qed.

Lemma compat_gen : 
  forall G ty aa e e' 
    (WR1: wreq G (tsubst_bvar 0 (Tname aa) ty) e e')
    (WF: wfprog G (Tforall ty) (Egen e) (Egen e'))
    (FRESH1 : forall x, ~ ftn_typ (G x) aa)
    (FRESH2 : ~ ftn_typ ty aa),
  wreq G (Tforall ty) (Egen e) (Egen e').
Proof.
  compat_init1.

  ret; eauto.
  s; repeat econs; apply (wg_le gOK). 
  instantiate (1 := fun x => 0); instantiate (1 := fun x => 0).
  complete w.
  rewrite (VoE_pf f1), (VoE_pf f2), <- x, <- x0; ss; clear x x0.
  econs; i; [|M|M]; Mdo (econs 2; eauto).

  assert (LC' : forall x, lc_typ 0 (if eq_nat_dec x aa then ty' else dd x)).
    by i; destruct (eq_nat_dec); eauto; apply dgOK.

  eapply E1 with (dd := fun x => if eq_nat_dec x aa then ty' else dd x); eauto; clear E1.
  * by eapply (wglobgen_mon _ g'OK); vauto.
  * destruct dgOK; econs; eauto.
    by i; rewrite tsubst_name_simpl1; eauto using vclos_mon, (wg_mon g'OK).
  * rewrite tsubst_name_bvar; s; destruct (eq_nat_dec aa aa); try done.
    by rewrite tsubst_name_simpl1; eauto.
Qed.
Hint Resolve compat_gen : compat.

Lemma compat_inst :
  forall G ty ty' e e' 
    (WR1: wreq G (Tforall ty) e e')
    (WF: wfprog G (tsubst_bvar 0 ty' ty) (Einst e) (Einst e'))
    (LCty': lc_typ 0 ty'),
  wreq G (tsubst_bvar 0 ty' ty) (Einst e) (Einst e').
Proof.
  compat_init1.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+1) (fun gg => nf1 gg+1).
  2: by kill_budget (nk'+1) (fun gg => nf1' gg+1).

  s; apply (wg_le gOK).
  complete w.

  ss; destruct VREL; depdes REL; simpl_inj; subst; ss.

  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= n); omega.
  2: by instantiate (1:= n'); omega.
  ii; econs 2; s; eauto using lreduces_nil. 

  rewrite tsubst_name_bvar in KOK; [|by apply dgOK].
  eapply wstuck_Inst with (ty' := tsubst_name dd ty'); eauto 10 using (wg_mon g'OK). 
  by eapply lc_tsubst; eauto; r; apply dgOK. 
Qed.
Hint Resolve compat_inst : compat.


(*----------------------------------------------------------------------------*)
(** ** Continuations *)
(*----------------------------------------------------------------------------*)

Lemma compat_callcc : 
  forall G ty x e e' 
    (WR1: wreq (sext x (Tcont ty) G) ty e e')
    (WF: wfprog G ty (Ecallcc x e) (Ecallcc x e')),
   wreq G ty (Ecallcc x e) (Ecallcc x e').
Proof.
  compat_init1; s.
  ii; eapply wexp_core_lred; try solve [eapply lreduces_step; eauto using lreduces_nil]; eauto.
  rewrite <-!subexp_sext_l
  ; eauto using vclos_closed, subclo_weakly, wfdg_closed, (@wg_clo w).
  rr in E1.
  erewrite (@subexp_agree _ (sext x (Econt K') _)), 
           (@subexp_agree _ (sext x (Econt K) _)). 
  eapply E1 with (gg := sext x ((_,_),(_,_)) gg); eauto.
  eapply wfdg_sext; eauto; vauto. 
  by unfold sext, compose; i; destruct eq_termvar.
  by unfold sext, compose; i; destruct eq_termvar.
Grab Existential Variables. 
  exact 0. exact 0. exact (fun x => 0). exact (fun x => 0). 
Qed.
Hint Resolve compat_callcc : compat.

Lemma compat_isolate : 
  forall G ty e e' 
    (WR1: wreq G (Tarrow ty Tint) e e')
    (WF: wfprog G (Tcont ty) (Eisolate e) (Eisolate e')),
   wreq G (Tcont ty) (Eisolate e) (Eisolate e').
Proof.
  compat_init1; s.
  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto| |].
  2: by kill_budget (nk+2) (fun gg => nf1 gg+2).
  2: by kill_budget (nk'+2) (fun gg => nf1' gg+2).

  s; apply (wg_le gOK).
  complete w; s.

  econs; i; [|M|M]; Mdo (by econs 2; eauto). 

  ss; assert (VR := VREL); destruct VR; depdes REL; simpl_inj; subst.

  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK), wextpub_trans).
  do 2 econs.
  instantiate (1 := n' + 1); instantiate (1 := n + 1).
  s; apply (wg_le g'OK).
  complete w.

  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= n); omega.
  2: by instantiate (1:= n'); omega.

  econs 2; eauto using lreduces_nil.
  eapply wstuck_App with (K1 := nil) (K2 := nil); s; eauto.
  eapply g'OK0; try eapply le_n; eauto.
  eapply LE0; try eapply le_n; s; eauto.
  eapply (wg_sc g'OK).
Grab Existential Variables. 
  exact 0. exact 0. exact 0. exact 0.
Qed.
Hint Resolve compat_isolate : compat.

Lemma compat_throw : 
  forall G ty ty' e1 e1' e2 e2' 
    (WR1: wreq G ty e1 e1')
    (WR2: wreq G (Tcont ty) e2 e2')
    (WF: wfprog G ty' (Ethrow e1 e2) (Ethrow e1' e2')),
   wreq G ty' (Ethrow e1 e2) (Ethrow e1' e2').
Proof.
  compat_init2; s.
  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E1; eauto; s| |].
  2: by kill_budget (nf2(fst<*>fst<*>gg)+nk+2) (fun gg => nf1 gg+nf2 gg+2).
  2: by kill_budget (nf2'(fst<*>snd<*>gg)+nk'+2) (fun gg => nf1' gg+nf2' gg+2).

  s; apply (wg_le gOK).
  complete w; s.
  econs; i; [|M|M]; Mdo eapply lred_log.
  2: by instantiate (1:= na - 1); omega.
  2: by instantiate (1:= nb - 1); omega.

  s; eapply wexp_app_scont; eauto.
  eapply wexpgen_mon; [eapply E2; eauto; s| |].
  { eapply wglobgen_mon; eauto; i; s; left; eapply paco7_mon; eauto. }
  { econs; [eapply wfdg_ddok|]; eauto using vclos_mon, wfdg_ggok, (wg_mon gOK). }
  2: by instantiate (1:= nv1+nk+1); omega.
  2: by instantiate (1:= nv2+nk'+1); omega.

  apply (wg_le g'OK).
  complete w; s.
  ss; destruct VREL0; depdes REL; simpl_inj; subst; s.
  econs; [|M|M]; Mdo (econs 2; eauto).
  ret; (inst; s; eauto 10 using vclos_mon, (wg_mon g'OK0), wextpub_trans).
Qed.
Hint Resolve compat_throw : compat.

(*----------------------------------------------------------------------------*)
(** ** Adequacy *)
(*----------------------------------------------------------------------------*)

Definition gemp (x: termvar) := Tunit.

Definition semp (x: termvar) := Eunit.

Theorem wreq_adequate : 
  forall e1 e2 (WEQ: wreq gemp Tint e1 e2) h1 h2,
     (diverges h1 (subst_exp semp e1) /\ diverges h2 (subst_exp semp e2))
  \/ (exists h1' h2' n, 
      reduces h1 (subst_exp semp e1) h1' (Eint n) /\ 
      reduces h2 (subst_exp semp e2) h2' (Eint n)).
Proof.
  i; rdes WEQ; exploit (REQ (fun x => True)).
    eexists (fun x => `x), (fun x => exist _ x I).
    by split; rr; i; [eauto using proj1_sig_inj|depdes H].
  clear REQ; i; des; rdes WEQ.
  rrdes INH; specialize (INH _ (wvfix_glob _)).
  destruct EQ.
  eapply mp; [eapply (WEXP s0 _ (wvfix_glob (wlift w)) 
    (fun _ => Tunit) (fun _ => ((0,Vunit),(0,Vunit)))); eauto|].
  { econs; s; eauto. }
  clear WEXP; intro WEXP.
  eapply wexp_no_external in WEXP; eauto using wvfix_glob.
  - eapply mp; [eapply (WEXP _ _ _ h1 _ _ h2); eauto|].
    unfold compose; s; clear WEXP; intro WEXP; depdes WEXP.
    + eauto using ldiv_div.
    + depdes EREL CREL; des; clarify.
      destruct VREL; depdes REL; simpl_injs; subst; ss.
      eauto 10 using lreds_reds. 
  - i; depdes PR; subst; eauto using wval_mon.
    depdes REL; done. 
Grab Existential Variables. 
eauto using hplus_hemp_l. eauto. exact 0. exact 0.
Qed.



(*----------------------------------------------------------------------------*)
(** ** Fundamental property (partial reflexivity) *)
(*----------------------------------------------------------------------------*)

Fixpoint g2f (G: termenv) : termvar -> typ :=
  match G with
  | nil => gemp
  | (x,ty) :: G' => sext x ty (g2f G')
  end.

Lemma wftyp_lclo: forall D ty, wftyp D ty -> lc_typ 0 ty.
Proof. i; inv H. Qed.

Lemma g2f_clo: forall G
    (CLO: forall x ty, In (x, ty) G -> lc_typ 0 ty),
  forall x, lc_typ 0 (g2f G x).
Proof.
  induction G; s; eauto.
  i; destruct a; unfold sext; destruct eq_termvar; subst; eauto.
Qed.

Lemma g2f_finite2: forall G,
  forall x, ~ In x (map fst G) -> g2f G x = Tunit.
Proof.
  induction G; eauto.
  des; destruct a as (x,ty).
  s; i; unfold sext; destruct eq_termvar; eauto.
  subst; exfalso; eauto.
Qed.

Lemma g2f_finite: forall G,
  exists xs, forall x, ~ In x xs -> g2f G x = Tunit.
Proof. eexists; eapply g2f_finite2. Qed.

Lemma tsubbvar_lctyp: forall ty ty' n
    (CLO: lc_typ (S n) ty) 
    (CLO': lc_typ n ty'), 
  lc_typ n (tsubst_bvar n ty' ty).
Proof.
  induction ty; s; i; des; eauto using lc_mon.
  ss; depdes CLO. 
  - by rewrite (proj2 (nat_compare_eq_iff n n) refl_equal).
  - rewrite (proj1 (nat_compare_lt bb n)); s; omega.
Qed.

Lemma tsubbvar_lctyp_r: forall ty ty' n
    (CLO: lc_typ n (tsubst_bvar n ty' ty)),
  lc_typ (S n) ty.
Proof.
  induction ty; s; i; des; eauto using lc_mon.
  destruct (nat_compare bb n) eqn: X.
  - apply nat_compare_eq_iff in X; subst; omega.
  - apply nat_compare_lt in X; omega.
  - apply nat_compare_gt in X; ss; omega.
Qed.

Lemma hastype_wfprog1: forall D G e ty,
  hastype D G e ty -> wfprog (g2f G) ty e Eunit.
Proof.
  induction 1; i
  ; try match goal with [H : wfenv _ _ |- _] => destruct H end
  ; try match goal with [H : wftyp _ _ _ |- _] => destruct H end
  ; econs; s; eauto using g2f_finite; esplits; eauto using g2f_clo, list_assoc_someD, wftyp_lclo with wfprog
  ; try by apply IHhastype.
  
  - eapply (wftyp_lclo H0).
  - eapply tsubbvar_lctyp; eauto; apply IHhastype.
  - destruct IHhastype; specialize (wfprog_envok x); ss.
    by unfold sext in wfprog_envok; rewrite eqtmv_true in wfprog_envok.
  - apply IHhastype1.
  - eapply tsubbvar_lctyp_r, IHhastype.
  - by eapply tsubbvar_lctyp; eauto; [apply IHhastype|inv H0].
  - eapply tsubbvar_lctyp_r, IHhastype.
Qed.

Lemma hastype_wfprog: forall D G e1 e2 ty,
  hastype D G e1 ty -> hastype D G e2 ty -> wfprog (g2f G) ty e1 e2.
Proof.
  i; apply hastype_wfprog1 in H; apply hastype_wfprog1 in H0.
  destruct H, H0; eauto.
Qed.

Lemma wfenv_wftyp : 
  forall D G (WF: wfenv D G) x ty
    (A : list_assoc eq_termvar x G = Some ty),
    wftyp D ty.
Proof.
  destruct 1; ii; apply list_assoc_someD in A; econs; eauto.
Qed.

Lemma hastype_wfenv: forall D G e ty,
  hastype D G e ty -> wfenv D G.
Proof. induction 1; eauto. Qed.

Lemma list_minus_eq_nilD: 
  forall A eqA (l l' : list A), list_minus eqA l l' = nil <-> (forall x (IN: In x l), In x l').
Proof.
  induction l; s; i; vauto.
  pose proof (list_mem_spec eqA a l'); 
    destruct (list_mem); try rewrite IHl; split; ii; des; subst; try done; eauto.
  destruct H; eauto.
Qed.

Lemma ftn_subbvarD n ty1 ty2 x :
  ftn_typ (tsubst_bvar n ty1 ty2) x -> ftn_typ ty1 x \/ ftn_typ ty2 x.
Proof.
  revert n; induction ty2; ss; i; des; eauto;
    try solve [eapply IHty2_1 in H; tauto | eapply IHty2_2 in H; tauto]. 
  destruct nat_compare; eauto.
Qed.

Lemma ftn_subbvarI n ty1 ty2 x :
  ftn_typ ty2 x -> ftn_typ (tsubst_bvar n ty1 ty2) x.
Proof.
  revert n; induction ty2; ss; i; des; eauto.
Qed.

Lemma hastype_wftyp: forall D G e ty,
  hastype D G e ty -> wftyp D ty.
Proof.
  induction 1; eauto using hastype_wfenv, wfenv_wftyp; vauto;
  econs; ss; rewrite ?list_minus_app;
    repeat match goal with | H: wftyp _ _ |- _ => inv H; clear H; ss; 
             rewrite ?list_minus_app in *;
               des end; 
    repeat match goal with | H : _ ++ _ = nil |- _ => apply app_eq_nil in H; des end; 
    repeat match goal with | H : _ = nil |- _ => rewrite H end; 
      try split; auto using tsubbvar_lctyp; eauto using tsubbvar_lctyp_r;
    rewrite list_minus_eq_nilD in *; ii.

    apply H1; rewrite <- ftn_typL_in_iff in *;
    eapply ftn_subbvarD in IN; ss; tauto. 

  by eapply hastype_wfenv in H; inv H; clear H; ss; eauto.
  by rewrite <- app_nil_end; eapply hastype_wfenv in H; inv H; clear H; ss; eauto.

  by inv H1; rewrite list_minus_eq_nilD in *; eauto.
 
  by rewrite <- ftn_typL_in_iff in *;
     eapply ftn_subbvarD in IN; ss; des; eauto using ftn_typL_in.

  by apply H2; rewrite <- ftn_typL_in_iff in *; eauto using ftn_subbvarI.
Qed.

Lemma g2f_in: forall G x ty
    (IN: list_assoc eq_termvar x G = Some ty),
  ty = g2f G x.
Proof.
  induction G; ss; eauto.
  i; destruct a; unfold sext; destruct (eq_termvar x t) eqn: X.
  - subst; rewrite X in IN; inversion IN; eauto.
  - setoid_rewrite eqtmv_false in IN; ss; eauto.
Grab Existential Variables. eauto.
Qed.

Lemma wftyp_gen_fv: forall D n ty
    (WF: wftyp_gen D n ty) aa (NOTIN: ~ In aa D),
  ftn_typ ty aa -> False.
Proof.
  i; destruct WF.
  apply ftn_typL_in in H.
  apply (f_equal (fun x => In aa x)) in H1; simpls.
  rewrite <- H1; clear H1.
  rewrite in_list_minus; split; try done.
  pose proof (list_mem_spec eq_typename aa D).
  by destruct list_mem; eauto; destruct NOTIN.
Qed.

Lemma wftyp_fv: forall D ty
    (WF: wftyp D ty) aa (NOTIN: ~ In aa D),
  ftn_typ ty aa -> False.
Proof.
  i; destruct WF.
  apply ftn_typL_in in H.
  apply (f_equal (fun x => In aa x)) in H1; simpls.
  rewrite <- H1; clear H1.
  rewrite in_list_minus; split; try done.
  pose proof (list_mem_spec eq_typename aa D).
  by destruct list_mem; eauto; destruct NOTIN.
Qed.

Lemma wfenv_fv: forall G D
    (WF: wfenv D G) aa (NOTIN: ~ In aa D),
  forall x, ftn_typ (g2f G x) aa -> False.
Proof.
  i; eapply wftyp_fv; eauto. 
  clear - WF; induction G; vauto.
  destruct a as [aa ty]; s; unfold sext.
  destruct eq_termvar; subst.
    by eapply wfenv_wftyp with (x:=aa); eauto; simpl; destruct eq_termvar.
  eapply IHG; econs; ii; inv WF; ss; eauto.
Qed.

Theorem wreq_refl : 
  forall D G e ty (WT: hastype D G e ty), wreq (g2f G) ty e e.
Proof.
  induction 1; eauto 10 using hastype_wfprog, wftyp_lclo, g2f_in with hastype compat.
  - eapply compat_gen; eauto 10 using hastype_wfprog, wftyp_lclo, g2f_in with hastype;
    eauto using wfenv_fv, wftyp_gen_fv. 
 - eapply compat_unpack; eauto 10 using hastype_wfprog, wftyp_lclo, g2f_in with hastype;
   eauto using hastype_wfenv, hastype_wftyp, wfenv_fv, wftyp_fv. 
Qed.

Corollary wreq_reflp : 
  forall D G p ty (WT: phastype D G p ty), wreq (g2f G) ty p p.
Proof. eauto using wreq_refl, hastype_erase. Qed.

(*----------------------------------------------------------------------------*)
(** ** Congruence *)
(*----------------------------------------------------------------------------*)

Theorem wreq_cong : 
   forall C D G ty D' G' ty' 
     (WTC: wfctx C D G ty D' G' ty')
(p1 p2: prg)
(* *)
     (WF1: phastype D G p1 ty)
     (WF2: phastype D G p2 ty)
(* *)
     (WEQ: wreq (g2f G) ty p1 p2),
   wreq (g2f G') ty' (app_ctx C p1) (app_ctx C p2).
Proof.
  induction 1; i; ss; eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
                  hastype, phastype_app_ctx, hastype_erase with compat.

  * eapply compat_case; eauto using wreq_reflp.  
    eapply wreq_reflp with (G := (x,ty1)::G'); eauto.
    eapply wreq_reflp with (G := (x,ty2)::G'); eauto.
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat.

  * eapply compat_case; eauto using wreq_reflp.  
    eapply wreq_reflp with (G := (x,ty2)::G'); eauto.
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat.

  * eapply compat_case; eauto using wreq_reflp.  
    eapply wreq_reflp with (G := (x,ty1)::G'); eauto.
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat.

  * eapply compat_gen; 
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat;
    eauto using wfenv_fv, wftyp_gen_fv. 

  * eapply compat_unpack; eauto using hastype_wfenv, hastype_wftyp, wfenv_fv, wftyp_fv;
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat;
    eauto using phastype_app_ctx, hastype_erase,
          hastype_wftyp, hastype_wfenv, wfenv_fv, wftyp_fv;
    eapply wreq_reflp with (G := (x,_)::G'); eauto.

  * eapply compat_unpack; eauto using hastype_wfenv, hastype_wftyp, wfenv_fv, wftyp_fv;
    eauto 18 using wreq_reflp, wftyp_lclo, hastype_wfprog, 
           hastype, phastype_app_ctx, hastype_erase with compat;
    eauto using phastype_app_ctx, hastype_erase,
          hastype_wftyp, hastype_wfenv, wfenv_fv, wftyp_fv;
    eapply wreq_reflp with (G := (x,_)::G'); eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Inclusion in contextual equivalence *)
(*----------------------------------------------------------------------------*)

Definition ctx_equiv D G ty p1 p2 :=
  forall c (WFC: wfctx c D G ty nil nil Tint) h,
    <<DIV: diverges h (app_ctx c p1)>> <-> <<DIV': diverges h (app_ctx c p2)>>.

Lemma lreds_val_noldiv :
  forall n h e n' h' (v: voe), lreduces n h e n' h' (e_of v) -> ~ ldiverges n h e.
Proof.
  intros; eapply ldiv_lreds in H; eauto.
  by apply ldiv_div in H; pinv H; eauto using not_reduce_val.
Grab Existential Variables. exact 0.
Qed.

Theorem wreq_soundness : 
  forall D G ty (p1 p2: prg),
    phastype D G p1 ty ->
    phastype D G p2 ty ->
    wreq (g2f G) ty p1 p2 ->
  ctx_equiv D G ty p1 p2.
Proof. 
  r; i; assert (X:= wreq_adequate (wreq_cong WFC H H0 H1) h h).
  rewrite !subexp_id_closed in X.
  - des; [intuition|].
    split; i; exfalso; eauto using lreds_val_noldiv, div_ldiv, reds_lreds.
  - eauto using phastype_app_ctx, phastype_fv.
  - eauto using phastype_app_ctx, phastype_fv.
Grab Existential Variables.
exact 0. exact 0.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Symmetry *)
(*----------------------------------------------------------------------------*)

Definition fsym (r : know) := fun ty n1 v1 n2 v2 => r ty n2 v2 n1 v1.

Definition hsym (r : hrel) : hrel := fun h1 h2 => r h2 h1.

Program Definition rsym (s: wst wref) : wst wref :=
  @exist _ _ (fun ty l l' => s ty l' l) _.
Next Obligation.
destruct s; simpls; rrdes w; hdes; repeat split; subst; eauto 15.
Defined.

Hint Unfold fsym hsym rsym.

Lemma rsym_mon: forall ps ps' (PE: wextpub wref ps' ps), 
  wextpub wref (rsym ps') (rsym ps).
Proof. repeat intro; simpls; eauto. Qed.

Program Definition wsym (w : world_local) : world_local :=
  {| wnam := wnam w
   ; wst := wst w
   ; wextpub := wextpub w
   ; wval r sr s := fsym (wval w (fun sr s => fsym (r (rsym sr) s)) (rsym sr) s)
   ; wheap r sr s := hsym (wheap w (fun sr s => fsym (r (rsym sr) s)) (rsym sr) s) |}.
Next Obligation. destruct w; eauto. Qed. 
Next Obligation. eauto 10 using wextpub_trans. Qed.
Next Obligation. eauto using wnam_supports. Qed.
Next Obligation. eauto 10 using wval_mon, rsym_mon. Qed.
Next Obligation. eapply wval_closed in V; des; eauto. Qed. 
Next Obligation. eauto using wheap_mon. Qed.

Definition gsym w (g : wst (wlift (wsym w)) -> know) : wst (wlift w) -> know :=
  fun s => let (sr, sl) := s in fsym (g (rsym sr, sl)).

Lemma vclos_sym: forall r ty n v n' v' (VC: vclos (fsym r) ty n' v' n v), vclos r ty n v n' v'.
Proof. intros; eapply (vclos_induction VC); intros; destruct IH; eauto. Qed.

Lemma rsym_inv: forall s, rsym (rsym s) = s.
Proof.
intros; destruct s; unfold rsym; simpls; erewrite (proof_irrelevance _ _ w); eauto.
Qed.

Lemma gsym_inv : forall w g sr s, fsym (g (sr, s)) = @gsym w g (rsym sr, s).
Proof. by i;s; rewrite ?rsym_inv. Qed.

Lemma wval_sym: forall w g sr s ty n1 v1 n2 v2
    (WV: wval (wlift w) (fun _ => gsym g) () (rsym sr,s) ty n2 v2 n1 v1),
  wval (wlift (wsym w)) (fun _ => g) () (sr,s) ty n1 v1 n2 v2.
Proof. 
  ss; i; unfold wlift_val in *.
  case_eq (is_ref v1 v2 ty) as EQ; case_eq (is_ref v2 v1 ty) as EQ'; [|M|M];
    Mdo (by exfalso; clear - EQ EQ'; destruct v1, v2; try done; destruct ty).
  ss; unfold wref_val in *; des; repeat eexists; eauto.
Qed.

Lemma wval_sym1: forall w g sr s ty n1 v1 n2 v2
    (WV: wval (wlift (wsym w)) (fun _ => g) () (rsym sr,s) ty n2 v2 n1 v1),
  wval (wlift w) (fun _ => gsym g) () (sr,s) ty n1 v1 n2 v2.
Proof. 
  ss; i; unfold wlift_val in *.
  case_eq (is_ref v1 v2 ty) as EQ; case_eq (is_ref v2 v1 ty) as EQ'; [|M|M|];
    Mdo (by exfalso; clear - EQ EQ'; destruct v1, v2; try done; destruct ty).
  ss; unfold wref_val in *; des; repeat eexists; eauto.
  ss; unfold fsym in *; rewrite rsym_inv in *.
  by eapply wval_mon; eauto; inst; i; ss; rewrite rsym_inv in *.
Qed.

Lemma wglob_sym: forall w g (gOK: wglob (wlift (wsym w)) g),
  wglob (wlift w) (gsym g).
Proof.
  destruct 1; econs; eauto; ii; destruct s as [sa sb]; try destruct s'.
    by des; eapply wg_clo; eauto.
    by eapply wg_mon; eauto; ss; des; eauto.
    clear - wg_le; econs; i; destruct (wg_le (rsym sa, sb)); eauto. 
      by eapply refn_le, wval_sym; rewrite rsym_inv.
      by eapply wval_sym1, refn_ref. 
      by eapply wval_sym1, refn_nam. 
   by eapply wg_sc.
Qed.

Lemma dwextpub_sym w sa sb sc sd :
  dwextpub (wlift (wsym w)) (sa, sb) (sc, sd) <->
  dwextpub (wlift w) (rsym sa, sb) (rsym sc, sd).
Proof. split; split; destruct H; unnw; i; ss; eauto. Qed.

Lemma wref_heap_sym: forall sr r h1 h2
    (WH: wref_heap (fun _ sr => fsym (r (rsym sr))) (rsym sr) h2 h1),
  wref_heap (fun _ sr => r sr) sr h1 h2.
Proof.
  unfold wref_heap; ii; ss; des; repeat split; ii; eauto.
  eapply HSAT in R; des; unnw; repeat eexists; eauto.
  eby eapply vclos_sym; rewrite rsym_inv in *.
Qed.


Lemma wref_heap_sym2: forall sr r h1 h2
    (WH: wref_heap (fun _ sr => r sr) sr h1 h2),
  wref_heap (fun _ sr => fsym (r (rsym sr))) (rsym sr) h2 h1.
Proof.
  unfold wref_heap; ii; ss; des; repeat split; ii; eauto.
  eapply HSAT in R; des; unnw; repeat eexists; eauto.
  eby eapply vclos_sym; rewrite rsym_inv in *.
Qed.

Lemma wheap_sym: forall w g sr s h1 h2,
  wheap (wsym w) (fun (sr : wst wref) s => g (sr, s)) sr s h1 h2
  <-> wheap w (fun (sr : wst wref) s => fsym (g (rsym sr, s))) (rsym sr) s h2 h1.
Proof.
  ss; unfold hsym; ii.
Qed.

Lemma dwheap_sym w g sa sb h1 h2 :
  dwheap (wlift (wsym w)) (fun _ => g) () (sa, sb) h1 h2 <->
  dwheap (wlift w) (fun _ => gsym g) () (rsym sa, sb) h2 h1.
Proof. 
  split; intro H; rr in H; rr; des; unnw.
  by rewrite wheap_sym in WH'; ss; eapply wref_heap_sym2 in WH; do 5 eexists; eauto.
  by do 5 eexists; [|rewrite wheap_sym; eauto]; eapply wref_heap_sym.
Qed.

Lemma wstuck_sym: forall rf rv n1 e1 n2 e2 (STK: wstuck (fsym rf) (fsym rv) n2 e2 n1 e1),
  wstuck rf rv n1 e1 n2 e2.
Proof. intros; destruct STK; subst; eauto 10 using vclos_sym. Qed.

Lemma wstuck_sym2: forall w g sa sb n1 e1 n2 e2 
  (STK: wstuck (wval (wlift (wsym w)) (fun _ => g) () (rsym sa,sb)) (g (rsym sa, sb)) n2 e2 n1 e1),
  wstuck (wval (wlift w) (fun _ : () => gsym g) () (sa,sb)) (gsym g (sa, sb)) n1 e1 n2 e2.
Proof.
  i; destruct STK; subst; try apply vclos_sym in VREL; eauto using wval_sym1.
Qed.

Lemma wexp_sym: forall w g sr s n e n' e' 
    (WE: wexp (wlift w) (gsym g) (rsym sr,s) n' e' n e),
  wexp (wlift (wsym w)) g (sr,s) n e n' e'.
Proof.
  rr; i; rr in WE; exploit (WE h' h); eauto.
    by rewrite dwheap_sym in *.
  inversion 1; [|destruct s']; i; vauto; econs 2; try eapply wstuck_sym; try solve [s; eauto]. 
    by rewrite dwextpub_sym, rsym_inv.
  by rewrite dwheap_sym, rsym_inv.
Qed.

Lemma winhabited_sym: forall w (INH: winhabited (wlift w)), winhabited (wlift (wsym w)).
Proof. 
  rr; i; rr in INH; i; des.
  destruct s0 as (sr, s); exists (rsym sr, s); i.
  by rewrite dwheap_sym, rsym_inv; eapply INH, wglob_sym.
Qed.

Lemma wconsistent_sym: forall w (CS: wconsistent (wlift w)), wconsistent (wlift (wsym w)).
Proof.
  rr; i; rr in CS; destruct s; eapply wglob_sym in gOK.
  exploit CS; eauto; clear CS.
    eby eapply wstuck_sym2; rewrite <- (rsym_inv w0) in PR.
  by destruct 1; eapply wexp_sym in R; econs; eauto.
Qed.

Lemma wstable_sym: forall w (WEQ : wstable w), wstable (wsym w).
Proof.
  ii; apply wglob_sym in gOK; rewrite wheap_sym in HR; ss; eapply wref_heap_sym2 in SHR.
  rr in WEQ.
  exploit (WEQ _ gOK _ _ _ _ HR (rsym sr')); s; eauto.
Qed.

Definition fswap {A B} (f: A -> B * B) (x : A) := (snd (f x), fst (f x)).

Theorem wreq_sym :
  forall G ty e1 e2 (WEQ: wreq G ty e1 e2), wreq G ty e2 e1.
Proof.
  ii; r in WEQ; des; split; ii.
  destruct WFP; vauto.
  eapply REQ in INF; des; r in WEQ; des.
  exists (wsym w); repeat split; unnw; eauto using wconsistent_sym, wstable_sym, winhabited_sym.
  destruct EQ; econs; i; destruct s; s; eapply wexp_sym.
  instantiate (1 := nf2); instantiate (1 := nf1).
  eapply wglob_sym in gOK.
  specialize (WEXP (rsym w0, w1) _ gOK dd (fswap gg)).
  eapply WEXP; s; rewrite ?rsym_inv; eauto. 
  clear - dgOK.
  destruct dgOK; econs; eauto.
  intros x; specialize (wfdg_ggok x). 
  by apply vclos_sym.
Qed.

