Require Import vgtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.

Open Scope string_scope.
Open Scope list_scope.

Set Implicit Arguments.

Section trans_part1. (*********************************************************************)

Variables
  (pG : termvar -> typ)
  (pty : typ)
  (names: typename -> Prop) 
  (Nci : countably_infinite names)
  (fGty : list typename).
  
Variables G_in_fGty :
  forall a x (IN: ftn_typ (pG x) a), In a fGty.
Variables ty_in_fGty :
  forall a (IN: ftn_typ pty a), In a fGty.

Definition names1 : typename -> Prop := 
  proj1_sig (cid (countably_infinite_split Nci)).

Definition names1r : typename -> Prop := 
  proj1_sig (cid (proj2_sig (cid (countably_infinite_split Nci)))).

Lemma names_split:
  <<CI1: countably_infinite names1>> /\ <<CI1r: countably_infinite names1r>> /\
  <<UN1r: names = nunion names1 names1r>> /\ <<DJ1r: ndisj names1 names1r>>.
Proof.
  apply (proj2_sig (cid (proj2_sig (cid (countably_infinite_split Nci))))).
Qed.

Definition names2 : typename -> Prop :=
  proj1_sig (cid 
    (countably_infinite_split_finite (proj1 (proj2 names_split)) fGty)).

Definition namesnew : typename -> Prop :=
  proj1_sig (cid 
    (proj2_sig (cid 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) fGty)))).

Lemma names1r_split : 
  <<CI2: countably_infinite names2>> /\ <<CInew: countably_infinite namesnew>> /\
  <<UN2new: names1r = nunion names2 namesnew>> /\ <<DJ2new: ndisj names2 namesnew>> /\
  <<FRESH: forall a (INL: In a fGty) (INN: namesnew a), False>>.
Proof.
  apply (proj2_sig (cid
    (proj2_sig (cid 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) fGty))))).
Qed.

Variables (lpw1 lpw2: world_local).

Definition pw1 := wlift lpw1.
Definition pw2 := wlift lpw2.
Definition pw_st := (wst pw1 * wst pw2) % type.

Variables (Nam1 : wnam pw1 <1= names1) (Nam2 : wnam pw2 <1= names2)
  (Wih1 : winhabited pw1) (Wih2 : winhabited pw2)
  (Wcs1 : wconsistent pw1) (Wcs2 : wconsistent pw2).

Fixpoint decomp proj ty : typ :=
  match ty with
    | Tname a => match LEM (namesnew a) with left pf => proj a | _ => Tname a end
    | Tbvar x => Tbvar x
    | Tunit => Tunit
    | Tint => Tint
    | Tbool => Tbool
    | Tprod ty1 ty2 => Tprod (decomp proj ty1) (decomp proj ty2)
    | Tsum ty1 ty2 => Tsum (decomp proj ty1) (decomp proj ty2)
    | Tarrow ty1 ty2 => Tarrow (decomp proj ty1) (decomp proj ty2)
    | Tforall ty1 => Tforall (decomp proj ty1)
    | Texists ty1 => Texists (decomp proj ty1)
    | Tmu ty1 => Tmu (decomp proj ty1)
    | Tref ty1 => Tref (decomp proj ty1)
    | Tcont ty1 => Tcont (decomp proj ty1)
  end.

Variables 
  (nfst nsnd : typename -> typ)
  (nfst_closed : forall a, lc_typ 0 (nfst a))
  (nsnd_closed : forall a, lc_typ 0 (nsnd a))
  (decomp_inj : forall ty1 ty2
                (EQ1 : decomp nfst ty1 = decomp nfst ty2)
                (EQ2 : decomp nsnd ty1 = decomp nsnd ty2),
              ty1 = ty2)
  (decomp_surj : forall ty1 ty2 (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2),
               exists ty, decomp nfst ty = ty1 /\ decomp nsnd ty = ty2).

Lemma Nci1 : countably_infinite names1.
Proof. apply names_split. Qed.

Lemma Nci2 : countably_infinite names2.
Proof. apply names1r_split. Qed.

Lemma Ncinew : countably_infinite namesnew.
Proof. apply names1r_split. Qed.

Lemma Nunion : forall a, names a <-> names1 a \/ names2 a \/ namesnew a.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by rewrite UN1r, UN2new; split; intros.
Qed.

Lemma Ndisj12 : ndisj names1 names2.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; left.
Qed.

Lemma Ndisj1new : ndisj names1 namesnew.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; right.
Qed.

Lemma Ndisj2new : ndisj names2 namesnew.
Proof.
  apply names1r_split.
Qed.

Lemma NdisjTy : forall a (IN : (ftn_typ pty) a), namesnew a -> False.
Proof.
  i; eapply names1r_split; eauto.
Qed.

Lemma NdisjGa : forall a x (IN: ftn_typ (pG x) a), namesnew a -> False.
Proof.
  i; eapply names1r_split; eauto.
Qed.

Lemma pw_st_countable: countable_set (pw_st).
Proof. eapply countable_prod; eauto using wst_countable. Qed.

Lemma eenc_definition:
  exists EEnc : pw_st -> typ -> nat -> val+cont -> nat -> val+cont -> nat,
    forall s s' ty ty' n1 kv1 n2 kv2 n1' kv1' n2' kv2' 
      (EQ : EEnc s ty n1 kv1 n2 kv2 = EEnc s' ty' n1' kv1' n2' kv2'),
    s = s' /\ ty = ty' /\ n1 = n1' /\ kv1 = kv1' /\ n2 = n2' /\ kv2 = kv2'.
Proof.
  assert (X := countable_prod pw_st_countable
               (countable_prod typ_countable (countable_prod 
               (countable_prod nat_countable 
               (countable_sum val_countable cont_countable)) 
               (countable_prod nat_countable 
               (countable_sum val_countable cont_countable))))).
  rdes X; exists (fun s ty n1 vk1 n2 vk2 => f (s,(ty,((n1,vk1),(n2,vk2))))).
  by intros; eapply X in EQ; inv EQ.
Qed.

Definition EEnc := proj1_sig (cid eenc_definition).

Lemma EEnc_inj :
  forall s s' ty ty' n1 vk1 n2 vk2 n1' vk1' n2' vk2' (EQ : EEnc s ty n1 vk1 n2 vk2 = EEnc s' ty' n1' vk1' n2' vk2'),
    s = s' /\ ty = ty' /\ n1 = n1' /\ vk1 = vk1' /\ n2 = n2' /\ vk2 = vk2'.
Proof. apply (proj2_sig (cid eenc_definition)). Qed.

Lemma decomp_fst_tsubst_bvar : forall ty ty' n, 
  decomp nfst (tsubst_bvar n ty' ty) = tsubst_bvar n (decomp nfst ty') (decomp nfst ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N']; symmetry. 
      by apply tsubst_bvar_id; eapply lc_mon with (n:=0); [omega|apply nfst_closed].
    by apply tsubst_bvar_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_snd_tsubst_bvar : forall ty ty' n, 
  decomp nsnd (tsubst_bvar n ty' ty) = tsubst_bvar n (decomp nsnd ty') (decomp nsnd ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N']; symmetry.
      by apply tsubst_bvar_id; eapply lc_mon with (n:=0); [omega|apply nsnd_closed].
    by apply tsubst_bvar_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_lc_typ : forall ty n,
  lc_typ n (decomp nfst ty) \/ lc_typ n (decomp nsnd ty) -> lc_typ n ty.
Proof. by induction ty; simpl; intros; auto; try destruct H; des; auto. Qed.

Lemma decomp_lc_typ' : forall ty n,
  lc_typ n ty -> lc_typ n (decomp nfst ty) /\ lc_typ n (decomp nsnd ty).
Proof.
  induction ty; simpls; intros; des; hdes; eauto 10.

  case (LEM (namesnew aa)); intros; eauto.
  by split; (eapply lc_mon with (n := 0); [omega|eauto using nfst_closed, nsnd_closed]).
Qed.

Lemma decomp_tsubst : forall d (dOK: tsubst_lclosed d)
    ty (tyN : forall a, ftn_typ ty a -> namesnew a -> False) proj,
  decomp proj (tsubst_name d ty) = tsubst_name (fun a => decomp proj (d a)) ty.
Proof.
  induction ty; intros; simpls.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
Qed.

Lemma decomp_id : forall ty (tyN : forall a, ftn_typ ty a -> namesnew a -> False) proj,
  decomp proj ty = ty.
Proof.
  induction ty; intros; simpl; auto.
  - destruct (LEM (namesnew aa)) as [N|N]; auto.
    by exfalso; eapply tyN; ss; eauto. 
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; ss; eauto; apply in_or_app; right; eauto.
    by eapply tyN; ss; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; ss; eauto; apply in_or_app; right; eauto.
    by eapply tyN; ss; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; ss; eauto; apply in_or_app; right; eauto.
    by eapply tyN; ss; eauto; apply in_or_app; left; eauto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
Qed.

Definition badval s1 s2 ty n1 v1 n3 v3 :=
  Vint (EEnc (s1, s2) ty n1 (inl v1) n3 (inl v3)).

Definition badcont s1 s2 ty n1 K1 n3 K3 :=
  [SKwriteR (Vint (EEnc (s1, s2) ty n1 (inr K1) n3 (inr K3)))].

Inductive know_fst (r: pw_st -> know) s2 s1 : know :=
| know_fst_v : forall s2' ty ty' n1 v1 n3 v3
    (IN : !V (r (s1,s2')) ty n1 v1 n3 v3)
    (TEQ: ty' = decomp nfst ty) 
    (EXT: wextpub pw2 s2 s2')
    (NN : forall a, names a -> ty <> Tname a)
    (NR : forall ty', ty <> Tref ty'),
  !V (know_fst r s2 s1) ty' 
    n1 v1 0 (badval s1 s2' ty n1 v1 n3 v3)
| know_fst_k : forall s2' ty ty' n1 K1 n3 K3 
    (IN : !K (r (s1,s2')) ty n1 K1 n3 K3)
    (TEQ: ty' = decomp nfst ty) 
    (EXT: wextpub pw2 s2 s2'),
  !K (know_fst r s2 s1) ty'
    (S n1) K1 0 (badcont s1 s2' ty n1 K1 n3 K3)
.
Hint Constructors know_fst.

Inductive know_snd (r: pw_st -> know) s1 s2 : know :=
| know_snd_v : forall s1' ty ty' n1 v1 n3 v3
    (IN : !V (r (s1',s2)) ty n1 v1 n3 v3)
    (TEQ: ty' = decomp nsnd ty) 
    (EXT: wextpub pw1 s1 s1')
    (NN : forall a, names a -> ty <> Tname a)
    (NR : forall ty', ty <> Tref ty'),
  !V (know_snd r s1 s2) ty'
     0 (badval s1' s2 ty n1 v1 n3 v3) n3 v3 
| know_snd_k : forall s1' ty ty' n1 K1 n3 K3 
    (IN : !K (r (s1',s2)) ty n1 K1 n3 K3)
    (TEQ: ty' = decomp nsnd ty) 
    (EXT: wextpub pw1 s1 s1'),
  !K (know_snd r s1 s2) ty'
    0 (badcont s1' s2 ty n1 K1 n3 K3) (S n3) K3
.
Hint Constructors know_snd.

Lemma know_fst_mon: forall r s1 s2 ty n1 vk1 n2 vk2
    (WV: know_fst r s2 s1 ty n1 vk1 n2 vk2)
    r' (RLE: r <6= r')
    s2' (EXT: wextpub pw2 s2' s2),
  know_fst r' s2' s1 ty n1 vk1 n2 vk2.
Proof.
  i; depdes WV; eauto using wextpub_trans.
Qed.

Lemma know_snd_mon: forall r s1 s2 ty n1 vk1 n2 vk2
    (WV: know_snd r s1 s2 ty n1 vk1 n2 vk2)
    r' (RLE: r <6= r')
    s1' (EXT: wextpub pw1 s1' s1),
  know_snd r' s1' s2 ty n1 vk1 n2 vk2.
Proof.
  i; depdes WV; eauto using wextpub_trans.
Qed.

Lemma know_fst_closed : forall r (CL : forall s, know_closed (r s)) s1 s2, 
  know_closed (know_fst r s2 s1).
Proof.
  r; i; destruct V; subst; eapply CL in IN; ss; des; eauto.
Qed.

Lemma know_snd_closed : forall r (CL : forall s, know_closed (r s)) s1 s2, 
  know_closed (know_snd r s1 s2).
Proof.
  r; i; destruct V; subst; eapply CL in IN; ss; des; eauto.
Qed.

Inductive comp2 A (r1 r2 : relation A) v1 v3 : Prop :=
_comp2 v2 (R1: r1 v1 v2) (R2: r2 v2 v3).
Hint Constructors comp2.

Inductive bullet A (r1 r2 : typ -> nat -> A -> nat -> A -> Prop) ty n1 vk1 n3 vk3 : Prop :=
_bullet n2 n2' vk2
  (R1: r1 (decomp nfst ty) n1  vk1 n2 vk2)
  (R2: r2 (decomp nsnd ty) n2' vk2 n3 vk3)
.
Hint Constructors bullet.

(* next: construction of w *)
Definition grel_fst r s2 := wvfix pw1 (know_fst r s2). 
Definition grel_snd r s1 := wvfix pw2 (know_snd r s1). 

Definition vkclos (r: know) : know := fun ty n1 vk1 n2 vk2 =>
  match vk1, vk2 with
  | inl v1, inl v2 => vclos r ty n1 v1 n2 v2
  | inr K1, inr K2 => r ty n1 vk1 n2 vk2
  | _, _ => False
  end.

Inductive pw_val r (s : pw_st) ty n1 vk1 n2 vk2 : Prop :=
_pw_val n1' n2'
  (R: bullet (vkclos (grel_fst r (snd s) (fst s))) 
             (vkclos (grel_snd r (fst s) (snd s))) ty n1' vk1 n2' vk2)
  (NAM: forall a (NAME: ty = Tname a) v1 (V1: vk1 = inl v1) v2 (V2: vk2 = inl v2), names a)
  (CLO: forall a, ~ (fv_exp (vk2exp vk1) a \/ fv_exp (vk2exp vk2) a))
  (NEQ1: n1 >= S n1') (NEQ1: n2 >= S n2').
Hint Constructors pw_val.

Definition pw_heap r (s : pw_st) : hrel :=
  comp2 (wheap pw1 (fun _ => grel_fst r (snd s)) () (fst s)) 
        (wheap pw2 (fun _ => grel_snd r (fst s)) () (snd s)).

Definition prodrel A B (r1: relation A) (r2: relation B) : relation (A*B) :=
  fun ab ab' => r1 (fst ab) (fst ab') /\ r2 (snd ab) (snd ab').

Lemma pw_val_mon: 
  forall r s ty n vk n' vk' (WV: pw_val r s ty n vk n' vk')
    r'  (LT: r <6= r')
    s' (R: prodrel (wextpub pw1) (wextpub pw2) s' s)
    n1  (LN: n <= n1)
    n1' (LN': n' <= n1'),
  pw_val r' s' ty n1 vk n1' vk'.
Proof.
  i; destruct WV; eapply _pw_val with (n1':=n1'0) (n2':=n2'); eauto; try omega.
  unfold grel_fst, grel_snd in *; destruct R.
  destruct R0; eapply _bullet with (vk2:=vk2).
  - destruct vk, vk2; try done.
    + eapply vclos_mon; [eapply vclos_mon_lv|]; eauto. 
      * eauto using wvfix_mon.
      * eauto using wvfix_mon, know_fst_mon.
    + s; eauto using wvfix_mon, know_fst_mon.
  - destruct vk2, vk'; try done.
    + eapply vclos_mon; [eapply vclos_mon_lv|]; eauto. 
      * eauto using wvfix_mon.
      * eauto using wvfix_mon, know_snd_mon.
    + s; eauto using wvfix_mon, know_snd_mon.
Qed.

Lemma pw_val_closed: forall r s, know_closed (pw_val r s).
Proof.
  ii; destruct V; eauto.
Qed.

Lemma pw_val_wf: forall r s ty n1 vk1 n2 vk2, 
  pw_val r s ty n1 vk1 n2 vk2 -> n1 > 0 /\ n2 > 0.
Proof.
  i; depdes H; omega.
Qed.

Lemma pw_val_supports: forall r s a n v n' v' 
  (WV: !V (pw_val r s) (Tname a) n v n' v'), names a.
Proof.
  i; destruct WV; eauto.
Qed.

Lemma pw_heap_mon: 
  forall r s h h' (PR: pw_heap r s h h')
    r' (LT: r <6= r'),
  pw_heap r' s h h'.
Proof.
  i; destruct PR; econs; unfold grel_fst, grel_snd in *.
  - eapply wheap_mon; eauto. 
    s; i; eapply wvfix_mon; eauto.
    s; i; depdes PR0; eauto using gmon_mon.
  - eapply wheap_mon; eauto. 
    s; i; eapply wvfix_mon; eauto.
    s; i; depdes PR0; eauto using gmon_mon.
Qed.

Definition pw : world.
  refine 
    {| wnam := names
     ; wst := wst pw1 * wst pw2
     ; wextpub := prodrel (wextpub pw1) (wextpub pw2)
     ; wval r := fun _ => pw_val (r ())
     ; wheap r := fun _ => pw_heap (r ())
     |};
  unfold prodrel in *; try by intros; des; unnw
  ; eauto using wextpub_trans, pw_val_closed, pw_heap_mon, pw_val_supports.
  - eauto using pw_st_countable.
  - eauto using pw_val_mon. 
Defined.

Lemma grel_fst_lerefn : forall g (G : wglob pw g) s2 s1,
  lerefn (wnam pw1) (wval pw1 (fun _ => grel_fst g s2) () s1) (grel_fst g s2 s1).
Proof.
  econs; i; [r; eauto| |].
  - destruct R; eauto using wval_mon; try done.
    depdes REL REL.
    destruct ty0; try (by inversion TEQ); [|by edestruct NR].
    s in TEQ; revert TEQ; destruct (LEM (namesnew aa)); i 
    ;[by edestruct NN; auto; apply Nunion; auto|by depdes TEQ].
  - destruct R; eauto using wval_mon; try done.
    depdes REL REL.
    destruct ty; try (by inversion TEQ).
    s in TEQ; revert TEQ; destruct (LEM (namesnew aa)); i
    ; [by edestruct NN; auto; apply Nunion; auto
      |by exfalso; eapply NN; [eapply Nunion|]; eauto].
Qed.

Lemma grel_snd_lerefn : forall g (G : wglob pw g) s1 s2,
  lerefn (wnam pw2) (wval pw2 (fun _ => grel_snd g s1) () s2) (grel_snd g s1 s2).
Proof.
  econs; i; [r; eauto| |].
  - destruct R; eauto using wval_mon; try done.
    depdes REL REL.
    destruct ty0; try (by inversion TEQ); [|by edestruct NR].
    s in TEQ; revert TEQ; destruct (LEM (namesnew aa)); i 
    ;[by edestruct NN; auto; apply Nunion; auto|by depdes TEQ].
  - destruct R; eauto using wval_mon; try done.
    depdes REL REL.
    destruct ty; try (by inversion TEQ).
    s in TEQ; revert TEQ; destruct (LEM (namesnew aa)); i
    ; [by edestruct NN; auto; apply Nunion; auto
      |by exfalso; eapply NN; [eapply Nunion|]; eauto].
Qed.

Lemma grel_fst_wglob : forall g (G : wglob pw g) s2, wglob pw1 (grel_fst g s2).
Proof.
  i; econs; unfold grel_fst; i
  ; eauto using wvfix_closed, know_fst_closed, (wg_clo G), wvfix_mon, grel_fst_lerefn.
Qed.

Lemma grel_snd_wglob : forall g (G : wglob pw g) s1, wglob pw2 (grel_snd g s1).
Proof.
  i; econs; unfold grel_snd; i
  ; eauto using wvfix_closed, know_snd_closed, (wg_clo G), wvfix_mon, grel_snd_lerefn.
Qed.

Lemma grel_fst_names: forall g s1 s2 a n1 v1 n2 v2
    (R: grel_fst g s2 s1 (Tname a) n1 (inl v1) n2 (inl v2)),
  ~ names2 a.
Proof.
  i; depdes R.
  - depdes REL; depdes REL; destruct ty; depdes TEQ; try done.
    s in x; destruct (LEM(namesnew aa)).
    + eapply (NN aa); eauto; apply Nunion; eauto.
    + depdes x; eapply (NN aa); eauto; apply Nunion; eauto.
  - apply wnam_supports, Nam1 in REL; eauto using Ndisj12.
Qed.

Lemma grel_snd_names: forall g s1 s2 a n1 v1 n2 v2
    (R: grel_snd g s1 s2 (Tname a) n1 (inl v1) n2 (inl v2)),
  ~ names1 a.
Proof.
  i; depdes R.
  - depdes REL; depdes REL; destruct ty; depdes TEQ; try done.
    s in x; destruct (LEM(namesnew aa)).
    + eapply (NN aa); eauto; apply Nunion; eauto.
    + depdes x; eapply (NN aa); eauto; apply Nunion; eauto.
  - apply wnam_supports, Nam2 in REL; eauto using Ndisj12.
Qed.

Lemma know_compose_v :
  forall g (G : wglob pw g) s1 s2 ty1 ty2 n1 v1 n2 n2' v2 n3 v3
    (R1: !V (gmon pw1 (know_fst g s2) s1) ty1 n1  v1 n2 v2)
    (R2: !V (gmon pw2 (know_snd g s1) s2) ty2 n2' v2 n3 v3),
  exists ty, !V (g (s1, s2)) ty n1 v1 n3 v3 /\
             ty1 = decomp nfst ty /\ ty2 = decomp nsnd ty.
Proof.
  i; depdes R1 REL R2 REL; try done.
  apply EEnc_inj in x; des; simpl_injs; subst; inv x; clear x.
  esplits; eauto. 
  eapply (wg_mon G); try split; eauto.
Qed.

Lemma know_compose_k :
  forall g (G : wglob pw g) s1 s2 ty1 ty2 n1 K1 n2 n2' K2 n3 K3
    (R1: !K (gmon pw1 (know_fst g s2) s1) ty1 n1  K1 n2 K2)
    (R2: !K (gmon pw2 (know_snd g s1) s2) ty2 n2' K2 n3 K3),
  exists ty, !K (g (s1, s2)) ty (n1-1) K1 (n3-1) K3 /\
             ty1 = decomp nfst ty /\ ty2 = decomp nsnd ty.
Proof.
  i; depdes R1 REL R2 REL; try done.
  apply EEnc_inj in x; des; simpl_injs; subst; inv x; clear x.
  esplits; eauto.
  eapply (wg_mon G); try split; eauto; omega.
Qed.

Lemma bullet_grel1 : forall g (G : wglob pw g) s1 s2 ty n1 vk1 n2 vk2
    (R: bullet (vkclos (grel_fst g s2 s1)) 
               (vkclos (grel_snd g s1 s2)) ty n1 vk1 n2 vk2),
  vkclos (g (s1, s2)) ty (S n1) vk1 (S n2) vk2.
Proof.
  i; destruct R; destruct vk1, vk2, vk0; try done.

  (* Values *)
  {
  remember (decomp nfst ty) in R1; move R1 before G; revert_until R1.
  pattern t, n1, v, n0, v1; apply (vclos_induction_strong R1).
  clear t n1 v n0 v1 R1; i; subst. 
  destruct ty0
  (* non-flexible types except names and cont values *)
  ; [|M|M|M|M|M|M| | |M| |M|]
  ; (Mdo by depdes IH; destruct R2; depdes REL; unfold vkclos in *; des
    ; try (destruct decomp_surj with ty'0 ty' as (ty'' & E1 & E2); eauto; subst)
    ; do 2 econs; try eapply VR0
    ; try rewrite decomp_fst_tsubst_bvar; try rewrite decomp_snd_tsubst_bvar
    ; eauto using decomp_lc_typ)
  (* flexible types *)
  ; [|M|M|M|]
  ; (Mdo by depdes IH; do 2 econs; apply G; econs; eauto; try done
    ; [econs; [|eauto]; unfold vkclos in *; s; eauto
    | i; destruct R2; depdes REL;
      eapply wvfix_closed in VR;  [|i; eapply know_snd_closed, G];
      eapply wvfix_closed in VR0; [|i; eapply know_fst_closed, G]; des; eauto]).

  (* names *)
  {
  simpl in IH; destruct (LEM (namesnew aa)) as [N|N] eqn: E.
  - do 2 econs; eapply (wg_le G); econs; eauto.
    + econs; eauto; s; rewrite E.
      econs; eapply vclos_gen_mon; eauto.
      i; apply PR.
    + i; depdes NAME; apply Nunion; eauto.
    + assert (R2':=vclos_closed R2); clear R2.

      assert (IH':=vclos_gen_closed IH); clear IH.
      i; des; [|by eapply R2'; [apply wvfix_closed, know_snd_closed, G|eauto]].
      ss; eapply IH'; eauto.
      * by eapply wvfix_closed; i; apply know_fst_closed, G.
      * i; des; eapply (vclos_closed V0); eauto
        ; apply wvfix_closed; i; apply know_fst_closed, G.
  - s in R2; rewrite E in *.
    depdes IH; destruct R2; depdes REL.
    assert (X:= grel_fst_names VR0); assert (Y:= grel_snd_names VR).
    depdes VR; [| by apply wnam_supports, Nam2 in REL].
    depdes VR0; [| by apply wnam_supports, Nam1 in REL0].
    depdes REL; try done; depdes REL0; try done; depdes REL; depdes REL0.
    apply EEnc_inj in x; des; simpl_injs; subst; depdes x2 x4.
    do 2 econs; replace (Tname aa) with ty. 
    + eapply (wg_mon G); eauto; split; eauto.
    + by eapply decomp_inj; s; rewrite E.
  }
  (* cont values *)
  {
  depdes IH; destruct R2; depdes REL; unfold vkclos in *; des
  ; try (destruct decomp_surj with ty'0 ty' as (ty'' & E1 & E2); eauto; subst).
  do 2 econs; eapply (wg_le G); econs; eauto.
  + eapply _bullet with (vk2:=inr _); eauto.
  + i; vauto.
  + eapply wvfix_closed in VR; [|by i; eapply know_snd_closed, G].
    eapply wvfix_closed in VR0; [|by i; eapply know_fst_closed, G].
    i; des; eauto.
  }
  }

  (* Continuations *)
  {
  eapply (wg_le G); econs; eauto; try done.
  eapply wvfix_closed in R1; [|i; eapply know_fst_closed, G].
  eapply wvfix_closed in R2; [|i; eapply know_snd_closed, G].
  i; des; eauto.
  }
Grab Existential Variables. exact 0. exact 0.
Qed.

Lemma bullet_grel2_v : forall g (G : wglob pw g) s1 s2 ty n1 v1 n2 v2
    (R: vclos (g (s1, s2)) ty n1 v1 n2 v2),
  bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty n1 v1 n2 v2.
Proof.
  i; pattern ty, n1, v1, n2, v2; apply (vclos_induction_strong R).
  clear ty n1 v1 n2 v2 R; i; subst.
  destruct ty
  ; [|M|M|M|M|M|M|M| |M|M|M|M]
  ; (Mdo by depdes IH; des
    ; try (destruct VR)
    ; try (destruct VR1)
    ; try (destruct VR2)
    ; try rewrite decomp_fst_tsubst_bvar in R1; try rewrite decomp_snd_tsubst_bvar in R2
    ; econs; s; do 2 econs; try(do 3 econs)
    ; eauto; try apply decomp_lc_typ').

  (* names *)
  - depdes IH; destruct (classic(names aa)).
    + apply (wg_le G) in VR; eauto; s in VR; depdes VR.
      destruct R; destruct vk2; [|done]; simpl in R1, R2. 
      destruct (LEM(namesnew aa)) as [N|N] eqn : E.
      * econs; s; rewrite E
        ; eapply vclos_mon_lv; try r; eauto using wvfix_mon; omega.
      * depdes R1 REL R2 REL. 
        exfalso; apply Nunion in H.
        des; eauto using grel_fst_names, grel_snd_names.
    + destruct (LEM(namesnew aa)) as [N|N] eqn : E.
        by exfalso; apply H, Nunion; eauto.
      econs; s; rewrite E
      ; by do 5 econs; eauto
         ; [s; rewrite E|ii; depdes H1; eauto|].

  (* refs *)
  - depdes IH; apply (wg_le G) in VR; s in VR; depdes VR.
    destruct R; destruct vk2; [|done]; simpl in R1, R2; depdes R1 REL R2 REL.
    econs; s; do 2 econs; eauto.

Grab Existential Variables.
exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
Qed.

Lemma bullet_grel2_k : forall g (G : wglob pw g) s1 s2 ty n1 K1 n2 K2
    (R: g (s1, s2) ty n1 (inr K1) n2 (inr K2)),
  bullet (!K (grel_fst g s2 s1)) (!K (grel_snd g s1 s2)) ty (S n1) K1 (S n2) K2.
Proof.
  i; do 3 econs; ss; eauto.
Qed.

Lemma atom_app_val_val : forall (f v: voe), is_atom (Eapp (e_of f) (e_of v)).
Proof.
  i; destruct f, v; subst.
  ii; apply NV; clear NV.
  destruct sk; inv EQ; eauto using val_is_voe.
Qed.

Lemma atom_inst_val : forall v, is_atom (Einst (e_of v)).
Proof.
  i; destruct v; subst.
  ii; apply NV; clear NV.
  destruct sk; inv EQ; eauto using val_is_voe.
Qed.

Lemma atom_write_val_val: forall i (v: voe), is_atom (Ewrite (Eint i) (e_of v)).
Proof.
  i; destruct v; subst.
  ii; apply NV; clear NV.
  destruct sk; inv EQ; eauto using val_is_voe.
Qed.

Lemma write_app_conflict: forall i (v f a: voe) K,
  ~ Ewrite (Eint i) (e_of v) = app_cont K (Eapp (e_of f) (e_of a)).
Proof.
  i; eapply app_cont_eq with (k:=[]) in FH
  ; eauto using atom_app_val_val, atom_write_val_val.
  des; depdes FH0.
Qed.

Lemma write_inst_conflict: forall i (v f: voe) K,
  ~ Ewrite (Eint i) (e_of v) = app_cont K (Einst (e_of f)).
Proof.
  i; eapply app_cont_eq with (k:=[]) in FH
  ; eauto using atom_inst_val, atom_write_val_val.
  des; depdes FH0.
Qed.

Lemma app_inst_conflict: forall (f a g: voe) K K',
  ~ app_cont K (Eapp (e_of f) (e_of a)) = app_cont K' (Einst (e_of g)).
Proof.
  i; eapply app_cont_eq in FH
  ; eauto using atom_app_val_val, atom_inst_val.
  des; depdes FH0.
Qed.

Lemma write_ksys_conflict: forall i (v v': voe),
  ~ Ewrite (Eint i) (e_of v) = app_cont [] (e_of v').
Proof.
  i; destruct v'; subst; destruct v_of; vauto.
Qed.

Lemma app_ksys_conflict: forall (f a v: voe) K,
  ~ app_cont K (Eapp (e_of f) (e_of a)) = app_cont [] (e_of v).
Proof.
  i; destruct v; subst; ss; cut (is_val v_of); eauto. 
  rewrite <- FH; i; eapply val_is_voe, value_cont in H; vauto.
Qed.

Lemma inst_ksys_conflict: forall (f v: voe) K,
  ~ app_cont K (Einst (e_of f)) = app_cont [] (e_of v).
Proof.
  i; destruct v; subst; ss; cut (is_val v_of); eauto. 
  rewrite <- FH; i; eapply val_is_voe, value_cont in H; vauto.
Qed.

Lemma not_reduce_app_int : forall h n (v : voe) h' e',
  ~ reduce h (Eapp (Eint n) (e_of v)) h' e'.
Proof.
  i; depdes FH; subst. 
  - by i; depdes RED; destruct SK; depdes x; eauto using not_reduce_val.
  - exploit (atom_generalize (@atom_app_val_val _ _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
  - exploit (atom_generalize (@atom_app_val_val _ _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
Qed.

Lemma not_reduce_inst_int : forall h n h' e',
  ~ reduce h (Einst (Eint n)) h' e'.
Proof.
  i; depdes FH; subst. 
  - by i; depdes RED; destruct SK; depdes x; eauto using not_reduce_val.
  - exploit (atom_generalize (@atom_inst_val _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
  - exploit (atom_generalize (@atom_inst_val _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
Qed.

Lemma not_reduce_write_to_int: forall h n (v : voe) h' e',
  ~ reduce h (Ewrite (Eint n) (e_of v)) h' e'.
Proof.
  i; depdes FH; subst. 
  - by i; depdes RED; destruct SK; depdes x; eauto using not_reduce_val.
  - exploit (atom_generalize (@atom_write_val_val _ _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
  - exploit (atom_generalize (@atom_write_val_val _ _) _ _ (eq_sym x)); eauto.
    i; subst; depdes x.
Qed.

Lemma wstuck_fst_not_reduce: forall g r s1 s2 n1 e1 n2 e2
    (STK: wstuck (gmon pw1 (know_fst g s2) s1 \5/ sysknow) r n1 e1 n2 e2 ),
  forall h2 h2' e2', ~ reduce h2 e2 h2' e2'.
Proof.
  i; apply wstuck_union in STK; des.
  - destruct STK; subst; depdes CREL; try done; depdes REL
    ; [|M|M]; (Mdo destruct f2; eapply not_reduce_cont, FH); ss; subst; s
    ; eauto using not_reduce_write_to_int, not_reduce_app_int, not_reduce_inst_int.
  - destruct STK; subst; depdes CREL; des; clarify; ss; eauto using not_reduce_val.
Qed.

Lemma wstuck_snd_not_reduce: forall g r s1 s2 n1 e1 n2 e2
    (STK: wstuck (gmon pw2 (know_snd g s2) s1 \5/ sysknow) r n1 e1 n2 e2),
  forall h1 h1' e1', ~ reduce h1 e1 h1' e1'.
Proof.
  i; apply wstuck_union in STK; des.
  - i; destruct STK; subst; depdes CREL; try done; depdes REL
    ; [|M|M]; (Mdo destruct f1; eapply not_reduce_cont, FH); ss; subst; s
    ; eauto using not_reduce_write_to_int, not_reduce_app_int, not_reduce_inst_int.
  - destruct STK; subst; depdes CREL; des; clarify; ss; eauto using not_reduce_val.
Qed.

Lemma wexp_trans : forall g (gOK: wglob pw g) s1 s2 n1 e1 n2 n2' e2 n3 e3
    (EQ1: wexp pw1 (grel_fst g s2) s1 n1 e1 n2 e2)
    (EQ2: wexp pw2 (grel_snd g s1) s2 n2' e2 n3 e3),
  wexp pw g (s1, s2) n1 e1 n3 e3.
Proof.
  rr; i; destruct HREL; rename v2 into h2.
  eapply wexp_core_external with (r:=gmon pw1 (know_fst g s2) \6/ fun _ => sysknow) in EQ1
  ; eauto using grel_fst_wglob.
  2: by i; depdes PR; subst; eauto 10 using wval_mon.
  eapply wexp_core_external with (r:=gmon pw2 (know_snd g s1) \6/ fun _ => sysknow) in EQ2
  ; eauto using grel_snd_wglob.
  2: by i; depdes PR; subst; eauto 10 using wval_mon.
  clear R1 R2; destruct EQ1, EQ2.

  (* div & div *)
  {
  by apply wexp_Diverge. 
  }

  (* div & stuck *)
  {
  exfalso; assert (X:= ldiv_lreds RED DIV' 0); apply ldiv_div in X; pdes X.
  eauto using wstuck_snd_not_reduce.
  }

  (* stuck & div *)
  {
  exfalso; assert (X:= ldiv_lreds RED' DIV 0); apply ldiv_div in X; pdes X.
  eauto using wstuck_fst_not_reduce.
  }

  (* stuck & stuck *)
  {
  apply hplus_hemp_rE in PLUS0; apply hplus_hemp_rE in PLUS2; subst.
  assert (STK1 := wstuck_fst_not_reduce EREL).
  assert (STK2 := wstuck_snd_not_reduce EREL0).
  assert (EQ := lreds_term_det RED' (STK1 _) RED0 (STK2 _)).
  des; subst; clear RED' RED0 STK1 STK2.
  eapply wexp_Stuck with (s':=(s',s'0)); eauto.
  - split; eauto.
  - destruct EREL; subst; des; depdes CREL; try depdes REL 
    ; destruct EREL0; subst; des; depdes CREL; try depdes REL
    ; try (by unfold badcont, sysknow in *; des; clarify; exfalso 
            ; eauto using write_app_conflict, write_inst_conflict, app_inst_conflict, 
                          write_ksys_conflict, app_ksys_conflict, inst_ksys_conflict).
    + (* return & return (bad continuations) *)
      repeat simpl_injs; subst.
      apply EEnc_inj in x; simpl_injs; des; subst; clarify.
      eapply wstuck_Return; [M|M| | | |]; Mdo by eauto.
      * eapply (wg_mon gOK); eauto; split; eauto.
      * eapply bullet_grel1 with (vk1:=inl _) (vk2:=inl _); eauto.
        eapply _bullet with (vk2:=inl _); eapply vclos_mon; eauto
        ; i; eapply wvfix_mon; eauto using know_fst_mon, know_snd_mon.
      * omega.
      * omega.
    + (* return & return (system continuations) *)
      des; clarify; ss; destruct VREL, VREL0; depdes REL REL0; simpl_injs; clarify.
      eapply wstuck_Return with (nv1:=ne1) (nk1:=0) (K1:=[]) (nv2:= ne3) (nk2:=0) (K2:=[])
      ; s; eauto using (wg_sc gOK); s; eauto; omega.
    + (* app & app *)
      repeat simpl_injs; subst.
      eapply app_cont_eq in EQ1; eauto using atom_app_val_val.
      des; depdes EQ0; apply EEnc_inj in x; des; repeat simpl_injs; clarify.
      assert (T : exists tty1 tty2, ty2 = Tarrow tty1 tty2).
        destruct ty2; ss; eauto.
        destruct (LEM (namesnew aa)) as [N'|N'] eqn : E'.
          by exfalso; eapply NN; [apply Nunion; right|]; eauto.
        by inversion TEQ.
      des; subst; simpl in TEQ, TEQ0; depdes TEQ.
      eapply wstuck_App; [M|M| | |]; Mdo by eauto.
      * eapply (wg_mon gOK); eauto; split; eauto.
      * eapply bullet_grel1 with (vk1:=inl _) (vk2:=inl _); eauto.
        eapply _bullet with (vk2:=inl _); eapply vclos_mon; eauto
        ; i; eapply wvfix_mon; eauto using know_fst_mon, know_snd_mon.
      * eapply bullet_grel1 with (vk1:=inr _) (vk2:=inr _); eauto
        ; eapply _bullet with (vk2:=inr _)
        ; eapply wvfix_mon; eauto using know_fst_mon, know_snd_mon.
    + (* inst & inst *)
      repeat simpl_injs; subst.
      eapply app_cont_eq in EQ1; eauto using atom_inst_val.
      des; depdes EQ0; apply EEnc_inj in x; des; repeat simpl_injs; clarify.
      assert (T : exists tty, ty2 = Tforall tty).
        destruct ty2; ss; eauto.
        destruct (LEM (namesnew aa)) as [N'|N'] eqn : E'.
          by exfalso; eapply NN; [apply Nunion; right|]; eauto.
        by inversion TEQ.
      des; subst; simpl in TEQ, TEQ0; depdes TEQ.
      assert (X := decomp_surj _ _ LC LC0); des; subst.
      eapply wstuck_Inst with (ty':=ty); [M|M|M| |]; Mdo (by eauto using decomp_lc_typ).
      * eapply (wg_mon gOK); eauto; split; eauto.
      * eapply bullet_grel1 with (vk1:=inr _) (vk2:=inr _); eauto.
        unfold grel_fst, grel_snd; eapply _bullet with (vk2:=inr _); s
        ; rewrite ?decomp_fst_tsubst_bvar,?decomp_snd_tsubst_bvar
        ; eauto using wvfix_mon, know_fst_mon, know_snd_mon.
  - econs; eapply wheap_mon; eauto; unfold grel_fst, grel_snd; i
    ; eauto using wvfix_mon, know_fst_mon, know_snd_mon.
  }
Qed.

Lemma pw_inhabited : winhabited pw.
Proof.
  destruct Wih1 as (s1,INH1), Wih2 as (s2,INH2).
  exists (s1,s2); i; econs. 
  - eapply INH1, (grel_fst_wglob gOK).
  - eapply INH2, (grel_snd_wglob gOK).
Qed.

Lemma pw_consistent : wconsistent pw.
Proof.
  ii; rename x0 into n1, x1 into e1, x2 into n3, x3 into e3.
  destruct s as (s1,s2).
  assert (P : exists n2 n2' e2, 
    n1 > 0 /\ n3 > 0 /\
    wstuck (grel_fst g s2 s1) (grel_fst g s2 s1) (n1-1) e1 n2 e2 /\
    wstuck (grel_snd g s1 s2) (grel_snd g s1 s2) n2' e2 (n3-1) e3).
  {
    depdes PR; subst.
    - (* return *)
      apply bullet_grel2_v in VREL; eauto; depdes VREL.
      inv CREL; depdes R; destruct vk0; try (by inversion R0).
      esplits; try omega; eapply wstuck_Return; eauto; omega.
    - (* app *)
      apply bullet_grel2_v in VREL; eauto; depdes VREL.
      apply (bullet_grel2_k gOK) in KREL; depdes KREL.
      inv CREL; depdes R; destruct vk1; try (by inversion R1).
      s in R4; depdes R4; depdes REL; s in R5; depdes R5; depdes REL.
      esplits; try omega; eapply wstuck_App; eauto
      ; [eapply (wg_mon (grel_fst_wglob gOK _))
        |eapply (wg_mon (grel_snd_wglob gOK _))]; s; eauto; omega.
    - (* inst *)
      apply (bullet_grel2_k gOK) in KREL; depdes KREL.
      inv CREL; depdes R; destruct vk0; try (by inversion R1).
      s in R0; depdes R0; depdes REL; s in R3; depdes R3; depdes REL.
      esplits; try omega.
      + eapply wstuck_Inst with (ty':=decomp nfst ty') (ty:=decomp nfst ty); eauto.
        * eapply decomp_lc_typ'; eauto.
        * eapply (wg_mon (grel_fst_wglob gOK _)); s; eauto; omega.
        * rewrite <- decomp_fst_tsubst_bvar.
          eapply (wg_mon (grel_fst_wglob gOK _)); s; eauto.
      + eapply wstuck_Inst with (ty':=decomp nsnd ty') (ty:=decomp nsnd ty); eauto.
        * eapply decomp_lc_typ'; eauto.
        * eapply (wg_mon (grel_snd_wglob gOK _)); s; eauto; omega.
        * rewrite <- decomp_snd_tsubst_bvar.
          eapply (wg_mon (grel_snd_wglob gOK _)); s; eauto.
  }

  des; econs.
  - eapply wexp_trans; eauto; ii; eapply wexp_Stuck; eauto using lreduces_nil; s; eauto.
  - econs; omega.
  - econs; omega.
Qed.

Lemma weq_trans : forall e1 e2 e3
    (WEQ1 : weq pw1 pG pty e1 e2)
    (WEQ2 : weq pw2 pG pty e2 e3),
  weq pw pG pty e1 e3.
Proof.
  i; cut (wequiv pw pG pty e1 e3).
    by s; eauto 10 using pw_inhabited, pw_consistent.

  rdes WEQ1; rdes WEQ2.
  destruct EQ, EQ0; econs; eauto.
  i; destruct s as (s1, s2).

  (**********************************************************************)
  (* in order to construct the mediating gammas, we use classical choice *)
  (* to do so, we first bring bullet_grel into the right form: *)
  assert (bullet_grel' :
      forall (t : typ * ((nat * val) * (nat * val))),
      exists nnv2 : (nat * nat) * val,
        match t with pair ty (pair (pair n1 v1) (pair n3 v3)) =>
        match nnv2 with pair (pair n2 n2') v2 =>
        forall (IN : vclos (g (s1, s2)) ty n1 v1 n3 v3),
          vclos (grel_fst g s2 s1) (decomp nfst ty) n1 v1 n2 v2 /\
          vclos (grel_snd g s1 s2) (decomp nsnd ty) n2' v2 n3 v3
        end end).
  {
    intros; destruct t as (ty,((n1,v1),(n3,v3))).
    destruct (classic (vclos (g (s1, s2)) ty n1 v1 n3 v3)) as [IN|IN]
    ; [| by exists (0,0,Vunit)].
    apply bullet_grel2_v in IN; auto.
    destruct IN; exists (n2,n2',vk2); eauto.
  }

  apply choice in bullet_grel'; destruct bullet_grel' as (f & Pf).
  set (gg1 := fun x => (
         fst(gg x),
         let nnv := f (tsubst_name dd (pG x),gg x) in (fst(fst nnv),snd nnv))).
  set (gg2 := fun x => (
         let nnv := f (tsubst_name dd (pG x),gg x) in (snd(fst nnv),snd nnv),
         snd(gg x))).
  set (dd1 := (decomp nfst) <*> dd).
  set (dd2 := (decomp nsnd) <*> dd).

  (**********************************************************************)

  eapply (bullet_grel2_k gOK) in KOK; eauto.
  destruct KOK.
  eapply wexp_trans; [eauto| | ].
  - eapply wexpgen_mon; [eapply (WEXP s1 _ (grel_fst_wglob gOK s2) dd1) with (gg:=gg1)| |].
    + econs.
      * i; eapply decomp_lc_typ'; eauto using wfdg_ddok.
      * i; assert (X: decomp nfst (tsubst_name dd (pG x)) = tsubst_name dd1 (pG x)).
        { apply decomp_tsubst
          ; [eauto using (wfdg_ddok dgOK)|i; eapply NdisjGa; eauto]. }
        rewrite <- X; clear X.
        unfold compose, gg1, gg2 in *.
        specialize (Pf (tsubst_name dd (pG x), gg x)).
        ss; remember (gg x) as ggx in *.
        destruct ggx as ((n1,v1),(n3,v3)).
        remember (f (tsubst_name dd (pG x),((n1,v1),(n3,v3)))) as nnv in *.
        destruct nnv as ((?,?),?); ss.
        assert (X:=wfdg_ggok dgOK x).
        by unfold compose in X; rewrite <-Heqggx in X; eapply Pf, X.
    + rewrite decomp_tsubst in R1; eauto using NdisjTy, wfdg_ddok.
    + instantiate (1:= fun xs => S (nf1 xs)).
      unfold gg1, compose; s; omega.
    + eauto.
  - eapply wexpgen_mon; [eapply (WEXP0 s2 _ (grel_snd_wglob gOK s1) dd2) with (gg:=gg2)| |].
    + econs.
      * i; eapply decomp_lc_typ'; eauto using wfdg_ddok.
      * i; assert (X: decomp nsnd (tsubst_name dd (pG x)) = tsubst_name dd2 (pG x)).
        { apply decomp_tsubst
          ; [eauto using (wfdg_ddok dgOK)|i; eapply NdisjGa; eauto]. }
        rewrite <- X; clear X.
        unfold compose, gg1, gg2 in *.
        specialize (Pf (tsubst_name dd (pG x), gg x)).
        ss; remember (gg x) as ggx in *.
        destruct ggx as ((n1,v1),(n3,v3)).
        remember (f (tsubst_name dd (pG x),((n1,v1),(n3,v3)))) as nnv in *.
        destruct nnv as ((?,?),?); ss.
        assert (X:=wfdg_ggok dgOK x).
        by unfold compose in X; rewrite <-Heqggx in X; eapply Pf, X.
    + rewrite decomp_tsubst in R2; eauto using NdisjTy, wfdg_ddok.
    + eauto.
    + instantiate (1:= fun xs => S (nf3 xs)); unfold gg2, compose; s; omega.
Qed.

End trans_part1.
