Require Import Program List Arith Even Div2 Relations.
Require Import hpattern vgtac basis countable lang subs model basiclemmas.
Require Import Classical FunctionalExtensionality.
Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** * wcomplete *)
(*----------------------------------------------------------------------------*)

Definition is_ref_name (vk1 vk2: val+cont) ty : bool :=
   match vk1, vk2, ty with
   | inl _, inl _, Tname _ => true
   | inl _, inl _, Tref  _ => true
   | _, _, _ => false
   end.

Definition woverwrite (w: world) lk r s ty n1 vk1 n2 vk2 :=
   if is_ref_name vk1 vk2 ty
   then wval w (fun _ => r) () s ty n1 vk1 n2 vk2
   else lk r s ty n1 vk1 n2 vk2.

Inductive wcgen (w: world) lk (g: wst w -> know) s : know :=
| _wcgen0: forall ty n1 vk1 n2 vk2
    (CLO : know_closed (singknow ty n1 vk1 n2 vk2))
    (CONS: forall (g': wst w -> know) s'
             (g'OK: wglobgen w (woverwrite w lk) g')
             (LES: wextpub w s' s)
             (LEG: g <6= g'),
           wstuck (singknow ty n1 vk1 n2 vk2) (g' s') <4= drop (wexp w g' s')),
  wcgen w lk g s ty n1 vk1 n2 vk2
.
Hint Constructors wcgen.

Definition wkcomp0 (w: world) g s ty n1 vk1 n2 vk2 : Prop :=
  paco7 (wcgen w) bot7 g s ty n1 vk1 n2 vk2.
Hint Unfold wkcomp0.

Definition wkcomp (w: world) := woverwrite w (wkcomp0 w).
Hint Unfold wkcomp.

Ltac unow := unfold wkcomp, woverwrite in *.

Lemma wglobgen_mon: forall (w: world) lk lk' g
    (gOK: wglobgen w (woverwrite w lk) g)
    (LE: lk' <7= lk),
  wglobgen w (woverwrite w lk') g.
Proof.
  i; destruct gOK; econs; eauto.
  unow; i; econs.
  - i; specialize (refn_le (wg_le s) ty n vk n' vk'). 
    destruct (is_ref_name _ _ _); eauto.
  - by i; s; apply (refn_ref (wg_le _)).
  - by i; s; apply (refn_nam (wg_le _)).
Qed.

Lemma wcgen_mon: forall (w: world), monotone7 (wcgen w).
Proof.
  ii; destruct IN; econs; eauto.
  i; eapply CONS; eauto using wglobgen_mon.
Qed.
Hint Resolve wcgen_mon : paco.

Lemma wkcomp_mon: forall (w: world)
    s r ty n vk n' vk' (WV: wkcomp w r s ty n vk n' vk')
    r'  (LT: r <6= r')
    s' (R: wextpub w s' s)
    n1  (LN: n <= n1)
    n1' (LN': n' <= n1'),
  wkcomp w r' s' ty n1 vk n1' vk'.
Proof.
  i; unow; destruct (is_ref_name _ _ _).
    by eauto using wval_mon.
  revert_until w; pcofix CIH; i.
  pdes WV; pfold; econs; [by ii; depdes V; eauto|].
  i; apply CONS; eauto using wextpub_trans.
  - eapply wglobgen_mon; eauto.
    by i; des; [|done]; left; eapply paco7_mon; eauto; i; done.
  - eapply wstuck_mon; eauto.
    i; depdes PR0; eauto; econs; omega.
Qed.

Lemma woverwrite_wval: forall (w: world),
  woverwrite w (fun r => wval w (fun _ => r) ()) = (fun r => wval w (fun _ => r) ()).
Proof.
  i; extensionalities r s; extensionalities ty n vk n' vk'.
  by unow; destruct (is_ref_name _ _ _).
Qed.

Lemma wkcomp_val : forall (w: world) (CS : wconsistent w) g,
  wval w (fun _=>g) () <6= wkcomp w g.
Proof.
  i; r; r; destruct (is_ref_name _ _ _); eauto.
  revert_until CS.
  assert (WCG : forall g, 
    wval w (fun _=>g) () <6= wcgen w (fun r' => wval w (fun _ => r') ()) g).
  { econs; [by ii; inversion V; subst; eapply wval_closed; eauto|].
    i; rewrite woverwrite_wval in g'OK; eapply CS; eauto.
    depdes PR0; subst; inversion CREL; subst; eauto using wval_mon.
  }

  pcofix CIH; i; pfold; eapply wcgen_mon; [by eauto using WCG|].
  i; ss; right; eapply CIH in PR0; eauto.
Qed.

Program Definition wcomp (w : world_local) : world_local :=
  {| wst := wst w; wnam := wnam w; wextpub := wextpub w
   ; wheap := wheap w
   ; wval g sr s := wkcomp (wlift w) (fun t => g (fst t) (snd t)) (sr,s) |}.
Next Obligation. eauto using wst_countable. Qed.
Next Obligation. eauto using wextpub_trans. Qed.
Next Obligation. by do 3 r in WV; apply wnam_supports in WV; rr in WV; des. Qed.
Next Obligation. eapply wkcomp_mon; eauto; s; eauto. Qed.
Next Obligation.
  unow; destruct (is_ref_name _ _ _).
  - eapply wval_closed in V; eauto. 
  - pdes V; eauto.
Qed.
Next Obligation. eauto using wheap_mon. Qed.

Lemma ref_is_ref_name: forall {vk vk' ty}, 
  is_ref vk vk' ty = true -> is_ref_name vk vk' ty = true.
Proof. i; destruct vk, vk', ty; eauto. Qed.

Lemma wlift_wcomp: forall (w: world_local), 
  (fun r => wval (wlift (wcomp w)) (fun _=>r) ()) =  wkcomp (wlift w).
Proof.
  i; extensionalities g s ty; extensionalities n vk n' vk'.
  s; unfold wlift_val.
  destruct (is_ref _ _ _) eqn: X.
    by unow; s; unfold wlift_val; rewrite (ref_is_ref_name X), X.
  s; rewrite <-surjective_pairing.
  by f_equal; extensionality s'; destruct s'.
Qed.

Lemma wkcomp_glob_gen :
  forall (w: world) g r
    (CS: wconsistent w) 
    (G:  wglobgen w  (woverwrite w (paco7 (wcgen w) r \7/ r)) g),
  wglob w g.
Proof.
  i; destruct G; econs; eauto.
  econs; i.
  - eapply wg_le; apply wkcomp_val in R; eauto.
    r; do 2 r in R; destruct is_ref_name; eauto.
    by left; eapply paco7_mon; eauto.
  - apply (refn_ref (wg_le _)) in R; unow; ss.
  - apply (refn_nam (wg_le _)) in R; unow; ss.
Qed.

Lemma wkcomp_glob :
  forall (w: world) g
    (CS : wconsistent w) 
    (G : wglobgen w (wkcomp w) g), 
  wglob w g.
Proof.
  i; eapply wkcomp_glob_gen; eauto.
  by eapply wglobgen_mon; i; des; eauto.
Qed.

Lemma wcomp_glob: forall (w: world_local) g
    (CS : wconsistent (wlift w))
    (gOK : wglob (wlift (wcomp w)) g),
  wglob (wlift w) g.
Proof.
  i; eapply wkcomp_glob; eauto.
  by rewrite <-wlift_wcomp.
Qed.

Lemma wkcomp_consistent: forall (w: world), wconsistentgen w (wkcomp w).
Proof.
  r; i; destruct PR; subst; pinv CREL; eapply CONS; s; eauto
  ; by eapply wglobgen_mon; eauto; i; des; eauto.
Qed.

Lemma lk_generalize: 
  forall (w: world) (lk: (wst w->know)->wst w->know) g s ty n1 vk1 n2 vk2
    (gOK: wglobgen w lk g)
    (GEN: forall g' s' 
            (g'OK: wglobgen w lk g') 
            (LEG: g <6= g')
            (LES: wextpub w s' s), 
          lk g' s' ty n1 vk1 n2 vk2),
  lk g s ty n1 vk1 n2 vk2.
Proof.
  i; destruct vk1, vk2; ss; eauto.
Qed.

