Require Import Program List Arith Even Div2 Relations String.
Require Import vgtac basis countable lang subs model basiclemmas.
Require Import Classical FunctionalExtensionality.

Open Scope string_scope.

Set Implicit Arguments.

Structure wmap (w1: world) (w2: world) := mkwmap
{ smap :> wst w1 -> wst w2
; wmap_nam : forall a, wnam w1 a -> wnam w2 a
; wmap_ext : forall s s' (EXT: wext w1 s' s), 
    wext w2 (smap s') (smap s)
; wmap_extpub : forall s s' (EXT: wextpub w1 s' s), 
    wextpub w2 (smap s') (smap s)
; wmap_val : forall s, 
    wval w1 () s <6= wval w2 () (smap s)
; wmap_heap : forall s g (gOK : wglob w2 g),
    wheap w1 () s (g (smap s)) <2= wheap w2 () (smap s) (g (smap s))
}.

Structure wiso (w1: world) (w2: world) := mkwiso
{ wmap1 :> wmap w1 w2
; wmap2 :> wmap w2 w1
; wiso_bij1 : forall s, wmap2 (wmap1 s) = s
; wiso_bij2 : forall s, wmap1 (wmap2 s) = s
}.

Lemma wiso_glob: forall (w1: world) (w2: world) (is: wiso w1 w2) g (gOK: wglob w2 g),
  wglob w1 (fun s => g (wmap1 is s)).
Proof.
  s; i; rdes gOK; econs; i; esplits; eauto. 
  - eauto using wmap_ext.
  - rhdes GSUP; r; esplits; i; [eauto using wmap_val| |].
    + rewrite <- (wiso_bij1 is s); apply wmap_val.
      apply GEREF in R; eapply wval_mon; eauto. 
      by i; rewrite wiso_bij2.
    + rewrite <- (wiso_bij1 is s); apply wmap_val.
      apply GENAM in R; [eapply wval_mon; eauto|by apply (wmap_nam (wmap1 is))].
      by i; rewrite wiso_bij2.
Qed. 

Section equiv. (******************************************************************************)

Variables w1 : world.
Variables w2 : world.
Variable is : wiso w1 w2.

Lemma wexp_iso : forall g (gOK: wglob w2 g) s0 s,
  wexp w1 (fun s => g (wmap1 is s)) (wmap2 is s0) (wmap2 is s) <5= wexp w2 g s0 s.
Proof.
  pcofix CIH.
  i; exploit PR; [|apply HPLUS|apply HPLUS'|].
    apply (wmap_heap (wmap2 is) s (wiso_glob is gOK)).
    by eapply wheap_mon; eauto; i; rewrite wiso_bij2.
  intro X; pfold; pdes X; [by eauto| |].
  - apply (wmap_ext is) in EXT; apply (wmap_extpub is) in EXTP.
    rewrite wiso_bij2 in EXT, EXTP.
    econs 2; try apply PLUS1; try apply PLUS2; try apply RED; try apply RED'
    ; eauto using wmap_heap.
  - apply (wmap_ext is) in EXT; rewrite wiso_bij2 in EXT.
    econs 3; try apply PLUS1; try apply PLUS2; try apply RED; try apply RED'
    ; eauto using wmap_heap.
    ii; right; eapply CIH; eauto.
    ii; eapply or_false, KREL; eauto using wiso_glob.
    + by apply (wmap_extpub (wmap2 is)) in EXTP; rewrite wiso_bij1 in EXTP.
    + by apply (wmap_ext (wmap2 is)) in EXT0; rewrite wiso_bij1 in EXT0.
    + by rewrite wiso_bij2.
Qed.

Lemma winhabited_iso : forall (INH: winhabited w1), winhabited w2.
Proof.
  ii; edestruct INH; eauto using wiso_glob.
  exists (wmap1 is x); i; eapply wmap_heap; eauto; eapply H.
Qed.

Lemma wconsistent_iso : forall (CS: wconsistent w1), wconsistent w2.
Proof.
  ii; edestruct CS.
  - eapply wiso_glob; eauto.
  - eapply wstuck_mon; eauto.
    + by i; eapply wmap_val, wval_mon; eauto; i; rewrite wiso_bij2.
    + by i; rewrite wiso_bij2.
  - econs; eauto using wexp_iso.
Qed.

Theorem weq_iso : forall D G ty nf1 e1 nf2 e2
   (EQ1 : weq D G w1 ty nf1 e1 nf2 e2),
   weq D G w2 ty nf1 e1 nf2 e2.
Proof.
  i; rdes EQ1; r; esplits; eauto using winhabited_iso, wconsistent_iso.
  destruct EQ; econs; eauto.
  i; eapply wexp_iso; eauto.
  eapply WEXP; eauto using wiso_glob. 
  i; rewrite wiso_bij2; eauto.
Qed.

End equiv. (******************************************************************************)


Program Definition symiso (w1: world) (w2: world) (is: wiso w1 w2) : wiso w2 w1 :=
{| wmap1 := {| smap s := wmap2 is s |}
 ; wmap2 := {| smap s := wmap1 is s |} |}.
Next Obligation. eauto using (wmap_nam (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_ext (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_extpub (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_val (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_heap (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_nam (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_ext (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_extpub (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_val (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_heap (wmap1 is)). Qed.
Next Obligation. by s; rewrite wiso_bij2. Qed.
Next Obligation. by s; rewrite wiso_bij1. Qed.

Program Definition swapiso (w1: world_local) (w2: world_local) 
  : wiso (wlift (wprod w1 w2)) (wlift (wprod w2 w1)) :=
{| wmap1 := {| smap s := (fst s, (snd (snd s), fst (snd s))) |}
 ; wmap2 := {| smap s := (fst s, (snd (snd s), fst (snd s))) |} |}.
Next Obligation. rrdes H; rrdes H; rr; eauto; right; rr; eauto. Qed.
Next Obligation. ss; des;  esplits; eauto. Qed.
Next Obligation. by destruct H; [left|right; destruct H;[right|left]]. Qed.
Next Obligation. ss; des; esplits; eauto. Qed.
Next Obligation. by s; destruct s as (?,(?,?)). Qed.
Next Obligation. by s; destruct s as (?,(?,?)). Qed.

Program Definition id_intro_iso w : wiso (wlift (wprod w wid)) (wlift w) :=
  {| wmap1 := {| smap s := (fst s, fst(snd s)) : wst (wlift w) |}
   ; wmap2 := {| smap s := (fst s, (snd s,())) |} |}.
Next Obligation. rrdes H; rrdes H; rr; eauto; right; rr; eauto. Qed.
Next Obligation. by destruct PR; [left|destruct H; [right|destruct x1]]. Qed.
Next Obligation. 
  rrdes PR; rr; esplits; eauto.
  rrdes WH'; rrdes WH'; subst.
  apply hplus_hemp_rE in HP0; apply hplus_hemp_rE in HP'0; subst; eauto.
Qed.
Next Obligation. ss; unfold nunion in*; des; eauto. Qed.
Next Obligation.
  rrdes PR; rr; esplits; eauto.
  rr; esplits; s; eauto.
Qed.
Next Obligation. by s; destruct s as (?,(?,())). Qed.
Next Obligation. by s; destruct s as (?,?). Qed.

Program Definition id_elim_iso : wiso wref (wlift wid) :=
  {| wmap1 := {| smap s := (s,()) : wst (wlift wid) |}
   ; wmap2 := {| smap s := fst s |} |}.
Next Obligation. rr; esplits; eauto; s; eauto. Qed.
Next Obligation. ss; unfold nunion in*; des; eauto. Qed.
Next Obligation. by destruct PR; eauto; destruct x1. Qed.
Next Obligation. 
  rrdes PR; rrdes WH'; subst.
  apply hplus_hemp_rE in HP; apply hplus_hemp_rE in HP'; subst; eauto.
Qed.
Next Obligation. by s; destruct s as (?,()). Qed.

Lemma wglob_wid: forall w g (gOK : wglob (wlift w) g) s0,
  wglob (wlift wid) (fun s => g (fst s,s0)).
Proof.
  intros; eapply (wiso_glob (id_intro_iso w)) in gOK.
  by eapply wglob_prod2 in gOK; simpls; eauto.
Qed.
