Require Import Program List Arith Even Div2 Relations String.
Require Import vgtac basis countable lang subs model basiclemmas.
Require Import Classical FunctionalExtensionality.

Open Scope string_scope.

Set Implicit Arguments.

Structure wmap (w1: world) (w2: world) := mkwmap
{ smap :> wst w1 -> wst w2
; wmap_nam : forall a, wnam w1 a -> wnam w2 a
; wmap_extpub : forall s s' (EXT: wextpub w1 s' s), 
    wextpub w2 (smap s') (smap s)
; wmap_val : forall s g (MON: gmon w2 g <6= g), 
    wval w1 (fun _ s => g (smap s)) () s <5= wval w2 (fun _ => g) () (smap s)
; wmap_heap : forall s g (gOK : wglob w2 g),
    wheap w1 (fun _ s => g (smap s)) () s <2= wheap w2 (fun _ => g) () (smap s)
}.

Structure wiso (w1: world) (w2: world) := mkwiso
{ wmap1 :> wmap w1 w2
; wmap2 :> wmap w2 w1
; wiso_bij1 : forall s, wmap2 (wmap1 s) = s
; wiso_bij2 : forall s, wmap1 (wmap2 s) = s
}.

Lemma wiso_gmon1: forall (w1: world) (w2: world) (is: wiso w1 w2) g (gOK: wglob w2 g),
  gmon w1 (fun s => g (wmap1 is s)) <6= (fun s => g (wmap1 is s)).
Proof. 
  i; depdes PR; eapply (wg_mon gOK); eauto using wmap_extpub.
Qed.

Lemma wiso_gmon2: forall (w1: world) (w2: world) (is: wiso w1 w2) g (gOK: wglob w2 g),
  gmon w2 g <6= g.
Proof. i; depdes PR; eauto using (wg_mon gOK). Qed.

Lemma wiso_glob: forall (w1: world) (w2: world) (is: wiso w1 w2) g (gOK: wglob w2 g),
  wglob w1 (fun s => g (wmap1 is s)).
Proof.
  s; i; econs; i. 
  - eauto using (wg_clo gOK).
  - eapply (wg_mon gOK); (inst; s; eauto using wmap_extpub).
  - i; destruct (wg_le gOK (wmap1 is s)).
    econs; i.
    + eapply wmap_val in R; eauto.
      eapply wiso_gmon2; eauto.
    + rewrite <- (wiso_bij1 is s). 
      apply wmap_val; [by eapply wiso_gmon1; eauto|].
      apply refn_ref in R; eapply wval_mon; eauto.
      by i; rewrite wiso_bij2.
    + rewrite <- (wiso_bij1 is s).
      apply wmap_val; [by eapply wiso_gmon1; eauto|].
      apply refn_nam in R; [eapply wval_mon; eauto|by apply (wmap_nam (wmap1 is))].
      by i; rewrite wiso_bij2.
  - apply (wg_sc gOK).
Qed. 

Section equiv. (******************************************************************************)

Variables w1 : world.
Variables w2 : world.
Variable is : wiso w1 w2.

Lemma wexp_iso : forall g (gOK: wglob w2 g) s,
  wexp w1 (fun s => g (wmap1 is s)) (wmap2 is s) <4= wexp w2 g s.
Proof.
  ii; exploit PR; [|apply HPLUS|apply HPLUS'|].
    apply (wmap_heap (wmap2 is)); [by apply wiso_glob|].
    by eapply wheap_mon; eauto; i; rewrite wiso_bij2.
  intro X; destruct X; [by eauto|].
  econs 2; [rewrite <- (wiso_bij2 is s)|apply EREL| 
           |apply PLUS1 | apply PLUS2|apply RED | apply RED']
  ; s; eauto using wmap_extpub, wmap_heap.
Qed.

Lemma winhabited_iso : forall (INH: winhabited w1), winhabited w2.
Proof.
  ii; edestruct INH; exists (wmap1 is x); i.
  eapply wmap_heap; eauto; eapply H; eapply (wiso_glob _ gOK).
Qed.

Lemma wconsistent_iso : forall (CS: wconsistent w1), wconsistent w2.
Proof.
  ii; edestruct CS.
  - eapply wiso_glob; eauto.
  - eapply wstuck_mon; eauto.
    + i; eapply wmap_val, wval_mon; eauto; i.
      * eapply (wiso_gmon1 is gOK); eauto.
      * rewrite wiso_bij2; eauto.
    + i; rewrite wiso_bij2; eauto.
  - econs; eauto using wexp_iso.
Qed.

Theorem weq_iso : forall G typ e1 e2
   (EQ1 : weq w1 G typ e1 e2),
   weq w2 G typ e1 e2.
Proof.
  i; rdes EQ1; r; esplits; eauto using winhabited_iso, wconsistent_iso.
  destruct EQ; econs; eauto.
  i; eapply wexp_iso; eauto.
  eapply WEXP; eauto; [by apply wiso_glob| |]; i; rewrite wiso_bij2; eauto.
Qed.

End equiv. (******************************************************************************)


Program Definition symiso (w1: world) (w2: world) (is: wiso w1 w2) : wiso w2 w1 :=
{| wmap1 := {| smap s := wmap2 is s |}
 ; wmap2 := {| smap s := wmap1 is s |} |}.
Next Obligation. eauto using (wmap_nam (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_extpub (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_val (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_heap (wmap2 is)). Qed.
Next Obligation. eauto using (wmap_nam (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_extpub (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_val (wmap1 is)). Qed.
Next Obligation. eauto using (wmap_heap (wmap1 is)). Qed.
Next Obligation. by s; rewrite wiso_bij2. Qed.
Next Obligation. by s; rewrite wiso_bij1. Qed.

Program Definition swapiso (w1: world_local) (w2: world_local) 
  : wiso (wlift (wprod w1 w2)) (wlift (wprod w2 w1)) :=
{| wmap1 := {| smap s := (fst s, (snd (snd s), fst (snd s))) |}
 ; wmap2 := {| smap s := (fst s, (snd (snd s), fst (snd s))) |} |}.
Next Obligation. rrdes H; rrdes H; rr; eauto; right; rr; eauto. Qed.
Next Obligation. 
  ss; unfold wlift_val in *; destruct (is_ref _ _ _)
  ; ss; rrdes PR; rr; esplits; eauto.
Qed.
Next Obligation. rrdes PR; rrdes WH'; rr; esplits; eauto; rr; esplits; eauto. Qed.
Next Obligation. rrdes H; rrdes H; rr; eauto; right; rr; eauto. Qed.
Next Obligation. 
  ss; unfold wlift_val in *; destruct (is_ref _ _ _)
  ; ss; rrdes PR; rr; esplits; eauto.
Qed.
Next Obligation. rrdes PR; rrdes WH'; rr; esplits; eauto; rr; esplits; eauto. Qed.
Next Obligation. by s; destruct s as (? & (? & ?)). Qed.
Next Obligation. by s; destruct s as (? & (? & ?)). Qed.

Program Definition wid : world_local := 
  {| wnam n := False
   ; wst := unit
   ; wextpub s s' := True
   ; wval r sp s ty n vk n' vk' := False
   ; wheap r sp s h h' := h = hemp /\ h' = hemp |}.
Next Obligation. by eexists (fun _ => 0); ii; destruct x, y. Qed.

Lemma wstable_id : wstable wid.
Proof. repeat red; eauto. Qed.

Lemma winhabited_id : winhabited (wlift wid).
Proof. exists (sref0,()); repeat exists hemp; eby repeat split; eauto. Qed.

Lemma wconsistent_id : wconsistent (wlift wid).
Proof. red; inversion 2; intros; simpls; des; clarify. Qed.

Hint Resolve  wstable_id winhabited_id wconsistent_id.

Program Definition id_intro_iso w : wiso (wlift (wprod w wid)) (wlift w) :=
  {| wmap1 := {| smap s := (fst s, fst(snd s)) : wst (wlift w) |}
   ; wmap2 := {| smap s := (fst s, (snd s,())) |} |}.
Next Obligation. rrdes H; rrdes H; rr; eauto; right; rr; eauto. Qed.
Next Obligation. 
  ss; unfold wlift_val in *; destruct is_ref; eauto.
  destruct PR; try done.
  eapply wval_mon; eauto.
  ss; i; depdes PR; eauto.
Qed.
Next Obligation. 
  rrdes PR; rr; esplits; eauto.
  rrdes WH'; rrdes WH'; subst.
  apply hplus_hemp_rE in HP0; apply hplus_hemp_rE in HP'0; subst. 
  eapply wheap_mon; eauto.
  ss; i; depdes PR; eauto using (wg_mon gOK).
Qed.
Next Obligation. ss; unfold nunion in*; des; eauto. Qed.
Next Obligation.
  ss; unfold wlift_val in *; destruct is_ref; eauto.
  left; eapply wval_mon; (inst; ss; eauto).
Qed.
Next Obligation. 
  rrdes PR; rr; esplits; eauto.
  rr; esplits; ss; eauto using hplus_hemp_l.
  eapply wheap_mon; (inst; ss; eauto).
Qed.
Next Obligation. by s; destruct s as (?,(?,())). Qed.
Next Obligation. by s; destruct s as (?,?). Qed.

Program Definition id_elim_iso : wiso wref (wlift wid) :=
  {| wmap1 := {| smap s := (s,()) : wst (wlift wid) |}
   ; wmap2 := {| smap s := fst s |} |}.
Next Obligation. rrdes PR; subst; rr; s; esplits; eauto. Qed.
Next Obligation. rr; esplits; ss; eauto. Qed.
Next Obligation. ss; unfold nunion in*; des; eauto. Qed.
Next Obligation. by ss; r in PR; destruct is_ref; eauto. Qed.
Next Obligation. 
  rrdes PR; rrdes WH'; subst.
  apply hplus_hemp_rE in HP; apply hplus_hemp_rE in HP'; subst; eauto.
Qed.
Next Obligation. by s; destruct s as (?,()). Qed.

Lemma wglob_wid: forall w g (gOK : wglob (wlift w) g) s0,
  wglob (wlift wid) (fun s => g (fst s,s0)).
Proof.
  i; assert (X:= wiso_glob (id_intro_iso w) gOK).
  eapply wglob_prod2 in X; [|done].
  eapply eqimpl, X.
  f_equal; extensionalities s ty n1 vk1 n2 vk2; apply prop_ext.
  destruct s as (s,()); split; i.
  - depdes H; s; s in REL; eapply (wg_mon gOK); s; eauto.
  - econs; s; eauto.
Qed.
